Package: Rtinycc
Title: Builds the 'TinyCC' Command-Line Interface and Library for 'C'
        Scripting in 'R'
Version: 0.1.12
Authors@R: c(
    person(given = "Sounkou Mahamane", family = "Toure", 
    email = "sounkoutoure@gmail.com", role = c("aut", "cre")),
    person(given = "Mike", family = "Cheng", role = c("cph"),
    comment = "Adapted parts of the vignette knitr engine from the callme package"),
    person(given = "Adam", family = "Dunkels", role = c("ctb"),
    comment = "Author of the bundled Protothreads headers"),
    person(family = "Swedish Institute of Computer Science", role = c("cph"),
    comment = "Copyright holder for the bundled Protothreads headers"),
    person(given = "Fabrice Bellard and TinyCC authors", role = c("cph"),
    comment = "TinyCC compiler authors and copyright holders")
    )
Description: Builds the 'TinyCC' (Tiny 'C' Compiler) command-line interface
    and library for package use in 'R'. The package compiles 'TinyCC' from source and
    provides R functions to interact with the compiler. 'TinyCC' can be
    used for header preprocessing, just-in-time compilation of 'C' code in
    'R', and lightweight 'C' scripting workflows.
License: GPL (>= 3)
Copyright: See inst/LICENSE.note for bundled TinyCC copyright and
        licensing details.
Depends: R (>= 4.4.0)
Imports: lambda.r
Suggests: bench, callme, knitr, rmarkdown, tinytest, treesitter.c (>=
        0.0.4)
VignetteBuilder: knitr
SystemRequirements: GNU make
Encoding: UTF-8
URL: https://github.com/sounkou-bioinfo/Rtinycc,
        https://sounkou-bioinfo.github.io/Rtinycc/
BugReports: https://github.com/sounkou-bioinfo/Rtinycc/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-18 20:49:42 UTC; sounkoutoure
Author: Sounkou Mahamane Toure [aut, cre],
  Mike Cheng [cph] (Adapted parts of the vignette knitr engine from the
    callme package),
  Adam Dunkels [ctb] (Author of the bundled Protothreads headers),
  Swedish Institute of Computer Science [cph] (Copyright holder for the
    bundled Protothreads headers),
  Fabrice Bellard and TinyCC authors [cph] (TinyCC compiler authors and
    copyright holders)
Maintainer: Sounkou Mahamane Toure <sounkoutoure@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-19 18:00:11 UTC
Built: R 4.6.1; aarch64-apple-darwin23; 2026-07-19 18:52:58 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",tcc_compiled)
S3method(as.character,tcc_cstring)
S3method(print,tcc_callback)
S3method(print,tcc_compiled)
S3method(print,tcc_cstring)
S3method(print,tcc_ffi)
export(blas_lapack_info)
export(get_external_ptr_addr)
export(tcc_add_file)
export(tcc_add_include_path)
export(tcc_add_library)
export(tcc_add_library_path)
export(tcc_add_symbol)
export(tcc_add_sysinclude_path)
export(tcc_bin_path)
export(tcc_bind)
export(tcc_call_symbol)
export(tcc_callback)
export(tcc_callback_async_drain)
export(tcc_callback_async_schedule)
export(tcc_callback_close)
export(tcc_callback_ptr)
export(tcc_callback_valid)
export(tcc_cli)
export(tcc_compile)
export(tcc_compile_string)
export(tcc_container_of)
export(tcc_cstring)
export(tcc_cstring_object)
export(tcc_data_ptr)
export(tcc_enum)
export(tcc_ffi)
export(tcc_field_addr)
export(tcc_free)
export(tcc_generate_bindings)
export(tcc_get_symbol)
export(tcc_global)
export(tcc_header)
export(tcc_include)
export(tcc_include_path)
export(tcc_include_paths)
export(tcc_introspect)
export(tcc_lib_path)
export(tcc_lib_paths)
export(tcc_library)
export(tcc_library_path)
export(tcc_link)
export(tcc_list_symbols)
export(tcc_malloc)
export(tcc_map_c_type_to_ffi)
export(tcc_null_ptr)
export(tcc_options)
export(tcc_output)
export(tcc_output_file)
export(tcc_path)
export(tcc_prefix)
export(tcc_ptr_addr)
export(tcc_ptr_free_set_null)
export(tcc_ptr_is_null)
export(tcc_ptr_is_owned)
export(tcc_ptr_set)
export(tcc_read_bytes)
export(tcc_read_cstring)
export(tcc_read_f32)
export(tcc_read_f64)
export(tcc_read_i16)
export(tcc_read_i32)
export(tcc_read_i64)
export(tcc_read_i8)
export(tcc_read_ptr)
export(tcc_read_u16)
export(tcc_read_u32)
export(tcc_read_u64)
export(tcc_read_u8)
export(tcc_recompile)
export(tcc_relocate)
export(tcc_run_cli)
export(tcc_set_options)
export(tcc_source)
export(tcc_state)
export(tcc_struct)
export(tcc_struct_raw_access)
export(tcc_symbol_is_valid)
export(tcc_sysinclude_paths)
export(tcc_treesitter_bindings)
export(tcc_treesitter_defines)
export(tcc_treesitter_enum_bindings)
export(tcc_treesitter_enum_members)
export(tcc_treesitter_enums)
export(tcc_treesitter_functions)
export(tcc_treesitter_global_types)
export(tcc_treesitter_globals)
export(tcc_treesitter_struct_accessors)
export(tcc_treesitter_struct_bindings)
export(tcc_treesitter_struct_members)
export(tcc_treesitter_structs)
export(tcc_treesitter_union_accessors)
export(tcc_treesitter_union_bindings)
export(tcc_treesitter_union_members)
export(tcc_treesitter_unions)
export(tcc_union)
export(tcc_write_bytes)
export(tcc_write_f32)
export(tcc_write_f64)
export(tcc_write_i16)
export(tcc_write_i32)
export(tcc_write_i64)
export(tcc_write_i8)
export(tcc_write_ptr)
export(tcc_write_u16)
export(tcc_write_u32)
export(tcc_write_u64)
export(tcc_write_u8)
importFrom(lambda.r,"%as%")
importFrom(lambda.r,UseFunction)
useDynLib(Rtinycc, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Rtinycc

Builds `TinyCC` `Cli` and Library For `C` Scripting in `R`

<!-- badges: start -->

[![R-CMD-check](https://github.com/sounkou-bioinfo/Rtinycc/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sounkou-bioinfo/Rtinycc/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/Rtinycc)](https://CRAN.R-project.org/package=Rtinycc)
[![Rtinycc status
badge](https://sounkou-bioinfo.r-universe.dev/Rtinycc/badges/version)](https://sounkou-bioinfo.r-universe.dev/Rtinycc)
<!-- badges: end -->

## Abstract

Rtinycc is an R interface to [TinyCC](https://github.com/TinyCC/tinycc),
providing both CLI access and a libtcc-backed in-memory compiler. It
includes an FFI inspired by [Bun’s
FFI](https://bun.com/docs/runtime/ffi) for binding C symbols with
predictable type conversions and pointer utilities. The package works on
unix-alikes and Windows and focuses on embedding TinyCC and enabling
JIT-compiled bindings directly from R. Combined with
[treesitter.c](https://github.com/sounkou-bioinfo/treesitter.c), which
provides C header parsers, it can be used to rapidly generate
declarative bindings.

## How it works

When you call `tcc_compile()`, Rtinycc generates C wrapper functions
whose signature follows the `.Call` convention (`SEXP` in, `SEXP` out).
These wrappers convert R types to C, call the target function, and
convert the result back. TCC compiles them in-memory – no shared library
is written to disk and no `R_init_*` registration is needed.

After `tcc_relocate()`, wrapper pointers are retrieved via
`tcc_get_symbol()`, which internally calls `RC_libtcc_get_symbol()`.
That function converts TCC’s raw `void*` into a `DL_FUNC` wrapped with
`R_MakeExternalPtrFn` (tagged `"native symbol"`). The symbol pointer
retains its owning `tcc_state`, so the relocated code remains alive
while the symbol is reachable. On the R side,
[`make_callable()`](R/ffi.R) creates a closure that passes this external
pointer to `.Call` (aliased as `.RtinyccCall` to keep `R CMD check`
happy). Non-memory states instead produce one-shot artifacts with
`tcc_output_file()` and cannot expose callable symbol pointers.

The design follows [CFFI’s](https://cffi.readthedocs.io/) API-mode
pattern: instead of computing struct layouts and calling conventions in
R (ABI-mode, like Python’s ctypes), the generated C code lets TCC handle
`sizeof`, `offsetof`, and argument passing. Rtinycc never replicates
platform-specific layout rules. The wrappers can also link against
external shared libraries whose symbols TCC resolves at relocation time.
For background on how this compares to a libffi approach, see the
[`RSimpleFFI`
README](https://github.com/sounkou-bioinfo/RSimpleFFI#readme).

On macOS the configure script strips `-flat_namespace` from TCC’s build
to avoid [BUS ERROR
issues](https://genomic.social/@bioinfhotep/115765645745231377). Without
it, TCC cannot resolve host symbols (e.g. `RC_free_finalizer`) through
the dynamic linker. Rtinycc works around this with
`RC_libtcc_add_host_symbols()`, which registers package-internal C
functions via `tcc_add_symbol()` before relocation. Any new C function
referenced by generated TCC code must be added there.

On Windows, the `configure.win` script generates a UCRT-backed
`msvcrt.def` so TinyCC resolves CRT symbols against `ucrtbase.dll` (R
4.2+ uses UCRT).

Ownership semantics are explicit. Pointers from `tcc_malloc()` are
tagged `rtinycc_owned` and can be released with `tcc_free()` (or by
their R finalizer). Generated struct constructors use a struct-specific
tag (`struct_<name>`) and host-allocated storage with
`RC_owned_native_finalizer`; free them with `struct_<name>_free()`, not
`tcc_free()`. Pointers from `tcc_data_ptr()` are tagged
`rtinycc_borrowed` and are never freed by Rtinycc. Array returns are
copied into a fresh R vector; set `free = TRUE` only when the C function
returns a `malloc`-owned buffer.

## Installation

``` r
install.packages(
      'Rtinycc', 
        repos = c('https://sounkou-bioinfo.r-universe.dev', 
                  'https://cloud.r-project.org')
        )
```

## Usage

### CLI

The CLI interface compiles C source files to standalone executables
using the bundled TinyCC toolchain.

``` r
library(Rtinycc)

src <- system.file("c_examples", "forty_two.c", package = "Rtinycc")
exe <- tempfile()
tcc_run_cli(c(
  "-B", tcc_prefix(),
  paste0("-I", tcc_include_paths()),
  paste0("-L", tcc_lib_paths()),
  src, "-o", exe
))
#> [1] 0
Sys.chmod(exe, mode = "0755")
system2(exe, stdout = TRUE)
#> [1] "42"
```

For in-memory workflows, prefer libtcc instead.

### In-memory compilation with libtcc

We can compile and call C functions entirely in memory. This is the
simplest path for quick JIT compilation.

``` r
state <- tcc_state(output = "memory")
tcc_compile_string(state, "int forty_two(){ return 42; }")
#> [1] 0
tcc_relocate(state)
#> [1] 0
tcc_call_symbol(state, "forty_two", return = "int")
#> [1] 42
```

For low-level pointer-style calls, `tcc_call_symbol()` can also use an R
`.C()`-like convention: call a `void` C routine with pointers to copied
argument buffers, then return a list of the modified values.

``` r
state <- tcc_state(output = "memory")
tcc_compile_string(state, "void add_one(int *x) { x[0] += 1; }")
#> [1] 0
tcc_relocate(state)
#> [1] 0
tcc_call_symbol(state, "add_one", x = as.integer(41))
#> $x
#> [1] 42
```

The lower-level API gives full control over include paths, libraries,
and the R C API. Using `#define _Complex` as a workaround for TCC’s lack
of [complex type
support](https://lists.gnu.org/archive/html/tinycc-devel/2022-04/msg00020.html),
we can link against R’s headers and call into `libR`.

``` r
state <- tcc_state(output = "memory")
tcc_add_include_path(state, R.home("include"))
#> [1] 0
tcc_add_library_path(state, R.home("lib"))
#> [1] 0

code <- '
#define _Complex
#include <R.h>
#include <Rinternals.h>

double call_r_sqrt(void) {
  SEXP fn   = PROTECT(Rf_findFun(Rf_install("sqrt"), R_BaseEnv));
  SEXP val  = PROTECT(Rf_ScalarReal(16.0));
  SEXP call = PROTECT(Rf_lang2(fn, val));
  SEXP out  = PROTECT(Rf_eval(call, R_GlobalEnv));
  double res = REAL(out)[0];
  UNPROTECT(4);
  return res;
}
'
tcc_compile_string(state, code)
#> [1] 0
tcc_relocate(state)
#> [1] 0
tcc_call_symbol(state, "call_r_sqrt", return = "double")
#> [1] 4
```

### Pointer utilities

Rtinycc ships a set of typed memory access functions similar to what the
[ctypesio](https://cran.r-project.org/package=ctypesio) package offers,
but designed around our FFI pointer model. Every scalar C type has a
corresponding `tcc_read_*` / `tcc_write_*` pair that operates at a byte
offset into any external pointer, so you can walk structs, arrays, and
output parameters without writing C helpers.

``` r
ptr <- tcc_cstring("hello")
tcc_read_cstring(ptr)
#> [1] "hello"
tcc_read_bytes(ptr, 5)
#> [1] 68 65 6c 6c 6f
tcc_ptr_addr(ptr, hex = TRUE)
#> [1] "0x5dfe4e9ec0f0"
tcc_ptr_is_null(ptr)
#> [1] FALSE
tcc_free(ptr)
#> NULL
```

Typed reads and writes cover the full scalar range (`i8`/`u8`,
`i16`/`u16`, `i32`/`u32`, `i64`/`u64`, `f32`/`f64`) plus pointer
dereferencing via `tcc_read_ptr` / `tcc_write_ptr`. All operations use a
byte offset and `memcpy` internally for alignment safety.

``` r
buf <- tcc_malloc(32)
tcc_write_i32(buf, 0L, 42L)
tcc_write_f64(buf, 8L, pi)
tcc_read_i32(buf, offset = 0L)
#> [1] 42
tcc_read_f64(buf, offset = 8L)
#> [1] 3.141593
tcc_free(buf)
#> NULL
```

Pointer-to-pointer workflows are supported for C APIs that return values
through output parameters.

``` r
ptr_ref <- tcc_malloc(.Machine$sizeof.pointer %||% 8L)
target <- tcc_malloc(8)
tcc_ptr_set(ptr_ref, target)
#> <pointer: 0x5dfe514cc430>
tcc_data_ptr(ptr_ref)
#> <pointer: 0x5dfe4b9cbef0>
tcc_ptr_set(ptr_ref, tcc_null_ptr())
#> <pointer: 0x5dfe514cc430>
tcc_free(target)
#> NULL
tcc_free(ptr_ref)
#> NULL
```

## Declarative FFI

A declarative interface inspired by [Bun’s
FFI](https://bun.com/docs/runtime/ffi) sits on top of the lower-level
API. We define types explicitly and Rtinycc generates the binding code,
compiling it in memory with TCC.

### Type system

The FFI exposes a small set of type mappings between R and C.
Conversions are explicit and predictable so callers know when data is
shared versus copied.

The scalar type names are C-facing, but the R-side carriers are not all
one-to-one with those C widths:

- `i8`, `i16`, `i32`, `u8`, and `u16` are mediated through R integer
  scalars
- `u32`, `i64`, `u64`, `f32`, and `f64` are mediated through R numeric
  (`double`) coercion and boxing
- `bool` uses R logical
- `cstring` uses an R character scalar

This means `u32` is routed through `double` to preserve the full
unsigned 32-bit range, and `i64` / `u64` are only exact up to `2^53` on
the R side.

Array arguments pass R vectors to C with zero copy: `raw` maps to
`uint8_t*`, `integer_array` to `int32_t*`, `numeric_array` to `double*`.

Pointer types include `ptr` (opaque external pointer), `sexp` (pass a
`SEXP` directly), and callback signatures like
`callback:double(double)`.

Variadic functions are supported in two forms: typed prefix tails
(`varargs`) and bounded dynamic tails (`varargs_types` +
`varargs_min`/`varargs_max`). Prefix mode is the cheaper runtime path
because dispatch is by tail arity only; bounded dynamic mode adds
per-call scalar type inference to select a compatible wrapper. For hot
loops, prefer fixed arity first, then prefix variadics with a tight
maximum tail size.

Array returns use
`returns = list(type = "integer_array", length_arg = 2, free = TRUE)` to
copy the result into a new R vector. The `length_arg` is the 1-based
index of the C argument that carries the array length. Set `free = TRUE`
when the C function returns a `malloc`-owned buffer.

### Simple functions

``` r
ffi <- tcc_ffi() |>
  tcc_source("
    int add(int a, int b) { return a + b; }
  ") |>
  tcc_bind(add = list(args = list("i32", "i32"), returns = "i32")) |>
  tcc_compile()

ffi$add(5L, 3L)
#> [1] 8
```

### Variadic calls (e.g. `Rprintf` style)

Rtinycc supports two ways to bind variadic tails. The legacy approach
uses `varargs` as a typed prefix tail, while the bounded dynamic
approach uses `varargs_types` together with `varargs_min` and
`varargs_max`. In the bounded mode, wrappers are generated across the
allowed arity and type combinations, and runtime dispatch selects the
matching wrapper from the scalar tail values provided at call time.

``` r
ffi_var <- tcc_ffi() |>
  tcc_header("#include <R_ext/Print.h>") |>
  tcc_source('
    #include <stdarg.h>

    int sum_fmt(int n, ...) {
      va_list ap;
      va_start(ap, n);
      int s = 0;
      for (int i = 0; i < n; i++) s += va_arg(ap, int);
      va_end(ap);
      Rprintf("sum_fmt(%d) = %d\\n", n, s);
      return s;
    }
  ') |>
  tcc_bind(
    Rprintf = list(
      args = list("cstring"),
      variadic = TRUE,
      varargs_types = list("i32"),
      varargs_min = 0L,
      varargs_max = 4L,
      returns = "void"
    ),
    sum_fmt = list(
      args = list("i32"),
      variadic = TRUE,
      varargs_types = list("i32"),
      varargs_min = 0L,
      varargs_max = 4L,
      returns = "i32"
    )
  ) |>
  tcc_compile()

ffi_var$Rprintf("Rprintf via bind: %d + %d = %d\n", 2L, 3L, 5L)
#> Rprintf via bind: 2 + 3 = 5
#> NULL
ffi_var$sum_fmt(0L)
#> sum_fmt(0) = 0
#> [1] 0
ffi_var$sum_fmt(2L, 10L, 20L)
#> sum_fmt(2) = 30
#> [1] 30
ffi_var$sum_fmt(4L, 1L, 2L, 3L, 4L)
#> sum_fmt(4) = 10
#> [1] 10
```

### Linking external libraries

We can bind directly to symbols in shared libraries. Here we link
against `libm`.

``` r
math <- tcc_ffi() |>
  tcc_library("m") |>
  tcc_bind(
    sqrt  = list(args = list("f64"), returns = "f64"),
    sin   = list(args = list("f64"), returns = "f64"),
    floor = list(args = list("f64"), returns = "f64")
  ) |>
  tcc_compile()

math$sqrt(16.0)
#> [1] 4
math$sin(pi / 2)
#> [1] 1
math$floor(3.7)
#> [1] 3
```

### Compiler options

Use `tcc_options()` to pass raw TinyCC options in the high-level FFI
pipeline. For low-level states, use `tcc_set_options()` directly.

``` r
ffi_opt_off <- tcc_ffi() |>
  tcc_options("-O0") |>
  tcc_source('
    int opt_macro() {
    #ifdef __OPTIMIZE__
      return 1;
    #else
      return 0;
    #endif
    }
  ') |>
  tcc_bind(opt_macro = list(args = list(), returns = "i32")) |>
  tcc_compile()

ffi_opt_on <- tcc_ffi() |>
  tcc_options(c("-Wall", "-O2")) |>
  tcc_source('
    int opt_macro() {
    #ifdef __OPTIMIZE__
      return 1;
    #else
      return 0;
    #endif
    }
  ') |>
  tcc_bind(opt_macro = list(args = list(), returns = "i32")) |>
  tcc_compile()

ffi_opt_off$opt_macro()
#> [1] 0
ffi_opt_on$opt_macro()
#> [1] 1
```

### Working with arrays

R vectors are passed to C with zero copy. Mutations in C are visible in
R.

``` r
ffi <- tcc_ffi() |>
  tcc_source("
    #include <stdlib.h>
    #include <string.h>

    int64_t sum_array(int32_t* arr, int32_t n) {
      int64_t s = 0;
      for (int i = 0; i < n; i++) s += arr[i];
      return s;
    }

    void bump_first(int32_t* arr) { arr[0] += 10; }

    int32_t* dup_array(int32_t* arr, int32_t n) {
      int32_t* out = malloc(sizeof(int32_t) * n);
      memcpy(out, arr, sizeof(int32_t) * n);
      return out;
    }
  ") |>
  tcc_bind(
    sum_array  = list(args = list("integer_array", "i32"), returns = "i64"),
    bump_first = list(args = list("integer_array"), returns = "void"),
    dup_array  = list(
      args = list("integer_array", "i32"),
      returns = list(type = "integer_array", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

x <- as.integer(1:100) # to avoid ALTREP
.Internal(inspect(x))
#> @5dfe4f87a180 13 INTSXP g0c0 [REF(65535)]  1 : 100 (compact)
ffi$sum_array(x, length(x))
#> [1] 5050

# Zero-copy: C mutation reflects in R
ffi$bump_first(x)
#> NULL
x[1]
#> [1] 11

# Array return: copied into a new R vector, C buffer freed
y <- ffi$dup_array(x, length(x))
y[1]
#> [1] 11

.Internal(inspect(x))
#> @5dfe4f87a180 13 INTSXP g0c0 [REF(65535)]  11 : 110 (expanded)
```

## Advanced FFI features

### Structs and unions

Complex C types are supported declaratively. Use `tcc_struct()` to
generate allocation and accessor helpers. Free instances when done.

``` r
ffi <- tcc_ffi() |>
  tcc_source('
    #include <math.h>
    struct point { double x; double y; };
    double distance(struct point* a, struct point* b) {
      double dx = a->x - b->x, dy = a->y - b->y;
      return sqrt(dx * dx + dy * dy);
    }
  ') |>
  tcc_library("m") |>
  tcc_struct("point", accessors = c(x = "f64", y = "f64")) |>
  tcc_bind(distance = list(args = list("ptr", "ptr"), returns = "f64")) |>
  tcc_compile()

p1 <- ffi$struct_point_new()
ffi$struct_point_set_x(p1, 0.0)
#> <pointer: 0x5dfe51808b90>
ffi$struct_point_set_y(p1, 0.0)
#> <pointer: 0x5dfe51808b90>

p2 <- ffi$struct_point_new()
ffi$struct_point_set_x(p2, 3.0)
#> <pointer: 0x5dfe4e2258c0>
ffi$struct_point_set_y(p2, 4.0)
#> <pointer: 0x5dfe4e2258c0>

ffi$distance(p1, p2)
#> [1] 5

ffi$struct_point_free(p1)
#> NULL
ffi$struct_point_free(p2)
#> NULL
```

### Enums

Enums are exposed as helper functions that return integer constants.

``` r
ffi <- tcc_ffi() |>
  tcc_source("enum color { RED = 0, GREEN = 1, BLUE = 2 };") |>
  tcc_enum("color", constants = c("RED", "GREEN", "BLUE")) |>
  tcc_compile()

ffi$enum_color_RED()
#> [1] 0
ffi$enum_color_BLUE()
#> [1] 2
```

### Bitfields

Bitfields are handled by TCC. Accessors read and write them like normal
fields.

``` r
ffi <- tcc_ffi() |>
  tcc_source("
    struct flags {
      unsigned int active : 1;
      unsigned int level  : 4;
    };
  ") |>
  tcc_struct("flags", accessors = c(active = "u8", level = "u8")) |>
  tcc_compile()

s <- ffi$struct_flags_new()
ffi$struct_flags_set_active(s, 1L)
#> <pointer: 0x5dfe4eb02090>
ffi$struct_flags_set_level(s, 9L)
#> <pointer: 0x5dfe4eb02090>
ffi$struct_flags_get_active(s)
#> [1] 1
ffi$struct_flags_get_level(s)
#> [1] 9
ffi$struct_flags_free(s)
#> NULL
```

### Global getters and setters

C globals can be exposed with explicit getter/setter helpers.

``` r
ffi <- tcc_ffi() |>
  tcc_source("
    int counter = 7;
    double pi_approx = 3.14159;
  ") |>
  tcc_global("counter", "i32") |>
  tcc_global("pi_approx", "f64") |>
  tcc_compile()

ffi$global_counter_get()
#> [1] 7
ffi$global_pi_approx_get()
#> [1] 3.14159
ffi$global_counter_set(42L)
#> [1] 42
ffi$global_counter_get()
#> [1] 42
```

### Callbacks

R functions can be registered as C function pointers via
`tcc_callback()` and passed to compiled code. Specify a
`callback:<signature>` argument in `tcc_bind()` so the trampoline is
generated automatically. Call `tcc_callback_close()` when you want
deterministic invalidation and earlier release of the preserved R
function. The legacy `threadsafe` argument is informational; worker
dispatch is selected by `callback_async:<signature>`.

``` r
cb <- tcc_callback(function(x) x * x, signature = "double (*)(double)")

code <- '
double apply_fn(double (*fn)(void* ctx, double), void* ctx, double x) {
  return fn(ctx, x);
}
'

ffi <- tcc_ffi() |>
  tcc_source(code) |>
  tcc_bind(
    apply_fn = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi$apply_fn(cb, tcc_callback_ptr(cb), 7.0)
#> [1] 49
tcc_callback_close(cb)
```

### Callback errors

If a callback throws an R error, the trampoline catches it, emits a
warning, and returns a type-appropriate sentinel instead of unwinding
through C. In practice this means NA-like numeric or integer values,
`NA` logical, `NULL` for `cstring`, or a null external pointer depending
on the declared return type.

``` r
cb_err <- tcc_callback(
  function(x) stop("boom"),
  signature = "double (*)(double)"
)

ffi_err <- tcc_ffi() |>
  tcc_source('
    double call_cb_err(double (*cb)(void* ctx, double), void* ctx, double x) {
      return cb(ctx, x);
    }
  ') |>
  tcc_bind(
    call_cb_err = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

warned <- FALSE
res <- withCallingHandlers(
  ffi_err$call_cb_err(cb_err, tcc_callback_ptr(cb_err), 1.0),
  warning = function(w) {
    warned <<- TRUE
    invokeRestart("muffleWarning")
  }
)
list(warned = warned, result = res)
#> $warned
#> [1] TRUE
#> 
#> $result
#> [1] NA
tcc_callback_close(cb_err)
```

### Async callbacks

For callback dispatch from worker threads, use
`callback_async:<signature>` in `tcc_bind()`. The async callback queue
is initialized automatically at package load.

When a bound function has any `callback_async:` argument, the generated
wrapper automatically runs your C function on a new thread while
draining callbacks on the main R thread. Your C code doesn’t need to
know about draining at all — just call the callback as normal and the
wrapper handles the rest. Interrupt requests are deferred until the
worker has returned and been joined, so native stack contexts cannot
escape while still in use.

**Void return (fire-and-forget):** the callback is enqueued from any
thread and executed on the main R thread automatically — on Windows via
R’s message pump, on Linux/macOS via R’s event loop `addInputHandler`.

**Non-void return (synchronous):** the worker thread blocks until the
main R thread executes the callback and returns the real result.
Supported return types: integer variants (`int`, `int32_t`, `i8`, `i16`,
`u8`, `u16`, `u32`, `i64`, `u64`), floating-point (`double`, `float`),
`bool`/`logical`, and pointer (`void*`, `T*`). Wide integer values must
remain within R’s exact numeric range (absolute value at most `2^53`).

Async scalar and string payloads are copied into queue tasks. Pointer
payloads copy only the address: C must keep the pointee alive until the
callback has executed. Closing a callback cancels queued work for that
callback; registry slot reuse cannot redirect old work to a newly
registered callback.

``` r
# Fire-and-forget: void callback accumulated from 100 worker threads
hits <- 0L
cb_async <- tcc_callback(
  function(x) { hits <<- hits + x; NULL },
  signature = "void (*)(int)"
)

code_async <- '
struct task { void (*cb)(void* ctx, int); void* ctx; int value; };

#ifdef _WIN32
#include <windows.h>

static DWORD WINAPI worker(LPVOID data) {
  struct task* t = (struct task*) data;
  t->cb(t->ctx, t->value);
  return 0;
}

int spawn_async(void (*cb)(void* ctx, int), void* ctx, int value) {
  if (!cb || !ctx) return -1;
  struct task t;
  t.cb = cb;
  t.ctx = ctx;
  t.value = value;
  HANDLE th = CreateThread(NULL, 0, worker, &t, 0, NULL);
  if (!th) return -2;
  WaitForSingleObject(th, INFINITE);
  CloseHandle(th);
  return 0;
}
#else
#include <pthread.h>

static void* worker(void* data) {
  struct task* t = (struct task*) data;
  t->cb(t->ctx, t->value);
  return NULL;
}

int spawn_async(void (*cb)(void* ctx, int), void* ctx, int value) {
  if (!cb || !ctx) return -1;
  const int n = 100;
  struct task tasks[100];
  pthread_t th[100];
  for (int i = 0; i < n; i++) {
    tasks[i].cb = cb;
    tasks[i].ctx = ctx;
    tasks[i].value = value;
    if (pthread_create(&th[i], NULL, worker, &tasks[i]) != 0) {
      for (int j = 0; j < i; j++) pthread_join(th[j], NULL);
      return -2;
    }
  }
  for (int i = 0; i < n; i++) pthread_join(th[i], NULL);
  return 0;
}
#endif
'

ffi_async <- tcc_ffi() |>
  tcc_source(code_async)
if (.Platform$OS.type != "windows") {
  ffi_async <- tcc_library(ffi_async, "pthread")
}
ffi_async <- ffi_async |>
  tcc_bind(
    spawn_async = list(
      args = list("callback_async:void(int)", "ptr", "i32"),
      returns = "i32"
    )
  ) |>
  tcc_compile()

rc <- ffi_async$spawn_async(cb_async, tcc_callback_ptr(cb_async), 2L)
hits
#> [1] 200
tcc_callback_close(cb_async)
```

Non-void return works the same way — the generated wrapper handles the
drain loop transparently:

``` r
cb_triple <- tcc_callback(
  function(x) x * 3L,
  signature = "int (*)(int)"
)

# Pure C: the worker calls the sync callback and returns its result.
# No drain logic needed — the generated wrapper handles it.
code_sync <- '
#ifdef _WIN32
#include <windows.h>

struct itask { int (*cb)(void*,int); void* ctx; int in; int out; };

static DWORD WINAPI iworker(LPVOID p) {
  struct itask* t = (struct itask*)p;
  t->out = t->cb(t->ctx, t->in);
  return 0;
}
int run_worker(int (*cb)(void*,int), void* ctx, int x) {
  struct itask t;
  t.cb = cb; t.ctx = ctx; t.in = x; t.out = -1;
  HANDLE th = CreateThread(NULL, 0, iworker, &t, 0, NULL);
  if (!th) return -1;
  WaitForSingleObject(th, INFINITE);
  CloseHandle(th);
  return t.out;
}
#else
#include <pthread.h>

struct itask { int (*cb)(void*,int); void* ctx; int in; int out; };

static void* iworker(void* p) {
  struct itask* t = (struct itask*)p;
  t->out = t->cb(t->ctx, t->in);
  return NULL;
}
int run_worker(int (*cb)(void*,int), void* ctx, int x) {
  struct itask t;
  t.cb = cb; t.ctx = ctx; t.in = x; t.out = -1;
  pthread_t th;
  if (pthread_create(&th, NULL, iworker, &t) != 0) return -1;
  pthread_join(th, NULL);
  return t.out;
}
#endif
'

ffi_sync <- tcc_ffi() |>
  tcc_source(code_sync)
if (.Platform$OS.type != "windows") {
  ffi_sync <- tcc_library(ffi_sync, "pthread")
}
ffi_sync <- ffi_sync |>
  tcc_bind(
    run_worker = list(args = list("callback_async:int(int)", "ptr", "i32"), returns = "i32")
  ) |>
  tcc_compile()

ffi_sync$run_worker(cb_triple, tcc_callback_ptr(cb_triple), 7L)  # 21
#> [1] 21
tcc_callback_close(cb_triple)
```

### SQLite: a complete example

This example ties together external library linking, callbacks, and
pointer dereferencing. We open an in-memory SQLite database, execute
queries, and collect rows through an R callback that reads `char**`
arrays using `tcc_read_ptr` and `tcc_read_cstring`.

``` r
ptr_size <- .Machine$sizeof.pointer

read_string_array <- function(ptr, n) {
  vapply(seq_len(n), function(i) {
    tcc_read_cstring(tcc_read_ptr(ptr, (i - 1L) * ptr_size))
  }, "")
}

cb <- tcc_callback(
  function(argc, argv, cols) {
    values <- read_string_array(argv, argc)
    names  <- read_string_array(cols, argc)
    cat(paste(names, values, sep = " = ", collapse = ", "), "\n")
    0L
  },
  signature = "int (*)(int, char **, char **)"
)

sqlite <- tcc_ffi() |>
  tcc_header("#include <sqlite3.h>") |>
  tcc_library("sqlite3") |>
  tcc_source('
    void* open_db() {
      sqlite3* db = NULL;
      sqlite3_open(":memory:", &db);
      return db;
    }
    int close_db(void* db) {
      return sqlite3_close((sqlite3*)db);
    }
  ') |>
  tcc_bind(
    open_db  = list(args = list(), returns = "ptr"),
    close_db = list(args = list("ptr"), returns = "i32"),
    sqlite3_libversion = list(args = list(), returns = "cstring"),
    sqlite3_exec = list(
      args = list("ptr", "cstring", "callback:int(int, char **, char **)", "ptr", "ptr"),
      returns = "i32"
    )
  ) |>
  tcc_compile()

sqlite$sqlite3_libversion()
#> [1] "3.45.1"

db <- sqlite$open_db()
sqlite$sqlite3_exec(db, "CREATE TABLE t (id INTEGER, name TEXT);", cb, tcc_callback_ptr(cb), tcc_null_ptr())
#> [1] 0
sqlite$sqlite3_exec(db, "INSERT INTO t VALUES (1, 'hello'), (2, 'world');", cb, tcc_callback_ptr(cb), tcc_null_ptr())
#> [1] 0
sqlite$sqlite3_exec(db, "SELECT * FROM t;", cb, tcc_callback_ptr(cb), tcc_null_ptr())
#> id = 1, name = hello 
#> id = 2, name = world
#> [1] 0
sqlite$close_db(db)
#> [1] 0
tcc_callback_close(cb)
```

### Stackless C Protothreads: streaming BCF/VCF records with htslib

Rtinycc bundles Adam Dunkels’
[Protothreads](http://dunkels.com/adam/pt/) library, allowing you to
JIT-compile cross-platform native iterators that yield back to R. This
demo binds [htslib](https://www.htslib.org/) using a protothread: R
resumes the C reader until the next `bcf1_t` is ready, then copies the
current fields into a regular R list. The C logic never blocks R’s event
loop, and R objects are created only after control returns safely to the
main R stack.

**Important Protothreads Limitations:** Because protothreads are
stackless (implemented using Duff’s device macro expansions), you must
observe a few strict rules in your C code:

1.  **Local variables are NOT preserved** across `PT_YIELD` calls. Store
    state in your context struct instead.
2.  **No nested blocking**: You cannot yield from inside a nested C
    function call.
3.  **No switch statements**: Because Protothreads uses `switch`
    internally, you cannot use `switch` within a `PT_THREAD` block.

The demo uses plain VCF text, opened directly by htslib through the same
API as BCF.

``` r
source("scripts/demo-streaming-bcf-reader-ffi.R")
run_streaming_bcf_demo()
#> Rtinycc version: 0.1.12
#> Demo: stackless Protothreads + htslib BCF/VCF API streaming reader
#> Note: htslib reads run using Protothreads. Warning: Local C variables are NOT preserved across yields!; R objects are built only after each yield.
#> 
#> Input: generated VCF text (opened directly by htslib)
#> Samples: sample1
#> 
#> == Streaming records one resume at a time ==
#> record 1: chr1:10 id=rs1 ref=A alt=C qual=50 alleles=[A,C]
#> record 2: chr1:20 id=. ref=G alt=A,T qual=99 alleles=[G,A,T]
#> record 3: chr1:30 id=rs3 ref=TT alt=T qual=. alleles=[TT,T]
#> done_after_collect=TRUE
```

The full demo source is foldable below.

<details>
<summary>
Click to show the complete
<code>scripts/demo-streaming-bcf-reader-ffi.R</code> script
</summary>

``` r
#!/usr/bin/env Rscript

suppressPackageStartupMessages(library(Rtinycc))


say <- function(...) cat(..., "\n", sep = "")

`%||%` <- function(x, y) if (is.null(x)) y else x

pkg_config_paths <- function(args, prefix) {
  out <- tryCatch(
    system2("pkg-config", c(args, "htslib"), stdout = TRUE, stderr = FALSE),
    error = function(e) character()
  )
  out <- unlist(strsplit(paste(out, collapse = " "), "[[:space:]]+"))
  out <- out[nzchar(out)]
  sub(paste0("^", prefix), "", out[startsWith(out, prefix)])
}

build_streaming_bcf_ffi <- function() {
  code <- '
#include <rtinycc/pt.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <htslib/hts.h>
#include <htslib/vcf.h>

enum {
  BCF_STREAM_ERR = -1,
  BCF_STREAM_EOF = 0,
  BCF_STREAM_RECORD = 1
};

typedef struct bcf_stream {
  struct pt pt;
  int yielded;
  int started;
  int done;
  int error;
  char err[256];
  htsFile *fp;
  bcf_hdr_t *hdr;
  bcf1_t *rec;
  char *alts;
  size_t alts_cap;
  int ret;
} bcf_stream_t;

static void bcf_stream_set_error(bcf_stream_t *st, const char *msg) {
  if (!st) return;
  if (!msg) msg = "unknown htslib error";
  snprintf(st->err, sizeof(st->err), "%s", msg);
  st->error = 1;
}

static int bcf_stream_resume_internal(bcf_stream_t *st) {
  PT_BEGIN(&st->pt);

  if (!st || !st->fp || !st->hdr || !st->rec) {
    if (st) bcf_stream_set_error(st, "stream is not open");
    if (st) st->done = 1;
    PT_EXIT(&st->pt);
  }

  for (;;) {
    st->ret = bcf_read(st->fp, st->hdr, st->rec);
    if (st->ret == 0) {
      if (bcf_unpack(st->rec, BCF_UN_STR) < 0) {
        bcf_stream_set_error(st, "bcf_unpack() failed");
        st->done = 1;
        st->yielded = BCF_STREAM_ERR;
        PT_EXIT(&st->pt);
      }
      st->yielded = BCF_STREAM_RECORD;
      PT_YIELD(&st->pt);
      if (st->error || st->done) PT_EXIT(&st->pt);
    } else {
      if (st->ret < -1) {
        bcf_stream_set_error(st, "bcf_read() failed");
      }
      st->done = 1;
      st->yielded = BCF_STREAM_EOF;
      PT_EXIT(&st->pt);
    }
  }

  PT_END(&st->pt);
}

void *bcf_stream_open(const char *path) {
  bcf_stream_t *st = (bcf_stream_t *) calloc(1, sizeof(bcf_stream_t));
  if (!st) return NULL;

  PT_INIT(&st->pt);

  if (!path || !path[0]) {
    bcf_stream_set_error(st, "path is empty");
    st->done = 1;
    return st;
  }

  st->fp = hts_open(path, "r");
  if (!st->fp) {
    bcf_stream_set_error(st, "hts_open() failed");
    st->done = 1;
    return st;
  }

  st->hdr = bcf_hdr_read(st->fp);
  if (!st->hdr) {
    bcf_stream_set_error(st, "bcf_hdr_read() failed");
    st->done = 1;
    return st;
  }

  st->rec = bcf_init();
  if (!st->rec) {
    bcf_stream_set_error(st, "bcf_init() failed");
    st->done = 1;
    return st;
  }

  return st;
}

int bcf_stream_resume(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;

  if (!st) return BCF_STREAM_ERR;
  if (st->error) return BCF_STREAM_ERR;
  if (st->done) return BCF_STREAM_EOF;

  st->started = 1;
  bcf_stream_resume_internal(st);

  if (st->error) return BCF_STREAM_ERR;
  if (st->done) return BCF_STREAM_EOF;
  return st->yielded;
}

int bcf_stream_done(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  return st ? st->done : 1;
}

const char *bcf_stream_error(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st) return "stream pointer is NULL";
  return st->err[0] ? st->err : NULL;
}

const char *bcf_stream_chrom(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->hdr || !st->rec) return NULL;
  return bcf_seqname(st->hdr, st->rec);
}

int64_t bcf_stream_pos1(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec) return -1;
  return (int64_t) st->rec->pos + 1;
}

int64_t bcf_stream_end1(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec) return -1;
  return (int64_t) st->rec->pos + (int64_t) st->rec->rlen;
}

const char *bcf_stream_id(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec || !st->rec->d.id || !st->rec->d.id[0]) return ".";
  return st->rec->d.id;
}

int bcf_stream_n_allele(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec) return 0;
  return st->rec->n_allele;
}

const char *bcf_stream_allele(void *ptr, int idx) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec || idx < 0 || idx >= st->rec->n_allele) return NULL;
  return st->rec->d.allele[idx];
}

const char *bcf_stream_ref(void *ptr) {
  return bcf_stream_allele(ptr, 0);
}

const char *bcf_stream_alt(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  size_t need = 2;
  int i;

  if (!st || !st->rec) return NULL;
  if (st->rec->n_allele <= 1) return ".";

  for (i = 1; i < st->rec->n_allele; ++i) {
    const char *a = st->rec->d.allele[i] ? st->rec->d.allele[i] : "";
    need += strlen(a) + 1;
  }

  if (need > st->alts_cap) {
    char *tmp = (char *) realloc(st->alts, need);
    if (!tmp) {
      bcf_stream_set_error(st, "failed to allocate ALT buffer");
      return NULL;
    }
    st->alts = tmp;
    st->alts_cap = need;
  }

  st->alts[0] = 0;
  for (i = 1; i < st->rec->n_allele; ++i) {
    if (i > 1) strcat(st->alts, ",");
    strcat(st->alts, st->rec->d.allele[i] ? st->rec->d.allele[i] : "");
  }

  return st->alts;
}

double bcf_stream_qual(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec || bcf_float_is_missing(st->rec->qual)) return NAN;
  return (double) st->rec->qual;
}

int bcf_stream_nsamples(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->hdr) return 0;
  return bcf_hdr_nsamples(st->hdr);
}

const char *bcf_stream_sample(void *ptr, int idx) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->hdr || idx < 0 || idx >= bcf_hdr_nsamples(st->hdr)) return NULL;
  return st->hdr->samples[idx];
}

void bcf_stream_close(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;

  if (!st) return;

  if (st->rec) bcf_destroy(st->rec);
  if (st->hdr) bcf_hdr_destroy(st->hdr);
  if (st->fp) hts_close(st->fp);
  free(st->alts);
  memset(st, 0, sizeof(bcf_stream_t));
  free(st);
}
'

  ffi <- tcc_ffi()
  ffi <- tcc_include(ffi, system.file("include", package = "Rtinycc"))

  for (path in pkg_config_paths("--cflags-only-I", "-I")) {
    ffi <- tcc_include(ffi, path)
  }
  for (path in pkg_config_paths("--libs-only-L", "-L")) {
    ffi <- tcc_library_path(ffi, path)
  }

  ffi |>
    tcc_source(code) |>
    tcc_library("hts") |>
    tcc_bind(
      bcf_stream_open = list(args = list("cstring"), returns = "ptr"),
      bcf_stream_resume = list(args = list("ptr"), returns = "i32"),
      bcf_stream_done = list(args = list("ptr"), returns = "bool"),
      bcf_stream_error = list(args = list("ptr"), returns = "cstring"),
      bcf_stream_chrom = list(args = list("ptr"), returns = "cstring"),
      bcf_stream_pos1 = list(args = list("ptr"), returns = "i64"),
      bcf_stream_end1 = list(args = list("ptr"), returns = "i64"),
      bcf_stream_id = list(args = list("ptr"), returns = "cstring"),
      bcf_stream_n_allele = list(args = list("ptr"), returns = "i32"),
      bcf_stream_allele = list(args = list("ptr", "i32"), returns = "cstring"),
      bcf_stream_ref = list(args = list("ptr"), returns = "cstring"),
      bcf_stream_alt = list(args = list("ptr"), returns = "cstring"),
      bcf_stream_qual = list(args = list("ptr"), returns = "f64"),
      bcf_stream_nsamples = list(args = list("ptr"), returns = "i32"),
      bcf_stream_sample = list(args = list("ptr", "i32"), returns = "cstring"),
      bcf_stream_close = list(args = list("ptr"), returns = "void")
    ) |>
    tcc_compile()
}

new_bcf_reader <- function(path, ffi = build_streaming_bcf_ffi()) {
  path <- normalizePath(path, mustWork = TRUE)
  ptr <- ffi$bcf_stream_open(path)
  if (tcc_ptr_is_null(ptr)) {
    stop("failed to allocate BCF stream", call. = FALSE)
  }

  err <- ffi$bcf_stream_error(ptr)
  if (!is.null(err)) {
    ffi$bcf_stream_close(ptr)
    stop(err, call. = FALSE)
  }

  reader <- new.env(parent = emptyenv())
  reader$ffi <- ffi
  reader$ptr <- ptr
  reader$closed <- FALSE
  class(reader) <- "bcf_stream_reader"

  reg.finalizer(reader, function(x) {
    if (!isTRUE(x$closed)) {
      x$ffi$bcf_stream_close(x$ptr)
      x$closed <- TRUE
    }
  }, onexit = TRUE)

  reader
}

close.bcf_stream_reader <- function(con, ...) {
  if (!inherits(con, "bcf_stream_reader")) {
    stop("expected a bcf_stream_reader", call. = FALSE)
  }
  if (!isTRUE(con$closed)) {
    con$ffi$bcf_stream_close(con$ptr)
    con$closed <- TRUE
  }
  invisible(NULL)
}

bcf_reader_samples <- function(reader) {
  stopifnot(inherits(reader, "bcf_stream_reader"), !isTRUE(reader$closed))
  n <- reader$ffi$bcf_stream_nsamples(reader$ptr)
  if (n <= 0L) return(character())
  vapply(seq_len(n) - 1L, function(i) reader$ffi$bcf_stream_sample(reader$ptr, i), "")
}

bcf_reader_next <- function(reader) {
  stopifnot(inherits(reader, "bcf_stream_reader"), !isTRUE(reader$closed))

  status <- reader$ffi$bcf_stream_resume(reader$ptr)
  if (identical(status, 0L)) {
    return(NULL)
  }
  if (identical(status, -1L)) {
    err <- reader$ffi$bcf_stream_error(reader$ptr) %||% "BCF stream failed"
    stop(err, call. = FALSE)
  }

  n_allele <- reader$ffi$bcf_stream_n_allele(reader$ptr)
  alleles <- if (n_allele > 0L) {
    vapply(seq_len(n_allele) - 1L, function(i) {
      reader$ffi$bcf_stream_allele(reader$ptr, i) %||% NA_character_
    }, "")
  } else {
    character()
  }

  list(
    chrom = reader$ffi$bcf_stream_chrom(reader$ptr),
    pos = reader$ffi$bcf_stream_pos1(reader$ptr),
    end = reader$ffi$bcf_stream_end1(reader$ptr),
    id = reader$ffi$bcf_stream_id(reader$ptr),
    ref = reader$ffi$bcf_stream_ref(reader$ptr),
    alt = reader$ffi$bcf_stream_alt(reader$ptr),
    qual = reader$ffi$bcf_stream_qual(reader$ptr),
    alleles = alleles
  )
}

bcf_reader_collect <- function(reader, n = Inf) {
  out <- list()
  i <- 0L
  repeat {
    if (i >= n) break
    rec <- bcf_reader_next(reader)
    if (is.null(rec)) break
    i <- i + 1L
    out[[i]] <- rec
  }
  out
}

write_demo_vcf <- function(path) {
  lines <- c(
    "##fileformat=VCFv4.3",
    "##contig=<ID=chr1,length=1000>",
    "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">",
    "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tsample1",
    "chr1\t10\trs1\tA\tC\t50\tPASS\t.\tGT\t0/1",
    "chr1\t20\t.\tG\tA,T\t99\tPASS\t.\tGT\t1/2",
    "chr1\t30\trs3\tTT\tT\t.\tPASS\t.\tGT\t0/1"
  )
  writeLines(lines, path)
  invisible(path)
}

make_demo_vcf <- function() {
  vcf <- tempfile(fileext = ".vcf")
  write_demo_vcf(vcf)
  vcf
}

run_streaming_bcf_demo <- function() {
  say("Rtinycc version: ", as.character(utils::packageVersion("Rtinycc")))
  say("Demo: stackless Protothreads + htslib BCF/VCF API streaming reader")
  say("Note: htslib reads run using Protothreads. Warning: Local C variables are NOT preserved across yields!; R objects are built only after each yield.")

  ffi <- build_streaming_bcf_ffi()
  path <- make_demo_vcf()

  say("")
  say("Input: generated VCF text (opened directly by htslib)")

  reader <- new_bcf_reader(path, ffi)
  on.exit(close(reader), add = TRUE)

  say("Samples: ", paste(bcf_reader_samples(reader), collapse = ", "))
  say("")
  say("== Streaming records one resume at a time ==")

  i <- 0L
  repeat {
    rec <- bcf_reader_next(reader)
    if (is.null(rec)) break
    i <- i + 1L
    say(
      "record ", i, ": ",
      rec$chrom, ":", rec$pos,
      " id=", rec$id,
      " ref=", rec$ref,
      " alt=", rec$alt,
      " qual=", if (is.nan(rec$qual)) "." else rec$qual,
      " alleles=[", paste(rec$alleles, collapse = ","), "]"
    )
  }

  say("done_after_collect=", isTRUE(ffi$bcf_stream_done(reader$ptr)))
  invisible(NULL)
}

if (identical(sys.nframe(), 0L)) {
  run_streaming_bcf_demo()
}
```

</details>

The embedded htslib Protothreads C source is extracted from the script
and rendered with the same Rtinycc C-code display helper used by the
vignettes.

<details>
<summary>
Click to show only the embedded C Protothreads / htslib source
</summary>

``` c
#include <rtinycc/pt.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <htslib/hts.h>
#include <htslib/vcf.h>

enum {
  BCF_STREAM_ERR = -1,
  BCF_STREAM_EOF = 0,
  BCF_STREAM_RECORD = 1
};

typedef struct bcf_stream {
  struct pt pt;
  int yielded;
  int started;
  int done;
  int error;
  char err[256];
  htsFile *fp;
  bcf_hdr_t *hdr;
  bcf1_t *rec;
  char *alts;
  size_t alts_cap;
  int ret;
} bcf_stream_t;

static void bcf_stream_set_error(bcf_stream_t *st, const char *msg) {
  if (!st) return;
  if (!msg) msg = "unknown htslib error";
  snprintf(st->err, sizeof(st->err), "%s", msg);
  st->error = 1;
}

static int bcf_stream_resume_internal(bcf_stream_t *st) {
  PT_BEGIN(&st->pt);

  if (!st || !st->fp || !st->hdr || !st->rec) {
    if (st) bcf_stream_set_error(st, "stream is not open");
    if (st) st->done = 1;
    PT_EXIT(&st->pt);
  }

  for (;;) {
    st->ret = bcf_read(st->fp, st->hdr, st->rec);
    if (st->ret == 0) {
      if (bcf_unpack(st->rec, BCF_UN_STR) < 0) {
        bcf_stream_set_error(st, "bcf_unpack() failed");
        st->done = 1;
        st->yielded = BCF_STREAM_ERR;
        PT_EXIT(&st->pt);
      }
      st->yielded = BCF_STREAM_RECORD;
      PT_YIELD(&st->pt);
      if (st->error || st->done) PT_EXIT(&st->pt);
    } else {
      if (st->ret < -1) {
        bcf_stream_set_error(st, "bcf_read() failed");
      }
      st->done = 1;
      st->yielded = BCF_STREAM_EOF;
      PT_EXIT(&st->pt);
    }
  }

  PT_END(&st->pt);
}

void *bcf_stream_open(const char *path) {
  bcf_stream_t *st = (bcf_stream_t *) calloc(1, sizeof(bcf_stream_t));
  if (!st) return NULL;

  PT_INIT(&st->pt);

  if (!path || !path[0]) {
    bcf_stream_set_error(st, "path is empty");
    st->done = 1;
    return st;
  }

  st->fp = hts_open(path, "r");
  if (!st->fp) {
    bcf_stream_set_error(st, "hts_open() failed");
    st->done = 1;
    return st;
  }

  st->hdr = bcf_hdr_read(st->fp);
  if (!st->hdr) {
    bcf_stream_set_error(st, "bcf_hdr_read() failed");
    st->done = 1;
    return st;
  }

  st->rec = bcf_init();
  if (!st->rec) {
    bcf_stream_set_error(st, "bcf_init() failed");
    st->done = 1;
    return st;
  }

  return st;
}

int bcf_stream_resume(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;

  if (!st) return BCF_STREAM_ERR;
  if (st->error) return BCF_STREAM_ERR;
  if (st->done) return BCF_STREAM_EOF;

  st->started = 1;
  bcf_stream_resume_internal(st);

  if (st->error) return BCF_STREAM_ERR;
  if (st->done) return BCF_STREAM_EOF;
  return st->yielded;
}

int bcf_stream_done(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  return st ? st->done : 1;
}

const char *bcf_stream_error(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st) return "stream pointer is NULL";
  return st->err[0] ? st->err : NULL;
}

const char *bcf_stream_chrom(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->hdr || !st->rec) return NULL;
  return bcf_seqname(st->hdr, st->rec);
}

int64_t bcf_stream_pos1(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec) return -1;
  return (int64_t) st->rec->pos + 1;
}

int64_t bcf_stream_end1(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec) return -1;
  return (int64_t) st->rec->pos + (int64_t) st->rec->rlen;
}

const char *bcf_stream_id(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec || !st->rec->d.id || !st->rec->d.id[0]) return ".";
  return st->rec->d.id;
}

int bcf_stream_n_allele(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec) return 0;
  return st->rec->n_allele;
}

const char *bcf_stream_allele(void *ptr, int idx) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec || idx < 0 || idx >= st->rec->n_allele) return NULL;
  return st->rec->d.allele[idx];
}

const char *bcf_stream_ref(void *ptr) {
  return bcf_stream_allele(ptr, 0);
}

const char *bcf_stream_alt(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  size_t need = 2;
  int i;

  if (!st || !st->rec) return NULL;
  if (st->rec->n_allele <= 1) return ".";

  for (i = 1; i < st->rec->n_allele; ++i) {
    const char *a = st->rec->d.allele[i] ? st->rec->d.allele[i] : "";
    need += strlen(a) + 1;
  }

  if (need > st->alts_cap) {
    char *tmp = (char *) realloc(st->alts, need);
    if (!tmp) {
      bcf_stream_set_error(st, "failed to allocate ALT buffer");
      return NULL;
    }
    st->alts = tmp;
    st->alts_cap = need;
  }

  st->alts[0] = 0;
  for (i = 1; i < st->rec->n_allele; ++i) {
    if (i > 1) strcat(st->alts, ",");
    strcat(st->alts, st->rec->d.allele[i] ? st->rec->d.allele[i] : "");
  }

  return st->alts;
}

double bcf_stream_qual(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->rec || bcf_float_is_missing(st->rec->qual)) return NAN;
  return (double) st->rec->qual;
}

int bcf_stream_nsamples(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->hdr) return 0;
  return bcf_hdr_nsamples(st->hdr);
}

const char *bcf_stream_sample(void *ptr, int idx) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;
  if (!st || !st->hdr || idx < 0 || idx >= bcf_hdr_nsamples(st->hdr)) return NULL;
  return st->hdr->samples[idx];
}

void bcf_stream_close(void *ptr) {
  bcf_stream_t *st = (bcf_stream_t *) ptr;

  if (!st) return;

  if (st->rec) bcf_destroy(st->rec);
  if (st->hdr) bcf_hdr_destroy(st->hdr);
  if (st->fp) hts_close(st->fp);
  free(st->alts);
  memset(st, 0, sizeof(bcf_stream_t));
  free(st);
}
```

</details>

This points toward a useful separate package idea: a small, generic
Rtinycc-based C coroutine/Protothreads layer for R. Such a package could
provide reusable pieces like `coro_new()`, `coro_resume()`,
finalizer-backed native handles, status/error conventions, and helpers
for writing streaming bindings around C libraries such as htslib,
SQLite, parsers, decompression libraries, or network clients. R would
see ordinary iterators; C would keep natural blocking or recursive
control flow without rewriting everything as a heap-allocated state
machine.

## Header parsing with treesitter.c

For header-driven bindings, we use `treesitter.c` to parse function
signatures and generate binding specifications automatically. For
struct, enum, and global helpers, `tcc_generate_bindings()` handles the
code generation.

The default mapper is conservative for pointers: `char*` is treated as
`ptr` because C does not guarantee NUL-terminated strings. If you know a
parameter is a C string, provide a custom mapper that returns `cstring`
for that type.

``` r
header <- '
double sqrt(double x);
double sin(double x);
struct point { double x; double y; };
enum status { OK = 0, ERROR = 1 };
int global_counter;
'

tcc_treesitter_functions(header)
#>   capture_name text start_line start_col params return_type
#> 1    decl_name sqrt          2         8 double      double
#> 2    decl_name  sin          3         8 double      double
tcc_treesitter_structs(header)
#>   capture_name  text start_line
#> 1  struct_name point          4
tcc_treesitter_enums(header)
#>   capture_name   text start_line
#> 1    enum_name status          5
tcc_treesitter_globals(header)
#>   capture_name           text start_line
#> 1  global_name global_counter          6

# Bind parsed functions to libm
symbols <- tcc_treesitter_bindings(header)
math <- tcc_link("m", symbols = symbols)
math$sqrt(16.0)
#> [1] 4

# Generate struct/enum/global helpers
ffi <- tcc_ffi() |>
  tcc_source(header) |>
  tcc_generate_bindings(
    header,
    functions = FALSE, structs = TRUE,
    enums = TRUE, globals = TRUE
  ) |>
  tcc_compile()

ffi$struct_point_new()
#> <pointer: 0x5dfe4ca39770>
ffi$enum_status_OK()
#> [1] 0
ffi$global_global_counter_get()
#> [1] 0
```

## io_uring Demo

`CSV` parser using [`io_uring`](https://en.wikipedia.org/wiki/Io_uring)
on linux

``` r
if (Sys.info()[["sysname"]] == "Linux") {
  c_file <- system.file("c_examples", "io_uring_csv.c", package = "Rtinycc")

  n_rows <- 20000L
  n_cols <- 8L
  block_size <- 1024L * 1024L

  set.seed(42)
  tmp_csv <- tempfile("rtinycc_io_uring_readme_", fileext = ".csv")
  on.exit(unlink(tmp_csv), add = TRUE)

  mat <- matrix(runif(n_rows * n_cols), ncol = n_cols)
  df <- as.data.frame(mat)
  names(df) <- paste0("V", seq_len(n_cols))
  utils::write.table(df, file = tmp_csv, sep = ",", row.names = FALSE, col.names = TRUE, quote = FALSE)
  csv_size_mb <- as.double(file.info(tmp_csv)$size) / 1024^2
  message(sprintf("CSV size: %.2f MB", csv_size_mb))

  io_uring_src <- paste(readLines(c_file, warn = FALSE), collapse = "\n")

  ffi <- tcc_ffi() |>
    tcc_source(io_uring_src) |>
    tcc_bind(
      csv_table_read = list(
        args = list("cstring", "i32", "i32"),
        returns = "sexp"
      ),
      csv_table_io_uring = list(
        args = list("cstring", "i32", "i32"),
        returns = "sexp"
      )
    ) |>
    tcc_compile()

  baseline <- utils::read.table(tmp_csv, sep = ",", header = TRUE)
  c_tbl <- ffi$csv_table_read(tmp_csv, block_size, n_cols)
  uring_tbl <- ffi$csv_table_io_uring(tmp_csv, block_size, n_cols)
  vroom_tbl <- vroom::vroom(
    tmp_csv,
    delim = ",",
    altrep = FALSE,
    col_types = vroom::cols(.default = "d"),
    progress = FALSE,
    show_col_types = FALSE
  )

  stopifnot(
    identical(dim(c_tbl), dim(baseline)),
    identical(dim(uring_tbl), dim(baseline)),
    identical(dim(vroom_tbl), dim(baseline)),
    isTRUE(all.equal(c_tbl, baseline, tolerance = 1e-8, check.attributes = FALSE)),
    isTRUE(all.equal(uring_tbl, baseline, tolerance = 1e-8, check.attributes = FALSE)),
    isTRUE(all.equal(vroom_tbl, baseline, tolerance = 1e-8, check.attributes = FALSE))
  )

  timings <- bench::mark(
    read_table_df = {
      x <- utils::read.table(tmp_csv, sep = ",", header = TRUE)
      nrow(x)
    },
    vroom_df_altrep_false = {
      x <- vroom::vroom(
        tmp_csv,
        delim = ",",
        altrep = FALSE,
        col_types = vroom::cols(.default = "d"),
        progress = FALSE,
        show_col_types = FALSE
      )
      nrow(x)
    },
    vroom_df_altrep_false_mat = {
      x <- vroom::vroom(
        tmp_csv,
        delim = ",",
        altrep = FALSE,
        col_types = vroom::cols(.default = "d"),
        progress = FALSE,
        show_col_types = FALSE
      )
      x <- as.matrix(x)
      nrow(x)
    },
    c_read_df = {
      x <- ffi$csv_table_read(tmp_csv, block_size, n_cols)
      nrow(x)
    },
    io_uring_df = {
      x <- ffi$csv_table_io_uring(tmp_csv, block_size, n_cols)
      nrow(x)
    },
    iterations = 2,
    memory = TRUE
  )

  
  print(timings)
  
  plot(timings, type = "boxplot") + bench::scale_x_bench_time(base = NULL)
}
#> CSV size: 2.75 MB
#> # A tibble: 5 × 13
#>   expression     min  median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time
#>   <bch:expr> <bch:t> <bch:t>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm>
#> 1 read_tabl… 47.22ms 49.14ms      20.3    6.33MB       0      2     0    98.28ms
#> 2 vroom_df_…  7.04ms  7.04ms     142.     1.22MB     142.     1     1     7.04ms
#> 3 vroom_df_…  8.95ms  9.29ms     108.     2.44MB       0      2     0    18.58ms
#> 4 c_read_df  21.26ms 21.95ms      45.6    1.22MB       0      2     0    43.89ms
#> 5 io_uring_… 21.14ms 21.21ms      47.1    1.22MB       0      2     0    42.43ms
#> # ℹ 4 more variables: result <list>, memory <list>, time <list>, gc <list>
```

<img src="man/figures/README-io_uring-demo-1.png" alt="" width="100%" />

## Known limitations

### `_Complex` types

TCC does not support C99 `_Complex` types. Generated code works around
this with `#define _Complex`, which suppresses the keyword. Apply the
same workaround in your own `tcc_source()` code when headers pull in
complex types.

### 64-bit integer precision

R represents `i64` and `u64` values as `double`, which loses precision
beyond $2^{53}$. Values that differ only past that threshold become
indistinguishable.

``` r
sprintf("2^53:     %.0f", 2^53)
#> [1] "2^53:     9007199254740992"
sprintf("2^53 + 1: %.0f", 2^53 + 1)
#> [1] "2^53 + 1: 9007199254740992"
identical(2^53, 2^53 + 1)
#> [1] TRUE
```

For exact 64-bit arithmetic, keep values in C-allocated storage and
manipulate them through pointers.

### Nested structs

Named nested struct fields can now be declared explicitly with
`struct:<name>`. Getters return borrowed nested views and setters copy
bytes from another struct object of the declared nested type.

``` r
ffi <- tcc_ffi() |>
  tcc_source('
    struct inner { int a; };
    struct outer { struct inner in; };
  ') |>
  tcc_struct("inner", accessors = c(a = "i32")) |>
  tcc_struct("outer", accessors = list(`in` = "struct:inner")) |>
  tcc_compile()

outer <- ffi$struct_outer_new()
inner <- ffi$struct_inner_new()
inner <- ffi$struct_inner_set_a(inner, 42L)
outer <- ffi$struct_outer_set_in(outer, inner)
inner_view <- ffi$struct_outer_get_in(outer)
ffi$struct_inner_get_a(inner_view)
#> [1] 42
ffi$struct_inner_free(inner)
#> NULL
ffi$struct_outer_free(outer)
#> NULL
```

For treesitter-generated bindings, nested struct fields inside structs
still fall back to ptr-like accessors. If you want a borrowed nested
view plus a copy-in setter, declare the nested field explicitly with
`struct:<name>`.

Bitfields are separate from ordinary addressable fields. They use scalar
helper accessors, but `field_addr()` and `container_of()` reject
bitfield members.

``` r
ffi <- tcc_ffi() |>
  tcc_source('struct flags { unsigned int flag : 1; };') |>
  tcc_struct(
    "flags",
    accessors = list(flag = list(type = "u8", bitfield = TRUE, width = 1))
  )

tcc_field_addr(ffi, "flags", "flag")
#> Error:
#> ! field_addr does not support bitfield members
tcc_container_of(ffi, "flags", "flag")
#> Error:
#> ! container_of does not support bitfield members
```

### Array fields in structs

Array fields require the `list(type = ..., size = N, array = TRUE)`
syntax in `tcc_struct()`, which generates element-wise accessors.

``` r
ffi <- tcc_ffi() |>
  tcc_source('struct buf { unsigned char data[16]; };') |>
  tcc_struct("buf", accessors = list(
    data = list(type = "u8", size = 16, array = TRUE)
  )) |>
  tcc_compile()

b <- ffi$struct_buf_new()
ffi$struct_buf_set_data_elt(b, 0L, 0xCAL)
#> <pointer: 0x5dfe51e41290>
ffi$struct_buf_set_data_elt(b, 1L, 0xFEL)
#> <pointer: 0x5dfe51e41290>
ffi$struct_buf_get_data_elt(b, 0L)
#> [1] 202
ffi$struct_buf_get_data_elt(b, 1L)
#> [1] 254
ffi$struct_buf_free(b)
#> NULL
```

## Serialization and fork safety

Compiled FFI objects are fork-safe: `parallel::mclapply()` and other
`fork()`-based parallelism work out of the box because TCC’s compiled
code lives in memory mappings that survive `fork()` via copy-on-write.

Serialization is also supported. Each `tcc_compiled` object stores its
FFI recipe internally, so after `saveRDS()` / `readRDS()` (or
`serialize()` / `unserialize()`), the first `$` access detects the dead
TCC state pointer and recompiles transparently.

``` r
ffi <- tcc_ffi() |>
  tcc_source("int square(int x) { return x * x; }") |>
  tcc_bind(square = list(args = list("i32"), returns = "i32")) |>
  tcc_compile()

ffi$square(7L)
#> [1] 49

tmp <- tempfile(fileext = ".rds")
saveRDS(ffi, tmp)
ffi2 <- readRDS(tmp)
unlink(tmp)

# Auto-recompiles on first access
ffi2$square(7L)
#> [Rtinycc] Recompiling FFI bindings after deserialization
#> [1] 49
```

For explicit control, use `tcc_recompile()`. Note that raw `tcc_state`
objects and bare pointers from `tcc_malloc()` do not carry a recipe and
remain dead after deserialization.

## Performance and benchmarking

`Rtinycc` is optimized for fast in-process compilation and convenient
FFI workflows, not for winning every microbenchmark against a
conventional precompiled `.Call()` shared library.

In practice, the usual pattern is:

- `Rtinycc` compiles tiny modules very quickly
- a regular `.Call()` module can have lower minimal per-call overhead
- array-oriented inputs can be zero-copy for already-materialized R
  vectors, but ALTREP inputs may materialize when C pointer access is
  requested
- return paths that copy native buffers back into fresh R vectors make
  that copy cost visible

If you want the benchmark details rather than the high-level summary,
see the `Compilation and Call Overhead` vignette.

## License

GPL-3

## References

- [TinyCC](https://github.com/TinyCC/tinycc)
- [Bun’s FFI](https://bun.com/docs/runtime/ffi)
- [CFFI](https://cffi.readthedocs.io/)
- [RSimpleFFI](https://github.com/sounkou-bioinfo/RSimpleFFI#readme)
- [CSlug](https://cslug.readthedocs.io/en/latest/)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_cstring <- Rtinycc::tcc_cstring
tcc_data_ptr <- Rtinycc::tcc_data_ptr
tcc_free <- Rtinycc::tcc_free
tcc_malloc <- Rtinycc::tcc_malloc
tcc_null_ptr <- Rtinycc::tcc_null_ptr
tcc_ptr_addr <- Rtinycc::tcc_ptr_addr
tcc_ptr_is_null <- Rtinycc::tcc_ptr_is_null
tcc_ptr_is_owned <- Rtinycc::tcc_ptr_is_owned
tcc_ptr_set <- Rtinycc::tcc_ptr_set
tcc_read_bytes <- Rtinycc::tcc_read_bytes
tcc_read_cstring <- Rtinycc::tcc_read_cstring
tcc_read_f64 <- Rtinycc::tcc_read_f64
tcc_read_i32 <- Rtinycc::tcc_read_i32
tcc_write_bytes <- Rtinycc::tcc_write_bytes
tcc_write_f64 <- Rtinycc::tcc_write_f64
tcc_write_i32 <- Rtinycc::tcc_write_i32

## -----------------------------------------------------------------------------
buf <- tcc_malloc(16)
str_ptr <- tcc_cstring("hello")

tcc_ptr_is_owned(buf)
tcc_ptr_is_owned(str_ptr)

## -----------------------------------------------------------------------------
tcc_free(buf)
tcc_free(str_ptr)

## -----------------------------------------------------------------------------
buf <- tcc_malloc(8)
tcc_write_bytes(buf, as.raw(1:8))
tcc_read_bytes(buf, 8)
tcc_free(buf)

## -----------------------------------------------------------------------------
buf <- tcc_malloc(16)
tcc_write_i32(buf, 0L, 42L)
tcc_write_f64(buf, 8L, 3.5)

tcc_read_i32(buf, offset = 0L)
tcc_read_f64(buf, offset = 8L)

tcc_free(buf)

## -----------------------------------------------------------------------------
ptr_size <- if (!is.null(.Machine$sizeof.pointer)) .Machine$sizeof.pointer else 8L
ptr_ref <- tcc_malloc(ptr_size)
target <- tcc_malloc(4)

tcc_ptr_set(ptr_ref, target)
tcc_ptr_addr(tcc_data_ptr(ptr_ref))
tcc_ptr_addr(target)

tcc_ptr_set(ptr_ref, tcc_null_ptr())
tcc_ptr_is_null(tcc_data_ptr(ptr_ref))

tcc_free(target)
tcc_free(ptr_ref)

## -----------------------------------------------------------------------------
str_ptr <- tcc_cstring("Hello, world")
tcc_read_cstring(str_ptr)
tcc_read_cstring(str_ptr, max_bytes = 5)
tcc_free(str_ptr)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_callback <- Rtinycc::tcc_callback
tcc_callback_async_drain <- Rtinycc::tcc_callback_async_drain
tcc_callback_close <- Rtinycc::tcc_callback_close
tcc_callback_ptr <- Rtinycc::tcc_callback_ptr
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_global <- Rtinycc::tcc_global
tcc_source <- Rtinycc::tcc_source
tcc_struct <- Rtinycc::tcc_struct

## -----------------------------------------------------------------------------
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>

    int* dup_array(int* x, int n) {
      if (n <= 0) return NULL;
      int* out = (int*)malloc(sizeof(int) * n);
      for (int i = 0; i < n; ++i) out[i] = x[i] * 2;
      return out;
    }
    "
  ) |>
  tcc_bind(
    dup_array = list(
      args = list("integer_array", "i32"),
      returns = list(type = "integer_array", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

ffi$dup_array(as.integer(c(1, 2, 3)), 3L)

## -----------------------------------------------------------------------------
cb <- tcc_callback(function(x) x * 3, "double (*)(double)")
cb_ptr <- tcc_callback_ptr(cb)

ffi_cb <- tcc_ffi() |>
  tcc_source(
    "
    double apply_cb(double (*cb)(void* ctx, double), void* ctx, double x) {
      return cb(ctx, x);
    }
    "
  ) |>
  tcc_bind(
    apply_cb = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi_cb$apply_cb(cb, cb_ptr, 5)
tcc_callback_close(cb)

## -----------------------------------------------------------------------------
ffi_global <- tcc_ffi() |>
  tcc_source(
    "
    int global_counter = 7;
    "
  ) |>
  tcc_global("global_counter", "i32") |>
  tcc_compile()

ffi_global$global_global_counter_get()
ffi_global$global_global_counter_set(9L)
ffi_global$global_global_counter_get()

## -----------------------------------------------------------------------------
ffi_struct <- tcc_ffi() |>
  tcc_source(
    "
    struct point {
      double x;
      double y;
    };
    "
  ) |>
  tcc_struct("point", accessors = c(x = "f64", y = "f64")) |>
  tcc_compile()

pt <- ffi_struct$struct_point_new()
pt <- ffi_struct$struct_point_set_x(pt, 1.5)
pt <- ffi_struct$struct_point_set_y(pt, 2.5)
ffi_struct$struct_point_get_x(pt)
ffi_struct$struct_point_free(pt)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_global <- Rtinycc::tcc_global
tcc_include_paths <- Rtinycc::tcc_include_paths
tcc_lib_paths <- Rtinycc::tcc_lib_paths
tcc_link <- Rtinycc::tcc_link
tcc_path <- Rtinycc::tcc_path
tcc_source <- Rtinycc::tcc_source

## -----------------------------------------------------------------------------
tcc_path()
tcc_include_paths()
tcc_lib_paths()

## -----------------------------------------------------------------------------
ffi <- tcc_ffi() |>
  tcc_source(
    "
    static double affine(double x, double slope, double intercept) {
      return slope * x + intercept;
    }
    "
  ) |>
  tcc_bind(
    affine = list(
      args = list("f64", "f64", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi$affine(2, 3, 4)

## -----------------------------------------------------------------------------
ffi_globals <- tcc_ffi() |>
  tcc_source(
    "
    int shared_counter = 1;

    void bump_counter(void) {
      shared_counter += 1;
    }
    "
  ) |>
  tcc_bind(
    bump_counter = list(args = list(), returns = "void")
  ) |>
  tcc_global("shared_counter", "i32") |>
  tcc_compile()

ffi_globals$global_shared_counter_get()
ffi_globals$bump_counter()
ffi_globals$global_shared_counter_get()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_generate_bindings <- Rtinycc::tcc_generate_bindings
tcc_map_c_type_to_ffi <- Rtinycc::tcc_map_c_type_to_ffi
tcc_source <- Rtinycc::tcc_source
tcc_struct <- Rtinycc::tcc_struct
tcc_treesitter_bindings <- Rtinycc::tcc_treesitter_bindings
tcc_treesitter_functions <- Rtinycc::tcc_treesitter_functions
tcc_treesitter_struct_accessors <- Rtinycc::tcc_treesitter_struct_accessors
has_treesitter <- requireNamespace("treesitter.c", quietly = TRUE) &&
  packageVersion("treesitter.c") >= "0.0.4"

## -----------------------------------------------------------------------------
has_treesitter

## -----------------------------------------------------------------------------
header <- paste(
  "double sqrt(double x);",
  "int add(int a, int b);",
  "struct point { double x; double y; };",
  "enum status { OK = 0, ERR = 1 };",
  sep = "\n"
)

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
tcc_treesitter_functions(header)

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
tcc_treesitter_bindings(header)

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
ffi <- tcc_ffi() |>
  tcc_source(
    "
    double sqrt(double x) { return x < 0 ? 0 : x; }
    int add(int a, int b) { return a + b; }
    struct point { double x; double y; };
    enum status { OK = 0, ERR = 1 };
    "
  )

ffi <- tcc_generate_bindings(
  ffi,
  header,
  functions = TRUE,
  structs = TRUE,
  enums = TRUE,
  unions = FALSE,
  globals = FALSE
)

compiled <- tcc_compile(ffi)

compiled$add(2L, 3L)
compiled$enum_status_OK()

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
tcc_treesitter_struct_accessors("struct point { double x; double y; };")

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
tcc_treesitter_struct_accessors(
  "struct flags { unsigned int flag : 1; unsigned int code : 6; };"
)

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
tcc_treesitter_struct_accessors(
  "struct child { int x; }; struct outer { struct child child; int y; };"
)

## -----------------------------------------------------------------------------
tcc_map_c_type_to_ffi("int")
tcc_map_c_type_to_ffi("double")
tcc_map_c_type_to_ffi("const char *")

## ----eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"----
string_mapper <- function(type) {
  if (trimws(type) == "const char *") {
    return("cstring")
  }
  tcc_map_c_type_to_ffi(type)
}

tcc_treesitter_bindings(
  "int puts(const char *s);",
  mapper = string_mapper
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_compile <- Rtinycc::tcc_compile
tcc_link <- Rtinycc::tcc_link
is_windows <- .Platform$OS.type == "windows"
find_system_math_library <- function() {
  if (is_windows) {
    return(NULL)
  }

  sysname <- as.character(unname(Sys.info()[["sysname"]]))
  paths <- unique(Rtinycc:::tcc_library_search_paths(sysname))
  # On Linux prefer the unversioned `libm.so`, which on glibc systems is a
  # GNU ld script that pulls in libm.so.6 plus libmvec.so.1; TinyCC handles
  # those scripts. Fall back to the SONAME-versioned file only if the dev
  # symlink is absent (some minimal containers).
  patterns <- switch(
    sysname,
    Linux = c("^libm\\.so$", "^libm\\.so(\\.[0-9]+)+$"),
    Darwin = c("^libSystem\\.B\\.dylib$", "^libm\\.dylib$"),
    character(0)
  )

  for (pattern in patterns) {
    for (path in paths) {
      if (!dir.exists(path)) {
        next
      }
      candidates <- list.files(
        path,
        pattern = pattern,
        full.names = TRUE
      )
      if (length(candidates) > 0L) {
        return(candidates[[1]])
      }
    }
  }

  NULL
}
math_lib_path <- find_system_math_library()
math_link_ok <- !is.null(math_lib_path) && tryCatch(
  {
    Rtinycc::tcc_link(
      math_lib_path,
      symbols = list(sqrt = list(args = list("f64"), returns = "f64"))
    )
    TRUE
  },
  error = function(e) FALSE
)

## ----eval = !is_windows && math_link_ok---------------------------------------
math <- tcc_link(
  math_lib_path,
  symbols = list(
    sqrt = list(args = list("f64"), returns = "f64"),
    cos = list(args = list("f64"), returns = "f64")
  )
)

math$sqrt(25)
math$cos(0)

## ----eval = !is_windows && math_link_ok---------------------------------------
math_helpers <- tcc_link(
  math_lib_path,
  symbols = list(
    sqrt = list(args = list("f64"), returns = "f64"),
    hypot2_checked = list(args = list("f64", "f64"), returns = "f64")
  ),
  user_code = "
    #include <math.h>

    double hypot2_checked(double x, double y) {
      if (x < 0 || y < 0) {
        return NAN;
      }
      return sqrt(x * x + y * y);
    }
  "
)

math_helpers$hypot2_checked(3, 4)
is.nan(math_helpers$hypot2_checked(-1, 4))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_callback_close <- Rtinycc::tcc_callback_close
tcc_compile <- Rtinycc::tcc_compile
tcc_cstring <- Rtinycc::tcc_cstring
tcc_data_ptr <- Rtinycc::tcc_data_ptr
tcc_ffi <- Rtinycc::tcc_ffi
tcc_get_symbol <- Rtinycc::tcc_get_symbol
tcc_link <- Rtinycc::tcc_link
tcc_malloc <- Rtinycc::tcc_malloc
tcc_relocate <- Rtinycc::tcc_relocate
tcc_set_options <- Rtinycc::tcc_set_options
tcc_source <- Rtinycc::tcc_source

## -----------------------------------------------------------------------------
ffi <- tcc_ffi() |>
  tcc_source("int add(int a, int b) { return a + b; }") |>
  tcc_bind(add = list(args = list("i32", "i32"), returns = "i32"))

names(ffi)

## -----------------------------------------------------------------------------
code <- Rtinycc:::generate_ffi_code(
  symbols = ffi$symbols,
  headers = ffi$headers,
  c_code = ffi$c_code,
  is_external = FALSE,
  structs = ffi$structs,
  unions = ffi$unions,
  enums = ffi$enums,
  globals = ffi$globals,
  container_of = ffi$container_of,
  field_addr = ffi$field_addr,
  struct_raw_access = ffi$struct_raw_access,
  introspect = ffi$introspect
)

grepl("SEXP R_wrap_add", code, fixed = TRUE)

## -----------------------------------------------------------------------------
Rtinycc:::generate_c_input("x", "arg1_", "i32")
Rtinycc:::generate_c_return("res", "f64")

# Simple working example of struct support in Rtinycc
# This demonstrates the API we implemented using base R pipe |>

library(Rtinycc)

# Create a simple struct example
ffi <- tcc_ffi() |>
  tcc_source(
    '
    struct point {
      double x;
      double y;
      int id;
    };
  '
  ) |>
  tcc_struct("point", accessors = c(x = "f64", y = "f64", id = "i32")) |>
  tcc_container_of("point", "x") |>
  tcc_field_addr("point", "x") |>
  tcc_introspect() |>
  tcc_source(
    '
    int test_point() { return 42; }
  '
  ) |>
  tcc_bind(test_point = list(args = list(), returns = "i32"))

cat("Compiling...\n")
compiled <- tcc_compile(ffi)

cat("Success! Available functions:\n")
cat("- test_point()\n")
cat("- point_new()\n")
cat("- point_free()\n")
cat("- point_get_x(), point_set_x()\n")
cat("- point_get_y(), point_set_y()\n")
cat("- point_get_id(), point_set_id()\n")
cat("- point_from_x() (container_of)\n")
cat("- point_sizeof(), point_alignof()\n")

# Test it
result <- compiled$test_point()
cat("\ntest_point() returned:", result, "\n")

# Create a point
p <- compiled$point_new()
p <- compiled$point_set_x(p, 3.14)
p <- compiled$point_set_y(p, 2.71)
p <- compiled$point_set_id(p, 42)

x <- compiled$point_get_x(p)
y <- compiled$point_get_y(p)
id <- compiled$point_get_id(p)

cat("\nPoint values: x =", x, ", y =", y, ", id =", id, "\n")

# Test container_of
x_ptr <- compiled$point_x_addr(p)
p2 <- compiled$point_from_x(x_ptr)
id2 <- compiled$point_get_id(p2)
cat("container_of test: recovered id =", id2, "\n")

# Test introspection
size <- compiled$point_sizeof()
cat("sizeof(struct point) =", size, "\n")

# Cleanup
compiled$point_free(p)

cat("\nAll tests passed!\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_free <- Rtinycc::tcc_free
tcc_malloc <- Rtinycc::tcc_malloc
tcc_source <- Rtinycc::tcc_source

## -----------------------------------------------------------------------------
ffi <- tcc_ffi() |>
  tcc_source(
    "
    int add_i32(int a, int b) { return a + b; }
    double mul_f64(double x, double y) { return x * y; }
    int negate_bool(_Bool x) { return !x; }
    "
  ) |>
  tcc_bind(
    add_i32 = list(args = list("i32", "i32"), returns = "i32"),
    mul_f64 = list(args = list("f64", "f64"), returns = "f64"),
    negate_bool = list(args = list("bool"), returns = "bool")
  ) |>
  tcc_compile()

ffi$add_i32(2L, 3L)
ffi$mul_f64(2, 4)
ffi$negate_bool(TRUE)

## -----------------------------------------------------------------------------
ffi_str <- tcc_ffi() |>
  tcc_source(
    "
    const char* echo_cstring(const char* s) { return s; }
    void* echo_ptr(void* p) { return p; }
    "
  ) |>
  tcc_bind(
    echo_cstring = list(args = list("cstring"), returns = "cstring"),
    echo_ptr = list(args = list("ptr"), returns = "ptr")
  ) |>
  tcc_compile()

ptr <- tcc_malloc(8)

ffi_str$echo_cstring("hello")
inherits(ffi_str$echo_ptr(ptr), "externalptr")

tcc_free(ptr)

## -----------------------------------------------------------------------------
ffi_arr <- tcc_ffi() |>
  tcc_source(
    "
    int first_int(int* x) { return x[0]; }
    double second_num(double* x) { return x[1]; }
    "
  ) |>
  tcc_bind(
    first_int = list(args = list("integer_array"), returns = "i32"),
    second_num = list(args = list("numeric_array"), returns = "f64")
  ) |>
  tcc_compile()

ffi_arr$first_int(as.integer(c(10, 20, 30)))
ffi_arr$second_num(c(1.5, 2.5))

## -----------------------------------------------------------------------------
ffi_sexp <- tcc_ffi() |>
  tcc_source(
    "
    #include <Rinternals.h>

    SEXP id_sexp(SEXP x) { return x; }
    "
  ) |>
  tcc_bind(
    id_sexp = list(args = list("sexp"), returns = "sexp")
  ) |>
  tcc_compile()

ffi_sexp$id_sexp(list(a = 1, b = 2))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
knitr::knit_engines$set(rtinycc = Rtinycc:::rtinycc_engine)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_source <- Rtinycc::tcc_source
has_callme <- requireNamespace("callme", quietly = TRUE)
has_bench <- requireNamespace("bench", quietly = TRUE)
has_profmem <- isTRUE(capabilities("profmem"))

## -----------------------------------------------------------------------------
build_rtinycc_module <- function() {
  tcc_ffi() |>
    tcc_source(rtinycc_code) |>
    tcc_bind(
      noop = list(args = list(), returns = "void"),
      fill_rand = list(args = list("numeric_array", "i32"), returns = "void"),
      rand_unif = list(
        args = list("i32"),
        returns = list(type = "numeric_array", length_arg = 1, free = TRUE)
      )
    ) |>
    tcc_compile()
}

build_callme_module <- function() {
  before <- names(getLoadedDLLs())
  mod <- callme::compile(callme_code, env = NULL, verbosity = 0)
  dlls <- getLoadedDLLs()
  new_names <- setdiff(names(dlls), before)
  new_names <- new_names[startsWith(new_names, "callme_")]
  attr(mod, "dll_paths") <- unname(vapply(
    dlls[new_names],
    function(x) x[["path"]],
    character(1)
  ))
  mod
}

unload_callme_dlls <- function(dll_paths) {
  dll_paths <- rev(unique(dll_paths))
  if (is.null(dll_paths) || !length(dll_paths)) {
    return(invisible(NULL))
  }
  for (dll_path in dll_paths) {
    if (is.character(dll_path) && nzchar(dll_path) && file.exists(dll_path)) {
      try(dyn.unload(dll_path), silent = TRUE)
    }
  }
  invisible(NULL)
}

build_and_dispose_callme_module <- function() {
  mod <- build_callme_module()
  dll_paths <- attr(mod, "dll_paths", exact = TRUE)
  rm(mod)
  gc()
  unload_callme_dlls(dll_paths)
  invisible(NULL)
}

callme_runtime_reason <- NULL
can_run_callme <- FALSE

if (!has_callme) {
  callme_runtime_reason <- "`callme` is not installed."
} else if (.Platform$OS.type == "windows") {
  callme_runtime_reason <- paste(
    "`callme` comparisons are skipped on Windows during vignette builds",
    "because the helper DLL compilation step is not reliable in CI."
  )
} else {
  callme_probe <- tryCatch(
    {
      build_and_dispose_callme_module()
      NULL
    },
    error = identity
  )

  if (inherits(callme_probe, "error")) {
    callme_runtime_reason <- paste(
      "`callme` comparisons were skipped because runtime compilation failed:",
      conditionMessage(callme_probe)
    )
  } else {
    can_run_callme <- TRUE
  }
}

can_run_benchmarks <- can_run_callme && has_bench && has_profmem

if (is.null(callme_runtime_reason) && !has_bench) {
  callme_runtime_reason <- "`bench` is not installed."
} else if (is.null(callme_runtime_reason) && !has_profmem) {
  callme_runtime_reason <- paste(
    "`bench` runtime comparisons are skipped because memory profiling",
    "is not available in this R build."
  )
} else if (is.null(callme_runtime_reason)) {
  callme_runtime_reason <- "Executable comparisons are enabled."
}

with_benchmark_modules <- function(fun) {
  rt_mod <- build_rtinycc_module()
  cm_mod <- build_callme_module()
  dll_paths <- attr(cm_mod, "dll_paths", exact = TRUE)

  on.exit({
    rm(rt_mod, cm_mod)
    gc()
    unload_callme_dlls(dll_paths)
  }, add = TRUE)

  fun(rt_mod, cm_mod)
}

median_elapsed <- function(expr, times = 3L) {
  expr <- substitute(expr)
  env <- parent.frame()
  stats::median(replicate(
    times,
    {
      gc()
      t0 <- proc.time()[["elapsed"]]
      eval(expr, envir = env)
      proc.time()[["elapsed"]] - t0
    }
  ))
}

run_noop <- function(fun, n) {
  for (i in seq_len(n)) {
    fun()
  }
  invisible(NULL)
}

run_rand <- function(fun, n, reps) {
  for (i in seq_len(reps)) {
    invisible(fun(n))
  }
  invisible(NULL)
}

run_fill <- function(fun, n, reps) {
  for (i in seq_len(reps)) {
    out <- numeric(n)
    invisible(fun(out, n))
  }
  invisible(NULL)
}

rtinycc_recipe <- tcc_ffi() |>
  tcc_source(rtinycc_code) |>
  tcc_bind(
    noop = list(args = list(), returns = "void"),
    fill_rand = list(args = list("numeric_array", "i32"), returns = "void"),
    rand_unif = list(
      args = list("i32"),
      returns = list(type = "numeric_array", length_arg = 1, free = TRUE)
    )
  )

generated_code <- Rtinycc:::generate_ffi_code(
  symbols = rtinycc_recipe$symbols,
  headers = rtinycc_recipe$headers,
  c_code = rtinycc_recipe$c_code,
  is_external = FALSE,
  structs = rtinycc_recipe$structs,
  unions = rtinycc_recipe$unions,
  enums = rtinycc_recipe$enums,
  globals = rtinycc_recipe$globals,
  container_of = rtinycc_recipe$container_of,
  field_addr = rtinycc_recipe$field_addr,
  struct_raw_access = rtinycc_recipe$struct_raw_access,
  introspect = rtinycc_recipe$introspect
)

## -----------------------------------------------------------------------------
has_callme

## -----------------------------------------------------------------------------
has_bench

## -----------------------------------------------------------------------------
has_profmem

## -----------------------------------------------------------------------------
can_run_callme

## -----------------------------------------------------------------------------
can_run_benchmarks

## -----------------------------------------------------------------------------
callme_runtime_reason

## ----eval = can_run_callme----------------------------------------------------
compile_times <- data.frame(
  implementation = c("Rtinycc", "callme"),
  seconds = c(
    median_elapsed(build_rtinycc_module(), times = 3L),
    median_elapsed(build_and_dispose_callme_module(), times = 3L)
  )
)

compile_times$milliseconds <- round(compile_times$seconds * 1000, 1)
compile_times

## ----results='asis'-----------------------------------------------------------
Rtinycc:::rtinycc_c_block(generated_code)

## ----eval = can_run_benchmarks------------------------------------------------
noop_bench <- with_benchmark_modules(function(rt_mod, cm_mod) {
  n_noop <- 1000L

  bench::mark(
    Rtinycc = run_noop(rt_mod$noop, n_noop),
    callme = run_noop(cm_mod$noop, n_noop),
    iterations = 20,
    check = TRUE,
    memory = has_profmem,
    filter_gc = FALSE
  )
})

noop_bench

## ----eval = can_run_benchmarks------------------------------------------------
fill_bench_n4096 <- with_benchmark_modules(function(rt_mod, cm_mod) {
  bench::mark(
    Rtinycc = run_fill(rt_mod$fill_rand, 4096L, 100L),
    callme = run_fill(cm_mod$fill_rand, 4096L, 100L),
    iterations = 20,
    check = FALSE,
    memory = has_profmem,
    filter_gc = FALSE
  )
})

fill_bench_n4096

## ----eval = can_run_benchmarks------------------------------------------------
rand_results <- with_benchmark_modules(function(rt_mod, cm_mod) {
  rand_bench_n1 <- bench::mark(
    Rtinycc = run_rand(rt_mod$rand_unif, 1L, 1000L),
    callme = run_rand(cm_mod$rand_unif, 1L, 1000L),
    iterations = 20,
    check = FALSE,
    memory = has_profmem,
    filter_gc = FALSE
  )

  rand_bench_n4096 <- bench::mark(
    Rtinycc = run_rand(rt_mod$rand_unif, 4096L, 100L),
    callme = run_rand(cm_mod$rand_unif, 4096L, 100L),
    iterations = 20,
    check = FALSE,
    memory = has_profmem,
    filter_gc = FALSE
  )

  list(rand_bench_n1 = rand_bench_n1, rand_bench_n4096 = rand_bench_n4096)
})

rand_results$rand_bench_n1
rand_results$rand_bench_n4096

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_callback <- Rtinycc::tcc_callback
tcc_callback_async_drain <- Rtinycc::tcc_callback_async_drain
tcc_callback_close <- Rtinycc::tcc_callback_close
tcc_callback_ptr <- Rtinycc::tcc_callback_ptr
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_source <- Rtinycc::tcc_source
tcc_struct <- Rtinycc::tcc_struct

## -----------------------------------------------------------------------------
ffi_struct <- tcc_ffi() |>
  tcc_source(
    "
    struct point {
      double x;
      double y;
    };

    double point_norm2(struct point* p) {
      return p->x * p->x + p->y * p->y;
    }
    "
  ) |>
  tcc_struct("point", accessors = c(x = "f64", y = "f64")) |>
  tcc_bind(
    point_norm2 = list(args = list("ptr"), returns = "f64")
  ) |>
  tcc_compile()

pt <- ffi_struct$struct_point_new()
pt <- ffi_struct$struct_point_set_x(pt, 3)
pt <- ffi_struct$struct_point_set_y(pt, 4)

ffi_struct$point_norm2(pt)
ffi_struct$struct_point_free(pt)

## -----------------------------------------------------------------------------
cb <- tcc_callback(
  function(x) x * 2,
  signature = "double (*)(double)"
)
cb_ptr <- tcc_callback_ptr(cb)

ffi_cb <- tcc_ffi() |>
  tcc_source(
    "
    double apply_cb(double (*cb)(void* ctx, double), void* ctx, double x) {
      return cb(ctx, x);
    }
    "
  ) |>
  tcc_bind(
    apply_cb = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi_cb$apply_cb(cb, cb_ptr, 5)
tcc_callback_close(cb)

# TinyCC SIMD detection and selected opcode use via `.byte`.
#
# TinyCC does not parse every SIMD mnemonic. For small fixed kernels, emit the
# instruction bytes directly in inline assembly. See simd_bytecode.c for the
# byte-to-mnemonic comments and ABI caveats.

library(Rtinycc)

source_path <- system.file("c_examples", "simd_bytecode.c", package = "Rtinycc")
if (!nzchar(source_path)) {
  source_path <- file.path("inst", "c_examples", "simd_bytecode.c")
}
code <- paste(readLines(source_path, warn = FALSE), collapse = "\n")

state <- tcc_state()
stopifnot(tcc_compile_string(state, code) == 0L)
stopifnot(tcc_relocate(state) == 0L)

cat("SSE2:        ", tcc_call_symbol(state, "rtcc_cpu_has_sse2", return = "int"), "\n", sep = "")
cat("AVX2:        ", tcc_call_symbol(state, "rtcc_cpu_has_avx2", return = "int"), "\n", sep = "")
cat("AVX-512F/BW/VL:", tcc_call_symbol(state, "rtcc_cpu_has_avx512f_bw_vl", return = "int"), "\n", sep = "")

sse2 <- tcc_call_symbol(
  state,
  "rtcc_sse2_add4_i32",
  a = as.integer(1:4),
  b = as.integer(c(10, 20, 30, 40)),
  out = integer(4)
)
print(sse2$out)

if (identical(tcc_call_symbol(state, "rtcc_cpu_has_avx2", return = "int"), 1L)) {
  avx2 <- tcc_call_symbol(
    state,
    "rtcc_avx2_add8_i32",
    a = as.integer(1:8),
    b = as.integer(seq(10, 80, by = 10)),
    out = integer(8)
  )
  print(avx2$out)
} else {
  message("Skipping AVX2 bytecode call because this CPU/runtime does not report AVX2.")
}

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

# FFI Class and High-Level API
# Bun-style FFI with API mode compilation

# FFI object class
new_rtinycc_symbol_return_spec <- function(
  type,
  type_info,
  length_arg = NULL,
  free = NULL
) {
  structure(
    list(
      type = type,
      type_info = type_info,
      length_arg = length_arg,
      free = free
    ),
    class = c("rtinycc_symbol_return_spec", "rtinycc_symbol_spec_component")
  )
}

new_rtinycc_bound_symbol <- function(
  name,
  args,
  arg_type_info,
  returns,
  return_spec,
  variadic = FALSE,
  varargs = NULL,
  varargs_types = NULL,
  varargs_type_info = NULL,
  varargs_min = NULL,
  varargs_max = NULL,
  varargs_mode = NULL,
  helper_kind = NULL,
  helper_operation = NULL
) {
  classes <- c("rtinycc_bound_symbol", "rtinycc_symbol_spec")
  if (!is.null(helper_kind)) {
    classes <- c(
      paste0("rtinycc_helper_symbol_", helper_kind),
      "rtinycc_helper_symbol",
      classes
    )
  }

  structure(
    list(
      name = name,
      args = args,
      arg_type_info = arg_type_info,
      returns = returns,
      return_spec = return_spec,
      variadic = variadic,
      varargs = varargs,
      varargs_types = varargs_types,
      varargs_type_info = varargs_type_info,
      varargs_min = varargs_min,
      varargs_max = varargs_max,
      varargs_mode = varargs_mode,
      helper_kind = helper_kind,
      helper_operation = helper_operation
    ),
    class = classes
  )
}

helper_symbol_kind <- function(x) {
  if (!inherits(x, "rtinycc_helper_symbol")) {
    stop("Expected rtinycc_helper_symbol object", call. = FALSE)
  }
  x$helper_kind
}

helper_symbol_operation <- function(x) {
  if (!inherits(x, "rtinycc_helper_symbol")) {
    stop("Expected rtinycc_helper_symbol object", call. = FALSE)
  }
  x$helper_operation
}

rtinycc_struct_field_spec <- function(ffi, struct_name, field_name) {
  if (is.null(ffi$structs) || is.null(ffi$structs[[struct_name]])) {
    return(NULL)
  }
  ffi$structs[[struct_name]][[field_name]] %||% NULL
}

rtinycc_is_bitfield_spec <- function(field_spec) {
  is.list(field_spec) && isTRUE(field_spec$bitfield)
}

rtinycc_nested_struct_type_name <- function(field_spec) {
  type_name <- if (is.list(field_spec)) {
    field_spec$type %||% NULL
  } else if (is.character(field_spec) && length(field_spec) == 1) {
    field_spec
  } else {
    NULL
  }
  if (
    is.null(type_name) || !is.character(type_name) || length(type_name) != 1
  ) {
    return(NULL)
  }
  if (!startsWith(type_name, "struct:")) {
    return(NULL)
  }
  sub("^struct:", "", type_name)
}

is_rtinycc_bound_symbol <- function(x) {
  inherits(x, "rtinycc_bound_symbol")
}

as_rtinycc_helper_symbol <- function(
  sym_name,
  sym,
  helper_kind,
  helper_operation
) {
  out <- as_rtinycc_bound_symbol(sym_name, sym)
  new_rtinycc_bound_symbol(
    name = out$name,
    args = out$args,
    arg_type_info = out$arg_type_info,
    returns = out$returns,
    return_spec = out$return_spec,
    variadic = out$variadic,
    varargs = out$varargs,
    varargs_types = out$varargs_types,
    varargs_type_info = out$varargs_type_info,
    varargs_min = out$varargs_min,
    varargs_max = out$varargs_max,
    varargs_mode = out$varargs_mode,
    helper_kind = helper_kind,
    helper_operation = helper_operation
  )
}

as_rtinycc_bound_symbol <- function(sym_name, sym) {
  if (is_rtinycc_bound_symbol(sym)) {
    return(sym)
  }

  if (is.null(sym$args)) {
    sym$args <- list()
  }

  variadic <- isTRUE(sym$variadic)
  has_varargs <- !is.null(sym$varargs)
  has_varargs_types <- !is.null(sym$varargs_types)

  if (
    !is.null(sym$variadic) &&
      (!is.logical(sym$variadic) || length(sym$variadic) != 1)
  ) {
    stop(
      "Symbol '",
      sym_name,
      "' variadic must be TRUE/FALSE",
      call. = FALSE
    )
  }

  if (variadic) {
    if (length(sym$args) == 0) {
      stop(
        "Symbol '",
        sym_name,
        "' variadic functions need at least one fixed argument",
        call. = FALSE
      )
    }

    if (has_varargs && has_varargs_types) {
      stop(
        "Symbol '",
        sym_name,
        "' cannot set both 'varargs' and 'varargs_types'",
        call. = FALSE
      )
    }

    if (!has_varargs && !has_varargs_types) {
      stop(
        "Symbol '",
        sym_name,
        "' variadic functions require 'varargs' or 'varargs_types'",
        call. = FALSE
      )
    }
  } else if (
    has_varargs ||
      has_varargs_types ||
      !is.null(sym$varargs_min) ||
      !is.null(sym$varargs_max)
  ) {
    stop(
      "Symbol '",
      sym_name,
      "' has variadic fields but variadic is not TRUE",
      call. = FALSE
    )
  }

  arg_type_info <- lapply(seq_along(sym$args), function(i) {
    check_ffi_type(
      sym$args[[i]],
      paste0("symbol '", sym_name, "' argument ", i)
    )
  })

  for (i in seq_along(arg_type_info)) {
    type_info <- arg_type_info[[i]]
    if (is_callback_async_type(type_info)) {
      sig <- parse_callback_type(type_info)
      uses_sexp <- identical(trimws(sig$return_type), "SEXP") ||
        identical(trimws(sig$return_type), "sexp") ||
        any(trimws(sig$arg_types) %in% c("SEXP", "sexp"))
      if (uses_sexp) {
        stop(
          "Symbol '",
          sym_name,
          "' async callback argument ",
          i,
          " cannot use SEXP arguments or return type",
          call. = FALSE
        )
      }
    }
  }

  if (!"returns" %in% names(sym)) {
    stop(
      "Symbol '",
      sym_name,
      "' missing 'returns' specification",
      call. = FALSE
    )
  }

  if (is.list(sym$returns)) {
    if (is.null(sym$returns$type)) {
      stop(
        "Symbol '",
        sym_name,
        "' return must include 'type'",
        call. = FALSE
      )
    }
    ret_type <- sym$returns$type
    ret_info <- check_ffi_type(
      ret_type,
      paste0("symbol '", sym_name, "' return")
    )
    if (!is.null(ret_info$kind) && ret_info$kind == "array") {
      if (is.null(sym$returns$length_arg)) {
        stop(
          "Symbol '",
          sym_name,
          "' array return requires 'length_arg'",
          call. = FALSE
        )
      }
      if (
        !is.numeric(sym$returns$length_arg) ||
          length(sym$returns$length_arg) != 1 ||
          is.na(sym$returns$length_arg) ||
          sym$returns$length_arg != as.integer(sym$returns$length_arg)
      ) {
        stop(
          "Symbol '",
          sym_name,
          "' array return length_arg must be a single integer",
          call. = FALSE
        )
      }
      if (!is.null(sym$args) && length(sym$args) > 0) {
        if (
          sym$returns$length_arg < 1 ||
            sym$returns$length_arg > length(sym$args)
        ) {
          stop(
            "Symbol '",
            sym_name,
            "' array return length_arg out of range",
            call. = FALSE
          )
        }
      }
    } else if (!is.null(sym$returns$length_arg) || !is.null(sym$returns$free)) {
      stop(
        "Symbol '",
        sym_name,
        "' non-array return cannot set length_arg/free",
        call. = FALSE
      )
    }
    return_spec <- new_rtinycc_symbol_return_spec(
      type = ret_type,
      type_info = ret_info,
      length_arg = sym$returns$length_arg,
      free = isTRUE(sym$returns$free)
    )
  } else {
    ret_info <- check_ffi_type(
      sym$returns,
      paste0("symbol '", sym_name, "' return")
    )
    return_spec <- new_rtinycc_symbol_return_spec(
      type = sym$returns,
      type_info = ret_info
    )
  }

  varargs_type_info <- NULL
  if (variadic) {
    if (has_varargs_types) {
      if (length(sym$varargs_types) == 0) {
        stop(
          "Symbol '",
          sym_name,
          "' varargs_types must be non-empty",
          call. = FALSE
        )
      }

      varargs_type_info <- lapply(seq_along(sym$varargs_types), function(i) {
        vtype <- sym$varargs_types[[i]]
        vinfo <- check_ffi_type(
          vtype,
          paste0("symbol '", sym_name, "' varargs_types ", i)
        )
        if (!is.null(vinfo$kind) && vinfo$kind != "scalar") {
          stop(
            "Symbol '",
            sym_name,
            "' varargs_types ",
            i,
            " must be a scalar FFI type",
            call. = FALSE
          )
        }
        if (
          ffi_type_family(vinfo) %in%
            c("callback", "callback_async") ||
            identical(vtype, "sexp")
        ) {
          stop(
            "Symbol '",
            sym_name,
            "' varargs_types ",
            i,
            " cannot be callback/sexp",
            call. = FALSE
          )
        }
        vinfo
      })

      if (is.null(sym$varargs_min)) {
        sym$varargs_min <- 0L
      }
      if (is.null(sym$varargs_max)) {
        sym$varargs_max <- sym$varargs_min
      }
      if (
        !is.numeric(sym$varargs_min) ||
          length(sym$varargs_min) != 1 ||
          is.na(sym$varargs_min) ||
          sym$varargs_min < 0 ||
          sym$varargs_min != as.integer(sym$varargs_min)
      ) {
        stop(
          "Symbol '",
          sym_name,
          "' varargs_min must be a non-negative integer",
          call. = FALSE
        )
      }
      if (
        !is.numeric(sym$varargs_max) ||
          length(sym$varargs_max) != 1 ||
          is.na(sym$varargs_max) ||
          sym$varargs_max < 0 ||
          sym$varargs_max != as.integer(sym$varargs_max)
      ) {
        stop(
          "Symbol '",
          sym_name,
          "' varargs_max must be a non-negative integer",
          call. = FALSE
        )
      }
      sym$varargs_min <- as.integer(sym$varargs_min)
      sym$varargs_max <- as.integer(sym$varargs_max)
      if (sym$varargs_min > sym$varargs_max) {
        stop(
          "Symbol '",
          sym_name,
          "' varargs_min must be <= varargs_max",
          call. = FALSE
        )
      }
      sym$varargs_mode <- "types"
      sym$varargs <- NULL
    } else {
      if (length(sym$varargs) == 0) {
        stop(
          "Symbol '",
          sym_name,
          "' variadic functions require non-empty 'varargs' type list",
          call. = FALSE
        )
      }
      max_varargs <- length(sym$varargs)
      if (is.null(sym$varargs_min)) {
        sym$varargs_min <- max_varargs
      }
      if (
        !is.numeric(sym$varargs_min) ||
          length(sym$varargs_min) != 1 ||
          is.na(sym$varargs_min) ||
          sym$varargs_min < 0 ||
          sym$varargs_min > max_varargs ||
          sym$varargs_min != as.integer(sym$varargs_min)
      ) {
        stop(
          "Symbol '",
          sym_name,
          "' varargs_min must be a single integer between 0 and length(varargs)",
          call. = FALSE
        )
      }
      sym$varargs_min <- as.integer(sym$varargs_min)
      sym$varargs_max <- max_varargs
      varargs_type_info <- lapply(seq_along(sym$varargs), function(i) {
        vtype <- sym$varargs[[i]]
        vinfo <- check_ffi_type(
          vtype,
          paste0("symbol '", sym_name, "' vararg ", i)
        )
        if (!is.null(vinfo$kind) && vinfo$kind != "scalar") {
          stop(
            "Symbol '",
            sym_name,
            "' vararg ",
            i,
            " must be a scalar FFI type",
            call. = FALSE
          )
        }
        if (
          ffi_type_family(vinfo) %in%
            c("callback", "callback_async") ||
            identical(vtype, "sexp")
        ) {
          stop(
            "Symbol '",
            sym_name,
            "' vararg ",
            i,
            " cannot be callback/sexp",
            call. = FALSE
          )
        }
        vinfo
      })
      sym$varargs_mode <- "prefix"
      sym$varargs_types <- NULL
    }
  } else {
    sym$varargs_min <- NULL
    sym$varargs_max <- NULL
    sym$varargs_types <- NULL
    sym$varargs_mode <- NULL
  }

  new_rtinycc_bound_symbol(
    name = sym_name,
    args = sym$args,
    arg_type_info = arg_type_info,
    returns = sym$returns,
    return_spec = return_spec,
    variadic = variadic,
    varargs = sym$varargs,
    varargs_types = sym$varargs_types,
    varargs_type_info = varargs_type_info,
    varargs_min = sym$varargs_min,
    varargs_max = sym$varargs_max,
    varargs_mode = sym$varargs_mode
  )
}

tcc_ffi_object <- function() {
  structure(
    list(
      state = NULL,
      symbols = list(),
      headers = character(0),
      c_code = character(0),
      options = character(0),
      libraries = character(0),
      lib_paths = character(0),
      include_paths = character(0),
      output = "memory",
      compiled = FALSE,
      wrapper_symbols = character(0),
      globals = list()
    ),
    class = "tcc_ffi"
  )
}

#' Create a new FFI compilation context
#'
#' Initialize a Bun-style FFI context for API-mode compilation.
#' This is the entry point for the modern FFI API.
#'
#' @return A tcc_ffi object with chaining support
#' @export
#' @examples
#' ffi <- tcc_ffi()
tcc_ffi <- function() {
  tcc_ffi_object()
}

#' Set output type for FFI compilation
#'
#' @param ffi A tcc_ffi object
#' @param output One of "memory", "dll", "exe"
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_output <- function(ffi, output = c("memory", "dll", "exe")) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  output <- match.arg(output)
  ffi$output <- output
  ffi
}

#' Add include path to FFI context
#'
#' @param ffi A tcc_ffi object
#' @param path Include directory path
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_include <- function(ffi, path) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  ffi$include_paths <- c(ffi$include_paths, path)
  ffi
}

#' Add library path to FFI context
#'
#' @param ffi A tcc_ffi object
#' @param path Library directory path
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_library_path <- function(ffi, path) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  ffi$lib_paths <- c(ffi$lib_paths, path)
  ffi
}

#' Add TinyCC compiler options to FFI context
#'
#' Append raw options that are passed to `tcc_set_options()` before compiling
#' generated wrappers (for example `"-O2"` or `"-Wall"`).
#'
#' @param ffi A tcc_ffi object
#' @param options Character vector of option fragments
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_options <- function(ffi, options) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  if (missing(options)) {
    stop("`options` must be provided", call. = FALSE)
  }
  if (is.null(options)) {
    ffi$options <- character(0)
    return(ffi)
  }

  opts <- as.character(options)
  if (!length(opts)) {
    stop("`options` must not be empty", call. = FALSE)
  }
  opts <- trimws(opts)
  if (any(!nzchar(opts))) {
    stop("`options` entries must be non-empty strings", call. = FALSE)
  }

  ffi$options <- c(ffi$options, opts)
  ffi
}

rtinycc_exact_library_name <- function(path) {
  paste0(":", basename(path))
}

rtinycc_add_library_file <- function(ffi, path) {
  ffi$lib_paths <- c(ffi$lib_paths, dirname(path))
  ffi$libraries <- c(ffi$libraries, rtinycc_exact_library_name(path))
  ffi
}

#' Add library to link against
#'
#' @param ffi A tcc_ffi object
#' @param library Library name (e.g., "m", "sqlite3") or a path to a
#'   shared library (e.g., "libm.so.6"). When a path or platform library file
#'   name is provided, the library directory is added automatically and TinyCC
#'   is asked to link that exact file name. This keeps versioned runtime
#'   libraries such as `libm.so.6` distinct from generic linker names such as
#'   `m`/`libm.so`.
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_library <- function(ffi, library) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  libs <- as.character(library)
  for (lib in libs) {
    if (!nzchar(lib)) {
      next
    }

    lib_name <- lib

    # If a path or a file name with extension is provided, resolve it and link
    # the exact file name via TinyCC's -l:<filename> path.  Do not collapse a
    # versioned runtime file such as libm.so.6 to -lm: some systems, including
    # CRAN's Fedora flavor, do not install the unversioned development symlink.
    if (file.exists(lib) || grepl("[/\\\\]", lib)) {
      if (!file.exists(lib)) {
        found_path <- tcc_find_library(lib)
        if (is.null(found_path)) {
          stop("Library not found: ", lib, call. = FALSE)
        }
        lib <- found_path
      }
      ffi <- rtinycc_add_library_file(ffi, lib)
      next
    } else if (grepl("\\.(so|dylib|dll)(\\..*)?$", lib, ignore.case = TRUE)) {
      found_path <- tcc_find_library(lib)
      if (is.null(found_path)) {
        stop("Library not found: ", lib, call. = FALSE)
      }
      ffi <- rtinycc_add_library_file(ffi, found_path)
      next
    }

    if (nzchar(lib_name)) {
      ffi$libraries <- c(ffi$libraries, lib_name)
    }
  }

  ffi
}

#' Declare a global variable getter
#'
#' Register a global C symbol so the compiled object exposes getter/setter
#' functions `global_<name>_get()` and `global_<name>_set()`.
#'
#' @details
#' Globals are limited to scalar FFI types. Array types are rejected.
#' Scalar conversions follow the same rules as wrapper arguments:
#'
#' - Integer inputs (`i8`, `i16`, `i32`, `u8`, `u16`) must be finite and
#'   within range; `NA` values error.
#' - Large integer types (`i64`, `u32`, `u64`) are mediated through R numeric
#'   (double). Values must be integer-valued and within range; for `i64`/`u64`
#'   only exact integers up to $2^53$ are accepted.
#' - Getter wrappers for `i64`/`u64` warn when the stored value exceeds R's
#'   exact integer range for numeric vectors.
#' - `bool` rejects `NA` logicals.
#'
#' Ownership notes:
#'
#' - `ptr` globals store the raw address from an external pointer. If the
#'   external pointer owns memory, keep it alive; otherwise the pointer may
#'   be freed while the global still points to it.
#' - `cstring` globals store a borrowed pointer to R's string data
#'   (UTF-8 translation). Do not free it; for C-owned strings prefer a `ptr`
#'   global and manage lifetime explicitly (e.g., with `tcc_cstring()`).
#'
#' @note
#' Global helpers are generated inside the compiled TCC unit. Recompiling
#' creates a new instance of the global variable; existing compiled objects
#' continue to refer to their own copy.
#'
#' @param ffi A tcc_ffi object
#' @param name Global symbol name
#' @param type FFI type for the global (scalar types only)
#' @return Updated tcc_ffi object (for chaining)
#' @export
#'
#' @examples
#' ffi <- tcc_ffi() |>
#'   tcc_source("int global_counter = 7;") |>
#'   tcc_global("global_counter", "i32") |>
#'   tcc_compile()
#' ffi$global_global_counter_get()
tcc_global <- function(ffi, name, type) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  if (!is.character(name) || length(name) != 1 || !nzchar(name)) {
    stop("Global name must be a non-empty string", call. = FALSE)
  }
  if (!is.character(type) || length(type) != 1 || !nzchar(type)) {
    stop("Global type must be a non-empty string", call. = FALSE)
  }

  type_info <- check_ffi_type(type, "global")
  if (!is.null(type_info$kind) && type_info$kind == "array") {
    stop("Global type cannot be an array type", call. = FALSE)
  }
  if (type == "void") {
    stop("Global type cannot be void", call. = FALSE)
  }

  if (is.null(ffi$globals)) {
    ffi$globals <- list()
  }
  ffi$globals[[name]] <- type
  ffi
}

#' Add C headers
#'
#' @param ffi A tcc_ffi object
#' @param header Header string or include directive
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_header <- function(ffi, header) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  ffi$headers <- c(ffi$headers, header)
  ffi
}

#' Add C source code
#'
#' @param ffi A tcc_ffi object
#' @param code C source code string
#' @return Updated tcc_ffi object (for chaining)
#' @export
tcc_source <- function(ffi, code) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }
  ffi$c_code <- c(ffi$c_code, as.character(code))
  ffi
}

#' Bind symbols with type specifications
#'
#' Define symbols with Bun-style type specifications for API mode.
#' This is the core of the declarative FFI API.
#'
#' @param ffi A tcc_ffi object
#' @param ... Named list of symbol definitions. Each definition is a list with:
#'   \itemize{
#'     \item args: List of fixed FFI argument types (e.g., list("i32", "f64"))
#'     \item returns: FFI type for return value (e.g., "f64", "cstring")
#'     \item variadic: Set TRUE for C varargs functions
#'     \item varargs: Legacy typed variadic tail (exact/prefix mode)
#'     \item varargs_types: Allowed scalar FFI types for true variadic tails
#'     \item varargs_min: Minimum number of trailing varargs
#'     \item varargs_max: Maximum number of trailing varargs (required for
#'       true variadic mode, defaults to \code{varargs_min})
#'     \item code: Optional C code for the symbol (for embedded functions)
#'   }
#'   Callback arguments should use the form \code{callback:<signature>} (e.g.,
#'   \code{callback:double(double)}). The generated trampoline expects a
#'   \code{tcc_callback_ptr(cb)} to the corresponding user-data parameter in
#'   the C API. For thread-safe scheduling, use
#'   \code{callback_async:<signature>} which enqueues the call on the main
#'   thread and returns a default value immediately.
#' @return Updated tcc_ffi object (for chaining)
#' @export
#' @examples
#' ffi <- tcc_ffi() |>
#'   tcc_bind(
#'     add = list(args = list("i32", "i32"), returns = "i32"),
#'     greet = list(args = list("cstring"), returns = "cstring")
#'   )
tcc_bind <- function(ffi, ...) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  symbols <- list(...)

  # Validate and normalize each symbol definition into a classed spec.
  for (sym_name in names(symbols)) {
    ffi$symbols[[sym_name]] <- as_rtinycc_bound_symbol(
      sym_name,
      symbols[[sym_name]]
    )
  }

  ffi
}

#' Compile FFI bindings
#'
#' Compile the defined symbols into callable functions.
#' This generates C wrapper code and compiles it with TinyCC.
#'
#' @param ffi A tcc_ffi object
#' @param verbose Print compilation info
#' @return A tcc_compiled object with callable functions
#' @export
tcc_compile <- function(ffi, verbose = FALSE) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  if (
    length(ffi$symbols) == 0 &&
      length(ffi$structs) == 0 &&
      length(ffi$unions) == 0 &&
      length(ffi$enums) == 0 &&
      length(ffi$globals) == 0
  ) {
    stop(
      "No symbols, structs, unions, enums, or globals defined. Use tcc_bind(), tcc_struct(), tcc_union(), tcc_enum(), or tcc_global() first.",
      call. = FALSE
    )
  }

  # Generate C code (always R API mode now)
  c_code <- generate_ffi_code(
    symbols = ffi$symbols,
    headers = ffi$headers,
    c_code = ffi$c_code,
    is_external = FALSE,
    structs = ffi$structs,
    unions = ffi$unions,
    enums = ffi$enums,
    globals = ffi$globals,
    container_of = ffi$container_of,
    field_addr = ffi$field_addr,
    struct_raw_access = ffi$struct_raw_access,
    introspect = ffi$introspect
  )

  if (verbose) {
    message("Generated C code:\n", c_code)
  }

  state <- tcc_ffi_create_state(ffi, output = ffi$output)
  tcc_ffi_compile_state(
    state = state,
    c_code = c_code,
    libraries = ffi$libraries,
    target = "FFI bindings"
  )

  # Create callable object
  compiled <- tcc_compiled_object(
    state,
    ffi$symbols,
    ffi$output,
    ffi$structs,
    ffi$unions,
    ffi$enums,
    ffi$globals,
    ffi$container_of,
    ffi$field_addr,
    ffi$struct_raw_access,
    ffi$introspect
  )

  # Store the recipe so the object can be recompiled after deserialization
  compiled$.ffi <- ffi

  compiled
}

tcc_runtime_library_paths <- function() {
  paths <- file.path(R.home("lib"))
  if (.Platform$OS.type == "windows") {
    # On Windows, R.dll lives in bin/<arch>; TCC needs it for R API symbols.
    paths <- c(paths, file.path(R.home(), "bin", .Platform$r_arch))
  }
  normalizePath(paths, winslash = "/", mustWork = FALSE)
}

tcc_ffi_create_state <- function(ffi, output = ffi$output) {
  state <- tcc_state(
    output = output,
    include_path = c(
      tcc_include_paths(),
      ffi$include_paths,
      file.path(R.home("include"))
    ),
    lib_path = c(
      tcc_lib_paths(),
      ffi$lib_paths,
      tcc_runtime_library_paths()
    )
  )

  if (!is.null(ffi$options) && length(ffi$options) > 0) {
    rc <- tcc_set_options(state, paste(ffi$options, collapse = " "))
    if (rc < 0) {
      stop("Failed to apply TinyCC compiler options", call. = FALSE)
    }
  }

  state
}

tcc_ffi_compile_state <- function(state, c_code, libraries, target) {
  for (lib in libraries) {
    result <- tcc_add_library(state, lib)
    if (result < 0) {
      stop(
        "Failed to add library '", lib, "' while compiling ", target,
        call. = FALSE
      )
    }
  }

  # Always link against R's shared library so TCC can resolve
  # R API symbols (and any symbols R itself exports).
  tcc_add_library(state, "R")

  result <- tcc_compile_string(state, c_code)
  if (result != 0) {
    stop("Failed to compile ", target, call. = FALSE)
  }

  # Register host symbols for macOS compatibility before relocation.
  .Call(RC_libtcc_add_host_symbols, state)

  result <- tcc_relocate(state)
  if (result != 0) {
    stop("Failed to relocate compiled code for ", target, call. = FALSE)
  }

  invisible(state)
}

infer_variadic_arg_type <- function(x, allowed_types) {
  if (inherits(x, "externalptr")) {
    if ("ptr" %in% allowed_types) {
      return("ptr")
    }
    stop(
      "external pointer variadic argument requires allowed type 'ptr'",
      call. = FALSE
    )
  }

  if (is.character(x) && length(x) == 1) {
    if ("cstring" %in% allowed_types) {
      return("cstring")
    }
    stop(
      "character variadic argument requires allowed type 'cstring'",
      call. = FALSE
    )
  }

  if (is.logical(x) && length(x) == 1) {
    if ("bool" %in% allowed_types) {
      return("bool")
    }
    stop(
      "logical variadic argument requires allowed type 'bool'",
      call. = FALSE
    )
  }

  if (is.raw(x) && length(x) == 1) {
    if ("u8" %in% allowed_types) {
      return("u8")
    }
    if ("i8" %in% allowed_types) {
      return("i8")
    }
    stop(
      "raw variadic argument requires allowed type 'u8' or 'i8'",
      call. = FALSE
    )
  }

  if (is.integer(x) && length(x) == 1) {
    pref <- c(
      "i32",
      "i64",
      "u32",
      "u64",
      "i16",
      "u16",
      "i8",
      "u8",
      "f64",
      "f32"
    )
    pick <- pref[pref %in% allowed_types]
    if (length(pick) > 0) {
      return(pick[[1]])
    }
    stop(
      "integer variadic argument has no compatible allowed type",
      call. = FALSE
    )
  }

  if (is.double(x) && length(x) == 1) {
    pref <- c("f64", "f32")
    pick <- pref[pref %in% allowed_types]
    if (length(pick) > 0) {
      return(pick[[1]])
    }
    stop(
      "numeric variadic argument has no compatible allowed floating type",
      call. = FALSE
    )
  }

  stop(
    "Unsupported variadic argument type; expected scalar logical/integer/numeric/character/raw/externalptr",
    call. = FALSE
  )
}

# Create compiled object with callable functions
tcc_compiled_object <- function(
  state,
  symbols,
  output,
  structs = NULL,
  unions = NULL,
  enums = NULL,
  globals = NULL,
  container_of = NULL,
  field_addr = NULL,
  struct_raw_access = NULL,
  introspect = NULL
) {
  # Build helper names for struct/union/enum helpers
  helper_names <- character()
  helper_specs <- list()

  # Struct helpers
  if (!is.null(structs)) {
    for (struct_name in names(structs)) {
      fields <- structs[[struct_name]]
      # new, free
      helper_names <- c(
        helper_names,
        paste0("struct_", struct_name, "_new"),
        paste0("struct_", struct_name, "_free")
      )
      helper_specs[[paste0("struct_", struct_name, "_new")]] <- list(
        args = list(),
        returns = "sexp"
      )
      helper_specs[[paste0("struct_", struct_name, "_free")]] <- list(
        args = list("sexp"),
        returns = "sexp"
      )
      # getters and setters
      for (field_name in names(fields)) {
        field_spec <- fields[[field_name]]
        is_array <- is.list(field_spec) && isTRUE(field_spec$array)

        helper_names <- c(
          helper_names,
          paste0("struct_", struct_name, "_get_", field_name)
        )
        helper_specs[[paste0(
          "struct_",
          struct_name,
          "_get_",
          field_name
        )]] <- c(
          list(
            args = list("sexp"),
            returns = "sexp"
          ),
          if (rtinycc_is_bitfield_spec(field_spec)) {
            list(.helper_operation = "bitfield_getter")
          } else if (!is.null(rtinycc_nested_struct_type_name(field_spec))) {
            list(.helper_operation = "nested_view")
          } else {
            list()
          }
        )

        if (is_array) {
          helper_names <- c(
            helper_names,
            paste0("struct_", struct_name, "_get_", field_name, "_elt"),
            paste0("struct_", struct_name, "_set_", field_name, "_elt")
          )
          helper_specs[[paste0(
            "struct_",
            struct_name,
            "_get_",
            field_name,
            "_elt"
          )]] <- list(
            args = list("sexp", "i32"),
            returns = "sexp"
          )
          helper_specs[[paste0(
            "struct_",
            struct_name,
            "_set_",
            field_name,
            "_elt"
          )]] <- list(
            args = list("sexp", "i32", "sexp"),
            returns = "sexp"
          )
        } else {
          helper_names <- c(
            helper_names,
            paste0("struct_", struct_name, "_set_", field_name)
          )
          helper_specs[[paste0(
            "struct_",
            struct_name,
            "_set_",
            field_name
          )]] <- c(
            list(
              args = list("sexp", "sexp"),
              returns = "sexp"
            ),
            if (rtinycc_is_bitfield_spec(field_spec)) {
              list(.helper_operation = "bitfield_setter")
            } else if (!is.null(rtinycc_nested_struct_type_name(field_spec))) {
              list(.helper_operation = "nested_setter")
            } else {
              list()
            }
          )
        }
      }
      # container_of helpers
      if (!is.null(container_of) && struct_name %in% names(container_of)) {
        for (member_name in container_of[[struct_name]]) {
          helper_names <- c(
            helper_names,
            paste0("struct_", struct_name, "_from_", member_name)
          )
          helper_specs[[paste0(
            "struct_",
            struct_name,
            "_from_",
            member_name
          )]] <- list(
            args = list("sexp"),
            returns = "sexp"
          )
        }
      }
      # field_addr helpers
      if (!is.null(field_addr) && struct_name %in% names(field_addr)) {
        for (field_name in field_addr[[struct_name]]) {
          helper_names <- c(
            helper_names,
            paste0("struct_", struct_name, "_", field_name, "_addr")
          )
          helper_specs[[paste0(
            "struct_",
            struct_name,
            "_",
            field_name,
            "_addr"
          )]] <- list(
            args = list("sexp"),
            returns = "sexp"
          )
        }
      }
      # raw_access helpers
      if (!is.null(struct_raw_access) && struct_name %in% struct_raw_access) {
        helper_names <- c(
          helper_names,
          paste0("struct_", struct_name, "_get_raw"),
          paste0("struct_", struct_name, "_set_raw")
        )
        helper_specs[[paste0("struct_", struct_name, "_get_raw")]] <- list(
          args = list("sexp", "i32"),
          returns = "sexp"
        )
        helper_specs[[paste0("struct_", struct_name, "_set_raw")]] <- list(
          args = list("sexp", "raw"),
          returns = "sexp"
        )
      }
      # introspection helpers
      if (!is.null(introspect) && introspect) {
        helper_names <- c(
          helper_names,
          paste0("struct_", struct_name, "_sizeof"),
          paste0("struct_", struct_name, "_alignof")
        )
        helper_specs[[paste0("struct_", struct_name, "_sizeof")]] <- list(
          args = list(),
          returns = "sexp"
        )
        helper_specs[[paste0("struct_", struct_name, "_alignof")]] <- list(
          args = list(),
          returns = "sexp"
        )
      }
    }
  }

  # Union helpers
  if (!is.null(unions)) {
    for (union_name in names(unions)) {
      union_def <- unions[[union_name]]
      # new, free
      helper_names <- c(
        helper_names,
        paste0("union_", union_name, "_new"),
        paste0("union_", union_name, "_free")
      )
      helper_specs[[paste0("union_", union_name, "_new")]] <- list(
        args = list(),
        returns = "sexp"
      )
      helper_specs[[paste0("union_", union_name, "_free")]] <- list(
        args = list("sexp"),
        returns = "sexp"
      )
      # getters and setters for members
      for (mem_name in names(union_def$members)) {
        mem_spec <- union_def$members[[mem_name]]
        is_nested_struct <- is.list(mem_spec) &&
          identical(mem_spec$type %||% NULL, "struct")

        helper_names <- c(
          helper_names,
          paste0("union_", union_name, "_get_", mem_name)
        )
        helper_specs[[paste0("union_", union_name, "_get_", mem_name)]] <- c(
          list(
            args = list("sexp"),
            returns = "sexp"
          ),
          if (is_nested_struct) {
            list(.helper_operation = "nested_view")
          } else {
            list()
          }
        )

        if (!is_nested_struct) {
          helper_names <- c(
            helper_names,
            paste0("union_", union_name, "_set_", mem_name)
          )
          helper_specs[[paste0(
            "union_",
            union_name,
            "_set_",
            mem_name
          )]] <- list(
            args = list("sexp", "sexp"),
            returns = "sexp"
          )
        }
      }
      # introspection
      if (!is.null(introspect) && introspect) {
        helper_names <- c(
          helper_names,
          paste0("union_", union_name, "_sizeof"),
          paste0("union_", union_name, "_alignof")
        )
        helper_specs[[paste0("union_", union_name, "_sizeof")]] <- list(
          args = list(),
          returns = "sexp"
        )
        helper_specs[[paste0("union_", union_name, "_alignof")]] <- list(
          args = list(),
          returns = "sexp"
        )
      }
    }
  }

  # Enum helpers
  if (!is.null(enums)) {
    for (enum_name in names(enums)) {
      enum_def <- enums[[enum_name]]
      # introspection
      if (!is.null(introspect) && introspect) {
        helper_names <- c(helper_names, paste0("enum_", enum_name, "_sizeof"))
        helper_specs[[paste0("enum_", enum_name, "_sizeof")]] <- list(
          args = list(),
          returns = "sexp"
        )
      }
      # constant export
      if (!is.null(enum_def$constants)) {
        for (const_name in enum_def$constants) {
          helper_names <- c(
            helper_names,
            paste0("enum_", enum_name, "_", const_name)
          )
          helper_specs[[paste0("enum_", enum_name, "_", const_name)]] <- list(
            args = list(),
            returns = "sexp",
            .helper_operation = "constant"
          )
        }
      }
    }
  }

  # Global helpers
  if (!is.null(globals) && length(globals) > 0) {
    for (global_name in names(globals)) {
      helper_names <- c(
        helper_names,
        paste0("global_", global_name, "_get"),
        paste0("global_", global_name, "_set")
      )
      helper_specs[[paste0("global_", global_name, "_get")]] <- list(
        args = list(),
        returns = "sexp"
      )
      helper_specs[[paste0("global_", global_name, "_set")]] <- list(
        args = list("sexp"),
        returns = "sexp"
      )
    }
  }

  if (length(helper_specs) > 0) {
    helper_specs <- stats::setNames(
      lapply(names(helper_specs), function(sym_name) {
        helper_kind <- if (startsWith(sym_name, "struct_")) {
          "struct"
        } else if (startsWith(sym_name, "union_")) {
          "union"
        } else if (startsWith(sym_name, "enum_")) {
          "enum"
        } else if (startsWith(sym_name, "global_")) {
          "global"
        } else {
          "helper"
        }
        helper_operation <- helper_specs[[sym_name]]$.helper_operation %||%
          if (grepl("_new$", sym_name)) {
            "constructor"
          } else if (grepl("_free$", sym_name)) {
            "destructor"
          } else if (grepl("_get_raw$", sym_name)) {
            "raw_get"
          } else if (grepl("_set_raw$", sym_name)) {
            "raw_set"
          } else if (
            grepl("_sizeof$", sym_name) || grepl("_alignof$", sym_name)
          ) {
            "introspection"
          } else if (grepl("_from_", sym_name, fixed = TRUE)) {
            "container_of"
          } else if (grepl("_addr$", sym_name)) {
            "field_addr"
          } else if (grepl("_set_.*_elt$", sym_name)) {
            "array_setter"
          } else if (grepl("_get_.*_elt$", sym_name)) {
            "array_getter"
          } else if (grepl("_set$", sym_name)) {
            "setter"
          } else if (grepl("_get$", sym_name)) {
            "getter"
          } else if (grepl("_set_", sym_name, fixed = TRUE)) {
            "setter"
          } else if (grepl("_get_", sym_name, fixed = TRUE)) {
            "getter"
          } else {
            "helper"
          }
        as_rtinycc_helper_symbol(
          sym_name,
          helper_specs[[sym_name]],
          helper_kind,
          helper_operation
        )
      }),
      names(helper_specs)
    )
  }

  # Create environment with callable functions
  env <- new.env(parent = emptyenv())
  bind_failures <- character(0)
  add_bind_failure <- function(msg) {
    bind_failures <<- c(bind_failures, msg)
  }

  for (sym_name in names(symbols)) {
    sym <- symbols[[sym_name]]

    if (isTRUE(sym$variadic)) {
      vararg_mode <- sym$varargs_mode %||% "prefix"
      fn_ptrs <- list()

      if (identical(vararg_mode, "types")) {
        allowed_types <- sym$varargs_types %||% list()
        min_varargs <- sym$varargs_min %||% 0L
        max_varargs <- sym$varargs_max %||% min_varargs

        for (n_varargs in seq.int(min_varargs, max_varargs)) {
          type_sequences <- generate_variadic_type_sequences(
            allowed_types,
            n_varargs
          )
          for (this_types in type_sequences) {
            wrapper_name <- variadic_wrapper_name_types(
              paste0("R_wrap_", sym_name),
              this_types
            )
            key <- variadic_signature_key(this_types)

            tryCatch(
              {
                fn_ptr <- tcc_get_symbol(state, wrapper_name)
                if (!tcc_symbol_is_valid(fn_ptr)) {
                  add_bind_failure(
                    paste0(
                      "Symbol '",
                      sym_name,
                      "' returned invalid pointer for '",
                      wrapper_name,
                      "'"
                    )
                  )
                  next
                }
                fn_ptrs[[key]] <- fn_ptr
              },
              error = function(e) {
                add_bind_failure(
                  paste0(
                    "Could not bind symbol '",
                    sym_name,
                    "' wrapper '",
                    wrapper_name,
                    "': ",
                    conditionMessage(e)
                  )
                )
              }
            )
          }
        }
      } else {
        max_varargs <- length(sym$varargs %||% list())
        min_varargs <- sym$varargs_min %||% max_varargs

        for (n_varargs in seq.int(min_varargs, max_varargs)) {
          wrapper_name <- variadic_wrapper_name(
            paste0("R_wrap_", sym_name),
            n_varargs
          )

          tryCatch(
            {
              fn_ptr <- tcc_get_symbol(state, wrapper_name)
              if (!tcc_symbol_is_valid(fn_ptr)) {
                add_bind_failure(
                  paste0(
                    "Symbol '",
                    sym_name,
                    "' returned invalid pointer for '",
                    wrapper_name,
                    "'"
                  )
                )
                next
              }
              fn_ptrs[[as.character(n_varargs)]] <- fn_ptr
            },
            error = function(e) {
              add_bind_failure(
                paste0(
                  "Could not bind symbol '",
                  sym_name,
                  "' wrapper '",
                  wrapper_name,
                  "': ",
                  conditionMessage(e)
                )
              )
            }
          )
        }
      }

      if (length(fn_ptrs) == 0) {
        add_bind_failure(
          paste0(
            "Could not bind any variadic wrappers for symbol '",
            sym_name,
            "'"
          )
        )
        next
      }

      env[[sym_name]] <- make_callable(fn_ptrs, sym, state)
      next
    }

    wrapper_name <- paste0("R_wrap_", sym_name)
    tryCatch(
      {
        fn_ptr <- tcc_get_symbol(state, wrapper_name)

        # Validate the pointer before creating callable
        if (!tcc_symbol_is_valid(fn_ptr)) {
          add_bind_failure(
            paste0(
              "Symbol '",
              sym_name,
              "' returned invalid pointer for '",
              wrapper_name,
              "'"
            )
          )
          next
        }
        env[[sym_name]] <- make_callable(fn_ptr, sym, state)
      },
      error = function(e) {
        add_bind_failure(
          paste0(
            "Could not bind symbol '",
            sym_name,
            "': ",
            conditionMessage(e)
          )
        )
      }
    )
  }

  for (sym_name in helper_names) {
    wrapper_name <- paste0("R_wrap_", sym_name)
    sym <- helper_specs[[sym_name]]
    if (is.null(sym)) {
      add_bind_failure(paste0("Unknown helper symbol '", sym_name, "'"))
      next
    }

    tryCatch(
      {
        fn_ptr <- tcc_get_symbol(state, wrapper_name)

        if (!tcc_symbol_is_valid(fn_ptr)) {
          add_bind_failure(
            paste0(
              "Symbol '",
              sym_name,
              "' returned invalid pointer for '",
              wrapper_name,
              "'"
            )
          )
          next
        }

        env[[sym_name]] <- make_callable(fn_ptr, sym, state)
      },
      error = function(e) {
        add_bind_failure(
          paste0(
            "Could not bind symbol '",
            sym_name,
            "': ",
            conditionMessage(e)
          )
        )
      }
    )
  }

  if (length(bind_failures) > 0) {
    stop(
      "Failed to bind compiled wrapper symbols:\n  - ",
      paste(unique(bind_failures), collapse = "\n  - "),
      call. = FALSE
    )
  }

  # Add state and metadata
  env$.state <- state
  env$.symbols <- symbols
  env$.output <- output
  env$.helpers <- list(
    structs = structs,
    unions = unions,
    enums = enums,
    globals = globals
  )
  env$.helper_specs <- helper_specs
  env$.valid <- TRUE

  structure(env, class = "tcc_compiled")
}

# Make a callable function from symbol pointer
# Uses .Call() directly with the external pointer - R can call external pointers!
make_callable <- function(fn_ptr, sym, state) {
  # Force evaluation to ensure pointer is captured
  force(fn_ptr)
  force(sym)
  force(state)

  arg_types <- sym$args %||% list()
  vararg_types <- sym$varargs %||% list()
  vararg_types_allowed <- sym$varargs_types %||% list()
  variadic <- isTRUE(sym$variadic)
  vararg_mode <- if (variadic) sym$varargs_mode %||% "prefix" else NULL
  fixed_n <- length(arg_types)
  varargs_min <- if (variadic) {
    sym$varargs_min %||% length(vararg_types)
  } else {
    0L
  }
  varargs_max <- if (variadic) {
    sym$varargs_max %||% length(vararg_types)
  } else {
    0L
  }
  min_n <- fixed_n + varargs_min
  max_n <- fixed_n + varargs_max
  fn_ptr_is_map <- is.list(fn_ptr)
  prefix_ptr_by_tail <- NULL
  types_ptr_env <- NULL
  if (variadic && fn_ptr_is_map) {
    if (identical(vararg_mode, "prefix")) {
      ptr_names <- suppressWarnings(as.integer(names(fn_ptr)))
      if (
        length(ptr_names) == length(fn_ptr) &&
          length(ptr_names) > 0L &&
          all(!is.na(ptr_names))
      ) {
        prefix_ptr_by_tail <- vector("list", varargs_max + 1L)
        for (i in seq_along(fn_ptr)) {
          tail_n <- ptr_names[[i]]
          if (!is.na(tail_n) && tail_n >= 0L && tail_n <= varargs_max) {
            prefix_ptr_by_tail[[tail_n + 1L]] <- fn_ptr[[i]]
          }
        }
      }
    } else if (identical(vararg_mode, "types")) {
      ptr_keys <- names(fn_ptr)
      if (!is.null(ptr_keys) && length(ptr_keys) == length(fn_ptr)) {
        types_ptr_env <- list2env(fn_ptr, hash = TRUE, parent = emptyenv())
      }
    }
  }
  sym_name <- sym$name

  if (!variadic) {
    expected_n <- fixed_n
    dot_syms <- if (expected_n > 0L) {
      lapply(seq_len(expected_n), function(i) as.name(paste0("..", i)))
    } else {
      list()
    }
    call_expr <- as.call(c(
      as.name(".RtinyccCall"),
      as.name(".call_ptr"),
      dot_syms
    ))
    body_expr <- bquote({
      n_args <- nargs()
      if (n_args != .(expected_n)) {
        stop(
          "Expected ",
          .(expected_n),
          " arguments, got ",
          n_args,
          call. = FALSE
        )
      }
      if (!tcc_symbol_is_valid(.call_ptr)) {
        stop(
          "Function pointer for '",
          .(sym_name),
          "' is no longer valid",
          call. = FALSE
        )
      }
      .(call_expr)
    })
    f <- eval(
      call("function", as.pairlist(alist(... = )), body_expr),
      envir = list2env(
        list(
          .call_ptr = fn_ptr,
          .RtinyccCall = .RtinyccCall,
          tcc_symbol_is_valid = tcc_symbol_is_valid
        ),
        parent = baseenv()
      )
    )

    # Store the pointer in the function's environment to prevent GC
    environment(f)$.fn_ptr <- fn_ptr
    environment(f)$.state <- state
    environment(f)$.arg_types <- arg_types
    environment(f)$.sym_name <- sym_name

    return(f)
  }

  # Create function that calls the wrapper via .Call()
  f <- function(...) {
    args <- list(...)
    n_args <- length(args)
    if (n_args < min_n || n_args > max_n) {
      stop(
        "Expected between ",
        min_n,
        " and ",
        max_n,
        " arguments, got ",
        n_args,
        call. = FALSE
      )
    }
    n_tail <- n_args - fixed_n

    if (identical(vararg_mode, "types")) {
      inferred_types <- character(n_tail)
      if (n_tail > 0L) {
        for (i in seq_len(n_tail)) {
          inferred_types[[i]] <- infer_variadic_arg_type(
            args[[fixed_n + i]],
            vararg_types_allowed
          )
        }
      }
      key <- variadic_signature_key(inferred_types)
      if (!is.null(types_ptr_env)) {
        call_ptr <- if (exists(key, envir = types_ptr_env, inherits = FALSE)) {
          types_ptr_env[[key]]
        } else {
          NULL
        }
      } else {
        call_ptr <- if (fn_ptr_is_map) fn_ptr[[key]] else fn_ptr
      }
    } else {
      if (!is.null(prefix_ptr_by_tail)) {
        idx <- n_tail + 1L
        call_ptr <- if (idx > 0L && idx <= length(prefix_ptr_by_tail)) {
          prefix_ptr_by_tail[[idx]]
        } else {
          NULL
        }
      } else {
        call_ptr <- if (fn_ptr_is_map) {
          fn_ptr[[as.character(n_tail)]]
        } else {
          fn_ptr
        }
      }
    }

    if (is.null(call_ptr)) {
      stop(
        "No compiled variadic wrapper for symbol '",
        sym_name,
        "' and the provided argument shape",
        call. = FALSE
      )
    }

    # Validate pointer is still valid before calling
    if (!tcc_symbol_is_valid(call_ptr)) {
      stop(
        "Function pointer for '",
        sym_name,
        "' is no longer valid",
        call. = FALSE
      )
    }

    # R's .Call() can invoke external pointers as functions.
    rtinycc_call(n_args, call_ptr, args)
  }

  # Store the pointer in the function's environment to prevent GC
  environment(f)$.fn_ptr <- fn_ptr
  environment(f)$.state <- state
  environment(f)$.arg_types <- arg_types
  environment(f)$.sym_name <- sym_name

  f
}

#' Print tcc_ffi object
#'
#' @param x A tcc_ffi object
#' @param ... Ignored
#' @return The input `tcc_ffi` object, invisibly. Called for its side
#'   effect of printing the configured output mode, registered symbols,
#'   and selected libraries/include paths.
#' @export
print.tcc_ffi <- function(x, ...) {
  cat("<tcc_ffi>\n")
  cat("  Output:", x$output, "\n")
  cat("  Symbols:", length(x$symbols), "defined\n")
  if (length(x$symbols) > 0) {
    for (name in names(x$symbols)) {
      sym <- x$symbols[[name]]
      args <- paste(sym$args %||% "void", collapse = ", ")
      cat("    ", name, "(", args, ") -> ", sym$returns, "\n", sep = "")
    }
  }
  if (length(x$libraries) > 0) {
    cat("  Libraries:", paste(x$libraries, collapse = ", "), "\n")
  }
  if (length(x$include_paths) > 0) {
    cat("  Include paths:", length(x$include_paths), "\n")
  }
  invisible(x)
}

#' Print tcc_compiled object
#'
#' @param x A tcc_compiled object
#' @param ... Ignored
#' @return The input `tcc_compiled` object, invisibly. Called for its side
#'   effect of printing the compilation output mode and the status of
#'   compiled callable symbols.
#' @export
print.tcc_compiled <- function(x, ...) {
  cat("<tcc_compiled>\n")
  cat("  Output:", x$.output, "\n")
  cat("  Symbols:", length(x$.symbols), "compiled\n")
  if (length(x$.symbols) > 0) {
    for (name in names(x$.symbols)) {
      if (exists(name, envir = x, inherits = FALSE)) {
        cat("    ", name, "() [callable]\n", sep = "")
      } else {
        cat("    ", name, "() [failed]\n", sep = "")
      }
    }
  }
  invisible(x)
}

#' Access a compiled FFI symbol
#'
#' Overrides \code{$} to detect dead pointers after deserialization
#' and recompile transparently from the stored recipe.
#'
#' @param x A tcc_compiled object
#' @param name Symbol name to access
#' @return The callable function or metadata field
#' @export
`$.tcc_compiled` <- function(x, name) {
  # Fast path: internal fields are always accessible
  if (startsWith(name, ".")) {
    return(.subset2(x, name))
  }

  # Detect dead pointers (nil after deserialization) and recompile.
  # tcc_symbol_is_valid() is a single C call; negligible overhead.
  state <- .subset2(x, ".state")
  if (is.null(state) || !tcc_symbol_is_valid(state)) {
    message("[Rtinycc] Recompiling FFI bindings after deserialization")
    recompile_into(x)
  }

  .subset2(x, name)
}

#' Recompile a tcc_compiled object
#'
#' Explicitly recompile from the stored FFI recipe.
#' Useful after deserialization (\code{readRDS}, \code{unserialize})
#' or to force a fresh compilation.
#'
#' @param compiled A tcc_compiled object
#' @return The recompiled tcc_compiled object (invisibly, same environment)
#' @export
tcc_recompile <- function(compiled) {
  if (!inherits(compiled, "tcc_compiled")) {
    stop("Expected tcc_compiled object", call. = FALSE)
  }
  recompile_into(compiled)
  invisible(compiled)
}

# Shared recompilation logic for both tcc_compile and tcc_link objects.
# Rebuilds the compiled code from the stored recipe and copies all
# bindings into the target environment.
recompile_into <- function(target) {
  link_args <- .subset2(target, ".link_args")
  if (!is.null(link_args)) {
    fresh <- do.call(tcc_link, link_args)
  } else {
    ffi <- .subset2(target, ".ffi")
    if (is.null(ffi)) {
      stop(
        "Cannot recompile: no FFI recipe stored in this object",
        call. = FALSE
      )
    }
    fresh <- tcc_compile(ffi)
  }
  for (nm in ls(fresh, all.names = TRUE)) {
    target[[nm]] <- fresh[[nm]]
  }
  target[[".valid"]] <- TRUE
  invisible(target)
}

# Helper for %||%
`%||%` <- function(x, y) if (is.null(x)) y else x

tcc_library_search_paths <- function(
  sysname = as.character(unname(Sys.info()[["sysname"]]))
) {
  platform_paths <- tcc_platform_lib_paths(sysname)

  env_vars <- unique(c(
    tcc_loader_env_key(sysname),
    "LIBRARY_PATH",
    if (identical(sysname, "Darwin")) {
      "DYLD_FALLBACK_LIBRARY_PATH"
    } else {
      character(0)
    }
  ))

  env_paths <- unlist(
    lapply(env_vars, function(key) {
      value <- Sys.getenv(key, unset = "")
      if (!nzchar(value)) {
        return(character(0))
      }
      strsplit(value, .Platform$path.sep, fixed = TRUE)[[1]]
    }),
    use.names = FALSE
  )

  unique(c(env_paths[nzchar(env_paths)], platform_paths))
}

# Search for library in platform-dependent paths
tcc_find_library <- function(name) {
  if (file.exists(name)) {
    return(name)
  }

  # Try platform-specific paths plus relevant loader/linker env paths.
  sysname <- as.character(unname(Sys.info()[["sysname"]]))
  paths <- tcc_library_search_paths(sysname)

  suffix_pattern <- c(
    Windows = "\\.dll$",
    Linux = "\\.so(\\..*)?$",
    Darwin = "\\.dylib(\\..*)?$"
  )[[sysname]]
  has_platform_suffix <- !is.null(suffix_pattern) &&
    grepl(suffix_pattern, name, ignore.case = identical(sysname, "Windows"))
  lib_name <- if (has_platform_suffix) {
    name
  } else {
    tcc_short_lib_filename(sysname, name)
  }

  for (path in paths) {
    if (dir.exists(path)) {
      full_path <- file.path(path, lib_name)
      if (file.exists(full_path)) {
        return(full_path)
      }
    }
  }

  # Not found
  NULL
}

#' Link an external shared library with Bun-style FFI bindings
#'
#' Link a system library (like libsqlite3) and generate type-safe
#' wrappers automatically using TinyCC JIT compilation (API mode).
#' Unlike dlopen(), this uses TinyCC to compile bindings that handle
#' type conversion between R and C automatically.
#'
#' @param path Library short name (e.g., `"m"`, `"sqlite3"`) or full path to
#'   the shared library. Short names stay on the normal linker-name path
#'   (`-l<name>`). File names such as `libm.so.6` or full paths are resolved
#'   through the configured library search paths when needed and linked as exact
#'   files rather than collapsed to generic names like `m`.
#' @param symbols Named list of symbol definitions with:
#'   \itemize{
#'     \item args: List of FFI types for arguments
#'     \item returns: FFI type for return value
#'   }
#' @param headers Optional C headers to include
#' @param libs Library names to link (e.g., "sqlite3")
#' @param lib_paths Additional library search paths
#' @param include_paths Additional include search paths
#' @param user_code Optional custom C code to include in the compilation
#' @param verbose Print debug information
#' @return A tcc_compiled object with callable functions
#' @export
#' @examples
#' \dontrun{
#' # Link SQLite with type-safe bindings
#' sqlite <- tcc_link(
#'   "libsqlite3.so",
#'   symbols = list(
#'     sqlite3_libversion = list(args = list(), returns = "cstring"),
#'     sqlite3_open = list(args = list("cstring", "ptr"), returns = "i32")
#'   ),
#'   libs = "sqlite3"
#' )
#'
#' # Call directly - type conversion happens automatically
#' sqlite$sqlite3_libversion()
#'
#' # Example with custom user code for helper functions
#' math_with_helpers <- tcc_link(
#'   "m",
#'   symbols = list(
#'     sqrt = list(args = list("f64"), returns = "f64"),
#'     safe_sqrt = list(args = list("f64"), returns = "f64")
#'   ),
#'   user_code = "
#'     #include <math.h>
#'
#'     // Helper function that validates input before calling sqrt
#'     double safe_sqrt(double x) {
#'       if (x < 0) {
#'         return NAN;
#'       }
#'       return sqrt(x);
#'     }
#'   ",
#'   libs = "m"
#' )
#' math_with_helpers$safe_sqrt(16.0)
#' math_with_helpers$safe_sqrt(-4.0) # Returns NaN for negative input
#' }
tcc_link <- function(
  path,
  symbols,
  headers = NULL,
  libs = character(0),
  lib_paths = character(0),
  include_paths = character(0),
  user_code = NULL,
  verbose = FALSE
) {
  # Find library if not absolute path. Keep short linker names such as
  # "m" or "sqlite3" on the -l<name> path. When the caller supplies a file
  # name or path, resolve it and link that exact file; versioned runtime files
  # such as libm.so.6 are not equivalent to the generic -lm linker name on all
  # systems.
  is_short_name <- !file.exists(path) &&
    !grepl("[/\\\\]", path) &&
    !grepl("\\.(so|dylib|dll)", path)

  if (!is_short_name && !file.exists(path) && !grepl("^/", path)) {
    found_path <- tcc_find_library(path)
    if (is.null(found_path)) {
      stop("Library not found: ", path, call. = FALSE)
    }
    path <- found_path
  }

  if (verbose) {
    message("Loading library: ", path)
    message("Symbols: ", paste(names(symbols), collapse = ", "))
  }

  # Create FFI context
  ffi <- tcc_ffi() |>
    tcc_output("memory")

  # Add library
  ffi$libraries <- libs
  ffi$lib_paths <- lib_paths
  ffi$include_paths <- include_paths

  # Process headers
  if (!is.null(headers)) {
    ffi$headers <- headers
  }

  # Process user code
  if (!is.null(user_code)) {
    ffi$c_code <- user_code
  }

  # Add the library path and link the library.
  if (!is_short_name) {
    ffi <- rtinycc_add_library_file(ffi, path)
  } else {
    # Short name like "m" — just pass to linker as -l<name>.
    ffi <- tcc_library(ffi, path)
  }

  # Store symbols as classed specs
  for (sym_name in names(symbols)) {
    ffi$symbols[[sym_name]] <- as_rtinycc_bound_symbol(
      sym_name,
      symbols[[sym_name]]
    )
  }

  # Mark as external library bindings
  attr(ffi, "external_lib") <- path
  attr(ffi, "is_external") <- TRUE

  # Generate C code with external declarations (always R API mode)
  c_code <- generate_ffi_code(
    symbols = ffi$symbols,
    headers = ffi$headers,
    c_code = ffi$c_code,
    is_external = TRUE,
    structs = ffi$structs,
    unions = ffi$unions,
    enums = ffi$enums,
    globals = ffi$globals,
    container_of = ffi$container_of,
    field_addr = ffi$field_addr,
    struct_raw_access = ffi$struct_raw_access,
    introspect = ffi$introspect
  )

  if (verbose) {
    message("Generated C code:\n", c_code)
  }

  state <- tcc_ffi_create_state(ffi, output = "memory")
  tcc_ffi_compile_state(
    state = state,
    c_code = c_code,
    libraries = ffi$libraries,
    target = paste0("FFI bindings for ", basename(path))
  )

  # Create compiled object
  compiled <- tcc_compiled_object(
    state,
    ffi$symbols,
    "memory",
    ffi$structs,
    ffi$unions,
    ffi$enums,
    ffi$globals,
    ffi$container_of,
    ffi$field_addr,
    ffi$struct_raw_access,
    ffi$introspect
  )

  # Store enough to re-link after deserialization
  compiled$.ffi <- ffi
  compiled$.link_args <- list(
    path = path,
    symbols = symbols,
    headers = headers,
    libs = libs,
    lib_paths = lib_paths,
    include_paths = include_paths,
    user_code = user_code
  )

  if (verbose) {
    message("Successfully loaded ", length(ffi$symbols), " symbols")
  }

  compiled
}

#' CString S3 Class
#'
#' Wrapper around a C string pointer with an optional cached R copy.
#' Ownership follows the underlying external pointer; this wrapper does not add
#' finalizer or freeing behavior on top of that pointer.
#'
#' @param ptr External pointer to C string
#' @param clone Whether to clone the string immediately (safe for R use)
#' @param owned Currently unused. Reserved for future finalizer support.
#' @return A tcc_cstring object
#' @export
tcc_cstring_object <- function(ptr, clone = TRUE, owned = FALSE) {
  if (!inherits(ptr, "externalptr")) {
    stop("Expected external pointer", call. = FALSE)
  }

  obj <- list(
    ptr = ptr,
    owned = owned,
    cached_string = NULL
  )

  if (clone) {
    # Immediately copy to R string
    addr <- get_external_ptr_addr(ptr)
    if (addr > 0) {
      # Read C string at address
      obj$cached_string <- tcc_read_cstring(ptr)
    }
  }

  # owned is currently reserved; the pointer itself may already carry
  # ownership/finalizer semantics (for example via tcc_cstring()).

  class(obj) <- "tcc_cstring"
  obj
}

#' Convert a `tcc_cstring` object to an R string
#'
#' @param x A `tcc_cstring` object.
#' @param ... Ignored.
#' @return A character scalar containing the string value. Returns the
#'   cached R copy when available; otherwise reads the current NUL-terminated
#'   C string from `x$ptr`.
#' @export
as.character.tcc_cstring <- function(x, ...) {
  if (!is.null(x$cached_string)) {
    return(x$cached_string)
  }
  tcc_read_cstring(x$ptr)
}

#' Print a `tcc_cstring` object
#'
#' @param x A `tcc_cstring` object.
#' @param ... Ignored.
#' @return The input `tcc_cstring` object, invisibly. Called for its side
#'   effect of printing the current string value.
#' @export
print.tcc_cstring <- function(x, ...) {
  str <- as.character(x)
  cat("<tcc_cstring> \"", str, "\"\n", sep = "")
  invisible(x)
}

# Helper to read C string from external pointer
read_c_string <- function(ptr) {
  tcc_read_cstring(ptr)
}

# ============================================================================
# Struct, Union, Enum Support
# ============================================================================

#' Declare struct for FFI helper generation
#'
#' Generate R-callable helpers for struct allocation, field access,
#' and pointer management. The struct must be defined in a header.
#'
#' @param ffi A tcc_ffi object
#' @param name Struct name (as defined in C header)
#' @param accessors Named list of field accessors where
#'   names are field names and values are FFI types (e.g., list(x="f64", y="f64")).
#'   Named nested struct fields can use `"struct:<name>"` to generate borrowed
#'   nested-view getters and copy-in setters (for example `child = "struct:child"`).
#' @return Updated tcc_ffi object
#' @export
#' @examples
#' \dontrun{
#' ffi <- tcc_ffi() |>
#'   tcc_header("#include <point.h>") |>
#'   tcc_struct("point", list(x = "f64", y = "f64", id = "i32"))
#' }
tcc_struct <- function(ffi, name, accessors) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  if (!is.character(name) || length(name) != 1) {
    stop("Struct name must be a single string", call. = FALSE)
  }

  if (!is.list(accessors) && !is.character(accessors)) {
    stop(
      "Accessors must be a named list or named character vector",
      call. = FALSE
    )
  }

  # Convert character vector to list if needed
  if (is.character(accessors)) {
    accessors <- as.list(accessors)
  }

  # Validate field types
  for (field_name in names(accessors)) {
    field_type <- accessors[[field_name]]
    # Handle complex field specs like list(type="cstring", size=20)
    if (is.list(field_type)) {
      type_name <- field_type$type %||% "ptr"
      if (isTRUE(field_type$array)) {
        if (is.null(field_type$size) || !is.numeric(field_type$size)) {
          stop(
            "Array field '",
            field_name,
            "' must include numeric 'size'",
            call. = FALSE
          )
        }
      }
    } else {
      type_name <- field_type
    }

    # Allow "struct:name" for nested structs
    if (!grepl("^struct:", type_name)) {
      check_ffi_type(
        type_name,
        paste0("struct '", name, "' field '", field_name, "'")
      )
    }
  }

  # Store struct definition
  if (is.null(ffi$structs)) {
    ffi$structs <- list()
  }
  ffi$structs[[name]] <- accessors

  ffi
}

#' Declare union for FFI helper generation
#'
#' Generate R-callable helpers for union allocation and member access.
#' The union must be defined in a header.
#'
#' @param ffi A tcc_ffi object
#' @param name Union name (as defined in C header)
#' @param members Named list of union members with FFI types
#' @param active Default active member for accessors
#' @return Updated tcc_ffi object
#' @export
#' @examples
#' \dontrun{
#' ffi <- tcc_ffi() |>
#'   tcc_union("data_variant",
#'     members = list(as_int = "i32", as_float = "f32"),
#'     active = "as_int"
#'   )
#' }
tcc_union <- function(ffi, name, members, active = NULL) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  # Validate member types
  for (mem_name in names(members)) {
    mem_type <- members[[mem_name]]
    if (is.list(mem_type)) {
      # Complex member like struct inside union
      if (!is.null(mem_type$type) && mem_type$type == "struct") {
        # Valid - nested struct
      } else {
        type_name <- mem_type$type %||% "ptr"
        check_ffi_type(
          type_name,
          paste0("union '", name, "' member '", mem_name, "'")
        )
      }
    } else {
      check_ffi_type(
        mem_type,
        paste0("union '", name, "' member '", mem_name, "'")
      )
    }
  }

  if (is.null(ffi$unions)) {
    ffi$unions <- list()
  }
  ffi$unions[[name]] <- list(members = members, active = active)

  ffi
}

#' Declare enum for FFI helper generation
#'
#' Generate R-callable helpers for enum constants and type conversions.
#' The enum must be defined in a header.
#'
#' @param ffi A tcc_ffi object
#' @param name Enum name (as defined in C header)
#' @param constants Character vector of constant names to export
#' @param export_constants Whether to export enum constants as R functions
#' @return Updated tcc_ffi object
#' @export
#' @examples
#' \dontrun{
#' ffi <- tcc_ffi() |>
#'   tcc_header("#include <errors.h>") |>
#'   tcc_enum("error_code", constants = c("OK", "ERROR"), export_constants = TRUE)
#' }
tcc_enum <- function(ffi, name, constants = NULL, export_constants = FALSE) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  if (is.null(ffi$enums)) {
    ffi$enums <- list()
  }
  ffi$enums[[name]] <- list(
    constants = constants,
    export_constants = export_constants || !is.null(constants)
  )

  ffi
}

#' Generate container_of helper for struct member
#'
#' Creates a function that recovers the parent struct pointer from
#' a pointer to one of its members. This is the classic Linux kernel
#' container_of macro made accessible from R.
#'
#' @param ffi A tcc_ffi object
#' @param struct_name Struct name
#' @param member_name Member field name to compute offset from
#' @return Updated tcc_ffi object
#' @export
#' @examples
#' \dontrun{
#' ffi <- tcc_ffi() |>
#'   tcc_struct("student", list(id = "i32", marks = "i32")) |>
#'   tcc_container_of("student", "marks") # Creates struct_student_from_marks()
#' }
tcc_container_of <- function(ffi, struct_name, member_name) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  field_spec <- rtinycc_struct_field_spec(ffi, struct_name, member_name)
  if (rtinycc_is_bitfield_spec(field_spec)) {
    stop("container_of does not support bitfield members", call. = FALSE)
  }

  if (is.null(ffi$container_of)) {
    ffi$container_of <- list()
  }
  ffi$container_of[[struct_name]] <- c(
    ffi$container_of[[struct_name]],
    member_name
  )

  ffi
}

#' Generate field address getter helpers
#'
#' Creates functions that return pointers to specific struct fields.
#' Useful for passing field pointers to C functions or for container_of.
#'
#' @param ffi A tcc_ffi object
#' @param struct_name Struct name
#' @param fields Character vector of field names
#' @return Updated tcc_ffi object
#' @export
#' @examples
#' \dontrun{
#' ffi <- tcc_ffi() |>
#'   tcc_struct("point", list(x = "f64", y = "f64")) |>
#'   tcc_field_addr("point", c("x", "y")) # point_x_addr(), point_y_addr()
#' }
tcc_field_addr <- function(ffi, struct_name, fields) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  for (field_name in fields) {
    field_spec <- rtinycc_struct_field_spec(ffi, struct_name, field_name)
    if (rtinycc_is_bitfield_spec(field_spec)) {
      stop("field_addr does not support bitfield members", call. = FALSE)
    }
  }

  if (is.null(ffi$field_addr)) {
    ffi$field_addr <- list()
  }
  ffi$field_addr[[struct_name]] <- c(
    ffi$field_addr[[struct_name]],
    fields
  )

  ffi
}

#' Enable raw byte access for struct
#'
#' Generates helper functions to read/write raw bytes from struct memory.
#' Useful for bitwise operations, debugging, or manual serialization.
#'
#' @param ffi A tcc_ffi object
#' @param struct_name Struct name
#' @return Updated tcc_ffi object
#' @export
tcc_struct_raw_access <- function(ffi, struct_name) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  if (is.null(ffi$struct_raw_access)) {
    ffi$struct_raw_access <- character()
  }
  ffi$struct_raw_access <- c(ffi$struct_raw_access, struct_name)

  ffi
}

#' Enable introspection helpers
#'
#' Generates sizeof, alignof, and offsetof helper functions for
#' structs, unions, and enums. Useful for debugging or when you need
#' to know C layout information from R.
#'
#' @param ffi A tcc_ffi object
#' @return Updated tcc_ffi object
#' @export
tcc_introspect <- function(ffi) {
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  ffi$introspect <- TRUE
  ffi
}

library(tinytest)
library(Rtinycc)

# ── Fork safety ──────────────────────────────────────────────────────────────
# TCC-compiled function pointers live in mmap'd memory owned by the parent
# process's TCC state.  After fork(), the child gets a copy-on-write mapping.
# Reads (calls) should work; the memory is still mapped.  But the TCC state
# external pointer itself might become invalid after serialization or if the
# parent frees it.

if (
  .Platform$OS.type != "windows" && requireNamespace("parallel", quietly = TRUE)
) {
  # --- Test 1: simple call across fork ----------------------------------
  ffi <- tcc_ffi() |>
    tcc_source("int add(int a, int b) { return a + b; }") |>
    tcc_bind(add = list(args = list("i32", "i32"), returns = "i32")) |>
    tcc_compile()

  # Verify it works in the parent
  expect_equal(ffi$add(2L, 3L), 5L, info = "Parent: add works before fork")

  # mclapply forks; each child should be able to call the compiled function
  results <- parallel::mclapply(
    1:4,
    function(i) {
      tryCatch(ffi$add(i, 10L), error = function(e) e)
    },
    mc.cores = 2
  )

  fork_ok <- vapply(results, is.integer, logical(1))
  expect_true(
    all(fork_ok),
    info = "Fork: compiled FFI callable works in forked children"
  )
  if (all(fork_ok)) {
    expect_equal(
      unlist(results),
      c(11L, 12L, 13L, 14L),
      info = "Fork: correct results from forked children"
    )
  }

  # --- Test 2: struct accessors across fork -----------------------------
  ffi2 <- tcc_ffi() |>
    tcc_source("struct pt { double x; double y; };") |>
    tcc_struct("pt", accessors = c(x = "f64", y = "f64")) |>
    tcc_compile()

  results2 <- parallel::mclapply(
    1:2,
    function(i) {
      tryCatch(
        {
          p <- ffi2$struct_pt_new()
          ffi2$struct_pt_set_x(p, as.double(i))
          val <- ffi2$struct_pt_get_x(p)
          ffi2$struct_pt_free(p)
          val
        },
        error = function(e) e
      )
    },
    mc.cores = 2
  )

  fork_struct_ok <- vapply(results2, is.double, logical(1))
  expect_true(
    all(fork_struct_ok),
    info = "Fork: struct accessors work in forked children"
  )

  # --- Test 2b: union accessors across fork ------------------------------
  ffi_union <- tcc_ffi() |>
    tcc_source("union duo { int i; float f; };") |>
    tcc_union(
      "duo",
      members = list(i = "i32", f = "f32"),
      active = "i"
    ) |>
    tcc_compile()

  results2b <- parallel::mclapply(
    1:2,
    function(i) {
      tryCatch(
        {
          u <- ffi_union$union_duo_new()
          ffi_union$union_duo_set_i(u, i * 10L)
          val <- ffi_union$union_duo_get_i(u)
          ffi_union$union_duo_free(u)
          val
        },
        error = function(e) e
      )
    },
    mc.cores = 2
  )

  fork_union_ok <- vapply(results2b, is.integer, logical(1))
  expect_true(
    all(fork_union_ok),
    info = "Fork: union accessors work in forked children"
  )
  if (all(fork_union_ok)) {
    expect_equal(
      unlist(results2b),
      c(10L, 20L),
      info = "Fork: union results correct"
    )
  }

  # --- Test 3: callback across fork (may be unsafe) ---------------------
  # Callbacks hold R function pointers which should survive fork
  # since R environments are duplicated by COW
  cb <- tcc_callback(function(x) x + 1, signature = "double (*)(double)")

  ffi3 <- tcc_ffi() |>
    tcc_source(
      "
      double apply_cb(double (*fn)(void* ctx, double), void* ctx, double x) {
        return fn(ctx, x);
      }
    "
    ) |>
    tcc_bind(
      apply_cb = list(
        args = list("callback:double(double)", "ptr", "f64"),
        returns = "f64"
      )
    ) |>
    tcc_compile()

  results3 <- parallel::mclapply(
    1:2,
    function(i) {
      tryCatch(
        ffi3$apply_cb(cb, tcc_callback_ptr(cb), as.double(i)),
        error = function(e) e
      )
    },
    mc.cores = 2
  )

  fork_cb_ok <- vapply(results3, is.double, logical(1))
  expect_true(
    all(fork_cb_ok),
    info = "Fork: callbacks work in forked children"
  )
  if (all(fork_cb_ok)) {
    expect_equal(
      unlist(results3),
      c(2.0, 3.0),
      info = "Fork: callback results correct"
    )
  }
  tcc_callback_close(cb)

  # --- Test 4: parent still works after children exit -------------------
  expect_equal(
    ffi$add(100L, 1L),
    101L,
    info = "Parent: compiled FFI still works after fork children exit"
  )
}


# ── Serialization: auto-recompile ─────────────────────────────────────────────
# tcc_compiled objects contain external pointers (TCC state, function pointers).
# These become nil after serialize/unserialize. The $.tcc_compiled method
# detects this and recompiles transparently from the stored FFI recipe.

ffi_ser <- tcc_ffi() |>
  tcc_source("int square(int x) { return x * x; }") |>
  tcc_bind(square = list(args = list("i32"), returns = "i32")) |>
  tcc_compile()

expect_equal(ffi_ser$square(7L), 49L, info = "Pre-serialization: square works")

# --- Test 5: serialize + unserialize auto-recompiles --------------------
raw_bytes <- serialize(ffi_ser, NULL)
ffi_restored <- unserialize(raw_bytes)

expect_true(
  inherits(ffi_restored, "tcc_compiled"),
  info = "Serialized object retains class"
)

# First access triggers recompilation; result should be correct
expect_equal(
  ffi_restored$square(7L),
  49L,
  info = "Deserialized FFI auto-recompiles and works"
)

# --- Test 6: saveRDS + readRDS auto-recompiles --------------------------
tmp <- tempfile(fileext = ".rds")
saveRDS(ffi_ser, tmp)
ffi_rds <- readRDS(tmp)
unlink(tmp)

expect_equal(
  ffi_rds$square(9L),
  81L,
  info = "readRDS FFI auto-recompiles and works"
)

# --- Test 7: original still works after serialization -------------------
expect_equal(
  ffi_ser$square(7L),
  49L,
  info = "Original FFI still works after serialization"
)

# --- Test 8: explicit tcc_recompile ------------------------------------
ffi_recomp <- unserialize(serialize(ffi_ser, NULL))
tcc_recompile(ffi_recomp)
expect_equal(
  ffi_recomp$square(11L),
  121L,
  info = "Explicit tcc_recompile works"
)

# --- Test 9: tcc_link round-trip ----------------------------------------
find_system_math_library <- function() {
  if (.Platform$OS.type == "windows") {
    return(NULL)
  }

  sysname <- as.character(unname(Sys.info()[["sysname"]]))
  paths <- unique(Rtinycc:::tcc_library_search_paths(sysname))
  # Prefer libm.so (GNU ld script on glibc) over the bare SONAME so TinyCC
  # can resolve through the script. Some minimal containers ship only the
  # versioned file, so keep that as a fallback.
  patterns <- switch(
    sysname,
    Linux = c("^libm\\.so$", "^libm\\.so(\\.[0-9]+)+$"),
    Darwin = c("^libSystem\\.B\\.dylib$", "^libm\\.dylib$"),
    character(0)
  )

  for (pattern in patterns) {
    for (path in paths) {
      if (!dir.exists(path)) {
        next
      }
      candidates <- list.files(
        path,
        pattern = pattern,
        full.names = TRUE
      )
      if (length(candidates) > 0L) {
        return(candidates[[1]])
      }
    }
  }

  NULL
}

math_lib_path <- find_system_math_library()
math <- if (!is.null(math_lib_path)) {
  tryCatch(
    tcc_link(
      math_lib_path,
      symbols = list(
        sqrt = list(args = list("f64"), returns = "f64")
      )
    ),
    error = function(e) NULL
  )
} else {
  NULL
}
if (!is.null(math)) {
  expect_equal(math$sqrt(25.0), 5.0, info = "tcc_link: original works")

  math2 <- unserialize(serialize(math, NULL))
  expect_equal(
    math2$sqrt(144.0),
    12.0,
    info = "tcc_link: auto-recompiles after deserialization"
  )

  if (identical(as.character(unname(Sys.info()[["sysname"]])), "Linux")) {
    tmp_lib_dir <- tempfile("rtinycc-libm-")
    dir.create(tmp_lib_dir)
    on.exit(unlink(tmp_lib_dir, recursive = TRUE), add = TRUE)

    # Regression guard: a full path to a versioned runtime shared object
    # must be linked as that exact file, not collapsed to -lm. We use the
    # actual SONAME file (resolving the ld script if necessary) so the
    # symlink/copy points at a real ELF object.
    src <- math_lib_path
    if (!isTRUE(file.info(src)$isdir)) {
      head_bytes <- tryCatch(
        readChar(src, 32, useBytes = TRUE),
        error = function(e) ""
      )
      if (!startsWith(as.character(head_bytes), "\x7fELF")) {
        soname <- file.path(dirname(src), "libm.so.6")
        if (file.exists(soname)) src <- soname
      }
    }

    isolated_math_path <- file.path(tmp_lib_dir, "libm_rtinycc_test.so.6")
    linked <- file.symlink(src, isolated_math_path)
    if (!isTRUE(linked)) {
      linked <- file.copy(src, isolated_math_path)
    }

    if (isTRUE(linked)) {
      math_exact <- tryCatch(
        tcc_link(
          isolated_math_path,
          symbols = list(
            sqrt = list(args = list("f64"), returns = "f64")
          )
        ),
        error = function(e) NULL
      )
      if (!is.null(math_exact)) {
        expect_equal(
          math_exact$sqrt(36.0),
          6.0,
          info = "tcc_link: versioned Linux shared object path links exactly"
        )
      }
    }
  }
}

# --- Test 10: raw pointers are still dead after deserialization ----------
ptr <- tcc_malloc(8)
raw_ptr <- serialize(ptr, NULL)
ptr_restored <- unserialize(raw_ptr)

res10 <- tryCatch(tcc_ptr_is_null(ptr_restored), error = function(e) e)
expect_true(
  inherits(res10, "error") || isTRUE(res10),
  info = "Deserialized pointer errors on use or reads as NULL"
)
tcc_free(ptr)

# --- Test 11: tcc_state external pointer still dead ---------------------
state <- tcc_state(output = "memory")
tcc_compile_string(state, "int one(void) { return 1; }")
tcc_relocate(state)
raw_state <- serialize(state, NULL)
state2 <- unserialize(raw_state)

res11 <- tryCatch(
  tcc_call_symbol(state2, "one", return = "int"),
  error = function(e) e
)
expect_true(
  inherits(res11, "error"),
  info = "Deserialized TCC state errors on use (no auto-recompile for raw state)"
)

# --- Test 12: callback tokens do not survive deserialization --------------
cb <- tcc_callback(function(x) x + 1L, signature = "int (*)(int)")
cb2 <- unserialize(serialize(cb, NULL))
expect_false(
  tcc_callback_valid(cb2),
  info = "Deserialized callback token is invalid in the new object"
)
expect_error(
  tcc_callback_ptr(cb2),
  info = "Deserialized callback cannot yield a fresh callback ptr"
)
tcc_callback_close(cb)

# --- Test 13: callback ptr wrappers do not survive deserialization --------
cb <- tcc_callback(function(x) x + 1L, signature = "int (*)(int)")
cb_ptr <- tcc_callback_ptr(cb)
cb_ptr2 <- unserialize(serialize(cb_ptr, NULL))
res13 <- tryCatch(tcc_ptr_is_null(cb_ptr2), error = function(e) e)
expect_true(
  inherits(res13, "error") || isTRUE(res13),
  info = "Deserialized callback ptr errors on use or reads as NULL"
)
tcc_callback_close(cb)

library(Rtinycc)

probe_optimize_src <- "
int opt_macro() {
#ifdef __OPTIMIZE__
  return 1;
#else
  return 0;
#endif
}
"

compile_probe_high <- function(ffi) {
  compiled <- ffi |>
    tcc_bind(opt_macro = list(args = list(), returns = "i32")) |>
    tcc_source(probe_optimize_src) |>
    tcc_compile()
  compiled$opt_macro()
}

compile_probe_low <- function(state) {
  expect_equal(tcc_compile_string(state, probe_optimize_src), 0L)
  expect_equal(tcc_relocate(state), 0L)
  tcc_call_symbol(state, "opt_macro", return = "int")
}

# High-level API: options are forwarded into TinyCC state.
expect_equal(
  compile_probe_high(tcc_ffi() |> tcc_options("-O0")),
  0L
)
expect_equal(
  compile_probe_high(tcc_ffi() |> tcc_options(c("-Wall", "-O2"))),
  1L
)
expect_equal(
  compile_probe_high(tcc_ffi() |> tcc_options("-Wall") |> tcc_options("-O2")),
  1L
)

# Input validation and reset behavior.
expect_error(
  tcc_options(tcc_ffi()),
  pattern = "must be provided"
)
expect_error(
  tcc_options(tcc_ffi(), ""),
  pattern = "non-empty"
)
expect_error(
  tcc_options(tcc_ffi(), character(0)),
  pattern = "must not be empty"
)

ffi_reset <- tcc_ffi() |>
  tcc_options("-O2") |>
  tcc_options(NULL)
expect_equal(compile_probe_high(ffi_reset), 0L)

# Low-level state helper exposes tcc_set_options directly.
state_opt <- tcc_state(output = "memory")
rc <- tcc_set_options(state_opt, "-O2")
expect_true(is.integer(rc) || is.numeric(rc))
expect_true(rc >= 0)
expect_equal(compile_probe_low(state_opt), 1L)

expect_error(
  tcc_set_options(state_opt, ""),
  pattern = "non-empty character scalar"
)

# Clean test of struct/union/enum support in Rtinycc
# This file tests the implemented functionality

library(tinytest)
library(Rtinycc)

# Test 1: Basic struct declaration
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct point { double x; double y; };")
    point_fields <- list(x = "f64", y = "f64")
    ffi <- tcc_struct(ffi, "point", point_fields)

    !is.null(ffi$structs) &&
      length(ffi$structs) == 1 &&
      names(ffi$structs)[1] == "point"
  },
  info = "Basic struct declaration works"
)

# Test 2: Struct with container_of
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct box { int id; double value; };")
    box_fields <- list(id = "i32", value = "f64")
    ffi <- tcc_struct(ffi, "box", box_fields)
    ffi <- tcc_container_of(ffi, "box", "value")

    !is.null(ffi$container_of) && "box" %in% names(ffi$container_of)
  },
  info = "container_of declaration works"
)

# Test 3: Struct with introspection
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct item { int x; };")
    ffi <- tcc_struct(ffi, "item", list(x = "i32"))
    ffi <- tcc_introspect(ffi)

    !is.null(ffi$introspect) && ffi$introspect == TRUE
  },
  info = "Introspection declaration works"
)

# Test 4: Full compilation test (minimal)
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "struct test_struct { int a; double b; }; int test_fn() { return 42; }"
    )
    test_fields <- list(a = "i32", b = "f64")
    ffi <- tcc_struct(ffi, "test_struct", test_fields)
    ffi <- tcc_container_of(ffi, "test_struct", "a")
    ffi <- tcc_introspect(ffi)
    ffi <- tcc_bind(ffi, test_fn = list(args = list(), returns = "i32"))

    compiled <- tcc_compile(ffi)

    has_new <- exists("struct_test_struct_new", envir = compiled)
    has_get <- exists("struct_test_struct_get_a", envir = compiled)
    has_from <- exists("struct_test_struct_from_a", envir = compiled)
    result <- compiled$test_fn()

    has_new && has_get && has_from && result == 42
  },
  info = "Full compilation with struct helpers works"
)

# Test 5: Union declaration
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "union data { int i; float f; };")
    ffi <- tcc_union(ffi, "data", list(i = "i32", f = "f32"), active = "i")

    !is.null(ffi$unions) && length(ffi$unions) == 1
  },
  info = "Union declaration works"
)

# Test 6: Enum declaration
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "enum status { OK = 0, ERROR = 1 };")
    ffi <- tcc_enum(ffi, "status", export_constants = FALSE)

    !is.null(ffi$enums) && length(ffi$enums) == 1
  },
  info = "Enum declaration works"
)

# Test 7: Field address declaration
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct node { int val; void* next; };")
    node_fields <- list(val = "i32")
    node_fields[["next"]] <- "ptr"
    ffi <- tcc_struct(ffi, "node", node_fields)
    ffi <- tcc_field_addr(ffi, "node", c("val", "next"))
    !is.null(ffi$field_addr) && "node" %in% names(ffi$field_addr)
  },
  info = "Field address declaration works"
)

# Codegen Soundness Framework
# ===========================================================================
#
# This file replaces grepl-based string tests with compile-and-call round-trip
# proofs. For every FFI type, we generate a C identity function, compile it
# with TinyCC, call it with known values, and verify the output matches.
#
# Sections:
#   1. Scalar input/output round-trips (every scalar FFI type)
#   2. Boundary value tests (min, max, zero, NA rejection)
#   3. Array round-trips (every array FFI type)
#   4. Struct field round-trips (every field type)
#   5. Callback return type round-trips
#   6. PROTECT balance / GC stress tests
#   7. Codegen structural properties (single-eval, no double free)
#
# ---------------------------------------------------------------------------

library(tinytest)
library(Rtinycc)

ffi_semantics <- Rtinycc:::rtinycc_ffi_semantics()
scalar_specs <- Rtinycc:::rtinycc_scalar_soundness_specs()

expect_equal(
  sort(names(ffi_semantics)),
  sort(Rtinycc:::VALID_FFI_TYPES),
  info = "ffi semantics covers every declared base ffi type"
)

# ===========================================================================
# 1. SCALAR ROUND-TRIP: compile identity(x) for each type, call, verify
# ===========================================================================

default_if_null <- function(x, y) {
  if (is.null(x)) {
    y
  } else {
    x
  }
}

eval_case_value <- function(value) {
  if (is.language(value)) {
    eval(value, envir = baseenv())
  } else {
    value
  }
}

make_identity <- function(c_type, ffi_type, c_body = NULL) {
  if (is.null(c_body)) {
    c_body <- sprintf("%s identity(%s x) { return x; }", c_type, c_type)
  }

  tcc_ffi() |>
    tcc_source(c_body) |>
    tcc_bind(identity = list(args = list(ffi_type), returns = ffi_type)) |>
    tcc_compile()
}

make_identity_from_spec <- function(spec) {
  identity <- spec$soundness$identity
  c_body <- default_if_null(
    identity$source,
    sprintf("%s identity(%s x) { return x; }", identity$c_type, identity$c_type)
  )

  make_identity(identity$c_type, spec$ffi_type, c_body = c_body)
}

expect_scalar_roundtrip <- function(spec, actual, expected, label) {
  info <- sprintf("%s round-trip: %s", spec$ffi_type, label)
  comparator <- default_if_null(spec$soundness$comparator, "equal")

  switch(
    comparator,
    equal = expect_equal(actual, expected, info = info),
    tolerance = expect_true(
      abs(actual - expected) < spec$soundness$tolerance,
      info = info
    ),
    stop("Unsupported scalar comparator: ", comparator, call. = FALSE)
  )
}

for (spec in scalar_specs) {
  ffi <- make_identity_from_spec(spec)

  for (case in spec$soundness$roundtrip) {
    input <- eval_case_value(case$value)
    expected <- eval_case_value(default_if_null(case$expected, case$value))
    actual <- ffi$identity(input)
    expect_scalar_roundtrip(spec, actual, expected, case$label)
  }
}

# --- ptr ---
ffi <- tcc_ffi() |>
  tcc_source("void* identity(void* x) { return x; }") |>
  tcc_bind(identity = list(args = list("ptr"), returns = "ptr")) |>
  tcc_compile()
buf <- tcc_malloc(8)
result <- ffi$identity(buf)
expect_equal(
  tcc_ptr_addr(result),
  tcc_ptr_addr(buf),
  info = "ptr round-trip: address preserved"
)
expect_false(
  tcc_ptr_is_owned(result),
  info = "ptr round-trip: returned wrapper is not owned"
)
expect_error(
  tcc_free(result),
  info = "ptr round-trip: explicit free refuses unowned wrapper"
)
tcc_free(buf)

# --- void ---
ffi <- tcc_ffi() |>
  tcc_source(
    "int g_side_effect = 0;\nvoid set_flag(int v) { g_side_effect = v; }"
  ) |>
  tcc_bind(set_flag = list(args = list("i32"), returns = "void")) |>
  tcc_compile()
result <- ffi$set_flag(42L)
expect_true(is.null(result), info = "void return: returns NULL/invisible")

# normalized symbol specs encode soundness-relevant distinctions before codegen
sym_ptr <- Rtinycc:::as_rtinycc_bound_symbol(
  "identity_ptr",
  list(args = list("ptr"), returns = "ptr")
)
expect_true(
  Rtinycc:::is_rtinycc_bound_symbol(sym_ptr),
  info = "normalized symbol spec is classed"
)
expect_identical(
  Rtinycc:::ffi_type_family(sym_ptr$arg_type_info[[1]]),
  "ptr",
  info = "normalized symbol spec preserves ptr arg family"
)
expect_identical(
  Rtinycc:::ffi_type_family(sym_ptr$return_spec$type_info),
  "ptr",
  info = "normalized symbol spec preserves ptr return family"
)

sym_arr <- Rtinycc:::as_rtinycc_bound_symbol(
  "copy_ints",
  list(
    args = list("integer_array", "i32"),
    returns = list(type = "integer_array", length_arg = 2, free = TRUE)
  )
)
expect_identical(
  sym_arr$return_spec$length_arg,
  2,
  info = "normalized symbol spec preserves array return length_arg"
)
expect_true(
  isTRUE(sym_arr$return_spec$free),
  info = "normalized symbol spec preserves array return free flag"
)

sym_cb <- Rtinycc:::as_rtinycc_bound_symbol(
  "call_cb",
  list(args = list("callback:SEXP(SEXP)", "ptr", "sexp"), returns = "sexp")
)
expect_identical(
  Rtinycc:::ffi_type_family(sym_cb$arg_type_info[[1]]),
  "callback",
  info = "normalized symbol spec preserves callback arg family"
)
expect_identical(
  Rtinycc:::ffi_type_family(sym_cb$arg_type_info[[3]]),
  "sexp",
  info = "normalized symbol spec distinguishes callback from sexp pass-through arg"
)

expect_error(
  Rtinycc:::as_rtinycc_bound_symbol(
    "bad_async_sexp",
    list(
      args = list("callback_async:SEXP(int)", "ptr", "i32"),
      returns = "sexp"
    )
  ),
  info = "normalized symbol spec rejects async callback SEXP return before codegen"
)
expect_error(
  Rtinycc:::as_rtinycc_bound_symbol(
    "bad_async_sexp_arg",
    list(
      args = list("callback_async:void(SEXP)", "ptr", "sexp"),
      returns = "void"
    )
  ),
  info = "normalized symbol spec rejects async callback SEXP arg before codegen"
)
expect_error(
  Rtinycc:::as_rtinycc_bound_symbol(
    "bad_nonarray_meta",
    list(
      args = list("i32"),
      returns = list(type = "i32", length_arg = 1, free = TRUE)
    )
  ),
  info = "normalized symbol spec rejects array metadata on non-array returns"
)

# ===========================================================================
# 2. BOUNDARY VALUE TESTS: NA rejection, range errors
# ===========================================================================

for (spec in scalar_specs) {
  if (!length(spec$soundness$reject)) {
    next
  }

  ffi <- make_identity_from_spec(spec)

  for (case in spec$soundness$reject) {
    expect_error(
      ffi$identity(eval_case_value(case$value)),
      info = sprintf("%s rejects %s", spec$ffi_type, case$label)
    )
  }
}

# Arity enforcement
ffi_i32 <- make_identity_from_spec(scalar_specs[["i32"]])
expect_error(ffi_i32$identity(), info = "rejects under-arity")
expect_error(ffi_i32$identity(1L, 2L), info = "rejects over-arity")

# ===========================================================================
# 3. ARRAY ROUND-TRIPS: pass array to C, get it back
# ===========================================================================

# integer_array: pass through C, memcpy back
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    #include <string.h>
    int* copy_ints(int* arr, int n) {
      int* out = (int*)malloc(sizeof(int) * n);
      memcpy(out, arr, sizeof(int) * n);
      return out;
    }
  "
  ) |>
  tcc_bind(
    copy_ints = list(
      args = list("integer_array", "i32"),
      returns = list(type = "integer_array", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

input_ints <- as.integer(c(-100, 0, 1, 42, .Machine$integer.max))
expect_equal(
  ffi$copy_ints(input_ints, length(input_ints)),
  input_ints,
  info = "integer_array round-trip preserves all values"
)

# numeric_array
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    #include <string.h>
    double* copy_doubles(double* arr, int n) {
      double* out = (double*)malloc(sizeof(double) * n);
      memcpy(out, arr, sizeof(double) * n);
      return out;
    }
  "
  ) |>
  tcc_bind(
    copy_doubles = list(
      args = list("numeric_array", "i32"),
      returns = list(type = "numeric_array", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

input_dbls <- c(-1e308, 0.0, 1e-308, 3.14, 1e308)
expect_equal(
  ffi$copy_doubles(input_dbls, length(input_dbls)),
  input_dbls,
  info = "numeric_array round-trip preserves all values"
)

# raw (uint8_t) array
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    #include <string.h>
    unsigned char* copy_raw(unsigned char* buf, int n) {
      unsigned char* out = (unsigned char*)malloc(n);
      memcpy(out, buf, n);
      return out;
    }
  "
  ) |>
  tcc_bind(
    copy_raw = list(
      args = list("raw", "i32"),
      returns = list(type = "raw", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

input_raw <- as.raw(c(0x00, 0x01, 0x7f, 0x80, 0xff))
expect_equal(
  ffi$copy_raw(input_raw, length(input_raw)),
  input_raw,
  info = "raw array round-trip preserves all values"
)

# logical_array
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    #include <string.h>
    int* copy_logicals(int* arr, int n) {
      int* out = (int*)malloc(sizeof(int) * n);
      memcpy(out, arr, sizeof(int) * n);
      return out;
    }
  "
  ) |>
  tcc_bind(
    copy_logicals = list(
      args = list("logical_array", "i32"),
      returns = list(type = "logical_array", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

input_lgl <- c(TRUE, FALSE, TRUE, FALSE)
expect_equal(
  ffi$copy_logicals(input_lgl, length(input_lgl)),
  input_lgl,
  info = "logical_array round-trip preserves all values"
)

# Array input wrappers reject wrong SEXPTYPE before borrowing vector storage.
ffi <- tcc_ffi() |>
  tcc_source(
    "
    void touch_raw(unsigned char* buf) { (void) buf; }
    void touch_ints(int* x) { (void) x; }
    void touch_nums(double* x) { (void) x; }
    void touch_lgl(int* x) { (void) x; }
  "
  ) |>
  tcc_bind(
    touch_raw = list(args = list("raw"), returns = "void"),
    touch_ints = list(args = list("integer_array"), returns = "void"),
    touch_nums = list(args = list("numeric_array"), returns = "void"),
    touch_lgl = list(args = list("logical_array"), returns = "void")
  ) |>
  tcc_compile()

expect_error(
  ffi$touch_raw(1:3),
  info = "raw input rejects non-raw vectors before borrowing"
)
expect_error(
  ffi$touch_ints(c(1, 2, 3)),
  info = "integer_array input rejects non-integer vectors before borrowing"
)
expect_error(
  ffi$touch_nums(1:3),
  info = "numeric_array input rejects non-numeric vectors before borrowing"
)
expect_error(
  ffi$touch_lgl(c(1L, 0L, 1L)),
  info = "logical_array input rejects non-logical vectors before borrowing"
)
# character_array type-check insertion is covered structurally in
# test_ffi_codegen.R. The generated wrapper uses the R-defined
# `const SEXP *` STRING_PTR_RO path (CHARSXP string-cell handles, not char**).
# A runtime compile round-trip is avoided here because TinyCC has shown
# platform-specific instability around small helpers that directly exercise
# this path on macOS CI.

# NULL array return -> R_NilValue
ffi <- tcc_ffi() |>
  tcc_source(
    "
    int* return_null(int n) { return 0; }
  "
  ) |>
  tcc_bind(
    return_null = list(
      args = list("i32"),
      returns = list(type = "integer_array", length_arg = 1, free = FALSE)
    )
  ) |>
  tcc_compile()
expect_true(
  is.null(ffi$return_null(5L)),
  info = "NULL array return gives R NULL"
)

# Zero-length array returns still round-trip as empty R vectors when C returns
# a non-NULL buffer. This guards the wrapper path used in the benchmark
# vignette, where length 0 should yield numeric(0) rather than NULL.
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    double* zero_vec(int n) {
      if (n != 0) return NULL;
      return (double*) malloc(sizeof(double));
    }
  "
  ) |>
  tcc_bind(
    zero_vec = list(
      args = list("i32"),
      returns = list(type = "numeric_array", length_arg = 1, free = TRUE)
    )
  ) |>
  tcc_compile()
expect_equal(
  ffi$zero_vec(0L),
  numeric(0),
  info = "zero-length array return with non-NULL buffer yields empty R vector"
)

# ===========================================================================
# 4. STRUCT FIELD ROUND-TRIPS: set field, read it back for each type
# ===========================================================================

ffi <- tcc_ffi() |>
  tcc_source(
    "
    struct all_types {
      int8_t   v_i8;
      int16_t  v_i16;
      int32_t  v_i32;
      int64_t  v_i64;
      uint8_t  v_u8;
      uint16_t v_u16;
      uint32_t v_u32;
      uint64_t v_u64;
      float    v_f32;
      double   v_f64;
      bool     v_bool;
    };
  "
  ) |>
  tcc_struct(
    "all_types",
    accessors = c(
      v_i8 = "i8",
      v_i16 = "i16",
      v_i32 = "i32",
      v_i64 = "i64",
      v_u8 = "u8",
      v_u16 = "u16",
      v_u32 = "u32",
      v_u64 = "u64",
      v_f32 = "f32",
      v_f64 = "f64",
      v_bool = "bool"
    )
  ) |>
  tcc_bind() |>
  tcc_compile()

s <- ffi$struct_all_types_new()

# i8 field
ffi$struct_all_types_set_v_i8(s, 42L)
expect_equal(
  ffi$struct_all_types_get_v_i8(s),
  42L,
  info = "struct i8 field round-trip"
)

# i16 field
ffi$struct_all_types_set_v_i16(s, -1000L)
expect_equal(
  ffi$struct_all_types_get_v_i16(s),
  -1000L,
  info = "struct i16 field round-trip"
)

# i32 field
ffi$struct_all_types_set_v_i32(s, .Machine$integer.max)
expect_equal(
  ffi$struct_all_types_get_v_i32(s),
  .Machine$integer.max,
  info = "struct i32 field round-trip"
)

# i64 field (through double)
ffi$struct_all_types_set_v_i64(s, 2^52)
expect_equal(
  ffi$struct_all_types_get_v_i64(s),
  2^52,
  info = "struct i64 field round-trip"
)

# u8 field
ffi$struct_all_types_set_v_u8(s, 255L)
expect_equal(
  ffi$struct_all_types_get_v_u8(s),
  255L,
  info = "struct u8 field round-trip"
)

# u16 field
ffi$struct_all_types_set_v_u16(s, 65535L)
expect_equal(
  ffi$struct_all_types_get_v_u16(s),
  65535L,
  info = "struct u16 field round-trip"
)

# u32 field (through double)
ffi$struct_all_types_set_v_u32(s, 4294967295)
expect_equal(
  ffi$struct_all_types_get_v_u32(s),
  4294967295,
  info = "struct u32 field round-trip"
)

# u64 field (through double)
ffi$struct_all_types_set_v_u64(s, 2^53)
expect_equal(
  ffi$struct_all_types_get_v_u64(s),
  2^53,
  info = "struct u64 field round-trip"
)

# f32 field
ffi$struct_all_types_set_v_f32(s, 3.14)
expect_true(
  abs(ffi$struct_all_types_get_v_f32(s) - 3.14) < 1e-5,
  info = "struct f32 field round-trip (f32 precision)"
)

# f64 field
ffi$struct_all_types_set_v_f64(s, 3.141592653589793)
expect_equal(
  ffi$struct_all_types_get_v_f64(s),
  3.141592653589793,
  info = "struct f64 field round-trip"
)

# bool field
ffi$struct_all_types_set_v_bool(s, TRUE)
expect_equal(
  ffi$struct_all_types_get_v_bool(s),
  TRUE,
  info = "struct bool field round-trip: TRUE"
)
ffi$struct_all_types_set_v_bool(s, FALSE)
expect_equal(
  ffi$struct_all_types_get_v_bool(s),
  FALSE,
  info = "struct bool field round-trip: FALSE"
)

ffi$struct_all_types_free(s)

# ===========================================================================
# 5. MULTI-ARG WRAPPERS: verify argument ordering is preserved
# ===========================================================================

ffi <- tcc_ffi() |>
  tcc_source(
    "
    double weighted_sum(double a, int32_t b, double c) {
      return a * 1.0 + (double)b * 10.0 + c * 100.0;
    }
  "
  ) |>
  tcc_bind(
    weighted_sum = list(
      args = list("f64", "i32", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

expect_equal(
  ffi$weighted_sum(1.0, 2L, 3.0),
  1.0 * 1.0 + 2.0 * 10.0 + 3.0 * 100.0,
  info = "multi-arg: argument order preserved"
)

# Verify with different values to rule out coincidence
expect_equal(
  ffi$weighted_sum(7.0, 0L, 0.5),
  7.0 * 1.0 + 0.0 * 10.0 + 0.5 * 100.0,
  info = "multi-arg: argument order confirmed"
)

# ===========================================================================
# 6. CALLBACK ROUND-TRIPS: verify C->R->C return conversion per type
# ===========================================================================

close_if_valid <- function(cb) {
  if (inherits(cb, "tcc_callback") && isTRUE(tcc_callback_valid(cb))) {
    tcc_callback_close(cb)
  }
}

# int callback
cb <- tcc_callback(function(x) x * 2L, signature = "int (*)(int)")
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    int call_int_cb(int (*cb)(void*, int), void* ctx, int x) {
      return cb(ctx, x);
    }
  "
  ) |>
  tcc_bind(
    call_int_cb = list(
      args = list("callback:int(int)", "ptr", "i32"),
      returns = "i32"
    )
  ) |>
  tcc_compile()

expect_equal(
  ffi$call_int_cb(cb, cb_ptr, 21L),
  42L,
  info = "callback int round-trip: 21*2=42"
)
close_if_valid(cb)

# double callback
cb <- tcc_callback(function(x) x + 0.5, signature = "double (*)(double)")
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    double call_dbl_cb(double (*cb)(void*, double), void* ctx, double x) {
      return cb(ctx, x);
    }
  "
  ) |>
  tcc_bind(
    call_dbl_cb = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

expect_true(
  abs(ffi$call_dbl_cb(cb, cb_ptr, 2.5) - 3.0) < 1e-12,
  info = "callback double round-trip: 2.5+0.5=3.0"
)
close_if_valid(cb)

# bool callback
cb <- tcc_callback(function(x) !x, signature = "bool (*)(bool)")
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    bool call_bool_cb(bool (*cb)(void*, bool), void* ctx, bool x) {
      return cb(ctx, x);
    }
  "
  ) |>
  tcc_bind(
    call_bool_cb = list(
      args = list("callback:bool(bool)", "ptr", "bool"),
      returns = "bool"
    )
  ) |>
  tcc_compile()

expect_equal(
  ffi$call_bool_cb(cb, cb_ptr, TRUE),
  FALSE,
  info = "callback bool round-trip: !TRUE=FALSE"
)
expect_equal(
  ffi$call_bool_cb(cb, cb_ptr, FALSE),
  TRUE,
  info = "callback bool round-trip: !FALSE=TRUE"
)
close_if_valid(cb)

# void callback (side-effect only)
side_effect <- 0L
cb <- tcc_callback(
  function(x) {
    side_effect <<- x
    NULL
  },
  signature = "void (*)(int)"
)
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    void call_void_cb(void (*cb)(void*, int), void* ctx, int x) {
      cb(ctx, x);
    }
  "
  ) |>
  tcc_bind(
    call_void_cb = list(
      args = list("callback:void(int)", "ptr", "i32"),
      returns = "void"
    )
  ) |>
  tcc_compile()

ffi$call_void_cb(cb, cb_ptr, 99L)
expect_equal(side_effect, 99L, info = "callback void: side effect executed")
close_if_valid(cb)

# pointer callback
ptr_arg_seen_externalptr <- FALSE
ptr_arg_seen_unowned <- FALSE
cb <- tcc_callback(
  function(x) {
    ptr_arg_seen_externalptr <<- inherits(x, "externalptr")
    ptr_arg_seen_unowned <<- !.Call("RC_ptr_is_owned", x, PACKAGE = "Rtinycc")
    x
  },
  signature = "void* (*)(void*)"
)
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    void* call_ptr_cb(void* (*cb)(void*, void*), void* ctx, void* x) {
      return cb(ctx, x);
    }
  "
  ) |>
  tcc_bind(
    call_ptr_cb = list(
      args = list("callback:void*(void*)", "ptr", "ptr"),
      returns = "ptr"
    )
  ) |>
  tcc_compile()

buf_cb <- tcc_malloc(8)
out_ptr <- ffi$call_ptr_cb(cb, cb_ptr, buf_cb)
expect_true(
  ptr_arg_seen_externalptr,
  info = "callback ptr round-trip: callback receives externalptr"
)
expect_true(
  ptr_arg_seen_unowned,
  info = "callback ptr round-trip: callback receives non-owned wrapper"
)
expect_equal(
  tcc_ptr_addr(out_ptr),
  tcc_ptr_addr(buf_cb),
  info = "callback ptr round-trip: address preserved"
)
expect_false(
  .Call("RC_ptr_is_owned", out_ptr, PACKAGE = "Rtinycc"),
  info = "callback ptr round-trip: returned wrapper is not owned"
)
expect_error(
  tcc_free(out_ptr),
  info = "callback ptr round-trip: explicit free refuses unowned wrapper"
)
tcc_free(buf_cb)
close_if_valid(cb)

# SEXP callback
payload <- list(alpha = 1:3, beta = "ok")
cb <- tcc_callback(function(x) x, signature = "SEXP (*)(SEXP)")
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    SEXP call_sexp_cb(SEXP (*cb)(void*, SEXP), void* ctx, SEXP x) {
      return cb(ctx, x);
    }
  "
  ) |>
  tcc_bind(
    call_sexp_cb = list(
      args = list("callback:SEXP(SEXP)", "ptr", "sexp"),
      returns = "sexp"
    )
  ) |>
  tcc_compile()

expect_identical(
  ffi$call_sexp_cb(cb, cb_ptr, payload),
  payload,
  info = "callback SEXP round-trip: object preserved exactly"
)
close_if_valid(cb)

# async pointer callback
ptr_async_seen_externalptr <- FALSE
ptr_async_seen_unowned <- FALSE
cb <- tcc_callback(
  function(x) {
    ptr_async_seen_externalptr <<- inherits(x, "externalptr")
    ptr_async_seen_unowned <<- !.Call("RC_ptr_is_owned", x, PACKAGE = "Rtinycc")
    x
  },
  signature = "void* (*)(void*)"
)
cb_ptr <- tcc_callback_ptr(cb)

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #define _Complex
    #ifdef _WIN32
    #include <windows.h>
    struct ptask { void* (*cb)(void*,void*); void* ctx; void* in; volatile void* out; volatile int done; };
    static struct ptask g_pt;
    static HANDLE g_pth = NULL;
    static DWORD WINAPI pworker(LPVOID p) {
      struct ptask* t = (struct ptask*) p;
      t->out = t->cb(t->ctx, t->in);
      t->done = 1;
      return 0;
    }
    int start_ptr_worker(void* (*cb)(void*,void*), void* ctx, void* x) {
      g_pt.cb = cb; g_pt.ctx = ctx; g_pt.in = x; g_pt.out = NULL; g_pt.done = 0;
      g_pth = CreateThread(NULL, 0, pworker, &g_pt, 0, NULL);
      return g_pth ? 0 : -1;
    }
    int is_ptr_worker_done(void) { return g_pt.done; }
    int join_ptr_worker(void) {
      if (g_pth) { WaitForSingleObject(g_pth, INFINITE); CloseHandle(g_pth); g_pth = NULL; }
      return g_pt.out == g_pt.in;
    }
    #else
    #include <pthread.h>
    struct ptask { void* (*cb)(void*,void*); void* ctx; void* in; volatile void* out; volatile int done; };
    static struct ptask g_pt;
    static pthread_t g_pth;
    static void* pworker(void* p) {
      struct ptask* t = (struct ptask*) p;
      t->out = t->cb(t->ctx, t->in);
      t->done = 1;
      return NULL;
    }
    int start_ptr_worker(void* (*cb)(void*,void*), void* ctx, void* x) {
      g_pt.cb = cb; g_pt.ctx = ctx; g_pt.in = x; g_pt.out = NULL; g_pt.done = 0;
      return pthread_create(&g_pth, NULL, pworker, &g_pt) == 0 ? 0 : -1;
    }
    int is_ptr_worker_done(void) { return g_pt.done; }
    int join_ptr_worker(void) { pthread_join(g_pth, NULL); return g_pt.out == g_pt.in; }
    #endif
  "
  )
if (.Platform$OS.type != "windows") {
  ffi <- tcc_library(ffi, "pthread")
}
ffi <- ffi |>
  tcc_bind(
    start_ptr_worker = list(
      args = list("callback_async:void*(void*)", "ptr", "ptr"),
      returns = "i32"
    ),
    is_ptr_worker_done = list(args = list(), returns = "i32"),
    join_ptr_worker = list(args = list(), returns = "i32")
  ) |>
  tcc_compile()

buf_async <- tcc_malloc(8)
ffi$start_ptr_worker(cb, cb_ptr, buf_async)
for (i in seq_len(50)) {
  tcc_callback_async_drain()
  if (ffi$is_ptr_worker_done() != 0L) {
    break
  }
  Sys.sleep(0.01)
}
expect_true(
  ptr_async_seen_externalptr,
  info = "async callback ptr: callback receives externalptr"
)
expect_true(
  ptr_async_seen_unowned,
  info = "async callback ptr: callback receives non-owned wrapper"
)
expect_equal(
  ffi$join_ptr_worker(),
  1L,
  info = "async callback ptr: returned native address preserved"
)
tcc_free(buf_async)
close_if_valid(cb)

# Wrapper absence invariants: covered wrapper families must not construct
# external pointers or register finalizers directly in generated code.
wrapper_absence_cases <- list(
  list(
    name = "ptr_identity",
    symbols = list(identity = list(args = list("ptr"), returns = "ptr")),
    c_code = "void* identity(void* x) { return x; }",
    required = list("RC_make_unowned_ptr("),
    forbidden = list("R_MakeExternalPtr(", "R_RegisterCFinalizerEx(")
  ),
  list(
    name = "callback_ptr_wrapper",
    symbols = list(
      call_ptr_cb = list(
        args = list("callback:void*(void*)", "ptr", "ptr"),
        returns = "ptr"
      )
    ),
    c_code = "void* call_ptr_cb(void* (*cb)(void*, void*), void* ctx, void* x) { return cb(ctx, x); }",
    required = list("RC_make_unowned_ptr("),
    forbidden = list("R_MakeExternalPtr(", "R_RegisterCFinalizerEx(")
  ),
  list(
    name = "callback_sexp_wrapper",
    symbols = list(
      call_sexp_cb = list(
        args = list("callback:SEXP(SEXP)", "ptr", "sexp"),
        returns = "sexp"
      )
    ),
    c_code = "SEXP call_sexp_cb(SEXP (*cb)(void*, SEXP), void* ctx, SEXP x) { return cb(ctx, x); }",
    required = list(
      "SEXP trampoline_R_wrap_call_sexp_cb_arg1(void* cb, SEXP arg1)"
    ),
    forbidden = list("R_MakeExternalPtr(", "R_RegisterCFinalizerEx(")
  )
)

for (case in wrapper_absence_cases) {
  code <- Rtinycc:::generate_ffi_code(
    symbols = case$symbols,
    c_code = case$c_code
  )
  for (pattern in case$required) {
    expect_true(
      grepl(pattern, code, fixed = TRUE),
      info = sprintf(
        "wrapper absence invariant (%s): requires %s",
        case$name,
        pattern
      )
    )
  }
  for (pattern in case$forbidden) {
    expect_false(
      grepl(pattern, code, fixed = TRUE),
      info = sprintf(
        "wrapper absence invariant (%s): forbids %s",
        case$name,
        pattern
      )
    )
  }
}

# ==========================================================================
# 7. GC STRESS: force GC between allocations to catch PROTECT bugs
# ==========================================================================

# Generate wrappers for several types, compile, then force GC heavily
# between calls. If PROTECT is wrong, this will crash or corrupt.
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    #include <string.h>
    int32_t add_i32(int32_t a, int32_t b) { return a + b; }
    double  add_f64(double a, double b) { return a + b; }
    const char* echo_str(const char* s) { return s; }
    int* make_seq(int n) {
      int* out = (int*)malloc(sizeof(int) * n);
      for (int i = 0; i < n; i++) out[i] = i;
      return out;
    }
  "
  ) |>
  tcc_bind(
    add_i32 = list(args = list("i32", "i32"), returns = "i32"),
    add_f64 = list(args = list("f64", "f64"), returns = "f64"),
    echo_str = list(args = list("cstring"), returns = "cstring"),
    make_seq = list(
      args = list("i32"),
      returns = list(type = "integer_array", length_arg = 1, free = TRUE)
    )
  ) |>
  tcc_compile()

gc_stress_ok <- TRUE
for (i in seq_len(100)) {
  # Force GC before each call
  gc(verbose = FALSE)

  r1 <- ffi$add_i32(i, -i)
  if (r1 != 0L) {
    gc_stress_ok <- FALSE
    break
  }

  gc(verbose = FALSE)
  r2 <- ffi$add_f64(as.double(i), -as.double(i))
  if (r2 != 0.0) {
    gc_stress_ok <- FALSE
    break
  }

  gc(verbose = FALSE)
  r3 <- ffi$echo_str(paste0("iter_", i))
  if (r3 != paste0("iter_", i)) {
    gc_stress_ok <- FALSE
    break
  }

  gc(verbose = FALSE)
  r4 <- ffi$make_seq(5L)
  if (!identical(r4, 0:4)) {
    gc_stress_ok <- FALSE
    break
  }
}
expect_true(
  gc_stress_ok,
  info = "GC stress: 100 iterations with forced GC survived"
)

# ===========================================================================
# 8. CODEGEN STRUCTURAL PROPERTIES: single-eval, PROTECT balance
# ===========================================================================

# Property: array return expressions must NOT be evaluated twice
# We test this by using a C function with side effects as the return expr
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>
    #include <string.h>
    static int call_count = 0;
    int* counted_alloc(int n) {
      call_count++;
      int* out = (int*)malloc(sizeof(int) * n);
      for (int i = 0; i < n; i++) out[i] = call_count;
      return out;
    }
    int get_call_count(void) { return call_count; }
  "
  ) |>
  tcc_bind(
    counted_alloc = list(
      args = list("i32"),
      returns = list(type = "integer_array", length_arg = 1, free = TRUE)
    ),
    get_call_count = list(args = list(), returns = "i32")
  ) |>
  tcc_compile()

result <- ffi$counted_alloc(3L)
count_after <- ffi$get_call_count()
expect_equal(
  count_after,
  1L,
  info = "array return: C function called exactly once (no double-eval)"
)
expect_equal(result, rep(1L, 3), info = "array return: values from single call")

# Call again to confirm counter increments properly
result2 <- ffi$counted_alloc(2L)
count_after2 <- ffi$get_call_count()
expect_equal(
  count_after2,
  2L,
  info = "array return: second call increments counter once"
)

# Property: scalar return with side effects also single-eval
ffi <- tcc_ffi() |>
  tcc_source(
    "
    static int scalar_calls = 0;
    int64_t counted_i64(void) {
      scalar_calls++;
      return (int64_t)scalar_calls;
    }
    int get_scalar_calls(void) { return scalar_calls; }
  "
  ) |>
  tcc_bind(
    counted_i64 = list(args = list(), returns = "i64"),
    get_scalar_calls = list(args = list(), returns = "i32")
  ) |>
  tcc_compile()

r <- ffi$counted_i64()
expect_equal(
  ffi$get_scalar_calls(),
  1L,
  info = "i64 return: C function called exactly once"
)
expect_equal(r, 1.0, info = "i64 return: correct value from single eval")

# cstring return single-eval
ffi <- tcc_ffi() |>
  tcc_source(
    "
    static int str_calls = 0;
    const char* counted_str(void) {
      str_calls++;
      return \"hello\";
    }
    int get_str_calls(void) { return str_calls; }
  "
  ) |>
  tcc_bind(
    counted_str = list(args = list(), returns = "cstring"),
    get_str_calls = list(args = list(), returns = "i32")
  ) |>
  tcc_compile()

r <- ffi$counted_str()
expect_equal(
  ffi$get_str_calls(),
  1L,
  info = "cstring return: C function called exactly once"
)
expect_equal(r, "hello", info = "cstring return: correct value")

# Regression: const-qualified char* returns must compile and box correctly.
ffi <- tcc_ffi() |>
  tcc_source(
    "
    const char* const_message(void) {
      return \"const-ok\";
    }
  "
  ) |>
  tcc_bind(const_message = list(args = list(), returns = "cstring")) |>
  tcc_compile()
expect_equal(
  ffi$const_message(),
  "const-ok",
  info = "const char* return compiles and boxes into an R string"
)

# wrapper argument single-eval
ffi <- tcc_ffi() |>
  tcc_source("int add3(int a, int b, int c) { return a + b + c; }") |>
  tcc_bind(
    add3 = list(args = list("i32", "i32", "i32"), returns = "i32")
  ) |>
  tcc_compile()

arg_force_count <- 0L
mk_arg <- function(value) {
  force(value)
  function() {
    arg_force_count <<- arg_force_count + 1L
    value
  }
}
arg1 <- mk_arg(1L)
arg2 <- mk_arg(2L)
arg3 <- mk_arg(3L)
expect_equal(
  ffi$add3(arg1(), arg2(), arg3()),
  6L,
  info = "wrapper arg single-eval: correct result"
)
expect_equal(
  arg_force_count,
  3L,
  info = "wrapper arg single-eval: each R argument forced once"
)

# ===========================================================================
# 9. CSTRING_ARRAY INPUT: verify element-by-element translation
# ===========================================================================

ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <string.h>
    int find_str(const char** arr, int n, const char* needle) {
      for (int i = 0; i < n; i++) {
        if (arr[i] && strcmp(arr[i], needle) == 0) return i;
      }
      return -1;
    }
  "
  ) |>
  tcc_bind(
    find_str = list(
      args = list("cstring_array", "i32", "cstring"),
      returns = "i32"
    )
  ) |>
  tcc_compile()

haystack <- c("alpha", "beta", "gamma", "delta")
expect_equal(
  ffi$find_str(haystack, 4L, "gamma"),
  2L,
  info = "cstring_array: finds element at index 2"
)
expect_equal(
  ffi$find_str(haystack, 4L, "missing"),
  -1L,
  info = "cstring_array: returns -1 for missing"
)
expect_equal(
  ffi$find_str(haystack, 4L, "alpha"),
  0L,
  info = "cstring_array: finds first element"
)

# ===========================================================================
# 10. ENUM ROUND-TRIP
# ===========================================================================

ffi <- tcc_ffi() |>
  tcc_source(
    "
    enum color { RED = 0, GREEN = 1, BLUE = 2 };
    enum color next_color(enum color c) {
      return (enum color)((c + 1) % 3);
    }
  "
  ) |>
  tcc_enum("color", constants = c("RED", "GREEN", "BLUE")) |>
  tcc_bind(
    next_color = list(
      args = list("enum:color"),
      returns = "enum:color"
    )
  ) |>
  tcc_compile()

expect_equal(ffi$next_color(0L), 1L, info = "enum round-trip: RED->GREEN")
expect_equal(ffi$next_color(1L), 2L, info = "enum round-trip: GREEN->BLUE")
expect_equal(ffi$next_color(2L), 0L, info = "enum round-trip: BLUE->RED")

# Verify enum constant getters
expect_equal(ffi$enum_color_RED(), 0L, info = "enum constant: RED=0")
expect_equal(ffi$enum_color_GREEN(), 1L, info = "enum constant: GREEN=1")
expect_equal(ffi$enum_color_BLUE(), 2L, info = "enum constant: BLUE=2")

# ===========================================================================
# 11. GLOBAL VARIABLE ROUND-TRIPS
# ===========================================================================

ffi <- tcc_ffi() |>
  tcc_source(
    "
    int32_t g_counter = 0;
    double  g_value = 0.0;
  "
  ) |>
  tcc_global("g_counter", "i32") |>
  tcc_global("g_value", "f64") |>
  tcc_compile()

# Initial values
expect_equal(
  ffi$global_g_counter_get(),
  0L,
  info = "global i32: initial value is 0"
)
expect_equal(
  ffi$global_g_value_get(),
  0.0,
  info = "global f64: initial value is 0.0"
)

# Set and get
ffi$global_g_counter_set(42L)
expect_equal(
  ffi$global_g_counter_get(),
  42L,
  info = "global i32 round-trip: set 42, get 42"
)

ffi$global_g_value_set(3.14)
expect_equal(
  ffi$global_g_value_get(),
  3.14,
  info = "global f64 round-trip: set 3.14, get 3.14"
)

# Windows smoke tests — run first (alphabetically) to diagnose JIT issues.
# Each test isolates a specific JIT capability so we can pinpoint failures.

library(tinytest)
library(Rtinycc)

# ---------- 1. Trivial: return a constant integer ----------------------------
expect_equal(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "int get_42(void) { return 42; }")
    ffi <- tcc_bind(ffi, get_42 = list(args = list(), returns = "i32"))
    compiled <- tcc_compile(ffi)
    compiled$get_42()
  },
  42L,
  info = "smoke: constant int return"
)

# ---------- 2. Two-argument function ----------------------------------------
expect_equal(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "int add(int a, int b) { return a + b; }")
    ffi <- tcc_bind(ffi, add = list(args = list("i32", "i32"), returns = "i32"))
    compiled <- tcc_compile(ffi)
    compiled$add(3L, 4L)
  },
  7L,
  info = "smoke: two-arg i32 add"
)

# ---------- 3. Double (f64) --------------------------------------------------
expect_equal(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "double mul(double a, double b) { return a * b; }")
    ffi <- tcc_bind(ffi, mul = list(args = list("f64", "f64"), returns = "f64"))
    compiled <- tcc_compile(ffi)
    compiled$mul(2.5, 4.0)
  },
  10.0,
  info = "smoke: f64 multiply"
)

# ---------- 4. Global variable -----------------------------------------------
expect_equal(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "int global_x = 99;")
    ffi <- tcc_global(ffi, "global_x", "i32")
    compiled <- tcc_compile(ffi)
    compiled$global_global_x_get()
  },
  99L,
  info = "smoke: global variable"
)

# ---------- 5. calloc + free via JIT (no R API, just CRT) --------------------
expect_equal(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "
#include <stdlib.h>
int test_calloc(void) {
    int *p = (int*)calloc(1, sizeof(int));
    if (!p) return -1;
    *p = 123;
    int v = *p;
    free(p);
    return v;
}
"
    )
    ffi <- tcc_bind(ffi, test_calloc = list(args = list(), returns = "i32"))
    compiled <- tcc_compile(ffi)
    compiled$test_calloc()
  },
  123L,
  info = "smoke: calloc/free from CRT"
)

# ---------- 6. R_MakeExternalPtr + R_NilValue (R API data imports) -----------
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "
#include <stddef.h>
void *RC_host_calloc_c(size_t n, size_t size);
void RC_host_free_c(void *ptr);
SEXP test_extptr(void) {
    int *p = (int*)RC_host_calloc_c(1, sizeof(int));
    if (!p) Rf_error(\"OOM\");
    *p = 42;
    return R_MakeExternalPtr(p, R_NilValue, R_NilValue);
}
void free_extptr(SEXP ext) {
    void *p = R_ExternalPtrAddr(ext);
    if (p) {
        RC_host_free_c(p);
        R_ClearExternalPtr(ext);
    }
}
"
    )
    ffi <- tcc_bind(
      ffi,
      test_extptr = list(args = list(), returns = "sexp"),
      free_extptr = list(args = list("sexp"), returns = "void")
    )
    compiled <- tcc_compile(ffi)
    res <- compiled$test_extptr()
    ok <- is(res, "externalptr")
    compiled$free_extptr(res)
    ok
  },
  info = "smoke: R_MakeExternalPtr + R_NilValue"
)

# ---------- 7. owned extptr helper via host finalizer registration -----------
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "
#include <stddef.h>
void *RC_host_calloc_c(size_t n, size_t size);
SEXP RC_make_owned_ptr(void *ptr, SEXP tag);
SEXP test_finalizer(void) {
    int *p = (int*)RC_host_calloc_c(1, sizeof(int));
    if (!p) Rf_error(\"OOM\");
    *p = 99;
    return RC_make_owned_ptr(p, R_NilValue);
}
"
    )
    ffi <- tcc_bind(ffi, test_finalizer = list(args = list(), returns = "sexp"))
    compiled <- tcc_compile(ffi)
    res <- compiled$test_finalizer()
    on.exit(
      if (inherits(res, "externalptr") && !tcc_ptr_is_null(res)) {
        .Call("RC_free", res, PACKAGE = "Rtinycc")
      },
      add = TRUE
    )
    is(res, "externalptr")
  },
  info = "smoke: host-owned extptr helper"
)

# ---------- 8. struct: compile only (no call) ---------------------------------
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct point { int x; int y; };")
    ffi <- tcc_struct(ffi, "point", accessors = c(x = "i32", y = "i32"))
    ffi <- tcc_bind(ffi)
    compiled <- tcc_compile(ffi)
    inherits(compiled, "tcc_compiled")
  },
  info = "smoke: struct compile succeeds"
)

# ---------- 9. struct: new() only --------------------------------------------
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct point { int x; int y; };")
    ffi <- tcc_struct(ffi, "point", accessors = c(x = "i32", y = "i32"))
    ffi <- tcc_bind(ffi)
    compiled <- tcc_compile(ffi)

    p <- compiled$struct_point_new()
    is(p, "externalptr")
  },
  info = "smoke: struct_point_new()"
)

# ---------- 10. struct: set_x() ---------------------------------------------
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct point { int x; int y; };")
    ffi <- tcc_struct(ffi, "point", accessors = c(x = "i32", y = "i32"))
    ffi <- tcc_bind(ffi)
    compiled <- tcc_compile(ffi)

    p <- compiled$struct_point_new()
    p <- compiled$struct_point_set_x(p, 10L)
    is(p, "externalptr")
  },
  info = "smoke: struct_point_set_x()"
)

# ---------- 11. struct: get_x() ---------------------------------------------
expect_equal(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct point { int x; int y; };")
    ffi <- tcc_struct(ffi, "point", accessors = c(x = "i32", y = "i32"))
    ffi <- tcc_bind(ffi)
    compiled <- tcc_compile(ffi)

    p <- compiled$struct_point_new()
    p <- compiled$struct_point_set_x(p, 10L)
    compiled$struct_point_get_x(p)
  },
  10L,
  info = "smoke: struct_point_get_x()"
)

# ---------- 12. struct: free() ----------------------------------------------
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "struct point { int x; int y; };")
    ffi <- tcc_struct(ffi, "point", accessors = c(x = "i32", y = "i32"))
    ffi <- tcc_bind(ffi)
    compiled <- tcc_compile(ffi)

    p <- compiled$struct_point_new()
    compiled$struct_point_free(p)
    TRUE
  },
  info = "smoke: struct_point_free()"
)

library(tinytest)
library(Rtinycc)

ffi_semantics <- Rtinycc:::rtinycc_ffi_semantics()
callback_semantics <- Rtinycc:::rtinycc_callback_semantics()
callback_abi_specs <- Rtinycc:::rtinycc_callback_abi_specs()
composite_semantics <- Rtinycc:::rtinycc_composite_semantics()
composite_codegen_specs <- Rtinycc:::rtinycc_composite_codegen_specs()

expect_equal(
  sort(names(callback_semantics)),
  c("async", "sync"),
  info = "callback semantics matrix covers sync and async callback modes"
)
expect_true(
  isTRUE(callback_semantics$sync$preserves_function),
  info = "sync callback semantics records preserved R function ownership"
)
expect_true(
  identical(callback_semantics$sync$invocation_thread, "main"),
  info = "sync callback semantics records main-thread invocation"
)
expect_true(
  isTRUE(callback_semantics$async$copies$cstring_payload),
  info = "async callback semantics records cstring payload copying"
)
expect_true(
  identical(callback_semantics$async$invocation_thread, "scheduled-on-main"),
  info = "async callback semantics records scheduled main-thread execution"
)

expect_true(
  length(callback_abi_specs$trampoline) >= 5L,
  info = "callback ABI matrix includes synchronous trampoline coverage"
)
expect_true(
  length(callback_abi_specs$wrapper) >= 3L,
  info = "callback ABI matrix includes wrapper-side signature coverage"
)
expect_true(
  length(callback_abi_specs$async_trampoline) >= 1L,
  info = "callback ABI matrix includes async trampoline coverage"
)

expected_composites <- c(
  "struct_owned",
  "struct_field_addr",
  "struct_container_of",
  "struct_raw_access",
  "struct_array_field",
  "union_owned",
  "union_nested_struct_view",
  "enum_i32",
  "global_scalar",
  "bitfield_native",
  "treesitter_header_bindings"
)

expect_true(
  all(expected_composites %in% names(composite_semantics)),
  info = "composite semantics matrix covers structs, unions, enums, globals, and bitfields"
)
expect_true(
  isTRUE(composite_semantics$struct_field_addr$protects_owner),
  info = "field_addr semantics records owner protection"
)
expect_true(
  identical(
    composite_semantics$struct_field_addr$lifetime_model,
    "preserved owner slot"
  ) &&
    isTRUE(composite_semantics$struct_field_addr$survives_gc_with_live_view),
  info = "field_addr semantics records GC-safe borrowed-view lifetime"
)
expect_true(
  isTRUE(composite_semantics$struct_container_of$protects_owner),
  info = "container_of semantics records owner protection"
)
expect_true(
  identical(
    composite_semantics$struct_container_of$lifetime_model,
    "preserved owner chain"
  ) &&
    isTRUE(composite_semantics$struct_container_of$survives_gc_with_live_view),
  info = "container_of semantics records GC-safe owner chain"
)
expect_true(
  isTRUE(composite_semantics$struct_raw_access$read_copy) &&
    isTRUE(composite_semantics$struct_raw_access$write_copy),
  info = "struct raw access semantics records explicit copy-in/copy-out"
)
expect_true(
  identical(
    composite_semantics$union_nested_struct_view$lifetime_model,
    "preserved owner slot"
  ) &&
    isTRUE(
      composite_semantics$union_nested_struct_view$survives_gc_with_live_view
    ),
  info = "union nested struct semantics records GC-safe borrowed-view lifetime"
)
expect_true(
  isTRUE(composite_semantics$union_nested_struct_view$protects_owner),
  info = "union nested struct semantics records owner protection"
)
expect_true(
  identical(composite_semantics$enum_i32$boundary_mode, "i32-like"),
  info = "enum semantics records integer-like wrapper boundary"
)
expect_true(
  isTRUE(composite_semantics$global_scalar$scalar_only) &&
    isTRUE(composite_semantics$global_scalar$arrays_forbidden),
  info = "global semantics records scalar-only restriction"
)
expect_true(
  identical(ffi_semantics$ptr$input$ownership, "caller-defined") &&
    identical(ffi_semantics$ptr$return$ownership, "unchanged") &&
    isTRUE(ffi_semantics$ptr$return$borrow) &&
    !isTRUE(ffi_semantics$ptr$return$copy),
  info = "ptr semantics records borrowed address wrapper with unchanged ownership"
)
expect_true(
  isTRUE(composite_semantics$bitfield_native$compiler_managed),
  info = "bitfield semantics records compiler-managed storage"
)
expect_true(
  isTRUE(
    composite_semantics$bitfield_native$address_helpers_forbidden %||% TRUE
  ),
  info = "bitfield semantics records address-style helper exclusion"
)
expect_true(
  isTRUE(composite_semantics$bitfield_native$container_of_forbidden %||% TRUE),
  info = "bitfield semantics records container_of exclusion"
)
expect_identical(
  composite_semantics$bitfield_native$treesitter_bitfield_type,
  "u8",
  info = "bitfield semantics records treesitter default ffi type"
)
expect_true(
  isTRUE(composite_semantics$bitfield_native$survives_forced_gc),
  info = "bitfield semantics records forced-GC stability expectation"
)
expect_true(
  all(
    c("functions", "structs", "unions", "enums", "globals") %in%
      composite_semantics$treesitter_header_bindings$generates
  ),
  info = "treesitter semantics records generated binding surface"
)
expect_identical(
  composite_semantics$treesitter_header_bindings$nested_struct_mode %||%
    "ptr-like",
  "ptr-like",
  info = "treesitter semantics records current nested struct accessor mode"
)

helper_classification_ffi <- tcc_ffi() |>
  tcc_source(
    "
    struct bitbox { unsigned int flag : 1; };
    struct inner { int x; };
    union wrapper { struct inner inner; int raw; };
    enum color { RED = 0, BLUE = 1 };
  "
  ) |>
  tcc_struct(
    "bitbox",
    accessors = list(flag = list(type = "u8", bitfield = TRUE, width = 1))
  ) |>
  tcc_struct("inner", accessors = c(x = "i32")) |>
  tcc_union(
    "wrapper",
    members = list(inner = list(type = "struct"), raw = "i32"),
    active = "raw"
  ) |>
  tcc_enum("color", constants = c("RED", "BLUE")) |>
  tcc_introspect() |>
  tcc_bind() |>
  tcc_compile()
helper_specs <- get(
  ".helper_specs",
  envir = helper_classification_ffi,
  inherits = FALSE
)
expect_identical(
  Rtinycc:::helper_symbol_operation(helper_specs$struct_bitbox_get_flag),
  "bitfield_getter",
  info = "helper metadata classifies bitfield getters explicitly"
)
expect_identical(
  Rtinycc:::helper_symbol_operation(helper_specs$struct_bitbox_set_flag),
  "bitfield_setter",
  info = "helper metadata classifies bitfield setters explicitly"
)
expect_identical(
  Rtinycc:::helper_symbol_operation(helper_specs$union_wrapper_get_inner),
  "nested_view",
  info = "helper metadata classifies nested union struct getters as nested views"
)

nested_struct_helper_ffi <- tcc_ffi() |>
  tcc_source(
    "
    struct child { int x; };
    struct outer { struct child child; int y; };
  "
  ) |>
  tcc_struct("child", accessors = c(x = "i32")) |>
  tcc_struct("outer", accessors = list(child = "struct:child", y = "i32")) |>
  tcc_bind() |>
  tcc_compile()
nested_struct_helper_specs <- get(
  ".helper_specs",
  envir = nested_struct_helper_ffi,
  inherits = FALSE
)
expect_identical(
  Rtinycc:::helper_symbol_operation(
    nested_struct_helper_specs$struct_outer_get_child
  ),
  "nested_view",
  info = "helper metadata classifies named nested struct getters as nested views"
)
expect_identical(
  Rtinycc:::helper_symbol_operation(
    nested_struct_helper_specs$struct_outer_set_child
  ),
  "nested_setter",
  info = "helper metadata classifies named nested struct setters explicitly"
)
expect_identical(
  Rtinycc:::helper_symbol_operation(helper_specs$enum_color_RED),
  "constant",
  info = "helper metadata classifies enum constant helpers explicitly"
)

render_composite_codegen <- function(spec) {
  do.call(Rtinycc:::generate_ffi_code, spec$generate_args)
}

for (spec in composite_codegen_specs) {
  code <- render_composite_codegen(spec)

  for (pattern in spec$patterns) {
    expect_true(
      grepl(pattern$pattern, code, fixed = isTRUE(pattern$fixed)),
      info = spec$info
    )
  }

  forbidden <- spec$forbidden
  if (is.null(forbidden)) {
    forbidden <- list()
  }

  for (pattern in forbidden) {
    expect_false(
      grepl(pattern$pattern, code, fixed = isTRUE(pattern$fixed)),
      info = spec$info
    )
  }
}

ptr_wrapper_code <- Rtinycc:::generate_ffi_code(
  symbols = list(
    identity = list(args = list("ptr"), returns = "ptr")
  ),
  c_code = "void* identity(void* x) { return x; }"
)
expect_true(
  grepl("RC_make_unowned_ptr(", ptr_wrapper_code, fixed = TRUE),
  info = "ptr wrapper codegen uses host unowned helper"
)
expect_false(
  grepl("return R_MakeExternalPtr(", ptr_wrapper_code, fixed = TRUE),
  info = "ptr wrapper codegen avoids direct external pointer construction"
)

expect_true(
  grepl(
    "RC_make_unowned_ptr(value, R_NilValue)",
    Rtinycc:::rtinycc_scalar_return_rule_body("ptr", "value"),
    fixed = TRUE
  ),
  info = "ptr scalar return codegen routes raw pointer boxing through host helper"
)

# Test enum support in FFI
# Tests: enum constants, enum as return/parameter type, enum introspection

library(tinytest)
library(Rtinycc)

# Helper operation-kind metadata for enum helpers
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source('enum probe { PROBE_ZERO = 0, PROBE_ONE = 1 };') |>
      tcc_enum("probe", constants = c("PROBE_ZERO", "PROBE_ONE")) |>
      tcc_introspect() |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    identical(
      Rtinycc:::helper_symbol_kind(helper_specs$enum_probe_PROBE_ZERO),
      "enum"
    ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$enum_probe_PROBE_ZERO),
        "constant"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$enum_probe_PROBE_ONE),
        "constant"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$enum_probe_sizeof),
        "introspection"
      )
  },
  info = "Enum helper specs carry operation-kind metadata"
)

# Test 1: Basic enum constants
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        '
      enum error_code { ERR_NONE = 0, ERR_INVALID = -1, ERR_TIMEOUT = -2, ERR_BUSY = 1, ERR_NOMEM = 2 };
    '
      ) |>
      tcc_enum(
        "error_code",
        constants = c(
          "ERR_NONE",
          "ERR_INVALID",
          "ERR_TIMEOUT",
          "ERR_BUSY",
          "ERR_NOMEM"
        )
      ) |>
      tcc_bind()

    compiled <- tcc_compile(ffi)

    err_none <- compiled$enum_error_code_ERR_NONE()
    err_invalid <- compiled$enum_error_code_ERR_INVALID()
    err_timeout <- compiled$enum_error_code_ERR_TIMEOUT()
    err_busy <- compiled$enum_error_code_ERR_BUSY()
    err_nomem <- compiled$enum_error_code_ERR_NOMEM()

    err_none == 0 &&
      err_invalid == -1 &&
      err_timeout == -2 &&
      err_busy == 1 &&
      err_nomem == 2
  },
  info = "Enum constants exported to R"
)

# Test 2: Enum as function parameter and return type
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        '
      enum error_code { ERR_NONE = 0, ERR_INVALID = -1, ERR_TIMEOUT = -2, ERR_BUSY = 1, ERR_NOMEM = 2 };
      
      enum error_code do_something(int input) {
        if (input < 0) return ERR_INVALID;
        if (input == 0) return ERR_NONE;
        if (input > 100) return ERR_NOMEM;
        return ERR_BUSY;
      }
      
      const char* error_to_string(enum error_code err) {
        switch(err) {
          case ERR_NONE: return "none";
          case ERR_INVALID: return "invalid";
          case ERR_TIMEOUT: return "timeout";
          case ERR_BUSY: return "busy";
          case ERR_NOMEM: return "nomem";
          default: return "unknown";
        }
      }
    '
      ) |>
      tcc_enum("error_code") |>
      tcc_bind(
        do_something = list(args = list("i32"), returns = "enum:error_code"),
        error_to_string = list(
          args = list("enum:error_code"),
          returns = "cstring"
        )
      )

    compiled <- tcc_compile(ffi)

    err1 <- compiled$do_something(-5)
    err2 <- compiled$do_something(0)
    err3 <- compiled$do_something(50)
    err4 <- compiled$do_something(200)

    str1 <- compiled$error_to_string(err1)
    str2 <- compiled$error_to_string(err2)
    str3 <- compiled$error_to_string(err3)
    str4 <- compiled$error_to_string(err4)

    err1 == -1 &&
      err2 == 0 &&
      err3 == 1 &&
      err4 == 2 &&
      str1 == "invalid" &&
      str2 == "none" &&
      str3 == "busy" &&
      str4 == "nomem"
  },
  info = "Enum as function parameter and return type"
)

# Test 3: Enum introspection
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source('enum error_code { ERR_NONE = 0, ERR_INVALID = -1 };') |>
      tcc_enum("error_code") |>
      tcc_introspect() |>
      tcc_bind()

    compiled <- tcc_compile(ffi)

    size <- compiled$enum_error_code_sizeof()
    size == 4
  },
  info = "Enum introspection - size"
)

# Test 4: Flag enum with bitwise operations
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        '
      enum permission { PERM_READ = 1, PERM_WRITE = 2, PERM_EXEC = 4, PERM_ALL = 7 };
    '
      ) |>
      tcc_enum(
        "permission",
        constants = c("PERM_READ", "PERM_WRITE", "PERM_EXEC", "PERM_ALL")
      ) |>
      tcc_bind()

    compiled <- tcc_compile(ffi)

    perm_read <- compiled$enum_permission_PERM_READ()
    perm_write <- compiled$enum_permission_PERM_WRITE()
    perm_exec <- compiled$enum_permission_PERM_EXEC()
    perm_all <- compiled$enum_permission_PERM_ALL()

    my_perms <- bitwOr(perm_read, perm_write)
    has_read <- bitwAnd(my_perms, perm_read) != 0
    has_exec <- bitwAnd(my_perms, perm_exec) != 0

    perm_read == 1 &&
      perm_write == 2 &&
      perm_exec == 4 &&
      perm_all == 7 &&
      my_perms == 3 &&
      has_read &&
      !has_exec
  },
  info = "Flag enum with R bitwise operations"
)

# Test 5: Multiple enums
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        '
      enum error_code { OK = 0, ERROR = 1 };
      enum status { IDLE = 0, RUNNING = 1 };
    '
      ) |>
      tcc_enum("error_code", constants = c("OK", "ERROR")) |>
      tcc_enum("status", constants = c("IDLE", "RUNNING")) |>
      tcc_bind()

    compiled <- tcc_compile(ffi)

    ok_val <- compiled$enum_error_code_OK()
    idle_val <- compiled$enum_status_IDLE()

    ok_val == 0 && idle_val == 0
  },
  info = "Multiple enums in same context"
)

# Variadic FFI bindings

ffi_var_spec <- tcc_ffi() |>
  tcc_bind(
    sum3 = list(
      args = list("i32"),
      variadic = TRUE,
      varargs = list("i32", "i32", "i32"),
      returns = "i32"
    ),
    add_mix = list(
      args = list("f64"),
      variadic = TRUE,
      varargs = list("i32", "f64"),
      returns = "f64"
    )
  )

expect_true(Rtinycc:::is_rtinycc_bound_symbol(ffi_var_spec$symbols$sum3))
expect_identical(ffi_var_spec$symbols$sum3$varargs_mode, "prefix")
expect_equal(length(ffi_var_spec$symbols$sum3$varargs_type_info), 3L)
expect_true(all(vapply(
  ffi_var_spec$symbols$sum3$varargs_type_info,
  Rtinycc:::is_rtinycc_ffi_type,
  logical(1)
)))
expect_identical(
  Rtinycc:::ffi_type_family(ffi_var_spec$symbols$add_mix$varargs_type_info[[
    2
  ]]),
  "f64"
)

ffi_var <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdarg.h>

    int sum3(int base, ...) {
      va_list ap;
      va_start(ap, base);
      int a = va_arg(ap, int);
      int b = va_arg(ap, int);
      int c = va_arg(ap, int);
      va_end(ap);
      return base + a + b + c;
    }

    double add_mix(double x, ...) {
      va_list ap;
      va_start(ap, x);
      int i = va_arg(ap, int);
      double d = va_arg(ap, double);
      va_end(ap);
      return x + (double)i + d;
    }
  "
  ) |>
  tcc_bind(
    sum3 = list(
      args = list("i32"),
      variadic = TRUE,
      varargs = list("i32", "i32", "i32"),
      returns = "i32"
    ),
    add_mix = list(
      args = list("f64"),
      variadic = TRUE,
      varargs = list("i32", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

expect_equal(ffi_var$sum3(10L, 1L, 2L, 3L), 16L)
expect_equal(ffi_var$add_mix(1.5, 2L, 3.25), 6.75)

expect_error(
  ffi_var$sum3(10L, 1L, 2L),
  info = "variadic binding enforces declared total argument count"
)

# Opt-in prefix arity: allow up to N typed varargs via varargs_min = 0
ffi_var_prefix <- tcc_ffi() |>
  tcc_source(
    "\
    #include <stdarg.h>\
\
    int sum_n(int n, ...) {\
      va_list ap;\
      va_start(ap, n);\
      int s = 0;\
      for (int i = 0; i < n; i++) s += va_arg(ap, int);\
      va_end(ap);\
      return s;\
    }\
  "
  ) |>
  tcc_bind(
    sum_n = list(
      args = list("i32"),
      variadic = TRUE,
      varargs = list("i32", "i32", "i32"),
      varargs_min = 0L,
      returns = "i32"
    )
  ) |>
  tcc_compile()

expect_equal(ffi_var_prefix$sum_n(0L), 0L)
expect_equal(ffi_var_prefix$sum_n(2L, 10L, 20L), 30L)
expect_equal(ffi_var_prefix$sum_n(3L, 1L, 2L, 3L), 6L)

expect_error(
  ffi_var_prefix$sum_n(4L, 1L, 2L, 3L, 4L),
  info = "variadic prefix mode enforces max declared tail"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(args = list("i32"), varargs = list("i32"), returns = "i32")
    ),
  info = "varargs requires variadic=TRUE"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(bad = list(args = list("i32"), variadic = TRUE, returns = "i32")),
  info = "variadic requires varargs or varargs_types"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(
        args = list("i32"),
        variadic = TRUE,
        varargs = list("integer_array"),
        returns = "i32"
      )
    ),
  info = "variadic varargs must be scalar types"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(
        args = list("i32"),
        variadic = TRUE,
        varargs = list("callback:void(int)"),
        returns = "i32"
      )
    ),
  info = "variadic varargs reject callback families at constructor normalization"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(
        args = list("i32"),
        variadic = TRUE,
        varargs_types = list("sexp"),
        returns = "i32"
      )
    ),
  info = "variadic varargs_types reject sexp at constructor normalization"
)

# True variadic mode with allowed types + min/max arity
ffi_var_types_spec <- tcc_ffi() |>
  tcc_bind(
    probe_types = list(
      args = list("i32"),
      variadic = TRUE,
      varargs_types = list("f64", "i32"),
      varargs_min = 1L,
      varargs_max = 2L,
      returns = "f64"
    )
  )
expect_identical(ffi_var_types_spec$symbols$probe_types$varargs_mode, "types")
expect_equal(ffi_var_types_spec$symbols$probe_types$varargs_min, 1L)
expect_equal(ffi_var_types_spec$symbols$probe_types$varargs_max, 2L)
expect_true(all(vapply(
  ffi_var_types_spec$symbols$probe_types$varargs_type_info,
  Rtinycc:::is_rtinycc_ffi_type,
  logical(1)
)))

ffi_var_types <- tcc_ffi() |>
  tcc_source(
    "\
    #include <stdarg.h>\
\
    double probe_types(int tag, ...) {\
      va_list ap;\
      va_start(ap, tag);\
      double out = 0.0;\
      if (tag == 1) out = (double)va_arg(ap, int);\
      else if (tag == 2) out = va_arg(ap, double);\
      else if (tag == 3) out = (double)va_arg(ap, int) + va_arg(ap, double);\
      va_end(ap);\
      return out;\
    }\
  "
  ) |>
  tcc_bind(
    probe_types = list(
      args = list("i32"),
      variadic = TRUE,
      varargs_types = list("f64", "i32"),
      varargs_min = 1L,
      varargs_max = 2L,
      returns = "f64"
    )
  ) |>
  tcc_compile()

expect_equal(ffi_var_types$probe_types(1L, 2L), 2)
expect_equal(ffi_var_types$probe_types(2L, 2.5), 2.5)
expect_equal(ffi_var_types$probe_types(3L, 1L, 2.5), 3.5)

expect_error(
  ffi_var_types$probe_types(0L),
  info = "true variadic mode enforces varargs_min"
)

expect_error(
  ffi_var_types$probe_types(4L, 1L, 2L, 3L),
  info = "true variadic mode enforces varargs_max"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(
        args = list("i32"),
        variadic = TRUE,
        varargs = list("i32"),
        varargs_types = list("i32"),
        returns = "i32"
      )
    ),
  info = "cannot set both varargs and varargs_types"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(
        args = list("i32"),
        variadic = TRUE,
        varargs_types = list("i32"),
        varargs_min = 2L,
        varargs_max = 1L,
        returns = "i32"
      )
    ),
  info = "varargs_min <= varargs_max"
)

expect_error(
  tcc_ffi() |>
    tcc_bind(
      bad = list(
        args = list("i32"),
        variadic = TRUE,
        varargs = list("i32", "i32"),
        varargs_min = 3L,
        returns = "i32"
      )
    ),
  info = "varargs_min must be between 0 and length(varargs)"
)

# Test union support in FFI
# Tests: basic unions, union introspection, multiple unions

library(tinytest)
library(Rtinycc)

force_gc <- function(rounds = 3L) {
  for (i in seq_len(rounds)) {
    gc(verbose = FALSE)
  }
  invisible(NULL)
}

# Test 1: Basic union
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "union data { int32_t i; float f; uint8_t b[4]; };")
    ffi <- tcc_union(
      ffi,
      "data",
      members = list(i = "i32", f = "f32", b = list(type = "raw", size = 4)),
      active = "i"
    )
    ffi <- tcc_bind(ffi)

    compiled <- tcc_compile(ffi)
    u <- compiled$union_data_new()
    u <- compiled$union_data_set_i(u, 42L)
    v <- compiled$union_data_get_i(u)
    compiled$union_data_free(u)
    v == 42L
  },
  info = "Basic union"
)

# Helper operation-kind metadata for union helpers
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source("union probe { int i; float f; unsigned char b[4]; };") |>
      tcc_union(
        "probe",
        members = list(i = "i32", f = "f32", b = list(type = "raw", size = 4)),
        active = "i"
      ) |>
      tcc_introspect() |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    identical(
      Rtinycc:::helper_symbol_kind(helper_specs$union_probe_new),
      "union"
    ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$union_probe_new),
        "constructor"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$union_probe_free),
        "destructor"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$union_probe_get_i),
        "getter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$union_probe_set_i),
        "setter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$union_probe_sizeof),
        "introspection"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$union_probe_alignof),
        "introspection"
      )
  },
  info = "Union helper specs carry operation-kind metadata"
)

# Test 2: Union introspection
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "union data { int i; float f; };")
    ffi <- tcc_union(
      ffi,
      "data",
      members = list(i = "i32", f = "f32"),
      active = "i"
    )
    ffi <- tcc_introspect(ffi)
    ffi <- tcc_bind(ffi)

    compiled <- tcc_compile(ffi)
    s <- compiled$union_data_sizeof()
    a <- compiled$union_data_alignof()
    s == 4 && a == 4
  },
  info = "Union introspection"
)

# Test 3: Multiple unions
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(ffi, "union u1 { int i; }; union u2 { int i; };")
    ffi <- tcc_union(ffi, "u1", members = list(i = "i32"), active = "i")
    ffi <- tcc_union(ffi, "u2", members = list(i = "i32"), active = "i")
    ffi <- tcc_bind(ffi)

    compiled <- tcc_compile(ffi)
    u1 <- compiled$union_u1_new()
    u2 <- compiled$union_u2_new()
    u1 <- compiled$union_u1_set_i(u1, 100L)
    u2 <- compiled$union_u2_set_i(u2, 200L)
    v1 <- compiled$union_u1_get_i(u1)
    v2 <- compiled$union_u2_get_i(u2)
    compiled$union_u1_free(u1)
    compiled$union_u2_free(u2)
    v1 == 100L && v2 == 200L
  },
  info = "Multiple unions"
)

# Test 4: Nested union struct getters are classified as nested views
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct inner {
        int x;
      };

      union wrapper {
        struct inner inner;
        int raw;
      };
    "
      ) |>
      tcc_struct("inner", accessors = c(x = "i32")) |>
      tcc_union(
        "wrapper",
        members = list(inner = list(type = "struct"), raw = "i32"),
        active = "raw"
      ) |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    identical(
      Rtinycc:::helper_symbol_operation(helper_specs$union_wrapper_get_inner),
      "nested_view"
    )
  },
  info = "Nested union struct getters carry nested-view helper metadata"
)

# Test 5: Nested struct view keeps union owner alive across GC
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct inner {
        int x;
        double y;
      };

      union wrapper {
        struct inner inner;
        int raw;
      };
    "
      ) |>
      tcc_struct("inner", accessors = c(x = "i32", y = "f64")) |>
      tcc_union(
        "wrapper",
        members = list(inner = list(type = "struct"), raw = "i32"),
        active = "raw"
      ) |>
      tcc_bind()

    compiled <- tcc_compile(ffi)
    ok <- TRUE

    for (i in seq_len(50)) {
      u <- compiled$union_wrapper_new()
      inner <- compiled$union_wrapper_get_inner(u)
      inner <- compiled$struct_inner_set_x(inner, as.integer(i))
      inner <- compiled$struct_inner_set_y(inner, i / 10)

      rm(u)
      force_gc()

      x_val <- compiled$struct_inner_get_x(inner)
      y_val <- compiled$struct_inner_get_y(inner)
      if (!identical(x_val, as.integer(i)) || abs(y_val - i / 10) > 1e-12) {
        ok <- FALSE
        break
      }

      rm(inner)
      force_gc()
    }

    ok
  },
  info = "Nested union struct view stays usable across GC after owner reference is dropped"
)

# Test 6: Nested struct view from union cannot be freed directly
expect_error(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct inner {
        int x;
      };

      union wrapper {
        struct inner inner;
        int raw;
      };
    "
      ) |>
      tcc_struct("inner", accessors = c(x = "i32")) |>
      tcc_union(
        "wrapper",
        members = list(inner = list(type = "struct"), raw = "i32"),
        active = "raw"
      ) |>
      tcc_bind()

    compiled <- tcc_compile(ffi)
    u <- compiled$union_wrapper_new()
    inner <- compiled$union_wrapper_get_inner(u)
    on.exit(rm(inner, u), add = TRUE)
    compiled$union_wrapper_free(inner)
  },
  pattern = "Cannot free borrowed view",
  info = "Nested union struct view must be released through its owner, not freed directly"
)

# Test bitfield support in FFI
# Bitfields are handled by the C compiler - we just access them like normal fields

library(tinytest)
library(Rtinycc)

force_gc <- function(rounds = 3L) {
  for (i in seq_len(rounds)) {
    gc(verbose = FALSE)
  }
  invisible(NULL)
}

# Bitfield helper metadata and exclusions
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct status {
        unsigned int flag : 1;
        unsigned int code : 6;
      };
    "
      ) |>
      tcc_struct(
        "status",
        accessors = list(
          flag = list(type = "u8", bitfield = TRUE, width = 1),
          code = list(type = "u8", bitfield = TRUE, width = 6)
        )
      ) |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    identical(
      Rtinycc:::helper_symbol_operation(helper_specs$struct_status_get_flag),
      "bitfield_getter"
    ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_status_set_flag),
        "bitfield_setter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_status_get_code),
        "bitfield_getter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_status_set_code),
        "bitfield_setter"
      )
  },
  info = "Bitfield helpers carry explicit bitfield operation kinds"
)

expect_error(
  tcc_ffi() |>
    tcc_source("struct status { unsigned int flag : 1; };") |>
    tcc_struct(
      "status",
      accessors = list(flag = list(type = "u8", bitfield = TRUE, width = 1))
    ) |>
    tcc_field_addr("status", "flag"),
  info = "field_addr rejects bitfield members"
)

expect_error(
  tcc_ffi() |>
    tcc_source("struct status { unsigned int flag : 1; };") |>
    tcc_struct(
      "status",
      accessors = list(flag = list(type = "u8", bitfield = TRUE, width = 1))
    ) |>
    tcc_container_of("status", "flag"),
  info = "container_of rejects bitfield members"
)

# Test 1: Basic bitfield access
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "
    struct status {
      unsigned int flag1 : 1;
      unsigned int flag2 : 1;
      unsigned int code : 6;
      unsigned int value : 8;
    };
  "
    )
    ffi <- tcc_struct(
      ffi,
      "status",
      accessors = c(flag1 = "u8", flag2 = "u8", code = "u8", value = "u8")
    )
    ffi <- tcc_bind(ffi)

    compiled <- tcc_compile(ffi)

    s <- compiled$struct_status_new()
    s <- compiled$struct_status_set_flag1(s, 1)
    s <- compiled$struct_status_set_flag2(s, 0)
    s <- compiled$struct_status_set_code(s, 42)
    s <- compiled$struct_status_set_value(s, 255)

    f1 <- compiled$struct_status_get_flag1(s)
    f2 <- compiled$struct_status_get_flag2(s)
    c <- compiled$struct_status_get_code(s)
    v <- compiled$struct_status_get_value(s)

    compiled$struct_status_free(s)

    f1 == 1 && f2 == 0 && c == 42 && v == 255
  },
  info = "Basic bitfield access"
)

# Test 2: Bitfield size (packed into int)
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "
    struct status {
      unsigned int flag1 : 1;
      unsigned int flag2 : 1;
      unsigned int code : 6;
      unsigned int value : 8;
    };
  "
    )
    ffi <- tcc_struct(ffi, "status", accessors = c(flag1 = "u8"))
    ffi <- tcc_introspect(ffi)
    ffi <- tcc_bind(ffi)

    compiled <- tcc_compile(ffi)

    size <- compiled$struct_status_sizeof()

    # Should be 4 bytes (all bitfields packed into one int)
    size == 4
  },
  info = "Bitfield struct size"
)

# Test 3: Bitfield max values (demonstrating masking)
expect_true(
  {
    ffi <- tcc_ffi()
    ffi <- tcc_source(
      ffi,
      "
    struct status {
      unsigned int flag : 1;  // max 1
      unsigned int code : 6;  // max 63
    };
  "
    )
    ffi <- tcc_struct(ffi, "status", accessors = c(flag = "u8", code = "u8"))
    ffi <- tcc_bind(ffi)

    compiled <- tcc_compile(ffi)

    s <- compiled$struct_status_new()
    s <- compiled$struct_status_set_flag(s, 2) # Should be masked to 0 (2 & 1)
    s <- compiled$struct_status_set_code(s, 100) # Should be masked to 36 (100 & 63)

    f <- compiled$struct_status_get_flag(s)
    c <- compiled$struct_status_get_code(s)

    compiled$struct_status_free(s)

    f == 0 && c == 36
  },
  info = "Bitfield value masking"
)

# Test 4: Bitfield accessors stay stable under forced GC
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct status {
        unsigned int flag : 1;
        unsigned int code : 6;
      };
    "
      ) |>
      tcc_struct("status", accessors = c(flag = "u8", code = "u8")) |>
      tcc_bind()

    compiled <- tcc_compile(ffi)
    ok <- TRUE

    for (i in seq_len(50)) {
      s <- compiled$struct_status_new()

      force_gc()
      s <- compiled$struct_status_set_flag(s, (i - 1L) %% 2L)
      force_gc()
      s <- compiled$struct_status_set_code(s, (i * 3L) %% 64L)
      force_gc()

      flag_val <- compiled$struct_status_get_flag(s)
      force_gc()
      code_val <- compiled$struct_status_get_code(s)

      compiled$struct_status_free(s)

      if (
        !identical(flag_val, as.integer((i - 1L) %% 2L)) ||
          !identical(code_val, as.integer((i * 3L) %% 64L))
      ) {
        ok <- FALSE
        break
      }
    }

    ok
  },
  info = "Bitfield accessors survive forced GC around setter/getter calls"
)

# Tests for global getters in FFI

library(tinytest)
library(Rtinycc)

expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source("int g_count = 1; double g_scale = 2.0;") |>
      tcc_global("g_count", "i32") |>
      tcc_global("g_scale", "f64") |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    identical(
      Rtinycc:::helper_symbol_kind(helper_specs$global_g_count_get),
      "global"
    ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$global_g_count_get),
        "getter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$global_g_count_set),
        "setter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$global_g_scale_get),
        "getter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$global_g_scale_set),
        "setter"
      )
  },
  info = "Global helper specs carry operation-kind metadata"
)

expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
        int global_counter = 7;
        double global_pi = 3.14159;
        void* global_name = 0;

        "
      ) |>
      tcc_global("global_counter", "i32") |>
      tcc_global("global_pi", "f64") |>
      tcc_global("global_name", "ptr")

    tmp <- tempfile()
    con <- file(tmp, open = "wt")
    sink(con)
    sink(con, type = "message")
    on.exit(
      {
        sink(type = "message")
        sink()
        close(con)
        unlink(tmp)
      },
      add = TRUE
    )
    compiled <- tcc_compile(ffi)

    counter <- compiled$global_global_counter_get()
    pi_val <- compiled$global_global_pi_get()
    name_ptr <- tcc_cstring("hello")
    compiled$global_global_name_set(name_ptr)
    name_ptr_out <- compiled$global_global_name_get()
    name <- tcc_read_cstring(name_ptr_out)

    compiled$global_global_counter_set(9L)
    compiled$global_global_pi_set(2.718)

    counter == 7L &&
      abs(pi_val - 3.14159) < 1e-8 &&
      name == "hello" &&
      compiled$global_global_counter_get() == 9L &&
      abs(compiled$global_global_pi_get() - 2.718) < 1e-8
  },
  info = "Global getters return expected values"
)

library(tinytest)
library(Rtinycc)

# High-level artifact modes used to proceed through relocation and expose
# non-callable addresses as ordinary R functions.
expect_error(
  tcc_output(tcc_ffi(), "dll"),
  "only supports output = 'memory'",
  info = "high-level FFI rejects shared-library output"
)
expect_error(
  tcc_output(tcc_ffi(), "exe"),
  "only supports output = 'memory'",
  info = "high-level FFI rejects executable output"
)

bad_output <- tcc_ffi() |>
  tcc_source("int one(void) { return 1; }") |>
  tcc_bind(one = list(args = list(), returns = "i32"))
bad_output$output <- "dll"
expect_error(
  tcc_compile(bad_output),
  "only creates callable in-memory bindings",
  info = "tcc_compile rejects stale or manually modified artifact recipes"
)

memory_state <- tcc_state(output = "memory")
expect_equal(tcc_compile_string(memory_state, "int one(void) { return 1; }"), 0L)
expect_error(
  tcc_output_file(memory_state, tempfile()),
  "requires a non-memory tcc_state",
  info = "memory states cannot be written as artifacts"
)
expect_error(
  tcc_get_symbol(memory_state, "one"),
  "requires a relocated tcc_state",
  info = "symbols cannot escape before executable memory is finalized"
)
expect_equal(tcc_relocate(memory_state), 0L)
expect_equal(tcc_call_symbol(memory_state, "one", return = "int"), 1L)
expect_error(
  tcc_relocate(memory_state),
  "cannot finalize a state more than once",
  info = "memory state finalization is single-shot"
)
expect_error(
  tcc_compile_string(memory_state, "int two(void) { return 2; }"),
  "cannot modify a finalized tcc_state",
  info = "relocated states reject further compilation"
)

object_path <- tempfile(fileext = ".o")
object_state <- tcc_state(output = "obj")
expect_equal(tcc_compile_string(object_state, "int one(void) { return 1; }"), 0L)
expect_equal(tcc_output_file(object_state, object_path), 0L)
expect_true(
  file.exists(object_path) && file.info(object_path)$size > 0,
  info = "low-level object output produces a non-empty artifact"
)
expect_error(
  tcc_output_file(object_state, object_path),
  "cannot finalize a state more than once",
  info = "artifact state finalization is single-shot"
)
expect_error(
  tcc_relocate(object_state),
  "requires a tcc_state with output = 'memory'",
  info = "non-memory states cannot enter the relocation path"
)
expect_error(
  tcc_get_symbol(object_state, "one"),
  "requires a tcc_state with output = 'memory'",
  info = "non-memory symbols cannot be exposed as callable pointers"
)
unlink(object_path)

expect_error(
  tcc_global(tcc_ffi(), "name", "cstring"),
  "use ptr with explicitly owned storage",
  info = "global cstring setters cannot retain borrowed R strings"
)
expect_error(
  tcc_struct(tcc_ffi(), "bad", accessors = c(name = "cstring")),
  "cstring pointer fields are unsafe",
  info = "struct cstring pointers require explicit ownership"
)
expect_error(
  tcc_union(tcc_ffi(), "bad", members = list(name = "cstring")),
  "cstring pointer members are unsafe",
  info = "union cstring pointers require explicit ownership"
)
expect_error(
  tcc_struct(
    tcc_ffi(),
    "bad_size",
    accessors = list(bytes = list(type = "u8", size = 1.5, array = TRUE))
  ),
  "positive integer 'size'",
  info = "composite array sizes must be positive integers"
)
expect_error(
  tcc_struct(
    tcc_ffi(),
    "bad_string_array",
    accessors = list(name = list(type = "cstring", size = 8, array = TRUE))
  ),
  "without array = TRUE",
  info = "fixed cstring fields do not use numeric array element accessors"
)

# One compilation covers package-header discovery, checked composite setters,
# type-tag enforcement, and symbol ownership.
compiled <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdint.h>
    #include <stdbool.h>
    #include <rtinycc/pt.h>

    struct checked_values {
      int8_t i8;
      int16_t i16;
      int32_t i32;
      int64_t i64;
      uint8_t u8;
      uint16_t u16;
      uint32_t u32;
      uint64_t u64;
      bool flag;
      uint8_t bytes[2];
      char name[8];
    };

    union checked_union {
      int32_t i32;
      uint8_t u8;
    };

    int forty_two(void) { return 42; }
    "
  ) |>
  tcc_struct(
    "checked_values",
    accessors = list(
      i8 = "i8",
      i16 = "i16",
      i32 = "i32",
      i64 = "i64",
      u8 = "u8",
      u16 = "u16",
      u32 = "u32",
      u64 = "u64",
      flag = "bool",
      bytes = list(type = "u8", size = 2, array = TRUE),
      name = list(type = "cstring", size = 8)
    )
  ) |>
  tcc_union(
    "checked_union",
    members = list(i32 = "i32", u8 = "u8")
  ) |>
  tcc_bind(forty_two = list(args = list(), returns = "i32")) |>
  tcc_compile()

s <- compiled$struct_checked_values_new()
setter_rejections <- list(
  list(compiled$struct_checked_values_set_i8, 128L, "i8 out of range"),
  list(compiled$struct_checked_values_set_i16, 32768L, "i16 out of range"),
  list(compiled$struct_checked_values_set_i32, NA_integer_, "integer value is NA"),
  list(compiled$struct_checked_values_set_i64, 1.5, "i64 requires integer value"),
  list(compiled$struct_checked_values_set_u8, -1L, "u8 out of range"),
  list(compiled$struct_checked_values_set_u16, 65536L, "u16 out of range"),
  list(compiled$struct_checked_values_set_u32, 1.5, "u32 requires integer value"),
  list(compiled$struct_checked_values_set_u64, -1, "u64 out of range"),
  list(compiled$struct_checked_values_set_flag, NA, "logical value is NA")
)
for (case in setter_rejections) {
  expect_error(case[[1]](s, case[[2]]), case[[3]])
}
expect_error(
  compiled$struct_checked_values_set_bytes_elt(s, 0L, 256L),
  "u8 out of range",
  info = "array-field setters use scalar range checks"
)
compiled$struct_checked_values_set_name(s, "abcdefghijk")
expect_equal(
  compiled$struct_checked_values_get_name(s),
  "abcdefg",
  info = "fixed cstring fields copy and terminate within their declared size"
)
expect_error(
  compiled$struct_checked_values_set_name(s, NA_character_),
  "cstring array field cannot be NA",
  info = "fixed cstring fields reject NA rather than storing borrowed data"
)

u <- compiled$union_checked_union_new()
expect_error(
  compiled$union_checked_union_set_u8(u, -1L),
  "u8 out of range",
  info = "union setters use scalar range checks"
)

foreign <- tcc_malloc(64)
expect_error(
  compiled$struct_checked_values_get_i8(foreign),
  "expected pointer tagged 'struct_checked_values'",
  info = "struct getter rejects a foreign pointer"
)
expect_error(
  compiled$struct_checked_values_free(foreign),
  "expected pointer tagged 'struct_checked_values'",
  info = "struct destructor refuses a foreign allocation"
)
expect_error(
  compiled$union_checked_union_get_i32(s),
  "expected pointer tagged 'union_checked_union'",
  info = "union getter rejects a struct pointer"
)
tcc_free(foreign)

compiled$struct_checked_values_free(s)
expect_error(
  compiled$struct_checked_values_get_i8(s),
  "NULL or already freed",
  info = "composite access rejects a released pointer"
)
expect_error(
  compiled$struct_checked_values_free(s),
  "NULL or already freed",
  info = "composite destructor rejects double free"
)
compiled$union_checked_union_free(u)

# The symbol itself must retain the TCC state after the compiled container and
# callable closure are dropped.
callable <- compiled$forty_two
symbol <- environment(callable)$.call_ptr
rm(callable, compiled, s, u)
for (i in 1:3) gc(verbose = FALSE)
expect_equal(
  base::.Call(symbol),
  42L,
  info = "tcc_symbol retains relocated code after its originating object is GC'd"
)

# Test FFI High-Level API
# Tests for tcc_ffi(), tcc_bind(), tcc_compile() with real compilation

library(Rtinycc)

# Test 1: Create FFI context
ffi <- tcc_ffi()
expect_true(inherits(ffi, "tcc_ffi"))
expect_equal(ffi$output, "memory")

# Test 2: Chain operations with native pipe
ffi <- tcc_ffi() |>
  tcc_output("memory") |>
  tcc_bind(
    add = list(args = list("i32", "i32"), returns = "i32")
  )

expect_equal(length(ffi$symbols), 1)
expect_true("add" %in% names(ffi$symbols))
expect_true(Rtinycc:::is_rtinycc_bound_symbol(ffi$symbols$add))
expect_equal(ffi$symbols$add$name, "add")
expect_true(all(vapply(
  ffi$symbols$add$arg_type_info,
  Rtinycc:::is_rtinycc_ffi_type,
  logical(1)
)))
expect_equal(
  Rtinycc:::ffi_type_family(ffi$symbols$add$return_spec$type_info),
  "i32"
)

# Test 3: Compile and call simple function
ffi <- tcc_ffi() |>
  tcc_bind(
    forty_two = list(args = list(), returns = "i32")
  ) |>
  tcc_source(
    "
    int forty_two() {
      return 42;
    }
  "
  ) |>
  tcc_compile()

expect_true(inherits(ffi, "tcc_compiled"))

# Test calling the compiled function
result <- ffi$forty_two()
expect_equal(result, 42L)

# Test 4: Compile and call function with arguments
ffi <- tcc_ffi() |>
  tcc_bind(
    add = list(args = list("i32", "i32"), returns = "i32")
  ) |>
  tcc_source(
    "
    int add(int a, int b) {
      return a + b;
    }
  "
  ) |>
  tcc_compile()

expect_true(inherits(ffi, "tcc_compiled"))

# Test calling with arguments
result <- ffi$add(5L, 3L)
expect_equal(result, 8L)

result <- ffi$add(10L, 20L)
expect_equal(result, 30L)
expect_error(
  ffi$add(1L),
  "Expected 2 arguments, got 1",
  info = "Fixed-arity wrappers enforce under-arity with stable error text"
)
expect_error(
  ffi$add(1L, 2L, 3L),
  "Expected 2 arguments, got 3",
  info = "Fixed-arity wrappers enforce over-arity with stable error text"
)

# Test 5: Compile and call function with double arguments
ffi <- tcc_ffi() |>
  tcc_bind(
    multiply = list(args = list("f64", "f64"), returns = "f64")
  ) |>
  tcc_source(
    "
    double multiply(double a, double b) {
      return a * b;
    }
  "
  ) |>
  tcc_compile()

expect_true(inherits(ffi, "tcc_compiled"))

result <- ffi$multiply(2.5, 4.0)
expect_equal(result, 10.0)

# Test 5b: scalar wrappers use the intended R-side carrier types
ffi_scalar_carriers <- tcc_ffi() |>
  tcc_bind(
    id_i32 = list(args = list("i32"), returns = "i32"),
    id_u16 = list(args = list("u16"), returns = "u16"),
    id_u32 = list(args = list("u32"), returns = "u32"),
    id_i64 = list(args = list("i64"), returns = "i64"),
    id_u64 = list(args = list("u64"), returns = "u64"),
    id_f32 = list(args = list("f32"), returns = "f32"),
    id_f64 = list(args = list("f64"), returns = "f64")
  ) |>
  tcc_source(
    "
    #include <stdint.h>

    int32_t id_i32(int32_t x) { return x; }
    uint16_t id_u16(uint16_t x) { return x; }
    uint32_t id_u32(uint32_t x) { return x; }
    int64_t id_i64(int64_t x) { return x; }
    uint64_t id_u64(uint64_t x) { return x; }
    float id_f32(float x) { return x; }
    double id_f64(double x) { return x; }
  "
  ) |>
  tcc_compile()

expect_identical(typeof(ffi_scalar_carriers$id_i32(7L)), "integer")
expect_identical(typeof(ffi_scalar_carriers$id_u16(65535L)), "integer")
expect_identical(typeof(ffi_scalar_carriers$id_u32(4294967295)), "double")
expect_identical(typeof(ffi_scalar_carriers$id_i64(2^53)), "double")
expect_identical(typeof(ffi_scalar_carriers$id_u64(2^53)), "double")
expect_identical(typeof(ffi_scalar_carriers$id_f32(1.25)), "double")
expect_identical(typeof(ffi_scalar_carriers$id_f64(1.25)), "double")

expect_equal(ffi_scalar_carriers$id_u32(4294967295), 4294967295)
expect_equal(ffi_scalar_carriers$id_i64(2^53), 2^53)
expect_equal(ffi_scalar_carriers$id_u64(2^53), 2^53)

expect_error(
  ffi_scalar_carriers$id_u32(1.5),
  "u32 requires integer value"
)
expect_error(
  ffi_scalar_carriers$id_i64(2^53 + 2),
  "i64 requires exact integer"
)
expect_error(
  ffi_scalar_carriers$id_u64(2^53 + 2),
  "u64 requires exact integer"
)

# Test 6: Test print methods
ffi <- tcc_ffi() |>
  tcc_bind(
    foo = list(args = list("i32"), returns = "f64")
  )

# Capture print output
output <- capture.output(print(ffi))
expect_true(any(grepl("tcc_ffi", output)))
expect_true(any(grepl("foo", output)))

# Test 7: Array return type
ffi_spec <- tcc_ffi() |>
  tcc_bind(
    dup_array = list(
      args = list("integer_array", "i32"),
      returns = list(type = "integer_array", length_arg = 2, free = TRUE)
    )
  )

expect_true(inherits(
  ffi_spec$symbols$dup_array$return_spec,
  "rtinycc_symbol_return_spec"
))
expect_equal(ffi_spec$symbols$dup_array$return_spec$type, "integer_array")
expect_equal(ffi_spec$symbols$dup_array$return_spec$length_arg, 2)
expect_true(isTRUE(ffi_spec$symbols$dup_array$return_spec$free))

ffi <- ffi_spec |>
  tcc_source(
    "
    #include <stdlib.h>
    int* dup_array(int* arr, int n) {
      if (n <= 0) return NULL;
      int* out = (int*)malloc(sizeof(int) * n);
      for (int i = 0; i < n; i++) out[i] = arr[i] * 2;
      return out;
    }
  "
  ) |>
  tcc_compile()

expect_true(inherits(ffi, "tcc_compiled"))
result <- ffi$dup_array(as.integer(c(1, 2, 3)), 3L)
expect_equal(result, c(2L, 4L, 6L))

# Test 8: helper specs are normalized to classed symbol specs too
ffi_helpers <- tcc_ffi() |>
  tcc_source("struct point { int x; int y; };") |>
  tcc_struct("point", accessors = c(x = "i32", y = "i32")) |>
  tcc_compile()
helper_specs <- get(".helper_specs", envir = ffi_helpers, inherits = FALSE)
expect_true(Rtinycc:::is_rtinycc_bound_symbol(helper_specs$struct_point_new))
expect_true(Rtinycc:::is_rtinycc_bound_symbol(helper_specs$struct_point_get_x))
expect_true(inherits(helper_specs$struct_point_new, "rtinycc_helper_symbol"))
expect_identical(
  Rtinycc:::helper_symbol_kind(helper_specs$struct_point_new),
  "struct"
)
expect_identical(
  Rtinycc:::helper_symbol_kind(helper_specs$struct_point_get_x),
  "struct"
)
expect_identical(
  Rtinycc:::helper_symbol_operation(helper_specs$struct_point_new),
  "constructor"
)
expect_identical(
  Rtinycc:::helper_symbol_operation(helper_specs$struct_point_get_x),
  "getter"
)
expect_identical(
  Rtinycc:::ffi_type_family(
    helper_specs$struct_point_get_x$return_spec$type_info
  ),
  "sexp"
)

# Test 9: Missing wrapper bindings fail fast (no partially-broken object)
state_bind_fail <- tcc_state(output = "memory")
expect_equal(
  tcc_compile_string(
    state_bind_fail,
    "
    int R_wrap_existing(void) {
      return 1;
    }
    "
  ),
  0L
)
expect_equal(tcc_relocate(state_bind_fail), 0L)

expect_error(
  Rtinycc:::tcc_compiled_object(
    state_bind_fail,
    symbols = list(missing = list(args = list(), returns = "i32")),
    output = "memory",
    structs = NULL,
    unions = NULL,
    enums = NULL,
    globals = NULL,
    container_of = NULL,
    field_addr = NULL,
    struct_raw_access = NULL,
    introspect = NULL
  ),
  "Failed to bind compiled wrapper symbols",
  info = "Missing wrappers should raise a hard binding error"
)

# Test struct support in FFI
# Tests: allocation, field access, container_of

library(tinytest)
library(Rtinycc)

force_gc <- function(rounds = 3L) {
  for (i in seq_len(rounds)) {
    gc(verbose = FALSE)
  }
  invisible(NULL)
}

# Test 1: Simple struct allocation and field access
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct point {
        double x;
        double y;
        int id;
      };

      struct point* create_point(double x, double y) {
        struct point* p = malloc(sizeof(struct point));
        p->x = x;
        p->y = y;
        p->id = 0;
        return p;
      }

      double point_distance(struct point* a, struct point* b) {
        double dx = a->x - b->x;
        double dy = a->y - b->y;
        return dx * dx + dy * dy;
      }
    "
      ) |>
      tcc_struct("point", accessors = c(x = "f64", y = "f64", id = "i32")) |>
      tcc_bind(
        create_point = list(args = list("f64", "f64"), returns = "ptr"),
        point_distance = list(args = list("ptr", "ptr"), returns = "f64")
      )

    compiled <- tcc_compile(ffi)

    p1 <- compiled$struct_point_new()
    p1 <- compiled$struct_point_set_x(p1, 0.0)
    p1 <- compiled$struct_point_set_y(p1, 0.0)
    p1 <- compiled$struct_point_set_id(p1, 1L)

    p2 <- compiled$struct_point_new()
    p2 <- compiled$struct_point_set_x(p2, 3.0)
    p2 <- compiled$struct_point_set_y(p2, 4.0)
    p2 <- compiled$struct_point_set_id(p2, 2L)

    x1 <- compiled$struct_point_get_x(p1)
    y2 <- compiled$struct_point_get_y(p2)
    dist <- compiled$point_distance(p1, p2)

    compiled$struct_point_free(p1)
    compiled$struct_point_free(p2)

    abs(x1 - 0.0) < 1e-10 && abs(y2 - 4.0) < 1e-10 && abs(dist - 25.0) < 1e-10
  },
  info = "Simple struct allocation, field access, and method call"
)

# Test 2: container_of - get parent struct from member pointer
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct student {
        char name[32];
        int id;
        double grade;
      };
    "
      ) |>
      tcc_struct("student", accessors = c(id = "i32", grade = "f64")) |>
      tcc_container_of("student", "id") |>
      tcc_field_addr("student", "id") |>
      tcc_bind()

    compiled <- tcc_compile(ffi)

    student <- compiled$struct_student_new()
    student <- compiled$struct_student_set_id(student, 42L)
    student <- compiled$struct_student_set_grade(student, 3.14)

    id_ptr <- compiled$struct_student_id_addr(student)
    student2 <- compiled$struct_student_from_id(id_ptr)
    id2 <- compiled$struct_student_get_id(student2)

    rm(student2, id_ptr)
    force_gc()
    compiled$struct_student_free(student)

    id2 == 42L
  },
  info = "container_of from member pointer to parent struct"
)

# Helper operation-kind metadata for struct helpers
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source("struct probe { int x; unsigned char data[4]; };") |>
      tcc_struct(
        "probe",
        accessors = list(
          x = "i32",
          data = list(type = "u8", size = 4, array = TRUE)
        )
      ) |>
      tcc_container_of("probe", "x") |>
      tcc_field_addr("probe", "x") |>
      tcc_struct_raw_access("probe") |>
      tcc_introspect() |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    identical(
      Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_new),
      "constructor"
    ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_free),
        "destructor"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_get_x),
        "getter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_set_x),
        "setter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(
          helper_specs$struct_probe_get_data_elt
        ),
        "array_getter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(
          helper_specs$struct_probe_set_data_elt
        ),
        "array_setter"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_x_addr),
        "field_addr"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_from_x),
        "container_of"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_get_raw),
        "raw_get"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_set_raw),
        "raw_set"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_probe_sizeof),
        "introspection"
      )
  },
  info = "Struct helper specs carry operation-kind metadata"
)

# Test 3: Struct with introspection
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct data {
        int a;
        double b;
        char c;
      };
    "
      ) |>
      tcc_struct("data", accessors = c(a = "i32", b = "f64")) |>
      tcc_introspect() |>
      tcc_bind()

    compiled <- tcc_compile(ffi)

    size <- compiled$struct_data_sizeof()
    align <- compiled$struct_data_alignof()

    size > 0 && align > 0
  },
  info = "Struct introspection - size and alignment"
)

# Test 4: Array element getters/setters
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source("struct buf { unsigned char data[8]; };") |>
      tcc_struct(
        "buf",
        accessors = list(data = list(type = "u8", size = 8, array = TRUE))
      )

    compiled <- tcc_compile(ffi)

    b <- compiled$struct_buf_new()
    compiled$struct_buf_set_data_elt(b, 0L, 200L)
    v <- compiled$struct_buf_get_data_elt(b, 0L)
    compiled$struct_buf_free(b)

    v == 200L
  },
  info = "Array element accessors"
)

# Test 5: Named nested struct fields expose nested views and copy-in setters
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct child {
        int x;
      };

      struct outer {
        struct child child;
        int y;
      };
    "
      ) |>
      tcc_struct("child", accessors = c(x = "i32")) |>
      tcc_struct(
        "outer",
        accessors = list(child = "struct:child", y = "i32")
      ) |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    child <- ffi$struct_child_new()
    child <- ffi$struct_child_set_x(child, 42L)

    outer <- ffi$struct_outer_new()
    outer <- ffi$struct_outer_set_child(outer, child)
    child_view <- ffi$struct_outer_get_child(outer)
    x_val <- ffi$struct_child_get_x(child_view)

    ffi$struct_child_free(child)
    ffi$struct_outer_free(outer)

    identical(x_val, 42L) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_outer_get_child),
        "nested_view"
      ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_outer_set_child),
        "nested_setter"
      )
  },
  info = "Named nested struct fields preserve nested-view getter and copy-in setter semantics"
)

# Test 6: Wide integer struct getters warn on precision loss
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct bigvals {
        int64_t signed_v;
        uint64_t unsigned_v;
      };

      void fill_bigvals(struct bigvals* p) {
        p->signed_v = (int64_t)9007199254740994LL;
        p->unsigned_v = (uint64_t)9007199254740994ULL;
      }
    "
      ) |>
      tcc_struct(
        "bigvals",
        accessors = c(signed_v = "i64", unsigned_v = "u64")
      ) |>
      tcc_bind(fill_bigvals = list(args = list("ptr"), returns = "void")) |>
      tcc_compile()

    p <- ffi$struct_bigvals_new()
    ffi$fill_bigvals(p)

    warned_i64 <- FALSE
    warned_u64 <- FALSE

    signed_v <- withCallingHandlers(
      ffi$struct_bigvals_get_signed_v(p),
      warning = function(w) {
        if (grepl("i64 precision loss", conditionMessage(w), fixed = TRUE)) {
          warned_i64 <<- TRUE
        }
        invokeRestart("muffleWarning")
      }
    )
    unsigned_v <- withCallingHandlers(
      ffi$struct_bigvals_get_unsigned_v(p),
      warning = function(w) {
        if (grepl("u64 precision loss", conditionMessage(w), fixed = TRUE)) {
          warned_u64 <<- TRUE
        }
        invokeRestart("muffleWarning")
      }
    )

    ffi$struct_bigvals_free(p)

    warned_i64 && warned_u64 && is.numeric(signed_v) && is.numeric(unsigned_v)
  },
  info = "Struct i64/u64 getters warn when R numeric loses precision"
)

# Test 7: field_addr borrowed view keeps owner alive across GC
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct student {
        int id;
        double grade;
      };
    "
      ) |>
      tcc_struct("student", accessors = c(id = "i32", grade = "f64")) |>
      tcc_field_addr("student", "id") |>
      tcc_bind()

    compiled <- tcc_compile(ffi)
    ok <- TRUE

    for (i in seq_len(50)) {
      student <- compiled$struct_student_new()
      student <- compiled$struct_student_set_id(student, as.integer(i))

      id_ptr <- compiled$struct_student_id_addr(student)
      rm(student)
      force_gc()

      id_val <- tcc_read_i32(id_ptr)
      if (!identical(id_val, as.integer(i))) {
        ok <- FALSE
        break
      }

      rm(id_ptr)
      force_gc()
    }

    ok
  },
  info = "field_addr borrowed view survives GC after owner reference is dropped"
)

# Test 8: container_of chain keeps recovered parent alive across GC
expect_true(
  {
    ffi <- tcc_ffi() |>
      tcc_source(
        "
      struct student {
        int id;
        double grade;
      };
    "
      ) |>
      tcc_struct("student", accessors = c(id = "i32", grade = "f64")) |>
      tcc_container_of("student", "id") |>
      tcc_field_addr("student", "id") |>
      tcc_bind()

    compiled <- tcc_compile(ffi)
    ok <- TRUE

    for (i in seq_len(50)) {
      student <- compiled$struct_student_new()
      student <- compiled$struct_student_set_id(student, as.integer(i))
      student <- compiled$struct_student_set_grade(student, i / 10)

      id_ptr <- compiled$struct_student_id_addr(student)
      recovered <- compiled$struct_student_from_id(id_ptr)

      rm(student, id_ptr)
      force_gc()

      id_val <- compiled$struct_student_get_id(recovered)
      grade_val <- compiled$struct_student_get_grade(recovered)
      if (
        !identical(id_val, as.integer(i)) || abs(grade_val - i / 10) > 1e-12
      ) {
        ok <- FALSE
        break
      }

      rm(recovered)
      force_gc()
    }

    ok
  },
  info = "container_of recovered parent survives GC through protected-slot owner chain"
)

# install.libs.R — copy Rtinycc.dll + companion DLLs into libs/

# Standard boilerplate: copy the package shared library
files <- Sys.glob(paste0("*", SHLIB_EXT))
dest <- file.path(R_PACKAGE_DIR, paste0("libs", R_ARCH))
dir.create(dest, recursive = TRUE, showWarnings = FALSE)
file.copy(files, dest, overwrite = TRUE)

# On Windows, also copy libtcc.dll next to Rtinycc.dll
if (.Platform$OS.type == "windows") {
  libtcc_dll <- file.path("..", "inst", "tinycc", "lib", "libtcc.dll")
  if (file.exists(libtcc_dll)) {
    file.copy(libtcc_dll, dest, overwrite = TRUE)
    message("Copied libtcc.dll to ", dest)
  }
}

if (file.exists("symbols.rds")) {
  file.copy("symbols.rds", dest, overwrite = TRUE)
}

# Minimal reprex: treesitter + Rtinycc segfault on Windows at R exit
#
# On Windows, R crashes (segfault) after this script finishes.
# treesitter alone does NOT crash. The crash requires both treesitter
# and Rtinycc (which loads libtcc.dll) to be loaded in the same session.
#
# Hypothesis: libtcc.dll's presence or unload order during exit interferes
# with treesitter's onexit=TRUE finalizers (ts_parser_delete, ts_tree_delete,
# ts_query_delete which call free()).
#
# Run with: Rscript treesitter-windows-segfault-reprex.R
#
# Try each variant to narrow down the trigger:
#   1. treesitter only            → no crash
#   2. treesitter + library(Rtinycc) → crash?
#   3. treesitter + Rtinycc JIT   → crash?

# --- Variant 1: treesitter only (known: no crash) ---
library(treesitter)
library(treesitter.c)

lang <- treesitter.c::language()
parser <- treesitter::parser(lang)
tree <- treesitter::parser_parse(parser, "int add(int a, int b);")
root <- treesitter::tree_root_node(tree)
query <- treesitter::query(lang, "(function_declarator) @fn")
caps <- treesitter::query_captures(query, root)

# --- Variant 2: uncomment to also load Rtinycc (loads libtcc.dll) ---
# library(Rtinycc)

# --- Variant 3: uncomment to also do JIT compilation ---
# library(Rtinycc)
# ffi <- tcc_ffi() |>
#   tcc_source("int add(int a, int b) { return a + b; }") |>
#   tcc_bind(add = list(args = list("i32", "i32"), returns = "i32")) |>
#   tcc_compile()
# stopifnot(ffi$add(2L, 3L) == 5L)

message("Done. If crash happens, it occurs after this line during R exit.")

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Adapted in part from callme::callme_engine() by Mike Cheng.
# The original implementation is MIT-licensed. This version stores the chunk
# body in a named R object for later vignette use instead of compiling it.

rtinycc_c_block <- function(code, options = list()) {
  if (!requireNamespace("knitr", quietly = TRUE)) {
    return(code)
  }

  opts <- utils::modifyList(
    list(
      echo = TRUE,
      eval = FALSE,
      engine = "c",
      class.source = c("rtinycc"),
      prompt = FALSE,
      comment = "",
      tidy = FALSE,
      highlight = TRUE
    ),
    options
  )
  knitr::asis_output(knitr::engine_output(opts, code, ""))
}

#' A knitr engine for reusable inline C source chunks in vignettes
#'
#' Supported custom chunk options:
#' - `object` or `name`: R object name receiving the chunk body as a single
#'   string. Defaults to the chunk label.
#' - `rcode`: whether to append a collapsible block showing the equivalent
#'   R assignment code. Defaults to `TRUE`.
#'
#' @param options Chunk options supplied by knitr.
#'
#' @return Rendered chunk text, or `NULL` if `knitr` is unavailable.
#' @keywords internal
#' @noRd
rtinycc_engine <- function(options) {
  if (!requireNamespace("knitr", quietly = TRUE)) {
    return(NULL)
  }

  `%||%` <- function(x, y) if (is.null(x)) y else x

  object_name <- options$object %||% options$name %||% options$label
  if (
    !is.character(object_name) ||
      length(object_name) != 1L ||
      !nzchar(object_name)
  ) {
    stop(
      "rtinycc_engine() requires a non-empty `object` option or chunk label.",
      call. = FALSE
    )
  }

  code <- paste(options$code, collapse = "\n")
  knit_env <- knitr::knit_global()
  assign(object_name, code, envir = knit_env)

  c_options <- options
  rendered <- rtinycc_c_block(code, c_options)

  if (isFALSE(options$rcode %||% TRUE)) {
    return(rendered)
  }

  rcode <- paste0(
    object_name,
    " <- ",
    paste(deparse(code, width.cutoff = 500L), collapse = "\n")
  )

  r_options <- options
  r_options$engine <- "r"
  r_options$class.source <- character()
  rendered_r <- knitr::engine_output(r_options, rcode, "")
  rendered_r <- paste(
    "<details>",
    "<summary>Click to show R code</summary>",
    rendered_r,
    "</details>",
    sep = "\n"
  )

  paste(rendered, rendered_r, sep = "\n")
}

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

# Canonical FFI semantics matrix.
#
# This is the internal source of truth for:
# - borrow vs copy behavior at the R/C boundary
# - ownership expectations for values crossing the wrapper
# - scalar codegen rule templates
# - generated soundness cases for scalar round-trip tests
#
# The matrix is intentionally explicit because copy and protection behavior is
# part of the package contract, not just an implementation detail.

RTINYCC_DOUBLE_INT_EXACT_LIMIT <- 9007199254740992

rtinycc_render_template <- function(lines, values) {
  rendered <- as.character(lines)

  for (name in names(values)) {
    rendered <- gsub(
      paste0("{{", name, "}}"),
      values[[name]],
      rendered,
      fixed = TRUE
    )
  }

  paste(rendered, collapse = "\n")
}

rtinycc_ffi_semantics <- function() {
  RTINYCC_FFI_SEMANTICS
}

rtinycc_callback_semantics <- function() {
  RTINYCC_CALLBACK_SEMANTICS
}

rtinycc_callback_abi_specs <- function() {
  RTINYCC_CALLBACK_ABI_SPECS
}

rtinycc_composite_semantics <- function() {
  RTINYCC_COMPOSITE_SEMANTICS
}

rtinycc_composite_codegen_specs <- function() {
  RTINYCC_COMPOSITE_CODEGEN_SPECS
}

rtinycc_soundness_case <- function(label, value, expected = NULL) {
  list(label = label, value = value, expected = expected)
}

RTINYCC_FFI_SEMANTICS <- list(
  i8 = list(
    ffi_type = "i8",
    c_type = "int8_t",
    kind = "scalar",
    input = list(
      mode = "copy_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "integer",
      checks = c("NA", "range"),
      notes = "Copied/coerced into a local int8_t before the C call."
    ),
    return = list(
      mode = "box_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar integer SEXP."
    ),
    codegen = list(
      input_form = "integer_range",
      input_range_min = "INT8_MIN",
      input_range_max = "INT8_MAX",
      input_range_error = "i8 out of range",
      return_form = "scalar_integer_cast"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "int8_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0L)),
        rtinycc_soundness_case("max", quote(127L)),
        rtinycc_soundness_case("min", quote(-128L)),
        rtinycc_soundness_case("typical", quote(42L))
      ),
      reject = list(
        rtinycc_soundness_case("NA_integer_", quote(NA_integer_)),
        rtinycc_soundness_case("128 (> INT8_MAX)", quote(128L)),
        rtinycc_soundness_case("-129 (< INT8_MIN)", quote(-129L))
      )
    )
  ),
  i16 = list(
    ffi_type = "i16",
    c_type = "int16_t",
    kind = "scalar",
    input = list(
      mode = "copy_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "integer",
      checks = c("NA", "range"),
      notes = "Copied/coerced into a local int16_t before the C call."
    ),
    return = list(
      mode = "box_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar integer SEXP."
    ),
    codegen = list(
      input_form = "integer_range",
      input_range_min = "INT16_MIN",
      input_range_max = "INT16_MAX",
      input_range_error = "i16 out of range",
      return_form = "scalar_integer_cast"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "int16_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0L)),
        rtinycc_soundness_case("max", quote(32767L)),
        rtinycc_soundness_case("min", quote(-32768L))
      ),
      reject = list(
        rtinycc_soundness_case("32768 (> INT16_MAX)", quote(32768L)),
        rtinycc_soundness_case("-32769 (< INT16_MIN)", quote(-32769L))
      )
    )
  ),
  i32 = list(
    ffi_type = "i32",
    c_type = "int32_t",
    kind = "scalar",
    input = list(
      mode = "copy_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "integer",
      checks = c("NA", "range"),
      notes = "Copied/coerced into a local int32_t before the C call."
    ),
    return = list(
      mode = "box_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar integer SEXP."
    ),
    codegen = list(
      input_form = "integer_range",
      input_range_min = "INT32_MIN",
      input_range_max = "INT32_MAX",
      input_range_error = "i32 out of range",
      return_form = "scalar_integer"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "int32_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0L)),
        rtinycc_soundness_case("INT_MAX", quote(.Machine$integer.max)),
        rtinycc_soundness_case("-INT_MAX", quote(-.Machine$integer.max))
      ),
      reject = list(
        rtinycc_soundness_case("NA_integer_", quote(NA_integer_))
      )
    )
  ),
  i64 = list(
    ffi_type = "i64",
    c_type = "int64_t",
    kind = "scalar",
    input = list(
      mode = "copy_exact_integer_numeric",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "numeric",
      checks = c("NA", "exact-integer", "range", "2^53-limit"),
      exact_integer_limit = RTINYCC_DOUBLE_INT_EXACT_LIMIT,
      notes = "Copied from R numeric into a local int64_t after exactness checks."
    ),
    return = list(
      mode = "box_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      exact_integer_limit = RTINYCC_DOUBLE_INT_EXACT_LIMIT,
      notes = "Returned as R numeric with a warning once precision exceeds exact double integer range."
    ),
    codegen = list(
      input_form = "exact_integer_real",
      input_exact_error = "i64 requires exact integer (|x| <= 2^53)",
      input_integer_error = "i64 requires integer value",
      input_range_min = "INT64_MIN",
      input_range_max = "INT64_MAX",
      input_range_error = "i64 out of range",
      return_form = "i64"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "int64_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0)),
        rtinycc_soundness_case("one", quote(1)),
        rtinycc_soundness_case("neg one", quote(-1)),
        rtinycc_soundness_case("2^53", quote(2^53)),
        rtinycc_soundness_case("-2^53", quote(-2^53))
      ),
      reject = list(
        rtinycc_soundness_case("NA_real_", quote(NA_real_)),
        rtinycc_soundness_case("Inf", quote(Inf)),
        rtinycc_soundness_case("non-integer", quote(1.5))
      )
    )
  ),
  u8 = list(
    ffi_type = "u8",
    c_type = "uint8_t",
    kind = "scalar",
    input = list(
      mode = "copy_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "integer",
      checks = c("NA", "range"),
      notes = "Copied/coerced into a local uint8_t before the C call."
    ),
    return = list(
      mode = "box_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar integer SEXP."
    ),
    codegen = list(
      input_form = "integer_range",
      input_range_min = "0",
      input_range_max = "UINT8_MAX",
      input_range_error = "u8 out of range",
      return_form = "scalar_integer_cast"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "uint8_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0L)),
        rtinycc_soundness_case("max", quote(255L)),
        rtinycc_soundness_case("one", quote(1L))
      ),
      reject = list(
        rtinycc_soundness_case("-1", quote(-1L)),
        rtinycc_soundness_case("256 (> UINT8_MAX)", quote(256L))
      )
    )
  ),
  u16 = list(
    ffi_type = "u16",
    c_type = "uint16_t",
    kind = "scalar",
    input = list(
      mode = "copy_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "integer",
      checks = c("NA", "range"),
      notes = "Copied/coerced into a local uint16_t before the C call."
    ),
    return = list(
      mode = "box_integer_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar integer SEXP."
    ),
    codegen = list(
      input_form = "integer_range",
      input_range_min = "0",
      input_range_max = "UINT16_MAX",
      input_range_error = "u16 out of range",
      return_form = "scalar_integer_cast"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "uint16_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0L)),
        rtinycc_soundness_case("max", quote(65535L))
      ),
      reject = list(
        rtinycc_soundness_case("-1", quote(-1L)),
        rtinycc_soundness_case("65536 (> UINT16_MAX)", quote(65536L))
      )
    )
  ),
  u32 = list(
    ffi_type = "u32",
    c_type = "uint32_t",
    kind = "scalar",
    input = list(
      mode = "copy_integer_numeric",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "numeric",
      checks = c("NA", "non-negative", "integer-valued", "range"),
      notes = "Copied from R numeric into a local uint32_t after range checks."
    ),
    return = list(
      mode = "box_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as R numeric because values can exceed INT_MAX."
    ),
    codegen = list(
      input_form = "real_integer_range",
      input_range_min = "0",
      input_range_max = "(double)UINT32_MAX",
      input_range_error = "u32 out of range",
      input_integer_error = "u32 requires integer value",
      return_form = "scalar_real_cast"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "uint32_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0)),
        rtinycc_soundness_case("UINT32_MAX", quote(4294967295)),
        rtinycc_soundness_case("one", quote(1))
      ),
      reject = list(
        rtinycc_soundness_case("-1", quote(-1)),
        rtinycc_soundness_case("non-integer", quote(1.5))
      )
    )
  ),
  u64 = list(
    ffi_type = "u64",
    c_type = "uint64_t",
    kind = "scalar",
    input = list(
      mode = "copy_exact_integer_numeric",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "numeric",
      checks = c("NA", "non-negative", "exact-integer", "2^53-limit"),
      exact_integer_limit = RTINYCC_DOUBLE_INT_EXACT_LIMIT,
      notes = "Copied from R numeric into a local uint64_t after exactness checks."
    ),
    return = list(
      mode = "box_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      exact_integer_limit = RTINYCC_DOUBLE_INT_EXACT_LIMIT,
      notes = "Returned as R numeric with a warning once precision exceeds exact double integer range."
    ),
    codegen = list(
      input_form = "exact_integer_real",
      input_nonnegative = TRUE,
      input_nonnegative_error = "u64 out of range",
      input_exact_error = "u64 requires exact integer (|x| <= 2^53)",
      input_integer_error = "u64 requires integer value",
      return_form = "u64"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "uint64_t"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0)),
        rtinycc_soundness_case("one", quote(1)),
        rtinycc_soundness_case("2^53", quote(2^53))
      ),
      reject = list(
        rtinycc_soundness_case("-1", quote(-1)),
        rtinycc_soundness_case("NA_real_", quote(NA_real_))
      )
    )
  ),
  f32 = list(
    ffi_type = "f32",
    c_type = "float",
    kind = "scalar",
    input = list(
      mode = "copy_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "numeric",
      checks = character(),
      notes = "Copied/coerced into a local float before the C call."
    ),
    return = list(
      mode = "box_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as R numeric after float to double widening."
    ),
    codegen = list(
      input_form = "real_scalar_cast",
      return_form = "scalar_real_cast"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "float"),
      comparator = "tolerance",
      tolerance = 1e-5,
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0.0)),
        rtinycc_soundness_case("pi approx (f32 precision)", quote(3.14)),
        rtinycc_soundness_case("neg one", quote(-1.0))
      ),
      reject = list()
    )
  ),
  f64 = list(
    ffi_type = "f64",
    c_type = "double",
    kind = "scalar",
    input = list(
      mode = "copy_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "numeric",
      checks = character(),
      notes = "Copied/coerced into a local double before the C call."
    ),
    return = list(
      mode = "box_numeric_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar real SEXP."
    ),
    codegen = list(
      input_form = "real_scalar",
      return_form = "scalar_real"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(c_type = "double"),
      roundtrip = list(
        rtinycc_soundness_case("zero", quote(0.0)),
        rtinycc_soundness_case("pi", quote(3.141592653589793)),
        rtinycc_soundness_case("large neg", quote(-1e308)),
        rtinycc_soundness_case("tiny pos", quote(1e-308))
      ),
      reject = list()
    )
  ),
  bool = list(
    ffi_type = "bool",
    c_type = "bool",
    kind = "scalar",
    input = list(
      mode = "copy_logical_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "wrapper-local",
      r_storage = "logical",
      checks = "NA",
      notes = "Copied/coerced into a local bool before the C call."
    ),
    return = list(
      mode = "box_logical_scalar",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Returned as a fresh scalar logical SEXP."
    ),
    codegen = list(
      input_form = "logical_scalar",
      return_form = "logical"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(
        c_type = "bool",
        source = "#include <stdbool.h>\nbool identity(bool x) { return x; }"
      ),
      roundtrip = list(
        rtinycc_soundness_case("TRUE", quote(TRUE)),
        rtinycc_soundness_case("FALSE", quote(FALSE))
      ),
      reject = list(
        rtinycc_soundness_case("NA", quote(NA))
      )
    )
  ),
  cstring = list(
    ffi_type = "cstring",
    c_type = "char*",
    kind = "scalar",
    input = list(
      mode = "borrow_utf8_pointer",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R/internal translation buffer",
      r_storage = "character",
      checks = character(),
      notes = "Borrowed UTF-8 pointer from Rf_translateCharUTF8() for the duration of the call."
    ),
    return = list(
      mode = "copy_string",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Copied into R-managed string memory with mkString()."
    ),
    codegen = list(
      input_form = "cstring",
      return_form = "cstring"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(
        c_type = "const char*",
        source = "const char* identity(const char* x) { return x; }"
      ),
      roundtrip = list(
        rtinycc_soundness_case("basic", quote("hello")),
        rtinycc_soundness_case("empty", quote("")),
        rtinycc_soundness_case("special chars", quote("hello world 123!@#"))
      ),
      reject = list()
    )
  ),
  ptr = list(
    ffi_type = "ptr",
    c_type = "void*",
    kind = "scalar",
    input = list(
      mode = "borrow_externalptr",
      borrow = TRUE,
      copy = FALSE,
      ownership = "caller-defined",
      r_storage = "externalptr",
      checks = character(),
      notes = "Wrapper passes the external pointer address through without copying pointee storage."
    ),
    return = list(
      mode = "external_pointer",
      borrow = TRUE,
      copy = FALSE,
      ownership = "unchanged",
      notes = "Returned as a raw external pointer; no pointee copy or ownership transfer is implied."
    ),
    codegen = list(
      input_form = "ptr",
      return_form = "ptr"
    )
  ),
  raw = list(
    ffi_type = "raw",
    c_type = "uint8_t*",
    kind = "array",
    input = list(
      mode = "borrow_vector_storage",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      r_storage = "raw",
      checks = character(),
      notes = "Zero-copy borrow of RAW(x); sound only while no fresh R allocations are inserted before the C call."
    ),
    return = list(
      mode = "copy_array",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Array returns allocate a fresh raw vector and memcpy() the C buffer into it."
    )
  ),
  integer_array = list(
    ffi_type = "integer_array",
    c_type = "int32_t*",
    kind = "array",
    input = list(
      mode = "borrow_vector_storage",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      r_storage = "integer",
      checks = character(),
      notes = "Zero-copy borrow of INTEGER(x); sound only while no fresh R allocations are inserted before the C call."
    ),
    return = list(
      mode = "copy_array",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Array returns allocate a fresh integer vector and memcpy() the C buffer into it."
    )
  ),
  numeric_array = list(
    ffi_type = "numeric_array",
    c_type = "double*",
    kind = "array",
    input = list(
      mode = "borrow_vector_storage",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      r_storage = "numeric",
      checks = character(),
      notes = "Zero-copy borrow of REAL(x); sound only while no fresh R allocations are inserted before the C call."
    ),
    return = list(
      mode = "copy_array",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Array returns allocate a fresh numeric vector and memcpy() the C buffer into it."
    )
  ),
  logical_array = list(
    ffi_type = "logical_array",
    c_type = "int*",
    kind = "array",
    input = list(
      mode = "borrow_vector_storage",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      r_storage = "logical",
      checks = character(),
      notes = "Zero-copy borrow of LOGICAL(x); sound only while no fresh R allocations are inserted before the C call."
    ),
    return = list(
      mode = "copy_array",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R",
      notes = "Array returns allocate a fresh logical vector and memcpy() the C buffer into it."
    )
  ),
  character_array = list(
    ffi_type = "character_array",
    c_type = "const SEXP*",
    kind = "array",
    input = list(
      mode = "borrow_string_cells",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      r_storage = "character",
      checks = character(),
      notes = paste(
        "Exposes STRING_PTR_RO(x), whose C type in R is const SEXP*;",
        "the pointed elements are CHARSXP string cells rather than char** C strings."
      )
    ),
    return = NULL
  ),
  cstring_array = list(
    ffi_type = "cstring_array",
    c_type = "const char**",
    kind = "array",
    input = list(
      mode = "copy_pointer_array",
      borrow = FALSE,
      copy = TRUE,
      ownership = "R_alloc-temporary",
      r_storage = "character",
      checks = "character",
      notes = "Allocates a temporary const char** with R_alloc() and fills it from translated string payloads; pointer array is copied, string payloads are borrowed."
    ),
    return = NULL
  ),
  sexp = list(
    ffi_type = "sexp",
    c_type = "SEXP",
    kind = "scalar",
    input = list(
      mode = "passthrough",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      r_storage = "ANY",
      checks = character(),
      notes = "Least transformed boundary mode: wrapper passes the SEXP through directly."
    ),
    return = list(
      mode = "passthrough",
      borrow = TRUE,
      copy = FALSE,
      ownership = "R",
      notes = "Returned directly as a SEXP."
    ),
    codegen = list(
      input_form = "sexp",
      return_form = "sexp"
    ),
    soundness = list(
      generated = TRUE,
      identity = list(
        c_type = "SEXP",
        source = "SEXP identity(SEXP x) { return x; }"
      ),
      roundtrip = list(
        rtinycc_soundness_case("integer", quote(42L)),
        rtinycc_soundness_case("string", quote("hello")),
        rtinycc_soundness_case("double", quote(3.14))
      ),
      reject = list()
    )
  ),
  void = list(
    ffi_type = "void",
    c_type = "void",
    kind = "scalar",
    input = NULL,
    return = list(
      mode = "side-effect-only",
      borrow = FALSE,
      copy = FALSE,
      ownership = "none",
      notes = "Wrapper evaluates the C expression for side effects and returns R NULL."
    ),
    codegen = list(
      return_form = "void"
    )
  ),
  callback = list(
    ffi_type = "callback",
    c_type = "void*",
    kind = "scalar",
    input = list(
      mode = "borrow_callback_pointer",
      borrow = TRUE,
      copy = FALSE,
      ownership = "callback-registry",
      r_storage = "externalptr/tcc_callback",
      checks = character(),
      notes = "Wrapper passes the trampoline/callback pointer through; callback lifetime is managed separately by the registry and preserved R function."
    ),
    return = NULL,
    codegen = list(
      input_form = "ptr"
    )
  )
)

RTINYCC_CALLBACK_SEMANTICS <- list(
  sync = list(
    kind = "callback_sync",
    preserves_function = TRUE,
    preservation_api = "R_PreserveObject",
    invocation_thread = "main",
    argument_transport = "stack-to-VECSXP",
    copies = list(
      scalars = TRUE,
      cstring_payload = FALSE,
      pointer_addresses = FALSE
    ),
    borrows = list(
      callback_token = TRUE,
      pointer_pointee = TRUE,
      cstring_payload = TRUE
    ),
    returns = list(
      default_on_error = "NA-like scalar or NULL pointer",
      mode = "converted_back_to_declared_c_type"
    ),
    notes = paste(
      "Synchronous trampolines allocate a VECSXP argument list on the main",
      "thread, invoke the preserved R function directly, then convert the",
      "SEXP result back to the declared C type."
    )
  ),
  async = list(
    kind = "callback_async",
    preserves_function = TRUE,
    preservation_api = "R_PreserveObject",
    invocation_thread = "scheduled-on-main",
    scheduler = c(
      "RC_callback_async_schedule_c",
      "RC_callback_async_schedule_sync_c"
    ),
    copies = list(
      scalar_values = TRUE,
      cstring_payload = TRUE,
      pointer_addresses = TRUE
    ),
    borrows = list(
      callback_token = TRUE,
      pointer_pointee = TRUE
    ),
    returns = list(
      default_on_error = "NA-like scalar or NULL pointer",
      nonvoid_mode = "sync-result-channel"
    ),
    notes = paste(
      "Async callbacks marshal arguments into cb_arg_t task payloads,",
      "duplicate cstring payloads for cross-thread safety, and reconstruct",
      "fresh R objects on the main thread before invocation."
    )
  )
)

RTINYCC_CALLBACK_ABI_SPECS <- list(
  trampoline = list(
    list(
      name = "tramp_i8",
      signature = "int8_t (*)(int8_t)",
      pattern = "int8_t tramp_i8\\(void\\* cb, int8_t arg1\\)",
      info = "Trampoline keeps int8_t argument ABI"
    ),
    list(
      name = "tramp_i16",
      signature = "int16_t (*)(int16_t)",
      pattern = "int16_t tramp_i16\\(void\\* cb, int16_t arg1\\)",
      info = "Trampoline keeps int16_t argument ABI"
    ),
    list(
      name = "tramp_f32",
      signature = "float (*)(float)",
      pattern = "float tramp_f32\\(void\\* cb, float arg1\\)",
      info = "Trampoline keeps float argument ABI"
    ),
    list(
      name = "tramp_bool_arg",
      signature = "bool (*)(bool)",
      pattern = "bool tramp_bool_arg\\(void\\* cb, bool arg1\\)",
      info = "Trampoline keeps bool argument ABI"
    ),
    list(
      name = "tramp_ptrptr_arg",
      signature = "void (*)(char **)",
      pattern = "void tramp_ptrptr_arg\\(void\\* cb, char \\*\\* arg1\\)",
      info = "Trampoline keeps pointer subtype in argument ABI"
    )
  ),
  async_trampoline = list(
    list(
      name = "tramp_async_f32",
      signature = "callback_async:f32(f32)",
      patterns = list(
        list(
          pattern = "float tramp_async_f32\\(void\\* cb, float arg1\\)",
          info = "Async trampoline normalizes f32 to float in the C signature"
        ),
        list(
          pattern = "return \\(float\\)result.v.d;",
          info = "Async trampoline normalizes f32 return casts to float"
        )
      )
    ),
    list(
      name = "tramp_async_ptr",
      signature = "callback_async:void*(void*)",
      patterns = list(
        list(
          pattern = "void\\* tramp_async_ptr\\(void\\* cb, void\\* arg1\\)",
          info = "Async pointer trampoline preserves pointer ABI"
        ),
        list(
          pattern = "args\\[0\\]\\.kind = CB_ARG_PTR;",
          info = "Async pointer trampoline stores pointer args in CB_ARG_PTR slot"
        ),
        list(
          pattern = "args\\[0\\]\\.v\\.p = arg1;",
          info = "Async pointer trampoline stores raw pointer value without pointee copy"
        ),
        list(
          pattern = "return \\(void\\*\\)result.v.p;",
          info = "Async pointer trampoline returns pointer results through cb_result_t pointer slot"
        )
      ),
      forbidden = list(
        list(
          pattern = "SET_VECTOR_ELT\\(args, 0, arg1\\);",
          info = "Async pointer trampoline does not treat pointer args as raw SEXP pass-through"
        ),
        list(
          pattern = "return result;",
          info = "Async pointer trampoline does not return a raw SEXP result"
        ),
        list(
          pattern = "R_ExternalPtrAddr\\(result\\)",
          info = "Async pointer trampoline does not use sync-callback external-pointer extraction"
        )
      )
    )
  ),
  wrapper = list(
    list(
      name = "call_bool_cb",
      args = list("callback:bool(bool)", "ptr", "bool"),
      returns = "bool",
      c_code = "
      bool call_bool_cb(bool (*cb)(void*, bool), void* ctx, bool x) {
        return cb(ctx, x);
      }
    ",
      patterns = list(
        list(
          pattern = "bool trampoline_R_wrap_call_bool_cb_arg1\\(void\\* cb, bool arg1\\)",
          info = "Bool callback wrapper preserves bool ABI"
        ),
        list(
          pattern = "bool \\(\\*arg1\\)\\(void\\*, bool\\) = trampoline_R_wrap_call_bool_cb_arg1;",
          info = "Bool callback wrapper declaration matches trampoline ABI"
        )
      )
    ),
    list(
      name = "call_i8_cb",
      args = list("callback:int8_t(int8_t)", "ptr", "i8"),
      returns = "i8",
      c_code = "
      int8_t call_i8_cb(int8_t (*cb)(void*, int8_t), void* ctx, int8_t x) {
        return cb(ctx, x);
      }
    ",
      patterns = list(
        list(
          pattern = "int8_t trampoline_R_wrap_call_i8_cb_arg1\\(void\\* cb, int8_t arg1\\)",
          info = "int8_t callback wrapper preserves narrow integer ABI"
        ),
        list(
          pattern = "int8_t \\(\\*arg1\\)\\(void\\*, int8_t\\) = trampoline_R_wrap_call_i8_cb_arg1;",
          info = "int8_t callback wrapper declaration matches trampoline ABI"
        )
      )
    ),
    list(
      name = "call_f32_cb",
      args = list("callback:float(float)", "ptr", "f32"),
      returns = "f32",
      c_code = "
      float call_f32_cb(float (*cb)(void*, float), void* ctx, float x) {
        return cb(ctx, x);
      }
    ",
      patterns = list(
        list(
          pattern = "float trampoline_R_wrap_call_f32_cb_arg1\\(void\\* cb, float arg1\\)",
          info = "float callback wrapper preserves float ABI"
        ),
        list(
          pattern = "float \\(\\*arg1\\)\\(void\\*, float\\) = trampoline_R_wrap_call_f32_cb_arg1;",
          info = "float callback wrapper declaration matches trampoline ABI"
        )
      )
    ),
    list(
      name = "call_ptr_cb",
      args = list("callback:void*(void*)", "ptr", "ptr"),
      returns = "ptr",
      c_code = "
      void* call_ptr_cb(void* (*cb)(void*, void*), void* ctx, void* x) {
        return cb(ctx, x);
      }
    ",
      patterns = list(
        list(
          pattern = "SET_VECTOR_ELT\\(args, 0, RC_make_unowned_ptr\\(arg1, R_NilValue\\)\\);",
          info = "Pointer callback trampoline boxes pointer args through host unowned helper"
        ),
        list(
          pattern = "return R_ExternalPtrAddr\\(result\\);",
          info = "Pointer callback trampoline converts R result back to native address"
        ),
        list(
          pattern = "return RC_make_unowned_ptr\\(call_ptr_cb\\(arg1, arg2, arg3\\), R_NilValue\\);",
          info = "Pointer callback wrapper return boxes raw pointers through host unowned helper"
        )
      ),
      forbidden = list(
        list(
          pattern = "R_MakeExternalPtr",
          info = "Pointer callback codegen avoids direct external pointer construction"
        )
      )
    ),
    list(
      name = "call_sexp_cb",
      args = list("callback:SEXP(SEXP)", "ptr", "sexp"),
      returns = "sexp",
      c_code = "
      SEXP call_sexp_cb(SEXP (*cb)(void*, SEXP), void* ctx, SEXP x) {
        return cb(ctx, x);
      }
    ",
      patterns = list(
        list(
          pattern = "SEXP trampoline_R_wrap_call_sexp_cb_arg1\\(void\\* cb, SEXP arg1\\)",
          info = "SEXP callback trampoline preserves SEXP ABI"
        ),
        list(
          pattern = "SET_VECTOR_ELT\\(args, 0, arg1\\);",
          info = "SEXP callback trampoline passes SEXP arguments through directly"
        ),
        list(
          pattern = "return result;",
          info = "SEXP callback trampoline returns the callback result SEXP directly"
        ),
        list(
          pattern = "SEXP \\(\\*arg1\\)\\(void\\*, SEXP\\) = trampoline_R_wrap_call_sexp_cb_arg1;",
          info = "SEXP callback wrapper declaration matches trampoline ABI"
        )
      ),
      forbidden = list(
        list(
          pattern = "RC_make_unowned_ptr\\(arg1, R_NilValue\\)",
          info = "SEXP callback trampoline does not pointer-box SEXP arguments"
        ),
        list(
          pattern = "R_ExternalPtrAddr\\(result\\)",
          info = "SEXP callback trampoline does not coerce callback results through external-pointer reads"
        )
      )
    )
  )
)

RTINYCC_COMPOSITE_SEMANTICS <- list(
  struct_owned = list(
    kind = "struct",
    helper = "tcc_struct",
    borrow = FALSE,
    copy = FALSE,
    ownership = "owned-native-storage",
    finalizer = TRUE,
    notes = paste(
      "Struct constructors allocate owned native storage and return an",
      "external pointer tagged with the struct type and a finalizer."
    )
  ),
  struct_field_addr = list(
    kind = "struct_view",
    helper = "tcc_field_addr",
    borrow = TRUE,
    copy = FALSE,
    ownership = "borrowed-from-struct",
    protects_owner = TRUE,
    lifetime_model = "preserved owner slot",
    survives_gc_with_live_view = TRUE,
    notes = paste(
      "Field-address helpers return borrowed external pointers, preserve the",
      "owner struct, and keep it in the protected slot so storage stays alive."
    )
  ),
  struct_container_of = list(
    kind = "struct_view",
    helper = "tcc_container_of",
    borrow = TRUE,
    copy = FALSE,
    ownership = "borrowed-from-member-owner-chain",
    protects_owner = TRUE,
    lifetime_model = "preserved owner chain",
    survives_gc_with_live_view = TRUE,
    notes = paste(
      "container_of recovers a parent struct pointer from a field pointer,",
      "preserves the incoming external pointer, and keeps it in the protected",
      "slot so the owner lifetime chain remains intact."
    )
  ),
  struct_raw_access = list(
    kind = "struct_raw_access",
    helper = "tcc_struct_raw_access",
    read_copy = TRUE,
    write_copy = TRUE,
    ownership = "struct-owned-storage",
    notes = paste(
      "Raw struct helpers copy bytes out to a fresh RAWSXP or copy bytes from",
      "a RAWSXP back into the struct buffer with memcpy()."
    )
  ),
  struct_array_field = list(
    kind = "struct_array_field",
    helper = "tcc_struct",
    borrow = FALSE,
    copy = TRUE,
    ownership = "struct-owned-storage",
    notes = paste(
      "Struct array field element helpers copy scalar elements between R and",
      "the native struct buffer; the struct storage itself remains owned by",
      "the struct external pointer."
    )
  ),
  union_owned = list(
    kind = "union",
    helper = "tcc_union",
    borrow = FALSE,
    copy = FALSE,
    ownership = "owned-native-storage",
    finalizer = TRUE,
    notes = paste(
      "Union constructors allocate owned native storage and expose member",
      "getters/setters over that shared buffer."
    )
  ),
  union_nested_struct_view = list(
    kind = "union_view",
    helper = "tcc_union",
    borrow = TRUE,
    copy = FALSE,
    ownership = "borrowed-from-union",
    protects_owner = TRUE,
    lifetime_model = "preserved owner slot",
    survives_gc_with_live_view = TRUE,
    notes = paste(
      "Nested struct getters on unions return borrowed member views and keep",
      "the owning union in the protected slot so the shared storage remains",
      "alive while the view exists."
    )
  ),
  enum_i32 = list(
    kind = "enum",
    helper = "tcc_enum",
    boundary_mode = "i32-like",
    borrow = FALSE,
    copy = TRUE,
    ownership = "R",
    notes = paste(
      "Enums cross the wrapper boundary as integer-like values and constant",
      "helpers box them into fresh scalar integer SEXPs."
    )
  ),
  global_scalar = list(
    kind = "global",
    helper = "tcc_global",
    scalar_only = TRUE,
    arrays_forbidden = TRUE,
    borrow = FALSE,
    copy = TRUE,
    ownership = "compiled-unit",
    notes = paste(
      "Global helpers are limited to scalar types and reuse the same wrapper",
      "input/output coercion rules as ordinary scalar bindings."
    )
  ),
  bitfield_native = list(
    kind = "bitfield",
    helper = "tcc_struct/tcc_treesitter_struct_accessors",
    compiler_managed = TRUE,
    default_ffi_type = "u8",
    treesitter_bitfield_type = "u8",
    include_bitfields = TRUE,
    address_helpers_forbidden = TRUE,
    container_of_forbidden = TRUE,
    survives_forced_gc = TRUE,
    notes = paste(
      "Bitfields are stored and masked by the C compiler; helper accessors",
      "treat them as scalar fields, treesitter defaults them to u8 unless",
      "the caller overrides bitfield_type, and address-style helpers are",
      "forbidden for bitfield members."
    )
  ),
  treesitter_header_bindings = list(
    kind = "header_codegen",
    helper = "tcc_generate_bindings",
    include_bitfields = TRUE,
    bitfield_type = "u8",
    nested_struct_mode = "ptr-like",
    generates = c("functions", "structs", "unions", "enums", "globals"),
    notes = paste(
      "Treesitter helpers project parsed header declarations into the same",
      "FFI helper surface, including optional bitfield accessors and global",
      "getter/setter bindings; treesitter-generated nested struct fields",
      "currently stay ptr-like unless a richer parser-level model is added",
      "explicitly."
    )
  )
)

RTINYCC_COMPOSITE_CODEGEN_SPECS <- list(
  list(
    name = "struct_owned_host_helper",
    info = "struct constructors use host owned-composite helper instead of direct extptr registration",
    generate_args = list(
      symbols = list(),
      c_code = "struct student { int id; double grade; };",
      structs = list(student = c(id = "i32", grade = "f64"))
    ),
    patterns = list(
      list(
        pattern = "return RC_make_owned_composite_ptr(p, Rf_install(\"struct_student\"));",
        fixed = TRUE
      )
    ),
    forbidden = list(
      list(
        pattern = "R_RegisterCFinalizerEx(ext, RC_owned_native_finalizer, FALSE);",
        fixed = TRUE
      ),
      list(
        pattern = "R_MakeExternalPtr(p, Rf_install(\"struct_student\"), R_NilValue)",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "struct_field_addr_owner",
    info = "field_addr helper preserves owner through borrowed-view helper",
    generate_args = list(
      symbols = list(),
      c_code = "struct student { int id; double grade; };",
      structs = list(student = c(id = "i32", grade = "f64")),
      field_addr = list(student = "id")
    ),
    patterns = list(
      list(
        pattern = "SEXP R_wrap_struct_student_id_addr(SEXP ext) {",
        fixed = TRUE
      ),
      list(
        pattern = "return RC_make_borrowed_view(field_ptr, Rf_install(\"rtinycc_borrowed\"), ext);",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "struct_container_of_owner",
    info = "container_of helper preserves owner chain through borrowed-view helper",
    generate_args = list(
      symbols = list(),
      c_code = "struct student { int id; double grade; };",
      structs = list(student = c(id = "i32", grade = "f64")),
      container_of = list(student = "id")
    ),
    patterns = list(
      list(
        pattern = "SEXP R_wrap_struct_student_from_id(SEXP ext) {",
        fixed = TRUE
      ),
      list(
        pattern = "return RC_make_borrowed_view(p, Rf_install(\"struct_student\"), ext);",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "struct_raw_access_copy",
    info = "struct raw access helpers use memcpy copy paths",
    generate_args = list(
      symbols = list(),
      c_code = "struct packet { unsigned char data[8]; };",
      structs = list(
        packet = list(data = list(type = "u8", size = 8, array = TRUE))
      ),
      struct_raw_access = "packet"
    ),
    patterns = list(
      list(
        pattern = "SEXP R_wrap_struct_packet_get_raw(SEXP ext, SEXP len) {",
        fixed = TRUE
      ),
      list(
        pattern = "memcpy(RAW(raw), p,",
        fixed = TRUE
      ),
      list(
        pattern = "SEXP R_wrap_struct_packet_set_raw(SEXP ext, SEXP raw) {",
        fixed = TRUE
      ),
      list(
        pattern = "memcpy(p, RAW(raw),",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "union_owned_host_helper",
    info = "union constructors use host owned-composite helper instead of direct extptr registration",
    generate_args = list(
      symbols = list(),
      c_code = "union wrapper { int raw; double value; };",
      unions = list(
        wrapper = list(
          members = list(raw = "i32", value = "f64"),
          active = "raw"
        )
      )
    ),
    patterns = list(
      list(
        pattern = "return RC_make_owned_composite_ptr(p, Rf_install(\"union_wrapper\"));",
        fixed = TRUE
      )
    ),
    forbidden = list(
      list(
        pattern = "R_RegisterCFinalizerEx(ext, RC_owned_native_finalizer, FALSE);",
        fixed = TRUE
      ),
      list(
        pattern = "R_MakeExternalPtr(p, Rf_install(\"union_wrapper\"), R_NilValue)",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "union_nested_struct_owner",
    info = "union nested struct getters return borrowed views that protect the union owner",
    generate_args = list(
      symbols = list(),
      c_code = "union wrapper { struct { int x; } inner; int raw; };",
      unions = list(
        wrapper = list(
          members = list(inner = list(type = "struct"), raw = "i32"),
          active = "raw"
        )
      )
    ),
    patterns = list(
      list(
        pattern = "return RC_make_borrowed_view(&p->inner, Rf_install(\"struct_inner\"), ext);",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "enum_constant_and_sizeof",
    info = "enum helpers expose constants and sizeof introspection",
    generate_args = list(
      symbols = list(),
      c_code = "enum color { RED = 1, BLUE = 2 };",
      enums = list(
        color = list(constants = c("RED", "BLUE"), export_constants = TRUE)
      ),
      introspect = TRUE
    ),
    patterns = list(
      list(
        pattern = "SEXP R_wrap_enum_color_RED(void) {",
        fixed = TRUE
      ),
      list(
        pattern = "return ScalarInteger(RED);",
        fixed = TRUE
      ),
      list(
        pattern = "SEXP R_wrap_enum_color_sizeof(void) {",
        fixed = TRUE
      )
    )
  ),
  list(
    name = "global_get_set_helpers",
    info = "global helpers generate scalar getter and setter wrappers",
    generate_args = list(
      symbols = list(),
      c_code = "int32_t global_counter = 0;",
      globals = list(global_counter = "i32")
    ),
    patterns = list(
      list(
        pattern = "SEXP R_wrap_global_global_counter_get(void) {",
        fixed = TRUE
      ),
      list(
        pattern = "SEXP R_wrap_global_global_counter_set(SEXP value_) {",
        fixed = TRUE
      ),
      list(
        pattern = "global_counter = value;",
        fixed = FALSE
      )
    )
  )
)

rtinycc_scalar_soundness_specs <- function() {
  Filter(
    function(spec) {
      !is.null(spec$soundness) && isTRUE(spec$soundness$generated)
    },
    RTINYCC_FFI_SEMANTICS
  )
}

rtinycc_scalar_input_rule_body <- function(type, arg_name, r_name) {
  spec <- RTINYCC_FFI_SEMANTICS[[type]]
  codegen <- spec$codegen

  if (is.null(codegen) || is.null(codegen$input_form)) {
    stop("No scalar input rule registered for type: ", type, call. = FALSE)
  }

  values <- c(
    list(
      arg_name = arg_name,
      r_name = r_name,
      c_type = spec$c_type
    ),
    codegen
  )

  switch(
    codegen$input_form,
    integer_range = rtinycc_render_template(
      c(
        "  int _{{arg_name}} = asInteger({{r_name}});",
        "  if (_{{arg_name}} == NA_INTEGER) Rf_error(\"integer value is NA\");",
        "  if (_{{arg_name}} < {{input_range_min}} || _{{arg_name}} > {{input_range_max}}) Rf_error(\"{{input_range_error}}\");",
        "  {{c_type}} {{arg_name}} = ({{c_type}})_{{arg_name}};"
      ),
      values
    ),
    exact_integer_real = rtinycc_render_template(
      c(
        "  double _{{arg_name}} = asReal({{r_name}});",
        "  if (ISNA(_{{arg_name}}) || ISNAN(_{{arg_name}})) Rf_error(\"numeric value is NA\");",
        if (isTRUE(codegen$input_nonnegative)) {
          "  if (_{{arg_name}} < 0) Rf_error(\"{{input_nonnegative_error}}\");"
        },
        "  if (fabs(_{{arg_name}}) > 9007199254740992.0) Rf_error(\"{{input_exact_error}}\");",
        "  if (trunc(_{{arg_name}}) != _{{arg_name}}) Rf_error(\"{{input_integer_error}}\");",
        if (
          !is.null(codegen$input_range_min) && !is.null(codegen$input_range_max)
        ) {
          "  if (_{{arg_name}} < (double){{input_range_min}} || _{{arg_name}} > (double){{input_range_max}}) Rf_error(\"{{input_range_error}}\");"
        },
        "  {{c_type}} {{arg_name}} = ({{c_type}})_{{arg_name}};"
      ),
      values
    ),
    real_integer_range = rtinycc_render_template(
      c(
        "  double _{{arg_name}} = asReal({{r_name}});",
        "  if (ISNA(_{{arg_name}}) || ISNAN(_{{arg_name}})) Rf_error(\"numeric value is NA\");",
        "  if (_{{arg_name}} < {{input_range_min}} || _{{arg_name}} > {{input_range_max}}) Rf_error(\"{{input_range_error}}\");",
        "  if (trunc(_{{arg_name}}) != _{{arg_name}}) Rf_error(\"{{input_integer_error}}\");",
        "  {{c_type}} {{arg_name}} = ({{c_type}})_{{arg_name}};"
      ),
      values
    ),
    real_scalar_cast = rtinycc_render_template(
      "  {{c_type}} {{arg_name}} = ({{c_type}})asReal({{r_name}});",
      values
    ),
    real_scalar = rtinycc_render_template(
      "  {{c_type}} {{arg_name}} = asReal({{r_name}});",
      values
    ),
    logical_scalar = rtinycc_render_template(
      c(
        "  int _{{arg_name}} = asLogical({{r_name}});",
        "  if (_{{arg_name}} == NA_LOGICAL) Rf_error(\"logical value is NA\");",
        "  bool {{arg_name}} = (bool)(_{{arg_name}} != 0);"
      ),
      values
    ),
    cstring = rtinycc_render_template(
      c(
        "  SEXP _{{arg_name}} = STRING_ELT({{r_name}}, 0);",
        "  const char* {{arg_name}} = (_{{arg_name}} == NA_STRING) ? NULL : Rf_translateCharUTF8(_{{arg_name}});"
      ),
      values
    ),
    ptr = rtinycc_render_template(
      "  void* {{arg_name}} = R_ExternalPtrAddr({{r_name}});",
      values
    ),
    sexp = rtinycc_render_template(
      "  SEXP {{arg_name}} = {{r_name}};",
      values
    ),
    stop("Unsupported scalar input form: ", codegen$input_form, call. = FALSE)
  )
}

rtinycc_scalar_return_rule_body <- function(type, value_expr) {
  spec <- RTINYCC_FFI_SEMANTICS[[type]]
  codegen <- spec$codegen

  if (is.null(codegen) || is.null(codegen$return_form)) {
    stop("No scalar return rule registered for type: ", type, call. = FALSE)
  }

  values <- c(
    list(value_expr = value_expr),
    codegen
  )

  switch(
    codegen$return_form,
    scalar_integer_cast = rtinycc_render_template(
      "return ScalarInteger((int){{value_expr}});",
      values
    ),
    scalar_integer = rtinycc_render_template(
      "return ScalarInteger({{value_expr}});",
      values
    ),
    scalar_real_cast = rtinycc_render_template(
      "return ScalarReal((double){{value_expr}});",
      values
    ),
    scalar_real = rtinycc_render_template(
      "return ScalarReal({{value_expr}});",
      values
    ),
    i64 = rtinycc_render_template(
      c(
        "  int64_t __rtinycc_ret = {{value_expr}};",
        "  if (__rtinycc_ret > INT64_C(9007199254740992) || __rtinycc_ret < -INT64_C(9007199254740992)) Rf_warning(\"i64 precision loss in R numeric\");",
        "  return ScalarReal((double)__rtinycc_ret);"
      ),
      values
    ),
    u64 = rtinycc_render_template(
      c(
        "  uint64_t __rtinycc_ret = {{value_expr}};",
        "  if (__rtinycc_ret > UINT64_C(9007199254740992)) Rf_warning(\"u64 precision loss in R numeric\");",
        "  return ScalarReal((double)__rtinycc_ret);"
      ),
      values
    ),
    logical = rtinycc_render_template(
      "return ScalarLogical((int){{value_expr}});",
      values
    ),
    cstring = rtinycc_render_template(
      c(
        "const char* __rtinycc_ret = {{value_expr}};",
        "if (__rtinycc_ret) {",
        "    SEXP out = PROTECT(mkString(__rtinycc_ret));",
        "    UNPROTECT(1);",
        "    return out;",
        "} else {",
        "    return R_NilValue;",
        "}"
      ),
      values
    ),
    ptr = rtinycc_render_template(
      "return RC_make_unowned_ptr({{value_expr}}, R_NilValue);",
      values
    ),
    sexp = rtinycc_render_template(
      "return {{value_expr}};",
      values
    ),
    void = rtinycc_render_template(
      c(
        "{{value_expr}};",
        "  return R_NilValue;"
      ),
      values
    ),
    stop("Unsupported scalar return form: ", codegen$return_form, call. = FALSE)
  )
}

rtinycc_register_scalar_codegen_rules <- function(where = parent.frame()) {
  for (type in names(RTINYCC_FFI_SEMANTICS)) {
    spec <- RTINYCC_FFI_SEMANTICS[[type]]

    if (!is.null(spec$codegen$input_form)) {
      eval(
        substitute(
          ffi_input_rule(TYPE, arg_name, r_name) %as%
            rtinycc_scalar_input_rule_body(TYPE, arg_name, r_name),
          list(TYPE = type)
        ),
        envir = where
      )
    }

    if (!is.null(spec$codegen$return_form)) {
      eval(
        substitute(
          ffi_return_rule(TYPE, value_expr) %as%
            rtinycc_scalar_return_rule_body(TYPE, value_expr),
          list(TYPE = type)
        ),
        envir = where
      )
    }
  }

  invisible()
}

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

# Internal lambda.r rules for codegen dispatch
#
# This file defines many small "rules" used by the code generator. Each
# rule is registered with a tiny DSL (the lambda.r style rules such as
# `ffi_input_rule`, `ffi_return_rule`, `r_to_c_return_lines_rule`, etc.)
# and returns either a string or a character vector containing C source
# lines to be injected into generated bindings. The sections below are
# grouped by purpose (input conversion, return conversion, array helpers,
# struct/union accessors, callback helpers, type mapping keys, etc.).
#
# NOTE: many groups include nearly-identical rules for multiple integer
# and floating types; these were written out explicitly for clarity.  See
# TODOs later in this file for suggestions where a small generator could
# reduce repetition by producing similar rules programmatically.

## ------------------------------------------------------------------
## FFI input conversion rules
##
## Produce C lines that convert an incoming R `SEXP` (named by `r_name`)
## into a C variable (named by `arg_name`) of the requested FFI type.
## These rules are used when generating the C thunk that bridges R -> C.
## ------------------------------------------------------------------
rtinycc_register_scalar_codegen_rules()

ffi_input_enum_rule(type_name, arg_name, r_name) %as%
  {
    sprintf("  int %s = asInteger(%s);", arg_name, r_name)
  }

ffi_input_callback_rule(type_name, arg_name, r_name) %as%
  {
    sprintf("  void* %s = R_ExternalPtrAddr(%s);", arg_name, r_name)
  }

ffi_input_rule("raw", arg_name, r_name) %as%
  {
    sprintf("  uint8_t* %s = RAW(%s);", arg_name, r_name)
  }

ffi_input_rule("integer_array", arg_name, r_name) %as%
  {
    sprintf("  int32_t* %s = INTEGER(%s);", arg_name, r_name)
  }

ffi_input_rule("numeric_array", arg_name, r_name) %as%
  {
    sprintf("  double* %s = REAL(%s);", arg_name, r_name)
  }

ffi_input_rule("logical_array", arg_name, r_name) %as%
  {
    sprintf("  int* %s = LOGICAL(%s);", arg_name, r_name)
  }

ffi_input_rule("character_array", arg_name, r_name) %as%
  {
    sprintf("  const SEXP* %s = STRING_PTR_RO(%s);", arg_name, r_name)
  }

ffi_input_rule("cstring_array", arg_name, r_name) %as%
  {
    rtinycc_render_template(
      c(
        "  if (!Rf_isString({{r_name}})) Rf_error(\"expected character vector\");",
        "  R_xlen_t _{{arg_name}}_n = XLENGTH({{r_name}});",
        "  const char** {{arg_name}} = (const char**) R_alloc((size_t)_{{arg_name}}_n, sizeof(const char*));",
        "  for (R_xlen_t _{{arg_name}}_i = 0; _{{arg_name}}_i < _{{arg_name}}_n; _{{arg_name}}_i++) {",
        "    SEXP _{{arg_name}}_elt = STRING_ELT({{r_name}}, _{{arg_name}}_i);",
        "    {{arg_name}}[_{{arg_name}}_i] = (_{{arg_name}}_elt == NA_STRING) ? NULL : Rf_translateCharUTF8(_{{arg_name}}_elt);",
        "  }"
      ),
      list(
        r_name = r_name,
        arg_name = arg_name
      )
    )
  }

ffi_input_rule(type, arg_name, r_name) %as%
  {
    stop("Unsupported FFI type: ", type, call. = FALSE)
  }

## ------------------------------------------------------------------
## Runtime `.RtinyccCall` dispatch rules
##
## Dispatch small arities directly and fall back to do.call for larger
## signatures.
## ------------------------------------------------------------------
rtinycc_call(0L, call_ptr, args) %as%
  {
    .RtinyccCall(call_ptr)
  }

rtinycc_call(1L, call_ptr, args) %as%
  {
    .RtinyccCall(call_ptr, args[[1]])
  }

rtinycc_call(2L, call_ptr, args) %as%
  {
    .RtinyccCall(call_ptr, args[[1]], args[[2]])
  }

rtinycc_call(3L, call_ptr, args) %as%
  {
    .RtinyccCall(call_ptr, args[[1]], args[[2]], args[[3]])
  }

rtinycc_call(4L, call_ptr, args) %as%
  {
    .RtinyccCall(call_ptr, args[[1]], args[[2]], args[[3]], args[[4]])
  }

rtinycc_call(5L, call_ptr, args) %as%
  {
    .RtinyccCall(
      call_ptr,
      args[[1]],
      args[[2]],
      args[[3]],
      args[[4]],
      args[[5]]
    )
  }

rtinycc_call(n_args, call_ptr, args) %as%
  {
    do.call(.RtinyccCall, c(list(call_ptr), args))
  }

## ------------------------------------------------------------------
## Runtime platform dispatch rules
##
## Keep platform-specific path/name/env branching declarative using
## lambda.r dispatch instead of large switch/if chains.
## ------------------------------------------------------------------
tcc_platform_lib_paths("Linux") %as%
  {
    c(
      "/usr/lib",
      "/usr/lib64",
      "/usr/local/lib",
      "/lib",
      "/lib64",
      "/lib32",
      "/usr/local/lib64",
      "/usr/lib/x86_64-linux-gnu",
      "/usr/lib/i386-linux-gnu",
      "/lib/x86_64-linux-gnu",
      "/lib32/x86_64-linux-gnu",
      "/lib/x86_64-linux-gnu/",
      # linux-unknown-gnu is for Alpine Linux with musl
      # which uses different library paths
      "/usr/lib/x86_64-linux-gnu",
      "/usr/lib/i386-linux-gnu",
      "/lib/x86_64-linux-gnu",
      "/lib32/x86_64-linux-gnu",
      "/usr/lib/x86_64-linux-musl",
      "/usr/lib/i386-linux-musl",
      "/lib/x86_64-linux-musl",
      "/lib32/x86_64-linux-musl",
      # amd/aarch64 multiarch paths
      "/usr/lib/amd64-linux-gnu",
      "/usr/lib/aarch64-linux-gnu"
      # manylinux paths
    )
  }

tcc_platform_lib_paths("Darwin") %as%
  {
    c(
      "/usr/lib",
      "/usr/local/lib",
      "/opt/homebrew/lib", # Apple Silicon Homebrew
      "/opt/local/lib", # MacPorts
      "/System/Library/Frameworks", # Apple system libs
      "/Library/Frameworks"
    )
  }

tcc_platform_lib_paths("Windows") %as%
  {
    sysroot <- Sys.getenv("SystemRoot", unset = "C:/Windows")
    r_bin <- file.path(R.home(), "bin")
    r_arch_bin <- file.path(R.home(), "bin", .Platform$r_arch)
    path_dirs <- strsplit(
      Sys.getenv("PATH", ""),
      .Platform$path.sep,
      fixed = TRUE
    )[[1]]

    unique(c(
      "C:/msys64/mingw64/lib", # MSYS2
      "C:/msys64/mingw32/lib",
      "C:/Rtools45/mingw_64/lib", # Rtools
      "C:/Rtools45/mingw_32/lib",
      "C:/Rtools44/mingw_64/lib", # Rtools
      "C:/Rtools44/mingw_32/lib",
      # Native Windows loader locations / common runtime dirs
      file.path(sysroot, "System32"),
      file.path(sysroot, "SysWOW64"),
      r_bin,
      r_arch_bin,
      # User/runtime PATH entries (non-Rtools setups)
      path_dirs
    ))
  }

tcc_platform_lib_paths(sysname) %as%
  {
    c("/usr/lib", "/usr/local/lib")
  }

tcc_short_lib_filename("Linux", name) %as%
  {
    paste0("lib", name, ".so")
  }

tcc_short_lib_filename("Darwin", name) %as%
  {
    paste0("lib", name, ".dylib")
  }

tcc_short_lib_filename("Windows", name) %as%
  {
    paste0(name, ".dll")
  }

tcc_short_lib_filename(sysname, name) %as%
  {
    paste0("lib", name, ".so")
  }

tcc_output_type_rule("memory") %as% 1L
tcc_output_type_rule("obj") %as% 3L
tcc_output_type_rule("dll") %as% 4L
tcc_output_type_rule("exe") %as% 2L
tcc_output_type_rule("preprocess") %as% 5L

tcc_loader_env_key("Windows") %as% "PATH"
tcc_loader_env_key("Darwin") %as% "DYLD_LIBRARY_PATH"
tcc_loader_env_key(sysname) %as% "LD_LIBRARY_PATH"

tcc_loader_env_sep("Windows") %as% .Platform$path.sep
tcc_loader_env_sep(sysname) %as% ":"

## ------------------------------------------------------------------
## FFI return conversion rules
##
## Produce C code that converts a C expression (`value_expr`) into an
## R `SEXP` return value. Used when generating the C -> R return thunk.
## ------------------------------------------------------------------
ffi_return_enum_rule(type_name, value_expr) %as%
  {
    sprintf("return ScalarInteger((int)%s);", value_expr)
  }

ffi_return_rule(type, value_expr) %as%
  {
    stop("Unsupported FFI return type: ", type, call. = FALSE)
  }

## ------------------------------------------------------------------
## Array return helpers
##
## Helper rules for allocating an R vector to hold array returns and for
## copying C buffers into the newly allocated R vector (`out`).
## ------------------------------------------------------------------
array_return_alloc_line_rule("raw", len_name) %as%
  {
    sprintf("SEXP out = PROTECT(allocVector(RAWSXP, %s));", len_name)
  }

array_return_alloc_line_rule("integer_array", len_name) %as%
  {
    sprintf("SEXP out = PROTECT(allocVector(INTSXP, %s));", len_name)
  }

array_return_alloc_line_rule("numeric_array", len_name) %as%
  {
    sprintf("SEXP out = PROTECT(allocVector(REALSXP, %s));", len_name)
  }

array_return_alloc_line_rule("logical_array", len_name) %as%
  {
    sprintf("SEXP out = PROTECT(allocVector(LGLSXP, %s));", len_name)
  }

array_return_alloc_line_rule(type, len_name) %as%
  {
    stop("Unsupported array return type: ", type, call. = FALSE)
  }

array_return_copy_line_rule("raw", len_name, value_expr) %as%
  {
    sprintf(
      "  if (%s > 0) memcpy(RAW(out), %s, sizeof(uint8_t) * %s);",
      len_name,
      value_expr,
      len_name
    )
  }

array_return_copy_line_rule("integer_array", len_name, value_expr) %as%
  {
    sprintf(
      "  if (%s > 0) memcpy(INTEGER(out), %s, sizeof(int32_t) * %s);",
      len_name,
      value_expr,
      len_name
    )
  }

array_return_copy_line_rule("numeric_array", len_name, value_expr) %as%
  {
    sprintf(
      "  if (%s > 0) memcpy(REAL(out), %s, sizeof(double) * %s);",
      len_name,
      value_expr,
      len_name
    )
  }

array_return_copy_line_rule("logical_array", len_name, value_expr) %as%
  {
    sprintf(
      "  if (%s > 0) memcpy(LOGICAL(out), %s, sizeof(int) * %s);",
      len_name,
      value_expr,
      len_name
    )
  }

array_return_copy_line_rule(type, len_name, value_expr) %as%
  {
    stop("Unsupported array return type: ", type, call. = FALSE)
  }

## ------------------------------------------------------------------
## Struct field getter rules
##
## Produce code that reads a field from a C struct (pointer `p`) and
## converts it into an R `SEXP` for consumer code. Rules are repeated
## for multiple primitive types for clarity and explicit handling.
## ------------------------------------------------------------------
struct_field_getter_rule("i8", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s);", field_name)
  }

struct_field_getter_rule("i16", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s);", field_name)
  }

struct_field_getter_rule("i32", field_name) %as%
  {
    sprintf("return ScalarInteger(p->%s);", field_name)
  }

struct_field_getter_rule("i64", field_name) %as%
  {
    c(
      sprintf("  int64_t __rtinycc_ret = p->%s;", field_name),
      "  if (__rtinycc_ret > INT64_C(9007199254740992) || __rtinycc_ret < -INT64_C(9007199254740992)) Rf_warning(\"i64 precision loss in R numeric\");",
      "  return ScalarReal((double)__rtinycc_ret);"
    )
  }

struct_field_getter_rule("u8", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s);", field_name)
  }

struct_field_getter_rule("u16", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s);", field_name)
  }

struct_field_getter_rule("u32", field_name) %as%
  {
    sprintf("return ScalarReal((double)p->%s);", field_name)
  }

struct_field_getter_rule("u64", field_name) %as%
  {
    c(
      sprintf("  uint64_t __rtinycc_ret = p->%s;", field_name),
      "  if (__rtinycc_ret > UINT64_C(9007199254740992)) Rf_warning(\"u64 precision loss in R numeric\");",
      "  return ScalarReal((double)__rtinycc_ret);"
    )
  }

struct_field_getter_rule("f32", field_name) %as%
  {
    sprintf("return ScalarReal((double)p->%s);", field_name)
  }

struct_field_getter_rule("f64", field_name) %as%
  {
    sprintf("return ScalarReal(p->%s);", field_name)
  }

struct_field_getter_rule("bool", field_name) %as%
  {
    sprintf("return ScalarLogical((int)p->%s);", field_name)
  }

struct_field_getter_rule("cstring", field_name) %as%
  {
    c(
      sprintf("  if (p->%s) {", field_name),
      sprintf("    SEXP out = PROTECT(mkString(p->%s));", field_name),
      "    UNPROTECT(1);",
      "    return out;",
      "  } else {",
      "    return R_NilValue;",
      "  }"
    )
  }

struct_field_getter_rule("ptr", field_name) %as%
  {
    sprintf(
      "return RC_make_borrowed_view(p->%s, Rf_install(\"rtinycc_borrowed\"), ext);",
      field_name
    )
  }

struct_field_getter_rule(type_name, field_name) %as%
  {
    sprintf(
      "return RC_make_borrowed_view(&p->%s, Rf_install(\"rtinycc_borrowed\"), ext);",
      field_name
    )
  }

## ------------------------------------------------------------------
## Struct array field getter rules
##
## Like the struct field getters, but for array fields inside structs
## (access element `idx`). Returns an R scalar for the selected element.
## ------------------------------------------------------------------
struct_array_field_getter_rule("i8", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("i16", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("i32", field_name) %as%
  {
    sprintf("return ScalarInteger(p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("i64", field_name) %as%
  {
    c(
      sprintf("  int64_t __rtinycc_ret = p->%s[idx];", field_name),
      "  if (__rtinycc_ret > INT64_C(9007199254740992) || __rtinycc_ret < -INT64_C(9007199254740992)) Rf_warning(\"i64 precision loss in R numeric\");",
      "  return ScalarReal((double)__rtinycc_ret);"
    )
  }

struct_array_field_getter_rule("u8", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("u16", field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("u32", field_name) %as%
  {
    sprintf("return ScalarReal((double)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("u64", field_name) %as%
  {
    c(
      sprintf("  uint64_t __rtinycc_ret = p->%s[idx];", field_name),
      "  if (__rtinycc_ret > UINT64_C(9007199254740992)) Rf_warning(\"u64 precision loss in R numeric\");",
      "  return ScalarReal((double)__rtinycc_ret);"
    )
  }

struct_array_field_getter_rule("f32", field_name) %as%
  {
    sprintf("return ScalarReal((double)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("f64", field_name) %as%
  {
    sprintf("return ScalarReal(p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule("bool", field_name) %as%
  {
    sprintf("return ScalarLogical((int)p->%s[idx]);", field_name)
  }

struct_array_field_getter_rule(type_name, field_name) %as%
  {
    sprintf("return ScalarInteger((int)p->%s[idx]);", field_name)
  }

## ------------------------------------------------------------------
## Struct array field setter rules
##
## Generate code to set an element of an array field inside a struct
## from an R `SEXP` value (named `val`).
## ------------------------------------------------------------------
struct_array_field_setter_rule("i8", field_name) %as%
  {
    sprintf("p->%s[idx] = (int8_t)asInteger(val);", field_name)
  }

struct_array_field_setter_rule("i16", field_name) %as%
  {
    sprintf("p->%s[idx] = (int16_t)asInteger(val);", field_name)
  }

struct_array_field_setter_rule("i32", field_name) %as%
  {
    sprintf("p->%s[idx] = asInteger(val);", field_name)
  }

struct_array_field_setter_rule("i64", field_name) %as%
  {
    sprintf("p->%s[idx] = (int64_t)REAL(val)[0];", field_name)
  }

struct_array_field_setter_rule("u8", field_name) %as%
  {
    sprintf("p->%s[idx] = (uint8_t)asInteger(val);", field_name)
  }

struct_array_field_setter_rule("u16", field_name) %as%
  {
    sprintf("p->%s[idx] = (uint16_t)asInteger(val);", field_name)
  }

struct_array_field_setter_rule("u32", field_name) %as%
  {
    sprintf("p->%s[idx] = (uint32_t)REAL(val)[0];", field_name)
  }

struct_array_field_setter_rule("u64", field_name) %as%
  {
    sprintf("p->%s[idx] = (uint64_t)REAL(val)[0];", field_name)
  }

struct_array_field_setter_rule("f32", field_name) %as%
  {
    sprintf("p->%s[idx] = (float)REAL(val)[0];", field_name)
  }

struct_array_field_setter_rule("f64", field_name) %as%
  {
    sprintf("p->%s[idx] = REAL(val)[0];", field_name)
  }

struct_array_field_setter_rule("bool", field_name) %as%
  {
    sprintf("p->%s[idx] = (bool)asLogical(val);", field_name)
  }

struct_array_field_setter_rule(type_name, field_name) %as%
  {
    sprintf("p->%s[idx] = (uint8_t)asInteger(val);", field_name)
  }

## ------------------------------------------------------------------
## Struct field setter rules
##
## Produce code that assigns an R value (`val`) into a struct field
## (handles scalar fields and fixed-size char arrays). These are the
## counterparts to the getter rules above.
## ------------------------------------------------------------------
struct_field_setter_rule("i8", field_name, size) %as%
  {
    sprintf("p->%s = (int8_t)asInteger(val);", field_name)
  }

struct_field_setter_rule("i16", field_name, size) %as%
  {
    sprintf("p->%s = (int16_t)asInteger(val);", field_name)
  }

struct_field_setter_rule("i32", field_name, size) %as%
  {
    sprintf("p->%s = asInteger(val);", field_name)
  }

struct_field_setter_rule("i64", field_name, size) %as%
  {
    sprintf("p->%s = (int64_t)REAL(val)[0];", field_name)
  }

struct_field_setter_rule("u8", field_name, size) %as%
  {
    sprintf("p->%s = (uint8_t)asInteger(val);", field_name)
  }

struct_field_setter_rule("u16", field_name, size) %as%
  {
    sprintf("p->%s = (uint16_t)asInteger(val);", field_name)
  }

struct_field_setter_rule("u32", field_name, size) %as%
  {
    sprintf("p->%s = (uint32_t)REAL(val)[0];", field_name)
  }

struct_field_setter_rule("u64", field_name, size) %as%
  {
    sprintf("p->%s = (uint64_t)REAL(val)[0];", field_name)
  }

struct_field_setter_rule("f32", field_name, size) %as%
  {
    sprintf("p->%s = (float)asReal(val);", field_name)
  }

struct_field_setter_rule("f64", field_name, size) %as%
  {
    sprintf("p->%s = asReal(val);", field_name)
  }

struct_field_setter_rule("bool", field_name, size) %as%
  {
    sprintf("p->%s = (bool)asLogical(val);", field_name)
  }

struct_field_setter_rule("ptr", field_name, size) %as%
  {
    sprintf("p->%s = R_ExternalPtrAddr(val);", field_name)
  }

struct_field_setter_rule("cstring", field_name, size) %as%
  {
    if (!is.null(size)) {
      return(c(
        "  const char *src = CHAR(STRING_ELT(val, 0));",
        sprintf("  strncpy(p->%s, src, %d);", field_name, size - 1),
        sprintf("  p->%s[%d] = '\\0';", field_name, size - 1)
      ))
    }
    sprintf("p->%s = CHAR(STRING_ELT(val, 0));", field_name)
  }

struct_field_setter_rule(type_name, field_name, size) %as%
  {
    sprintf("// Cannot set field of type %s", type_name)
  }

## ------------------------------------------------------------------
## Union field helper rules
##
## Unions are represented by mapping to the corresponding struct rules
## wherever possible; setting array members in unions is unsupported and
## emits placeholder comments. These rules delegate to struct helpers.
## ------------------------------------------------------------------
union_field_getter_rule(type_name, mem_name) %as%
  {
    struct_field_getter_rule(type_name, mem_name)
  }

union_field_setter_rule("raw", mem_name, size) %as%
  {
    sprintf("// Cannot set union member array type raw")
  }

union_field_setter_rule("integer_array", mem_name, size) %as%
  {
    sprintf("// Cannot set union member array type integer_array")
  }

union_field_setter_rule("numeric_array", mem_name, size) %as%
  {
    sprintf("// Cannot set union member array type numeric_array")
  }

union_field_setter_rule("logical_array", mem_name, size) %as%
  {
    sprintf("// Cannot set union member array type logical_array")
  }

union_field_setter_rule("character_array", mem_name, size) %as%
  {
    sprintf("// Cannot set union member array type character_array")
  }

union_field_setter_rule("cstring_array", mem_name, size) %as%
  {
    sprintf("// Cannot set union member array type cstring_array")
  }

union_field_setter_rule(type_name, mem_name, size) %as%
  {
    struct_field_setter_rule(type_name, mem_name, size)
  }

## ------------------------------------------------------------------
## Callback argument kind/value helpers
##
## Map a callback argument type to an internal CB_ARG_* kind and to the
## field name that holds the value inside the generic callback value
## union struct used by the runtime. These are small lookup rules used
## when generating callback trampoline code.
## ------------------------------------------------------------------
cb_arg_kind_rule("ptr") %as%
  {
    "CB_ARG_PTR"
  }

cb_arg_kind_rule("cstring") %as%
  {
    "CB_ARG_CSTRING"
  }

cb_arg_kind_rule("real") %as%
  {
    "CB_ARG_REAL"
  }

cb_arg_kind_rule("logical") %as%
  {
    "CB_ARG_LOGICAL"
  }

cb_arg_kind_rule("int") %as%
  {
    "CB_ARG_INT"
  }

cb_arg_value_rule("ptr", arg_expr) %as%
  {
    list(field = "p", expr = arg_expr)
  }

cb_arg_value_rule("cstring", arg_expr) %as%
  {
    list(field = "s", expr = arg_expr)
  }

cb_arg_value_rule("real", arg_expr) %as%
  {
    list(field = "d", expr = arg_expr)
  }

cb_arg_value_rule("logical", arg_expr) %as%
  {
    list(field = "i", expr = arg_expr)
  }

cb_arg_value_rule("int", arg_expr) %as%
  {
    list(field = "i", expr = arg_expr)
  }

## ------------------------------------------------------------------
## C default return helpers
##
## When a callback or conversion fails we need a sensible C-level default
## return statement for a given C type. These helpers emit appropriately
## indented return lines (used by many `r_to_c_return_lines_rule` cases).
## ------------------------------------------------------------------
c_default_return_rule("void", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return;")
  }

c_default_return_rule("sexp", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return R_NilValue;")
  }

c_default_return_rule("ptr", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return NULL;")
  }

c_default_return_rule("cstring", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return NULL;")
  }

c_default_return_rule("real", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return NA_REAL;")
  }

c_default_return_rule("logical", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return -1;")
  }

c_default_return_rule("int", indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return NA_INTEGER;")
  }

## ------------------------------------------------------------------
## C <-> SEXP type name helpers
##
## Map between C types used in generated stubs and the canonical names
## used for selecting conversion rules and SEXP constructors.
## ------------------------------------------------------------------
c_sexp_type_rule("ptr") %as%
  {
    "void*"
  }

c_sexp_type_rule("int") %as%
  {
    "int"
  }

c_sexp_type_rule("longlong") %as%
  {
    "long long"
  }

c_sexp_type_rule("real") %as%
  {
    "double"
  }

c_sexp_type_rule("cstring") %as%
  {
    "const char*"
  }

c_sexp_type_rule("bool") %as%
  {
    "bool"
  }

## ------------------------------------------------------------------
## Canonical R-facing type keys
##
## Map C types into the compact R-facing type keys (e.g. "i32", "f64",
## "ptr") used across many rule lookups.
## ------------------------------------------------------------------
c_r_type_rule("ptr") %as%
  {
    "ptr"
  }

c_r_type_rule("i32") %as%
  {
    "i32"
  }

c_r_type_rule("f64") %as%
  {
    "f64"
  }

c_r_type_rule("cstring") %as%
  {
    "cstring"
  }

c_r_type_rule("bool") %as%
  {
    "bool"
  }

c_r_type_rule("void") %as%
  {
    "void"
  }

c_r_type_rule("sexp") %as%
  {
    "sexp"
  }

## ------------------------------------------------------------------
## SEXP constructor helpers
##
## Provide the name of the constructor (or the full call form) used to
## create a SEXP from a native C value during return conversion.
## ------------------------------------------------------------------
sexp_constructor_rule("ptr") %as%
  {
    "R_MakeExternalPtr"
  }

sexp_constructor_rule("int") %as%
  {
    "ScalarInteger"
  }

sexp_constructor_rule("real") %as%
  {
    "ScalarReal"
  }

sexp_constructor_rule("cstring") %as%
  {
    "mkString"
  }

sexp_constructor_rule("bool") %as%
  {
    "ScalarLogical"
  }

sexp_constructor_rule("sexp") %as%
  {
    "identity"
  }

## ------------------------------------------------------------------
## SEXP constructor call forms
##
## Produce the string of C code to call the SEXP constructor for a given
## expression (used when building return expressions inline).
## ------------------------------------------------------------------
sexp_constructor_call_rule("ptr", arg_expr) %as%
  {
    sprintf("RC_make_unowned_ptr(%s, R_NilValue)", arg_expr)
  }

sexp_constructor_call_rule("cstring", arg_expr) %as%
  {
    sprintf(
      "({ const char* __rtinycc_tmp = %s; __rtinycc_tmp ? mkString(__rtinycc_tmp) : R_NilValue; })",
      arg_expr
    )
  }

sexp_constructor_call_rule("int", arg_expr) %as%
  {
    sprintf("ScalarInteger(%s)", arg_expr)
  }

sexp_constructor_call_rule("real", arg_expr) %as%
  {
    sprintf("ScalarReal(%s)", arg_expr)
  }

sexp_constructor_call_rule("bool", arg_expr) %as%
  {
    sprintf("ScalarLogical(%s)", arg_expr)
  }

sexp_constructor_call_rule("sexp", arg_expr) %as%
  {
    sprintf("%s", arg_expr)
  }

## ------------------------------------------------------------------
## R -> C converter helpers
##
## Provide the simple converter function name used to extract C values
## from R `SEXP` objects (e.g. `asInteger`, `asReal`, `R_ExternalPtrAddr`).
## ------------------------------------------------------------------
r_to_c_converter_rule("ptr") %as%
  {
    "R_ExternalPtrAddr"
  }

r_to_c_converter_rule("int") %as%
  {
    "asInteger"
  }

r_to_c_converter_rule("real") %as%
  {
    "asReal"
  }

r_to_c_converter_rule("cstring") %as%
  {
    "CHAR"
  }

r_to_c_converter_rule("bool") %as%
  {
    "asLogical"
  }

r_to_c_converter_rule("sexp") %as%
  {
    "identity"
  }

## ------------------------------------------------------------------
## R -> C return conversion lines
##
## These rules emit one or more lines of C that validate and convert
## a returned R `SEXP` (from a callback) into the expected C return
## value. They handle NA/NA_REAL checks, range checks, and emit warnings
## and default returns on error.
## ------------------------------------------------------------------
r_to_c_return_lines_rule("void", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return;")
  }

r_to_c_return_lines_rule("sexp", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return ", result_var, ";")
  }

r_to_c_return_lines_rule("ptr", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    paste0(pad, "return R_ExternalPtrAddr(", result_var, ");")
  }

r_to_c_return_lines_rule("cstring", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    c(
      paste0(pad, "if (", result_var, " == R_NilValue) return NULL;"),
      paste0(
        pad,
        "if (!Rf_isString(",
        result_var,
        ") || XLENGTH(",
        result_var,
        ") < 1) {"
      ),
      paste0(pad, "  Rf_warning(\"callback returned non-string\");"),
      paste0(pad, "  return NULL;"),
      paste0(pad, "}"),
      paste0(pad, "SEXP _str = STRING_ELT(", result_var, ", 0);"),
      paste0(pad, "if (_str == NA_STRING) return NULL;"),
      paste0(pad, "return Rf_translateCharUTF8(_str);")
    )
  }

r_to_c_return_lines_rule("bool", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asLogical(", result_var, ");"),
      paste0(pad, "if (_v == NA_LOGICAL) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA logical\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (_v != 0);")
    )
  }

r_to_c_return_lines_rule("i8", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asInteger(", result_var, ");"),
      paste0(pad, "if (_v == NA_INTEGER) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA integer\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < INT8_MIN || _v > INT8_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range i8\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (int8_t)_v;")
    )
  }

r_to_c_return_lines_rule("i16", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asInteger(", result_var, ");"),
      paste0(pad, "if (_v == NA_INTEGER) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA integer\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < INT16_MIN || _v > INT16_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range i16\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (int16_t)_v;")
    )
  }

r_to_c_return_lines_rule("i32", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asInteger(", result_var, ");"),
      paste0(pad, "if (_v == NA_INTEGER) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA integer\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (int32_t)_v;")
    )
  }

r_to_c_return_lines_rule("int", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asInteger(", result_var, ");"),
      paste0(pad, "if (_v == NA_INTEGER) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA integer\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return _v;")
    )
  }

r_to_c_return_lines_rule("u8", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asInteger(", result_var, ");"),
      paste0(pad, "if (_v == NA_INTEGER) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA integer\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < 0 || _v > UINT8_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range u8\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (uint8_t)_v;")
    )
  }

r_to_c_return_lines_rule("u16", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "int _v = asInteger(", result_var, ");"),
      paste0(pad, "if (_v == NA_INTEGER) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA integer\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < 0 || _v > UINT16_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range u16\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (uint16_t)_v;")
    )
  }

r_to_c_return_lines_rule("u32", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "double _v = asReal(", result_var, ");"),
      paste0(pad, "if (ISNA(_v) || ISNAN(_v)) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < 0 || _v > (double)UINT32_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range u32\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (trunc(_v) != _v) {"),
      paste0(pad, "  Rf_warning(\"callback returned non-integer numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (uint32_t)_v;")
    )
  }

r_to_c_return_lines_rule("i64", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "double _v = asReal(", result_var, ");"),
      paste0(pad, "if (ISNA(_v) || ISNAN(_v)) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (fabs(_v) > 9007199254740992.0) {"),
      paste0(
        pad,
        "  Rf_warning(\"callback i64 precision loss in R numeric\");"
      ),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (trunc(_v) != _v) {"),
      paste0(pad, "  Rf_warning(\"callback returned non-integer numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < (double)INT64_MIN || _v > (double)INT64_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range i64\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (int64_t)_v;")
    )
  }

r_to_c_return_lines_rule("u64", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "double _v = asReal(", result_var, ");"),
      paste0(pad, "if (ISNA(_v) || ISNAN(_v)) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v < 0) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range u64\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (fabs(_v) > 9007199254740992.0) {"),
      paste0(
        pad,
        "  Rf_warning(\"callback u64 precision loss in R numeric\");"
      ),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (trunc(_v) != _v) {"),
      paste0(pad, "  Rf_warning(\"callback returned non-integer numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "if (_v > (double)UINT64_MAX) {"),
      paste0(pad, "  Rf_warning(\"callback returned out-of-range u64\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (uint64_t)_v;")
    )
  }

r_to_c_return_lines_rule("double", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "double _v = asReal(", result_var, ");"),
      paste0(pad, "if (ISNA(_v) || ISNAN(_v)) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return _v;")
    )
  }

r_to_c_return_lines_rule("float", c_type, result_var, indent) %as%
  {
    pad <- paste(rep(" ", indent), collapse = "")
    default_line <- get_c_default_return(c_type, indent = indent)
    c(
      paste0(pad, "double _v = asReal(", result_var, ");"),
      paste0(pad, "if (ISNA(_v) || ISNAN(_v)) {"),
      paste0(pad, "  Rf_warning(\"callback returned NA numeric\");"),
      default_line,
      paste0(pad, "}"),
      paste0(pad, "return (float)_v;")
    )
  }

## ------------------------------------------------------------------
## Callback kind key mapping
##
## Map C types (and pointer-ness) to the compact kinds used by the
## callback trampoline generator (ptr, cstring, real, logical, int).
## ------------------------------------------------------------------
cb_kind_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
cb_kind_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
cb_kind_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
cb_kind_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
cb_kind_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
cb_kind_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
cb_kind_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
cb_kind_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
cb_kind_key_rule("double", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("float", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("f64", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("f32", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("int64_t", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("i64", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("uint32_t", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("u32", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("uint64_t", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("u64", FALSE) %as%
  {
    "real"
  }
cb_kind_key_rule("bool", FALSE) %as%
  {
    "logical"
  }
cb_kind_key_rule("_Bool", FALSE) %as%
  {
    "logical"
  }
cb_kind_key_rule(type_name, is_ptr) %as%
  {
    "int"
  }

## ------------------------------------------------------------------
## C-to-R type key map for arguments
##
## These rules map C type spellings (and pointer-ness) to the canonical
## R-facing type key used across code generation (e.g. "i32", "f64",
## "ptr"). This centralizes many small spelling variants in one place.
## ------------------------------------------------------------------
c_r_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
c_r_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
c_r_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
c_r_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
c_r_key_rule("int", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("int32_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("i32", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("int16_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("i16", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("int8_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("i8", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("long", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("long long", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("int64_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("i64", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("uint8_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("u8", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("uint16_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("u16", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("uint32_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("u32", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("uint64_t", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("u64", FALSE) %as%
  {
    "i32"
  }
c_r_key_rule("double", FALSE) %as%
  {
    "f64"
  }
c_r_key_rule("float", FALSE) %as%
  {
    "f64"
  }
c_r_key_rule("f64", FALSE) %as%
  {
    "f64"
  }
c_r_key_rule("f32", FALSE) %as%
  {
    "f64"
  }
c_r_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
c_r_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
c_r_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
c_r_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
c_r_key_rule("bool", FALSE) %as%
  {
    "bool"
  }
c_r_key_rule("_Bool", FALSE) %as%
  {
    "bool"
  }
c_r_key_rule("void", FALSE) %as%
  {
    "void"
  }
c_r_key_rule("SEXP", FALSE) %as%
  {
    "sexp"
  }
c_r_key_rule("sexp", FALSE) %as%
  {
    "sexp"
  }
c_r_key_rule(type_name, is_ptr) %as%
  {
    "ptr"
  }

## ------------------------------------------------------------------
## SEXP constructor key mapping
##
## Map C type names to the key used to select a SEXP constructor (e.g.
## mkString, ScalarReal). Handles pointer cases specially.
## ------------------------------------------------------------------
sexp_ctor_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
sexp_ctor_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
sexp_ctor_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
sexp_ctor_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
sexp_ctor_key_rule("int", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("int32_t", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("i32", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("int16_t", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("i16", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("int8_t", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("i8", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("uint8_t", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("u8", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("uint16_t", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("u16", FALSE) %as%
  {
    "int"
  }
sexp_ctor_key_rule("long", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("long long", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("int64_t", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("i64", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("uint32_t", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("u32", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("uint64_t", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("u64", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("double", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("float", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("f64", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("f32", FALSE) %as%
  {
    "real"
  }
sexp_ctor_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
sexp_ctor_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
sexp_ctor_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
sexp_ctor_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
sexp_ctor_key_rule("bool", FALSE) %as%
  {
    "bool"
  }
sexp_ctor_key_rule("_Bool", FALSE) %as%
  {
    "bool"
  }
sexp_ctor_key_rule("SEXP", FALSE) %as%
  {
    "sexp"
  }
sexp_ctor_key_rule("sexp", FALSE) %as%
  {
    "sexp"
  }
sexp_ctor_key_rule(type_name, is_ptr) %as%
  {
    "ptr"
  }

## ------------------------------------------------------------------
## R -> C key map
##
## Map C type spellings to the key that selects the appropriate R -> C
## converter (e.g. asInteger, asReal, R_ExternalPtrAddr). Pointer types
## map to "ptr".
## ------------------------------------------------------------------
r_to_c_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
r_to_c_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
r_to_c_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
r_to_c_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
r_to_c_key_rule("int", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("int32_t", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("i32", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("int16_t", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("i16", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("int8_t", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("i8", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("uint8_t", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("u8", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("uint16_t", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("u16", FALSE) %as%
  {
    "int"
  }
r_to_c_key_rule("long", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("long long", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("int64_t", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("i64", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("uint32_t", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("u32", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("uint64_t", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("u64", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("double", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("float", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("f64", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("f32", FALSE) %as%
  {
    "real"
  }
r_to_c_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
r_to_c_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
r_to_c_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
r_to_c_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
r_to_c_key_rule("bool", FALSE) %as%
  {
    "bool"
  }
r_to_c_key_rule("_Bool", FALSE) %as%
  {
    "bool"
  }
r_to_c_key_rule("SEXP", FALSE) %as%
  {
    "sexp"
  }
r_to_c_key_rule("sexp", FALSE) %as%
  {
    "sexp"
  }
r_to_c_key_rule(type_name, is_ptr) %as%
  {
    "ptr"
  }

## ------------------------------------------------------------------
## Default type key mapping
##
## A general-purpose fallback mapping used when no more specific key
## mapping is appropriate. This helps many rule lookups remain simple.
## ------------------------------------------------------------------
default_key_rule("void", FALSE) %as%
  {
    "void"
  }
default_key_rule("SEXP", FALSE) %as%
  {
    "sexp"
  }
default_key_rule("sexp", FALSE) %as%
  {
    "sexp"
  }
default_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
default_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
default_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
default_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
default_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
default_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
default_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
default_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
default_key_rule("double", FALSE) %as%
  {
    "real"
  }
default_key_rule("float", FALSE) %as%
  {
    "real"
  }
default_key_rule("f64", FALSE) %as%
  {
    "real"
  }
default_key_rule("f32", FALSE) %as%
  {
    "real"
  }
default_key_rule("bool", FALSE) %as%
  {
    "logical"
  }
default_key_rule("_Bool", FALSE) %as%
  {
    "logical"
  }
default_key_rule(type_name, is_ptr) %as%
  {
    "int"
  }

## ------------------------------------------------------------------
## SEXP type key mapping
##
## Determine which SEXP type (int/real/longlong/ptr/cstring/etc.) should
## be used for a C type when generating wrappers that will box/unbox
## values to/from R.
## ------------------------------------------------------------------
sexp_type_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
sexp_type_key_rule("int", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("int32_t", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("i32", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("int16_t", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("i16", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("int8_t", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("i8", FALSE) %as%
  {
    "int"
  }
sexp_type_key_rule("long", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("long long", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("int64_t", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("i64", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("uint8_t", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("u8", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("uint16_t", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("u16", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("uint32_t", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("u32", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("uint64_t", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("u64", FALSE) %as%
  {
    "longlong"
  }
sexp_type_key_rule("double", FALSE) %as%
  {
    "real"
  }
sexp_type_key_rule("float", FALSE) %as%
  {
    "real"
  }
sexp_type_key_rule("f64", FALSE) %as%
  {
    "real"
  }
sexp_type_key_rule("f32", FALSE) %as%
  {
    "real"
  }
sexp_type_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
sexp_type_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
sexp_type_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
sexp_type_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
sexp_type_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
sexp_type_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
sexp_type_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
sexp_type_key_rule("bool", FALSE) %as%
  {
    "bool"
  }
sexp_type_key_rule("_Bool", FALSE) %as%
  {
    "bool"
  }
sexp_type_key_rule("SEXP", FALSE) %as%
  {
    "sexp"
  }
sexp_type_key_rule("sexp", FALSE) %as%
  {
    "sexp"
  }
sexp_type_key_rule(type_name, is_ptr) %as%
  {
    "ptr"
  }

## ------------------------------------------------------------------
## Return type key mapping
##
## Map C return types to the compact keys used by the ffi_return_rule
## lookup (e.g. "i8", "i32", "double", "cstring", "ptr").
## ------------------------------------------------------------------
return_key_rule("void", FALSE) %as%
  {
    "void"
  }
return_key_rule("SEXP", FALSE) %as%
  {
    "sexp"
  }
return_key_rule("sexp", FALSE) %as%
  {
    "sexp"
  }
return_key_rule(type_name, TRUE) %as%
  {
    "ptr"
  }
return_key_rule("void*", FALSE) %as%
  {
    "ptr"
  }
return_key_rule("void *", FALSE) %as%
  {
    "ptr"
  }
return_key_rule("ptr", FALSE) %as%
  {
    "ptr"
  }
return_key_rule("char*", FALSE) %as%
  {
    "cstring"
  }
return_key_rule("const char*", FALSE) %as%
  {
    "cstring"
  }
return_key_rule("string", FALSE) %as%
  {
    "cstring"
  }
return_key_rule("cstring", FALSE) %as%
  {
    "cstring"
  }
return_key_rule("bool", FALSE) %as%
  {
    "bool"
  }
return_key_rule("_Bool", FALSE) %as%
  {
    "bool"
  }
return_key_rule("int8_t", FALSE) %as%
  {
    "i8"
  }
return_key_rule("i8", FALSE) %as%
  {
    "i8"
  }
return_key_rule("int16_t", FALSE) %as%
  {
    "i16"
  }
return_key_rule("i16", FALSE) %as%
  {
    "i16"
  }
return_key_rule("int32_t", FALSE) %as%
  {
    "i32"
  }
return_key_rule("i32", FALSE) %as%
  {
    "i32"
  }
return_key_rule("int", FALSE) %as%
  {
    "int"
  }
return_key_rule("long", FALSE) %as%
  {
    "int"
  }
return_key_rule("short", FALSE) %as%
  {
    "int"
  }
return_key_rule("uint8_t", FALSE) %as%
  {
    "u8"
  }
return_key_rule("u8", FALSE) %as%
  {
    "u8"
  }
return_key_rule("uint16_t", FALSE) %as%
  {
    "u16"
  }
return_key_rule("u16", FALSE) %as%
  {
    "u16"
  }
return_key_rule("uint32_t", FALSE) %as%
  {
    "u32"
  }
return_key_rule("u32", FALSE) %as%
  {
    "u32"
  }
return_key_rule("int64_t", FALSE) %as%
  {
    "i64"
  }
return_key_rule("i64", FALSE) %as%
  {
    "i64"
  }
return_key_rule("long long", FALSE) %as%
  {
    "i64"
  }
return_key_rule("uint64_t", FALSE) %as%
  {
    "u64"
  }
return_key_rule("u64", FALSE) %as%
  {
    "u64"
  }
return_key_rule("double", FALSE) %as%
  {
    "double"
  }
return_key_rule("f64", FALSE) %as%
  {
    "double"
  }
return_key_rule("float", FALSE) %as%
  {
    "float"
  }
return_key_rule("f32", FALSE) %as%
  {
    "float"
  }
return_key_rule(type_name, is_ptr) %as%
  {
    "ptr"
  }

## ------------------------------------------------------------------
## FFI C type canonicalization
##
## Map many C type spellings into the canonical internal FFI type keys
## used throughout this file (i8/i16/i32/i64/u8/u16/u32/u64/f32/f64/ptr
## etc.). This central lookup keeps downstream rule selection uniform.
## ------------------------------------------------------------------
ffi_c_type_map_rule(type_name, TRUE, is_ptr) %as%
  {
    "ptr"
  }
ffi_c_type_map_rule(type_name, FALSE, TRUE) %as%
  {
    "ptr"
  }
ffi_c_type_map_rule("void", FALSE, FALSE) %as%
  {
    "void"
  }
ffi_c_type_map_rule("bool", FALSE, FALSE) %as%
  {
    "bool"
  }
ffi_c_type_map_rule("_Bool", FALSE, FALSE) %as%
  {
    "bool"
  }
ffi_c_type_map_rule("SEXP", FALSE, FALSE) %as%
  {
    "sexp"
  }
ffi_c_type_map_rule("sexp", FALSE, FALSE) %as%
  {
    "sexp"
  }
ffi_c_type_map_rule("int", FALSE, FALSE) %as%
  {
    "i32"
  }
ffi_c_type_map_rule("signed", FALSE, FALSE) %as%
  {
    "i32"
  }
ffi_c_type_map_rule("int32_t", FALSE, FALSE) %as%
  {
    "i32"
  }
ffi_c_type_map_rule("__int32", FALSE, FALSE) %as%
  {
    "i32"
  }
ffi_c_type_map_rule("signed int", FALSE, FALSE) %as%
  {
    "i32"
  }
ffi_c_type_map_rule("short", FALSE, FALSE) %as%
  {
    "i16"
  }
ffi_c_type_map_rule("short int", FALSE, FALSE) %as%
  {
    "i16"
  }
ffi_c_type_map_rule("signed short", FALSE, FALSE) %as%
  {
    "i16"
  }
ffi_c_type_map_rule("signed short int", FALSE, FALSE) %as%
  {
    "i16"
  }
ffi_c_type_map_rule("int16_t", FALSE, FALSE) %as%
  {
    "i16"
  }
ffi_c_type_map_rule("char", FALSE, FALSE) %as%
  {
    "i8"
  }
ffi_c_type_map_rule("signed char", FALSE, FALSE) %as%
  {
    "i8"
  }
ffi_c_type_map_rule("int8_t", FALSE, FALSE) %as%
  {
    "i8"
  }
ffi_c_type_map_rule("long", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("long int", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("long long", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("long long int", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("signed long", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("signed long int", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("signed long long", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("signed long long int", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("int64_t", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("__int64", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("intptr_t", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("ptrdiff_t", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("ssize_t", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("off_t", FALSE, FALSE) %as%
  {
    "i64"
  }
ffi_c_type_map_rule("unsigned int", FALSE, FALSE) %as%
  {
    "u32"
  }
ffi_c_type_map_rule("unsigned", FALSE, FALSE) %as%
  {
    "u32"
  }
ffi_c_type_map_rule("uint32_t", FALSE, FALSE) %as%
  {
    "u32"
  }
ffi_c_type_map_rule("unsigned __int32", FALSE, FALSE) %as%
  {
    "u32"
  }
ffi_c_type_map_rule("unsigned short", FALSE, FALSE) %as%
  {
    "u16"
  }
ffi_c_type_map_rule("unsigned short int", FALSE, FALSE) %as%
  {
    "u16"
  }
ffi_c_type_map_rule("uint16_t", FALSE, FALSE) %as%
  {
    "u16"
  }
ffi_c_type_map_rule("unsigned char", FALSE, FALSE) %as%
  {
    "u8"
  }
ffi_c_type_map_rule("uint8_t", FALSE, FALSE) %as%
  {
    "u8"
  }
ffi_c_type_map_rule("unsigned long", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("unsigned long int", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("unsigned long long", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("unsigned long long int", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("uint64_t", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("unsigned __int64", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("size_t", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("uintptr_t", FALSE, FALSE) %as%
  {
    "u64"
  }
ffi_c_type_map_rule("double", FALSE, FALSE) %as%
  {
    "f64"
  }
ffi_c_type_map_rule("long double", FALSE, FALSE) %as%
  {
    "f64"
  }
ffi_c_type_map_rule("float", FALSE, FALSE) %as%
  {
    "f32"
  }
ffi_c_type_map_rule(type_name, is_char_ptr, is_ptr) %as%
  {
    "ptr"
  }

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

# C Wrapper Code Generator for API Mode
# Generates C code that converts R SEXP to C types and back

normalize_rtinycc_return_spec <- function(x) {
  if (inherits(x, "rtinycc_symbol_return_spec")) {
    return(x)
  }
  if (is.list(x)) {
    if (is.null(x$type)) {
      stop("Return spec missing 'type'", call. = FALSE)
    }
    return(new_rtinycc_symbol_return_spec(
      type = x$type,
      type_info = check_ffi_type(x$type, "return value"),
      length_arg = x$length_arg,
      free = isTRUE(x$free)
    ))
  }
  new_rtinycc_symbol_return_spec(
    type = x,
    type_info = check_ffi_type(x, "return value")
  )
}

rtinycc_return_type_name <- function(x) {
  normalize_rtinycc_return_spec(x)$type
}

rtinycc_return_type_info <- function(x) {
  normalize_rtinycc_return_spec(x)$type_info
}

# Generate C code to extract R SEXP to C type
generate_c_input <- function(arg_name, r_name, ffi_type) {
  type_name <- if (is_rtinycc_ffi_type(ffi_type)) ffi_type$name else ffi_type
  check_ffi_type(type_name, paste0("argument '", arg_name, "'"))
  special <- ffi_input_special_rule(type_name, arg_name, r_name)
  if (!is.null(special)) {
    return(special)
  }
  ffi_input_rule(type_name, arg_name, r_name)
}

# Generate C code to convert C return value to R SEXP
generate_c_return <- function(value_expr, ffi_type, arg_names = character()) {
  return_spec <- normalize_rtinycc_return_spec(ffi_type)
  type_info <- return_spec$type_info

  if (!is.null(type_info$kind) && type_info$kind == "array") {
    len_arg <- return_spec$length_arg
    if (is.null(len_arg) || !is.numeric(len_arg)) {
      stop("Array return requires numeric 'length_arg'", call. = FALSE)
    }
    if (len_arg < 1 || len_arg > length(arg_names)) {
      stop("length_arg out of range", call. = FALSE)
    }
    len_name <- arg_names[[as.integer(len_arg)]]
    value_var <- "__rtinycc_ret"

    alloc_line <- array_return_alloc_line(return_spec$type, len_name)
    copy_line <- array_return_copy_line(return_spec$type, len_name, value_var)

    free_line <- if (isTRUE(return_spec$free)) {
      sprintf("  if (%s) free(%s);", value_var, value_var)
    } else {
      NULL
    }

    return(paste(
      c(
        sprintf("%s %s = %s;", type_info$c_type, value_var, value_expr),
        sprintf("if (!%s) return R_NilValue;", value_var),
        alloc_line,
        copy_line,
        free_line,
        "  UNPROTECT(1);",
        "  return out;"
      ),
      collapse = "\n"
    ))
  }

  special <- ffi_return_special_rule(return_spec$type, value_expr)
  if (!is.null(special)) {
    return(special)
  }
  ffi_return_rule(return_spec$type, value_expr)
}

# Callback helper: generate a unique trampoline name per wrapper argument
callback_trampoline_name <- function(wrapper_name, arg_index) {
  paste0("trampoline_", wrapper_name, "_arg", arg_index)
}

variadic_wrapper_name <- function(wrapper_name, n_varargs) {
  paste0(wrapper_name, "__v", as.integer(n_varargs))
}

variadic_type_token <- function(x) {
  gsub("[^A-Za-z0-9_]", "_", x)
}

variadic_signature_key <- function(vararg_types) {
  if (length(vararg_types) == 0) {
    return("__none__")
  }
  paste(vararg_types, collapse = "|")
}

variadic_wrapper_name_types <- function(wrapper_name, vararg_types) {
  suffix <- if (length(vararg_types) == 0) {
    "none"
  } else {
    paste(
      vapply(vararg_types, variadic_type_token, character(1)),
      collapse = "__"
    )
  }
  paste0(wrapper_name, "__v", length(vararg_types), "__", suffix)
}

generate_variadic_type_sequences <- function(allowed_types, n_varargs) {
  if (n_varargs == 0) {
    return(list(list()))
  }
  if (length(allowed_types) == 0) {
    return(list())
  }

  idx_grid <- expand.grid(
    rep(list(seq_along(allowed_types)), n_varargs),
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  )
  lapply(seq_len(nrow(idx_grid)), function(i) {
    as.list(unname(unlist(allowed_types[as.integer(idx_grid[i, ])])))
  })
}

# Callback helper: declare a function pointer bound to a trampoline
callback_funptr_decl <- function(arg_name, sig, trampoline_name) {
  arg_types <- sig$arg_types
  c_args <- if (length(arg_types) > 0) {
    paste(
      c("void*", vapply(arg_types, callback_c_decl_type, character(1))),
      collapse = ", "
    )
  } else {
    "void*"
  }
  sprintf(
    "  %s (*%s)(%s) = %s;",
    callback_c_decl_type(sig$return_type),
    arg_name,
    c_args,
    trampoline_name
  )
}

# Generate full C wrapper function (SEXP-based only)
generate_c_wrapper <- function(
  symbol_name,
  wrapper_name,
  arg_types,
  vararg_types = list(),
  return_type,
  c_code = NULL,
  is_external = FALSE
) {
  n_args <- length(arg_types) + length(vararg_types)
  return_info <- rtinycc_return_type_info(return_type)

  # Detect if any argument is callback_async
  all_arg_types_combined <- c(arg_types, vararg_types)
  has_async <- any(vapply(
    all_arg_types_combined,
    is_callback_async_type,
    logical(1)
  ))

  # Always use R API mode: SEXP arguments and return
  # Build argument list
  if (n_args > 0) {
    fixed_arg_names <- names(arg_types)
    if (is.null(fixed_arg_names) || all(fixed_arg_names == "")) {
      fixed_arg_names <- paste0("arg", seq_along(arg_types))
    }
    vararg_names <- if (length(vararg_types) > 0) {
      paste0("vararg", seq_along(vararg_types))
    } else {
      character(0)
    }
    arg_names <- c(fixed_arg_names, vararg_names)
    all_arg_types <- c(arg_types, vararg_types)
    r_arg_names <- paste0("arg", seq_len(n_args), "_")

    # Input conversions from SEXP to C types
    input_lines <- character(0)
    for (i in seq_len(n_args)) {
      ffi_type <- all_arg_types[[i]]
      arg_name <- arg_names[[i]]
      r_name <- r_arg_names[[i]]

      if (is_callback_type(ffi_type)) {
        sig <- parse_callback_type(ffi_type)
        if (is.null(sig)) {
          stop(
            "callback type requires signature, e.g. callback:double(int)",
            call. = FALSE
          )
        }
        tramp_name <- callback_trampoline_name(
          paste0("R_wrap_", symbol_name),
          i
        )
        input_lines <- c(
          input_lines,
          sprintf(
            "  if (!Rf_inherits(%s, \"tcc_callback\")) Rf_error(\"expected tcc_callback for argument '%s'\");",
            r_name,
            arg_name
          ),
          callback_funptr_decl(arg_name, sig, tramp_name)
        )
      } else {
        input_lines <- c(
          input_lines,
          generate_c_input(arg_name, r_name, ffi_type)
        )
      }
    }

    input_conversions <- paste(input_lines, collapse = "\n")

    # Call the actual function
    call_expr <- sprintf(
      "%s(%s)",
      symbol_name,
      paste(arg_names, collapse = ", ")
    )
  } else {
    input_conversions <- "  // No arguments"
    r_arg_names <- character(0)
    arg_names <- character(0)
    all_arg_types <- list()
    call_expr <- sprintf("%s()", symbol_name)
  }

  # ---- Async wrapper path: run function on separate thread, drain on main ----
  if (has_async) {
    return(generate_async_exec_wrapper(
      symbol_name = symbol_name,
      wrapper_name = wrapper_name,
      n_args = n_args,
      arg_names = arg_names,
      all_arg_types = all_arg_types,
      r_arg_names = r_arg_names,
      input_conversions = input_conversions,
      call_expr = call_expr,
      return_type = return_type
    ))
  }

  # ---- Normal (synchronous) wrapper path ----
  # Return conversion from C type to SEXP
  return_conversion <- generate_c_return(call_expr, return_type, arg_names)

  # Build the wrapper with SEXP signature
  wrapper <- c(
    sprintf(
      "SEXP %s(%s) {",
      wrapper_name,
      if (n_args > 0) {
        paste0("SEXP ", r_arg_names, collapse = ", ")
      } else {
        "void"
      }
    ),
    input_conversions,
    "",
    "  // Call and return",
    paste("  ", strsplit(return_conversion, "\n")[[1]], collapse = "\n"),
    "}"
  )

  paste(wrapper, collapse = "\n")
}

# Generate async exec wrapper: struct + thread fn + SEXP wrapper.
# The generated SEXP wrapper:
#   1. Converts SEXP args to C types on the main thread
#   2. Packs them into a context struct
#   3. Calls RC_callback_async_exec_c(thread_fn, &ctx) which spawns a thread
#      for the real C call and drains callbacks on the main thread
#   4. Extracts the result from the struct and returns as SEXP
generate_async_exec_wrapper <- function(
  symbol_name,
  wrapper_name,
  n_args,
  arg_names,
  all_arg_types,
  r_arg_names,
  input_conversions,
  call_expr,
  return_type
) {
  return_info <- rtinycc_return_type_info(return_type)
  is_void <- identical(return_info$c_type, "void")

  # --- Build struct fields ---
  struct_fields <- character(0)
  for (i in seq_len(n_args)) {
    ffi_type <- all_arg_types[[i]]
    aname <- arg_names[[i]]
    if (is_callback_type(ffi_type)) {
      sig <- parse_callback_type(ffi_type)
      cb_arg_types <- sig$arg_types
      c_args <- if (length(cb_arg_types) > 0) {
        paste(c("void*", cb_arg_types), collapse = ", ")
      } else {
        "void*"
      }
      struct_fields <- c(
        struct_fields,
        sprintf("  %s (*%s)(%s);", sig$return_type, aname, c_args)
      )
    } else {
      type_info <- if (is_rtinycc_ffi_type(ffi_type)) {
        ffi_type
      } else {
        check_ffi_type(ffi_type, paste0("argument '", aname, "'"))
      }
      struct_fields <- c(
        struct_fields,
        sprintf("  %s %s;", type_info$c_type, aname)
      )
    }
  }
  if (!is_void) {
    struct_fields <- c(
      struct_fields,
      sprintf("  %s result;", return_info$c_type)
    )
  }

  struct_name <- paste0("_async_ctx_", wrapper_name)
  thread_fn_name <- paste0("_async_fn_", wrapper_name)

  # --- Build struct definition ---
  struct_code <- c(
    sprintf("typedef struct {"),
    struct_fields,
    sprintf("} %s;", struct_name)
  )

  # --- Build thread function ---
  # Unpacks the struct and calls the real C function
  thread_unpack <- character(0)
  thread_call_args <- character(0)
  for (i in seq_len(n_args)) {
    aname <- arg_names[[i]]
    thread_call_args <- c(thread_call_args, sprintf("ctx->%s", aname))
  }
  thread_call <- sprintf(
    "%s(%s)",
    symbol_name,
    paste(thread_call_args, collapse = ", ")
  )
  if (is_void) {
    thread_body <- sprintf("  %s;", thread_call)
  } else {
    thread_body <- sprintf("  ctx->result = %s;", thread_call)
  }

  thread_fn <- c(
    sprintf("static void %s(void *_p) {", thread_fn_name),
    sprintf("  %s *ctx = (%s *)_p;", struct_name, struct_name),
    thread_body,
    "}"
  )

  # --- Build SEXP wrapper function ---
  # Pack struct fields from the local C variables
  pack_lines <- character(0)
  for (i in seq_len(n_args)) {
    aname <- arg_names[[i]]
    pack_lines <- c(
      pack_lines,
      sprintf("  _ctx.%s = %s;", aname, aname)
    )
  }

  # Return conversion uses _ctx.result as the value expression
  if (is_void) {
    return_code <- "  return R_NilValue;"
  } else {
    return_conversion <- generate_c_return(
      "_ctx.result",
      return_type,
      arg_names
    )
    return_code <- paste0(
      "  ",
      strsplit(return_conversion, "\n")[[1]],
      collapse = "\n"
    )
  }

  sexp_wrapper <- c(
    sprintf(
      "SEXP %s(%s) {",
      wrapper_name,
      if (n_args > 0) {
        paste0("SEXP ", r_arg_names, collapse = ", ")
      } else {
        "void"
      }
    ),
    input_conversions,
    "",
    "  // Pack context struct for async exec",
    sprintf("  %s _ctx;", struct_name),
    sprintf("  memset(&_ctx, 0, sizeof(_ctx));"),
    pack_lines,
    "",
    "  // Run on thread, drain callbacks on main thread",
    sprintf("  RC_callback_async_exec_c(%s, &_ctx);", thread_fn_name),
    "",
    "  // Return result",
    return_code,
    "}"
  )

  # --- Combine all parts ---
  all_code <- c(struct_code, "", thread_fn, "", sexp_wrapper)
  paste(all_code, collapse = "\n")
}

# Generate header for external library (declare external functions)
generate_external_declarations <- function(symbols) {
  decls <- c()
  for (sym_name in names(symbols)) {
    sym <- symbols[[sym_name]]
    return_info <- if (!is.null(sym$return_spec)) {
      sym$return_spec$type_info
    } else {
      rtinycc_return_type_info(sym$returns)
    }

    arg_types <- sym$args %||% list()
    arg_info <- sym$arg_type_info %||% list()
    if (length(arg_types) > 0) {
      if (!length(arg_info)) {
        arg_info <- lapply(
          arg_types,
          check_ffi_type,
          context = paste0("symbol '", sym_name, "' argument")
        )
      }
      arg_decls <- paste(
        vapply(arg_info, function(x) x$c_type, character(1)),
        collapse = ", "
      )
    } else {
      arg_decls <- "void"
    }

    if (isTRUE(sym$variadic)) {
      decl <- sprintf(
        "extern %s %s(%s, ...);",
        return_info$c_type,
        sym_name,
        arg_decls
      )
    } else {
      decl <- sprintf(
        "extern %s %s(%s);",
        return_info$c_type,
        sym_name,
        arg_decls
      )
    }
    decls <- c(decls, decl)
  }

  paste(decls, collapse = "\n")
}

# Generate all wrappers for a set of symbols
generate_wrappers <- function(
  symbols,
  prefix = "R_wrap_",
  is_external = FALSE
) {
  wrappers <- c()

  for (sym_name in names(symbols)) {
    sym <- symbols[[sym_name]]
    base_wrapper_name <- paste0(prefix, sym_name)

    if (isTRUE(sym$variadic)) {
      vararg_mode <- sym$varargs_mode %||% "prefix"

      if (identical(vararg_mode, "types")) {
        allowed_types <- sym$varargs_types %||% list()
        min_varargs <- sym$varargs_min %||% 0L
        max_varargs <- sym$varargs_max %||% min_varargs

        for (n_varargs in seq.int(min_varargs, max_varargs)) {
          type_sequences <- generate_variadic_type_sequences(
            allowed_types,
            n_varargs
          )

          for (this_varargs in type_sequences) {
            wrapper <- generate_c_wrapper(
              symbol_name = sym_name,
              wrapper_name = variadic_wrapper_name_types(
                base_wrapper_name,
                this_varargs
              ),
              arg_types = sym$args,
              vararg_types = this_varargs,
              return_type = sym$returns,
              is_external = is_external
            )

            wrappers <- c(wrappers, wrapper, "", "")
          }
        }
      } else {
        all_varargs <- sym$varargs %||% list()
        max_varargs <- length(all_varargs)
        min_varargs <- sym$varargs_min %||% max_varargs

        for (n_varargs in seq.int(min_varargs, max_varargs)) {
          this_varargs <- if (n_varargs > 0) {
            all_varargs[seq_len(n_varargs)]
          } else {
            list()
          }

          wrapper <- generate_c_wrapper(
            symbol_name = sym_name,
            wrapper_name = variadic_wrapper_name(base_wrapper_name, n_varargs),
            arg_types = sym$args,
            vararg_types = this_varargs,
            return_type = sym$returns,
            is_external = is_external
          )

          wrappers <- c(wrappers, wrapper, "", "")
        }
      }
    } else {
      wrapper <- generate_c_wrapper(
        symbol_name = sym_name,
        wrapper_name = base_wrapper_name,
        arg_types = sym$args,
        vararg_types = sym$varargs %||% list(),
        return_type = sym$returns,
        is_external = is_external
      )

      wrappers <- c(wrappers, wrapper, "", "")
    }
  }

  paste(wrappers, collapse = "\n")
}

# ============================================================================
# Struct Helper Generation
# ============================================================================

# Generate struct helper functions (new, free, get/set, container_of, etc.)
generate_struct_helpers <- function(
  structs,
  container_of,
  field_addr,
  raw_access,
  introspect
) {
  if (is.null(structs) || length(structs) == 0) {
    return("")
  }

  helpers <- c("/* Struct helper functions */", "")

  for (struct_name in names(structs)) {
    fields <- structs[[struct_name]]

    # Constructor
    helpers <- c(helpers, generate_struct_new(struct_name))

    # Free function
    helpers <- c(helpers, generate_struct_free(struct_name))

    # Field getters and setters
    for (field_name in names(fields)) {
      field_spec <- fields[[field_name]]
      is_array <- is.list(field_spec) && isTRUE(field_spec$array)
      helpers <- c(
        helpers,
        generate_struct_getter(struct_name, field_name, field_spec)
      )
      if (is_array) {
        helpers <- c(
          helpers,
          generate_struct_array_getter(struct_name, field_name, field_spec),
          generate_struct_array_setter(struct_name, field_name, field_spec)
        )
      } else {
        helpers <- c(
          helpers,
          generate_struct_setter(struct_name, field_name, field_spec)
        )
      }
    }

    # container_of helpers
    if (!is.null(container_of) && struct_name %in% names(container_of)) {
      for (member_name in container_of[[struct_name]]) {
        helpers <- c(helpers, generate_container_of(struct_name, member_name))
      }
    }

    # Field address helpers
    if (!is.null(field_addr) && struct_name %in% names(field_addr)) {
      for (field_name in field_addr[[struct_name]]) {
        helpers <- c(helpers, generate_field_addr(struct_name, field_name))
      }
    }

    # Raw access helpers
    if (!is.null(raw_access) && struct_name %in% raw_access) {
      helpers <- c(helpers, generate_struct_raw_access(struct_name))
    }

    # Introspection helpers
    if (!is.null(introspect) && introspect) {
      helpers <- c(helpers, generate_struct_introspection(struct_name, fields))
    }

    helpers <- c(helpers, "")
  }

  paste(helpers, collapse = "\n")
}

# Generate constructor
generate_struct_new <- function(struct_name) {
  c(
    sprintf("SEXP R_wrap_struct_%s_new(void) {", struct_name),
    sprintf(
      "  struct %s *p = calloc(1, sizeof(struct %s));",
      struct_name,
      struct_name
    ),
    "  if (!p) Rf_error(\"Out of memory\");",
    sprintf(
      "  return RC_make_owned_composite_ptr(p, Rf_install(\"struct_%s\"));",
      struct_name
    ),
    "}",
    ""
  )
}

# Generate free function
generate_struct_free <- function(struct_name) {
  c(
    sprintf("SEXP R_wrap_struct_%s_free(SEXP ext) {", struct_name),
    "  if (R_ExternalPtrProtected(ext) != R_NilValue) {",
    "    Rf_error(\"Cannot free borrowed view; free the owning object instead\");",
    "  }",
    "  RC_free_finalizer(ext);",
    "  return R_NilValue;",
    "}",
    ""
  )
}

# Generate field getter
generate_struct_getter <- function(struct_name, field_name, field_spec) {
  if (is.list(field_spec) && isTRUE(field_spec$array)) {
    return(c(
      sprintf(
        "SEXP R_wrap_struct_%s_get_%s(SEXP ext) {",
        struct_name,
        field_name
      ),
      sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
      sprintf("  if (!p) Rf_error(\"Null pointer\");"),
      sprintf(
        "  return RC_make_borrowed_view(p->%s, Rf_install(\"rtinycc_borrowed\"), ext);",
        field_name
      ),
      "}",
      ""
    ))
  }
  if (is.list(field_spec)) {
    type_name <- field_spec$type %||% "ptr"
  } else {
    type_name <- field_spec
  }

  if (startsWith(type_name, "struct:")) {
    nested_struct_name <- sub("^struct:", "", type_name)
    return(c(
      sprintf(
        "SEXP R_wrap_struct_%s_get_%s(SEXP ext) {",
        struct_name,
        field_name
      ),
      sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
      sprintf("  if (!p) Rf_error(\"Null pointer\");"),
      sprintf(
        "  return RC_make_borrowed_view(&p->%s, Rf_install(\"struct_%s\"), ext);",
        field_name,
        nested_struct_name
      ),
      "}",
      ""
    ))
  }

  return_code <- struct_field_getter_lines(type_name, field_name)

  c(
    sprintf(
      "SEXP R_wrap_struct_%s_get_%s(SEXP ext) {",
      struct_name,
      field_name
    ),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    paste("  ", return_code, collapse = "\n"),
    "}",
    ""
  )
}

generate_struct_array_getter <- function(struct_name, field_name, field_spec) {
  type_name <- field_spec$type %||% "u8"
  size <- field_spec$size
  if (is.null(size) || !is.numeric(size)) {
    stop("Array field '", field_name, "' missing size", call. = FALSE)
  }

  return_code <- struct_array_field_getter_lines(type_name, field_name)

  c(
    sprintf(
      "SEXP R_wrap_struct_%s_get_%s_elt(SEXP ext, SEXP idx_) {",
      struct_name,
      field_name
    ),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    "  int idx = asInteger(idx_);",
    sprintf(
      "  if (idx < 0 || idx >= %d) Rf_error(\"index out of bounds\");",
      as.integer(size)
    ),
    paste("  ", return_code, collapse = "\n"),
    "}",
    ""
  )
}

generate_struct_array_setter <- function(struct_name, field_name, field_spec) {
  type_name <- field_spec$type %||% "u8"
  size <- field_spec$size
  if (is.null(size) || !is.numeric(size)) {
    stop("Array field '", field_name, "' missing size", call. = FALSE)
  }

  setter_code <- struct_array_field_setter_lines(type_name, field_name)

  c(
    sprintf(
      "SEXP R_wrap_struct_%s_set_%s_elt(SEXP ext, SEXP idx_, SEXP val) {",
      struct_name,
      field_name
    ),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    "  int idx = asInteger(idx_);",
    sprintf(
      "  if (idx < 0 || idx >= %d) Rf_error(\"index out of bounds\");",
      as.integer(size)
    ),
    paste("  ", setter_code, collapse = "\n"),
    "  return ext;",
    "}",
    ""
  )
}

# Generate field setter
generate_struct_setter <- function(struct_name, field_name, field_spec) {
  if (is.list(field_spec)) {
    type_name <- field_spec$type %||% "ptr"
    size <- field_spec$size
  } else {
    type_name <- field_spec
    size <- NULL
  }

  if (startsWith(type_name, "struct:")) {
    nested_struct_name <- sub("^struct:", "", type_name)
    return(c(
      sprintf(
        "SEXP R_wrap_struct_%s_set_%s(SEXP ext, SEXP val) {",
        struct_name,
        field_name
      ),
      sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
      sprintf("  if (!p) Rf_error(\"Null pointer\");"),
      sprintf(
        "  struct %s *child = R_ExternalPtrAddr(val);",
        nested_struct_name
      ),
      "  if (!child) Rf_error(\"Null pointer\");",
      sprintf(
        "  memcpy(&p->%s, child, sizeof(struct %s));",
        field_name,
        nested_struct_name
      ),
      "  return ext;",
      "}",
      ""
    ))
  }

  setter_code <- struct_field_setter_lines(type_name, field_name, size)

  c(
    sprintf(
      "SEXP R_wrap_struct_%s_set_%s(SEXP ext, SEXP val) {",
      struct_name,
      field_name
    ),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    paste("  ", setter_code, collapse = "\n"),
    "  return ext;", # Return ext for chaining
    "}",
    ""
  )
}

# Generate container_of helper
generate_container_of <- function(struct_name, member_name) {
  c(
    sprintf(
      "SEXP R_wrap_struct_%s_from_%s(SEXP ext) {",
      struct_name,
      member_name
    ),
    sprintf("  void *member_ptr = R_ExternalPtrAddr(ext);"),
    sprintf("  if (!member_ptr) Rf_error(\"Null pointer\");"),
    sprintf(
      "  struct %s *p = (struct %s *)((char *)member_ptr - offsetof(struct %s, %s));",
      struct_name,
      struct_name,
      struct_name,
      member_name
    ),
    sprintf(
      "  return RC_make_borrowed_view(p, Rf_install(\"struct_%s\"), ext);",
      struct_name
    ),
    "}",
    ""
  )
}

# Generate field address getter
generate_field_addr <- function(struct_name, field_name) {
  c(
    sprintf(
      "SEXP R_wrap_struct_%s_%s_addr(SEXP ext) {",
      struct_name,
      field_name
    ),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    sprintf("  void *field_ptr = &p->%s;", field_name),
    sprintf(
      "  return RC_make_borrowed_view(field_ptr, Rf_install(\"rtinycc_borrowed\"), ext);"
    ),
    "}",
    ""
  )
}

# Generate raw access helpers
generate_struct_raw_access <- function(struct_name) {
  c(
    sprintf("SEXP R_wrap_struct_%s_get_raw(SEXP ext, SEXP len) {", struct_name),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    "  int n = asInteger(len);",
    "  SEXP raw = PROTECT(allocVector(RAWSXP, n));",
    sprintf(
      "  memcpy(RAW(raw), p, (n < sizeof(struct %s)) ? n : sizeof(struct %s));",
      struct_name,
      struct_name
    ),
    "  UNPROTECT(1);",
    "  return raw;",
    "}",
    "",
    sprintf("SEXP R_wrap_struct_%s_set_raw(SEXP ext, SEXP raw) {", struct_name),
    sprintf("  struct %s *p = R_ExternalPtrAddr(ext);", struct_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    "  int n = LENGTH(raw);",
    sprintf(
      "  memcpy(p, RAW(raw), (n < sizeof(struct %s)) ? n : sizeof(struct %s));",
      struct_name,
      struct_name
    ),
    "  return R_NilValue;",
    "}",
    ""
  )
}

# Generate introspection helpers
generate_struct_introspection <- function(struct_name, fields) {
  c(
    sprintf("SEXP R_wrap_struct_%s_sizeof(void) {", struct_name),
    sprintf("  return ScalarInteger(sizeof(struct %s));", struct_name),
    "}",
    "",
    sprintf("SEXP R_wrap_struct_%s_alignof(void) {", struct_name),
    sprintf("  return ScalarInteger(_Alignof(struct %s));", struct_name),
    "}",
    ""
  )
}

# ============================================================================
# Union Helper Generation
# ============================================================================

generate_union_helpers <- function(unions, introspect) {
  if (is.null(unions) || length(unions) == 0) {
    return("")
  }

  helpers <- c("/* Union helper functions */", "")

  for (union_name in names(unions)) {
    union_def <- unions[[union_name]]
    members <- union_def$members

    # Constructor
    helpers <- c(helpers, generate_union_new(union_name))

    # Free function
    helpers <- c(helpers, generate_union_free(union_name))

    # Member getters and setters
    for (mem_name in names(members)) {
      mem_spec <- members[[mem_name]]
      helpers <- c(
        helpers,
        generate_union_getter(union_name, mem_name, mem_spec)
      )
      helpers <- c(
        helpers,
        generate_union_setter(union_name, mem_name, mem_spec)
      )
    }

    # Introspection
    if (!is.null(introspect) && introspect) {
      helpers <- c(helpers, generate_union_introspection(union_name))
    }

    helpers <- c(helpers, "")
  }

  paste(helpers, collapse = "\n")
}

generate_union_new <- function(union_name) {
  c(
    sprintf("SEXP R_wrap_union_%s_new(void) {", union_name),
    sprintf(
      "  union %s *p = calloc(1, sizeof(union %s));",
      union_name,
      union_name
    ),
    "  if (!p) Rf_error(\"Out of memory\");",
    sprintf(
      "  return RC_make_owned_composite_ptr(p, Rf_install(\"union_%s\"));",
      union_name
    ),
    "}",
    ""
  )
}

generate_union_free <- function(union_name) {
  c(
    sprintf("SEXP R_wrap_union_%s_free(SEXP ext) {", union_name),
    "  if (R_ExternalPtrProtected(ext) != R_NilValue) {",
    "    Rf_error(\"Cannot free borrowed view; free the owning object instead\");",
    "  }",
    "  RC_free_finalizer(ext);",
    "  return R_NilValue;",
    "}",
    ""
  )
}

generate_union_getter <- function(union_name, mem_name, mem_spec) {
  if (
    is.list(mem_spec) && !is.null(mem_spec$type) && mem_spec$type == "struct"
  ) {
    struct_name <- mem_spec$struct_name %||% mem_name
    # Nested struct in union
    return(c(
      sprintf("SEXP R_wrap_union_%s_get_%s(SEXP ext) {", union_name, mem_name),
      sprintf("  union %s *p = R_ExternalPtrAddr(ext);", union_name),
      sprintf("  if (!p) Rf_error(\"Null pointer\");"),
      sprintf(
        "  return RC_make_borrowed_view(&p->%s, Rf_install(\"struct_%s\"), ext);",
        mem_name,
        struct_name
      ),
      "}",
      ""
    ))
  }

  type_name <- if (is.list(mem_spec)) mem_spec$type else mem_spec
  return_code <- union_field_getter_lines(type_name, mem_name)

  c(
    sprintf("SEXP R_wrap_union_%s_get_%s(SEXP ext) {", union_name, mem_name),
    sprintf("  union %s *p = R_ExternalPtrAddr(ext);", union_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    paste("  ", return_code, collapse = "\n"),
    "}",
    ""
  )
}

array_return_alloc_line <- function(base_type, len_name) {
  array_return_alloc_line_rule(base_type, len_name)
}

array_return_copy_line <- function(base_type, len_name, value_expr) {
  array_return_copy_line_rule(base_type, len_name, value_expr)
}

struct_field_getter_lines <- function(type_name, field_name) {
  struct_field_getter_rule(type_name, field_name)
}

struct_array_field_getter_lines <- function(type_name, field_name) {
  struct_array_field_getter_rule(type_name, field_name)
}

struct_array_field_setter_lines <- function(type_name, field_name) {
  struct_array_field_setter_rule(type_name, field_name)
}

struct_field_setter_lines <- function(type_name, field_name, size) {
  struct_field_setter_rule(type_name, field_name, size)
}

union_field_getter_lines <- function(type_name, mem_name) {
  union_field_getter_rule(type_name, mem_name)
}

union_field_setter_lines <- function(type_name, mem_name, size) {
  union_field_setter_rule(type_name, mem_name, size)
}

ffi_input_array_check_rule <- function(ffi_type, arg_name, r_name) {
  check_line <- switch(
    ffi_type,
    raw = sprintf(
      "  if (TYPEOF(%s) != RAWSXP) Rf_error(\"expected raw vector for argument '%s'\");",
      r_name,
      arg_name
    ),
    integer_array = sprintf(
      "  if (TYPEOF(%s) != INTSXP) Rf_error(\"expected integer vector for argument '%s'\");",
      r_name,
      arg_name
    ),
    numeric_array = sprintf(
      "  if (TYPEOF(%s) != REALSXP) Rf_error(\"expected numeric vector for argument '%s'\");",
      r_name,
      arg_name
    ),
    logical_array = sprintf(
      "  if (TYPEOF(%s) != LGLSXP) Rf_error(\"expected logical vector for argument '%s'\");",
      r_name,
      arg_name
    ),
    character_array = sprintf(
      "  if (!Rf_isString(%s)) Rf_error(\"expected character vector for argument '%s'\");",
      r_name,
      arg_name
    ),
    NULL
  )

  if (is.null(check_line)) {
    return(NULL)
  }

  paste(c(check_line, ffi_input_rule(ffi_type, arg_name, r_name)), collapse = "\n")
}

ffi_input_special_rule <- function(ffi_type, arg_name, r_name) {
  special_map <- list(
    raw = function() ffi_input_array_check_rule(ffi_type, arg_name, r_name),
    integer_array = function() ffi_input_array_check_rule(ffi_type, arg_name, r_name),
    numeric_array = function() ffi_input_array_check_rule(ffi_type, arg_name, r_name),
    logical_array = function() ffi_input_array_check_rule(ffi_type, arg_name, r_name),
    character_array = function() ffi_input_array_check_rule(ffi_type, arg_name, r_name),
    "enum:" = function() ffi_input_enum_rule(ffi_type, arg_name, r_name),
    "callback" = function() ffi_input_callback_rule(ffi_type, arg_name, r_name)
  )
  matched <- names(special_map)[vapply(
    names(special_map),
    startsWith,
    logical(1),
    x = ffi_type
  )]
  if (length(matched) == 0) {
    return(NULL)
  }
  special_map[[matched[[1]]]]()
}

ffi_return_special_rule <- function(ffi_type, value_expr) {
  special_map <- list(
    "enum:" = function() ffi_return_enum_rule(ffi_type, value_expr)
  )
  matched <- names(special_map)[vapply(
    names(special_map),
    startsWith,
    logical(1),
    x = ffi_type
  )]
  if (length(matched) == 0) {
    return(NULL)
  }
  special_map[[matched[[1]]]]()
}

generate_union_setter <- function(union_name, mem_name, mem_spec) {
  if (
    is.list(mem_spec) && !is.null(mem_spec$type) && mem_spec$type == "struct"
  ) {
    # TODO : this looks fake, we could allocate memcopy another memory address into it
    return(c()) # Cannot set whole struct easily
  }

  type_name <- if (is.list(mem_spec)) mem_spec$type else mem_spec

  size <- if (is.list(mem_spec)) mem_spec$size else NULL
  setter_code <- union_field_setter_lines(type_name, mem_name, size)

  c(
    sprintf(
      "SEXP R_wrap_union_%s_set_%s(SEXP ext, SEXP val) {",
      union_name,
      mem_name
    ),
    sprintf("  union %s *p = R_ExternalPtrAddr(ext);", union_name),
    sprintf("  if (!p) Rf_error(\"Null pointer\");"),
    paste("  ", setter_code, collapse = "\n"),
    "  return ext;",
    "}",
    ""
  )
}

generate_union_introspection <- function(union_name) {
  c(
    sprintf("SEXP R_wrap_union_%s_sizeof(void) {", union_name),
    sprintf("  return ScalarInteger(sizeof(union %s));", union_name),
    "}",
    "",
    sprintf("SEXP R_wrap_union_%s_alignof(void) {", union_name),
    sprintf("  return ScalarInteger(_Alignof(union %s));", union_name),
    "}",
    ""
  )
}

# ============================================================================
# Enum Helper Generation
# ============================================================================

generate_enum_helpers <- function(enums, introspect) {
  if (is.null(enums) || length(enums) == 0) {
    return("")
  }

  helpers <- c("/* Enum helper functions */", "")

  for (enum_name in names(enums)) {
    enum_def <- enums[[enum_name]]

    # Sizeof helper
    if (!is.null(introspect) && introspect) {
      helpers <- c(
        helpers,
        c(
          sprintf("SEXP R_wrap_enum_%s_sizeof(void) {", enum_name),
          sprintf("  return ScalarInteger(sizeof(enum %s));", enum_name),
          "}",
          ""
        )
      )
    }

    # Constant export helpers
    if (!is.null(enum_def$constants)) {
      for (const_name in enum_def$constants) {
        helpers <- c(
          helpers,
          c(
            sprintf("SEXP R_wrap_enum_%s_%s(void) {", enum_name, const_name),
            sprintf("  return ScalarInteger(%s);", const_name),
            "}",
            ""
          )
        )
      }
    }

    helpers <- c(helpers, "")
  }

  paste(helpers, collapse = "\n")
}

# Generate global getter helpers
generate_global_helpers <- function(globals) {
  if (is.null(globals) || length(globals) == 0) {
    return("")
  }

  helpers <- c("/* Global getter helpers */", "")

  for (global_name in names(globals)) {
    ffi_type <- globals[[global_name]]
    type_info <- check_ffi_type(ffi_type, "global")
    if (!is.null(type_info$kind) && type_info$kind == "array") {
      stop("Global type cannot be an array type", call. = FALSE)
    }
    if (ffi_type == "void") {
      stop("Global type cannot be void", call. = FALSE)
    }

    c_type <- type_info$c_type
    get_ret <- generate_c_return(global_name, ffi_type)
    get_ret_lines <- paste0("  ", strsplit(get_ret, "\n")[[1]])

    set_arg <- "value_"
    set_input <- generate_c_input("value", set_arg, ffi_type)
    set_input_lines <- strsplit(set_input, "\n")[[1]]
    set_ret <- generate_c_return(global_name, ffi_type)
    set_ret_lines <- paste0("  ", strsplit(set_ret, "\n")[[1]])

    helpers <- c(
      helpers,
      sprintf("extern %s %s;", c_type, global_name),
      sprintf("SEXP R_wrap_global_%s_get(void) {", global_name),
      get_ret_lines,
      "}",
      "",
      sprintf("SEXP R_wrap_global_%s_set(SEXP %s) {", global_name, set_arg),
      set_input_lines,
      sprintf("  %s = value;", global_name),
      set_ret_lines,
      "}",
      ""
    )
  }

  paste(helpers, collapse = "\n")
}

# ============================================================================
# Main Code Generation
# ============================================================================

# Generate complete C code for FFI compilation (always R API mode)
generate_ffi_code <- function(
  symbols,
  headers = NULL,
  c_code = NULL,
  is_external = FALSE,
  structs = NULL,
  unions = NULL,
  enums = NULL,
  globals = NULL,
  container_of = NULL,
  field_addr = NULL,
  struct_raw_access = NULL,
  introspect = NULL
) {
  if (!is.null(symbols) && length(symbols) > 0) {
    symbols <- stats::setNames(
      lapply(names(symbols), function(sym_name) {
        as_rtinycc_bound_symbol(sym_name, symbols[[sym_name]])
      }),
      names(symbols)
    )
  }

  parts <- c()

  # TinyCC workaround: Define _Complex as empty since TinyCC doesn't support C99 complex types
  # This is required when using R headers which use _Complex
  parts <- c(
    parts,
    "/* TinyCC workaround: _Complex not supported */",
    "#define _Complex",
    ""
  )

  # Always include R headers for SEXP-based wrappers
  parts <- c(
    parts,
    "#include <R.h>",
    "#include <Rinternals.h>",
    "#ifndef STRING_PTR_RO",
    "#define STRING_PTR_RO STRING_PTR",
    "#endif",
    "void RC_free_finalizer(SEXP ext);",
    "void RC_owned_native_finalizer(SEXP ext);",
    "SEXP RC_make_borrowed_view(void *ptr, SEXP tag, SEXP owner);",
    "SEXP RC_make_unowned_ptr(void *ptr, SEXP tag);",
    "SEXP RC_make_owned_ptr(void *ptr, SEXP tag);",
    "SEXP RC_make_owned_composite_ptr(void *ptr, SEXP tag);",
    ""
  )

  parts <- c(
    parts,
    "#include <stdint.h>",
    "#include <stdbool.h>",
    "#include <stddef.h>",
    "#include <limits.h>",
    "#include <math.h>",
    "#include <string.h>",
    ""
  )

  cb_tramps <- generate_callback_trampolines(symbols)
  if (nzchar(cb_tramps$code)) {
    parts <- c(
      parts,
      "",
      "/* Callback trampoline support */",
      "typedef struct { int id; int refs; int origin_id; } callback_token_t;",
      if (cb_tramps$needs_sync) {
        "SEXP RC_invoke_callback_id(int, SEXP);"
      } else {
        NULL
      },
      if (cb_tramps$needs_async) {
        c(
          "typedef enum {",
          "  CB_ARG_INT = 0,",
          "  CB_ARG_REAL = 1,",
          "  CB_ARG_LOGICAL = 2,",
          "  CB_ARG_PTR = 3,",
          "  CB_ARG_CSTRING = 4",
          "} cb_arg_kind_t;",
          "",
          "typedef struct {",
          "  cb_arg_kind_t kind;",
          "  union { int i; double d; void* p; char* s; } v;",
          "} cb_arg_t;",
          "",
          "int RC_callback_async_schedule_c(int id, int n_args, const cb_arg_t *args);",
          "/* Drain pending async callbacks from main-thread C code */",
          "void RC_callback_async_drain_c(void);",
          "/* Run func(arg) on a new thread, drain callbacks on the main thread */",
          "void RC_callback_async_exec_c(void (*func)(void *), void *arg);"
        )
      } else {
        NULL
      },
      if (cb_tramps$needs_async_sync) {
        c(
          "",
          "typedef enum {",
          "  CB_RESULT_VOID = 0,",
          "  CB_RESULT_INT = 1,",
          "  CB_RESULT_REAL = 2,",
          "  CB_RESULT_LOGICAL = 3,",
          "  CB_RESULT_PTR = 4",
          "} cb_result_kind_t;",
          "",
          "typedef struct {",
          "  cb_result_kind_t kind;",
          "  union { int i; double d; void* p; } v;",
          "} cb_result_t;",
          "",
          "int RC_callback_async_schedule_sync_c(int id, int n_args, const cb_arg_t *args, cb_result_t *result);",
          "/* Drain loop: services callbacks via select()/MsgWait until *done_flag != 0. */",
          "void RC_callback_async_drain_loop_c(volatile int *done_flag);"
        )
      } else {
        NULL
      },
      "",
      cb_tramps$code,
      ""
    )
  }

  # Custom headers
  if (!is.null(headers) && length(headers) > 0) {
    parts <- c(parts, "/* User headers */", headers, "")
  }

  # External declarations
  if (is_external) {
    parts <- c(
      parts,
      "/* External library declarations */",
      generate_external_declarations(symbols),
      ""
    )
  }

  # User C code
  if (!is.null(c_code) && length(c_code) > 0) {
    c_code_lines <- c_code[nzchar(c_code)]
    if (length(c_code_lines) > 0) {
      parts <- c(parts, "/* User code */", c_code_lines, "")
    }
  }

  # Struct helpers
  struct_code <- generate_struct_helpers(
    structs,
    container_of,
    field_addr,
    struct_raw_access,
    introspect
  )
  if (nzchar(struct_code)) {
    parts <- c(parts, struct_code, "")
  }

  # Union helpers
  union_code <- generate_union_helpers(unions, introspect)
  if (nzchar(union_code)) {
    parts <- c(parts, union_code, "")
  }

  # Enum helpers
  enum_code <- generate_enum_helpers(enums, introspect)
  if (nzchar(enum_code)) {
    parts <- c(parts, enum_code, "")
  }

  # Global helpers
  global_code <- generate_global_helpers(globals)
  if (nzchar(global_code)) {
    parts <- c(parts, global_code, "")
  }

  # Wrapper functions for bound symbols
  parts <- c(
    parts,
    "/* R callable wrappers for bound symbols */",
    generate_wrappers(symbols, is_external = is_external)
  )

  paste(parts, collapse = "\n")
}

# Generate trampoline functions for callback arguments
generate_callback_trampolines <- function(symbols) {
  if (is.null(symbols) || length(symbols) == 0) {
    return(list(code = "", needs_sync = FALSE, needs_async = FALSE))
  }

  trampolines <- character(0)
  needs_sync <- FALSE
  needs_async <- FALSE
  needs_async_sync <- FALSE
  for (sym_name in names(symbols)) {
    sym <- symbols[[sym_name]]
    arg_types <- sym$args
    if (length(arg_types) == 0) {
      next
    }
    for (i in seq_along(arg_types)) {
      ffi_type <- arg_types[[i]]
      if (is_callback_type(ffi_type)) {
        sig <- parse_callback_type(ffi_type)
        if (is.null(sig)) {
          stop(
            "callback type requires signature, e.g. callback:double(int)",
            call. = FALSE
          )
        }
        wrapper_name <- paste0("R_wrap_", sym_name)
        tramp_name <- callback_trampoline_name(wrapper_name, i)
        if (is_callback_async_type(ffi_type)) {
          needs_async <- TRUE
          if (!identical(sig$return_type, "void")) {
            needs_async_sync <- TRUE
          }
          trampolines <- c(
            trampolines,
            generate_async_trampoline(tramp_name, sig),
            ""
          )
        } else {
          needs_sync <- TRUE
          trampolines <- c(
            trampolines,
            generate_trampoline(tramp_name, sig),
            ""
          )
        }
      }
    }
  }

  list(
    code = paste(trampolines, collapse = "\n"),
    needs_sync = needs_sync,
    needs_async = needs_async,
    needs_async_sync = needs_async_sync
  )
}

# Test FFI Code Generation
# Tests for C wrapper code generation

library(Rtinycc)
library(tinytest)

callback_abi_specs <- Rtinycc:::rtinycc_callback_abi_specs()

# Test 1: Generate C input code for R API mode (SEXP conversion)
code <- Rtinycc:::generate_c_input("x", "arg1_", "i32")
expect_true(grepl("int _x = asInteger", code))
expect_true(grepl("int32_t x = (int32_t)_x", code, fixed = TRUE))

code <- Rtinycc:::generate_c_input("y", "arg2_", "f64")
expect_true(grepl("double y = asReal", code))

code <- Rtinycc:::generate_c_input("u32_value", "arg4_", "u32")
expect_true(grepl("double _u32_value = asReal", code))
expect_true(grepl("uint32_t u32_value = \\(uint32_t\\)_u32_value", code))

code <- Rtinycc:::generate_c_input("i64_value", "arg5_", "i64")
expect_true(grepl("double _i64_value = asReal", code))
expect_true(grepl("int64_t i64_value = \\(int64_t\\)_i64_value", code))

code <- Rtinycc:::generate_c_input("u64_value", "arg6_", "u64")
expect_true(grepl("double _u64_value = asReal", code))
expect_true(grepl("uint64_t u64_value = \\(uint64_t\\)_u64_value", code))

code <- Rtinycc:::generate_c_input("flag", "arg3_", "bool")
expect_true(grepl("int _flag = asLogical", code))
expect_true(grepl("bool flag = (bool)(_flag != 0)", code, fixed = TRUE))

# Test 2: Generate C input for array types
code <- Rtinycc:::generate_c_input("buf", "arg1_", "raw")
expect_true(grepl("TYPEOF\\(arg1_\\) != RAWSXP", code))
expect_true(grepl("uint8_t[*] buf = RAW", code))

code <- Rtinycc:::generate_c_input("arr", "arg1_", "integer_array")
expect_true(grepl("TYPEOF\\(arg1_\\) != INTSXP", code))
expect_true(grepl("int32_t[*] arr = INTEGER", code))

code <- Rtinycc:::generate_c_input("nums", "arg1_", "numeric_array")
expect_true(grepl("TYPEOF\\(arg1_\\) != REALSXP", code))
expect_true(grepl("double[*] nums = REAL", code))

code <- Rtinycc:::generate_c_input("flags", "arg1_", "logical_array")
expect_true(grepl("TYPEOF\\(arg1_\\) != LGLSXP", code))
expect_true(grepl("int[*] flags = LOGICAL", code))

code <- Rtinycc:::generate_c_input("chars", "arg1_", "character_array")
expect_true(grepl("Rf_isString\\(arg1_\\)", code))
expect_true(grepl("const SEXP[*] chars = STRING_PTR_RO", code))
expect_true(grepl("STRING_PTR_RO", code, fixed = TRUE))

# Test 3: Generate C return code
rc <- Rtinycc:::generate_c_return("result", "i32")
expect_true(grepl("ScalarInteger", rc))

rc <- Rtinycc:::generate_c_return("result", "f64")
expect_true(grepl("ScalarReal", rc))

rc <- Rtinycc:::generate_c_return("result", "u32")
expect_true(grepl("ScalarReal", rc))

rc <- Rtinycc:::generate_c_return("result", "i64")
expect_true(grepl("ScalarReal", rc))

rc <- Rtinycc:::generate_c_return("result", "u64")
expect_true(grepl("ScalarReal", rc))

rc <- Rtinycc:::generate_c_return("result", "cstring")
expect_true(grepl("mkString", rc))

# Test 3b: Array return expressions are evaluated once and reused
rc <- Rtinycc:::generate_c_return(
  "rand_unif(arg1)",
  list(type = "numeric_array", length_arg = 1, free = TRUE),
  arg_names = c("arg1")
)
expect_true(grepl("double\\* __rtinycc_ret = rand_unif\\(arg1\\);", rc))
expect_true(grepl("memcpy\\(REAL\\(out\\), __rtinycc_ret,", rc))
expect_true(grepl("if \\(__rtinycc_ret\\) free\\(__rtinycc_ret\\);", rc))
expect_false(grepl("memcpy\\(REAL\\(out\\), rand_unif\\(arg1\\)", rc))
expect_false(grepl("free\\(rand_unif\\(arg1\\)\\)", rc))

# Test 3c: Scalar return expressions with warnings/branching are evaluated once
rc <- Rtinycc:::generate_c_return("next_i64()", "i64")
expect_true(grepl("int64_t __rtinycc_ret = next_i64\\(\\);", rc))
expect_false(grepl("next_i64\\(\\).+next_i64\\(\\)", rc))

rc <- Rtinycc:::generate_c_return("next_u64()", "u64")
expect_true(grepl("uint64_t __rtinycc_ret = next_u64\\(\\);", rc))
expect_false(grepl("next_u64\\(\\).+next_u64\\(\\)", rc))

rc <- Rtinycc:::generate_c_return("next_str()", "cstring")
expect_true(grepl("char\\* __rtinycc_ret = next_str\\(\\);", rc))
expect_true(grepl("mkString\\(__rtinycc_ret\\)", rc))
expect_false(grepl("next_str\\(\\).+next_str\\(\\)", rc))

sexp_call <- Rtinycc:::get_sexp_constructor_call("const char*", "next_str()")
expect_true(grepl("__rtinycc_tmp = next_str\\(\\);", sexp_call))
expect_false(grepl("next_str\\(\\).+next_str\\(\\)", sexp_call))

# Test 4: Generate full wrapper function
wrapper <- Rtinycc:::generate_c_wrapper(
  symbol_name = "my_add",
  wrapper_name = "R_wrap_my_add",
  arg_types = list("i32", "i32"),
  return_type = "i32",
  is_external = FALSE
)
expect_true(grepl("SEXP R_wrap_my_add", wrapper))
expect_true(grepl("int32_t", wrapper))
expect_true(grepl("my_add", wrapper))
expect_true(grepl("ScalarInteger", wrapper))

# Test 5: Generate wrapper with no arguments
wrapper <- Rtinycc:::generate_c_wrapper(
  symbol_name = "get_value",
  wrapper_name = "R_wrap_get_value",
  arg_types = list(),
  return_type = "i32",
  is_external = FALSE
)
expect_true(grepl("SEXP R_wrap_get_value", wrapper))
expect_true(grepl("void", wrapper))

# Test 6: Generate external declarations
decls <- Rtinycc:::generate_external_declarations(list(
  foo = list(args = list("i32", "f64"), returns = "cstring")
))
expect_true(grepl("extern char[*] foo", decls))

# Test 7: Generate complete FFI code
symbols <- list(
  add = list(args = list("i32", "i32"), returns = "i32"),
  greet = list(args = list("cstring"), returns = "cstring")
)

full_code <- Rtinycc:::generate_ffi_code(
  symbols = symbols,
  headers = c("#include <stdio.h>"),
  c_code = "int add(int a, int b) { return a + b; }",
  is_external = FALSE
)

expect_true(grepl("#define _Complex", full_code))
expect_true(grepl("#include <R.h>", full_code))
expect_true(grepl("#include <Rinternals.h>", full_code))
expect_true(grepl("#ifndef STRING_PTR_RO", full_code))
expect_true(grepl("#include <stdint.h>", full_code))
expect_true(grepl("#include <stdbool.h>", full_code))
expect_true(grepl("R_wrap_add", full_code))
expect_true(grepl("R_wrap_greet", full_code))
expect_true(grepl("SEXP R_wrap_add", full_code))

for (case in callback_abi_specs$wrapper) {
  cb_code <- Rtinycc:::generate_ffi_code(
    symbols = setNames(
      list(list(args = case$args, returns = case$returns)),
      case$name
    ),
    c_code = case$c_code,
    is_external = FALSE
  )

  for (pattern in case$patterns) {
    expect_true(grepl(pattern$pattern, cb_code), info = pattern$info)
  }
}

# Test 8: tcc_source stores code as chunks without introducing a leading blank line
ffi <- tcc_ffi() |>
  tcc_source("int a(void) { return 1; }") |>
  tcc_source("int b(void) { return 2; }")
expect_equal(
  ffi$c_code,
  c("int a(void) { return 1; }", "int b(void) { return 2; }")
)

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

#' Treesitter helpers for header-driven bindings
#'
#' Convenience wrappers around `treesitter.c` that map C types to the
#' Rtinycc FFI type system and return `tcc_bind()`-ready specs.
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to treesitter.c helpers.
#' @return A data frame of nodes or a named list of binding specs.
#' @name tcc_treesitter_helpers
#' @keywords internal
#'
#' @examples
#' \dontrun{
#' header <- "double sqrt(double x);\nint add(int a, int b);"
#' tcc_treesitter_functions(header)
#' tcc_treesitter_bindings(header)
#' }
#'
#' @section treesitter.c:
#' These helpers require the optional treesitter.c package.
#'
NULL

.tcc_treesitter_root <- function(header) {
  if (!requireNamespace("treesitter.c", quietly = TRUE)) {
    stop("treesitter.c is required for header parsing", call. = FALSE)
  }
  treesitter.c::parse_header_text(header)
}

#' Parse function declarations with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_function_nodes()`.
#' @return A data frame of function nodes.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "double sqrt(double x);"
#' tcc_treesitter_functions(header)
#' }
tcc_treesitter_functions <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_function_nodes(
    root,
    extract_params = TRUE,
    extract_return = TRUE,
    ...
  )
}

#' Parse struct declarations with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_struct_nodes()`.
#' @return A data frame of struct nodes.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "struct point { double x; double y; };"
#' tcc_treesitter_structs(header)
#' }
tcc_treesitter_structs <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_struct_nodes(root, ...)
}

#' Parse union declarations with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_union_nodes()`.
#' @return A data frame of union nodes.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "union data { int i; double d; };"
#' tcc_treesitter_unions(header)
#' }
tcc_treesitter_unions <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_union_nodes(root, ...)
}

#' Parse union members with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_union_members_from_root()`.
#' @return A data frame of union members.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "union data { int i; double d; };"
#' tcc_treesitter_union_members(header)
#' }
tcc_treesitter_union_members <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_union_members_from_root(root, ...)
}

#' Parse enum declarations with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_enum_nodes()`.
#' @return A data frame of enum nodes.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "enum status { OK = 0, ERR = 1 };"
#' tcc_treesitter_enums(header)
#' }
tcc_treesitter_enums <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_enum_nodes(root, ...)
}

#' Parse enum members with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_enum_members_from_root()`.
#' @return A data frame of enum members.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "enum status { OK = 0, ERR = 1 };"
#' tcc_treesitter_enum_members(header)
#' }
tcc_treesitter_enum_members <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_enum_members_from_root(root, ...)
}

#' Parse struct members (including bitfields) with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_struct_members()`.
#' @return A data frame of struct members.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "struct point { double x; double y; };"
#' tcc_treesitter_struct_members(header)
#' }
tcc_treesitter_struct_members <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_struct_members(root, ...)
}

#' Parse global declarations with treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_globals_from_root()`.
#' @return A data frame of global names.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "int global_counter;"
#' tcc_treesitter_globals(header)
#' }
tcc_treesitter_globals <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_globals_from_root(root, ...)
}

#' Parse global declarations with types using treesitter.c
#'
#' @param header Character scalar containing C declarations.
#' @param ... Additional arguments passed to `treesitter.c::get_globals_with_types_from_root()`.
#' @return A data frame of global names and C types.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "int global_counter;"
#' tcc_treesitter_global_types(header)
#' }
tcc_treesitter_global_types <- function(header, ...) {
  root <- .tcc_treesitter_root(header)
  treesitter.c::get_globals_with_types_from_root(root, ...)
}

#' Extract macro defines from a header file
#'
#' @param file Path to a header file.
#' @param use_cpp Logical; use the C preprocessor if available.
#' @param cc Compiler string; passed to `system2()` if `use_cpp = TRUE`.
#' @param ccflags Additional flags for the compiler.
#' @return Character vector of macro names defined in `file`.
#' @export
#'
#' @examples
#' \dontrun{
#' tcc_treesitter_defines("/usr/include/math.h")
#' }
tcc_treesitter_defines <- function(
  file,
  use_cpp = TRUE,
  cc = treesitter.c::r_cc(),
  ccflags = NULL
) {
  if (!requireNamespace("treesitter.c", quietly = TRUE)) {
    stop("treesitter.c is required for header parsing", call. = FALSE)
  }
  treesitter.c::get_defines_from_file(
    file,
    use_cpp = use_cpp,
    cc = cc,
    ccflags = ccflags
  )
}

#' Map a C type string to an Rtinycc FFI type
#'
#' @param c_type C type string (e.g., "int", "double", "char *").
#' @return A single FFI type string.
#' @export
#'
#' @details
#' This mapper is intentionally conservative for pointer types. `char*` is
#' treated as a raw pointer (`ptr`) because C does not guarantee NUL-terminated
#' strings. If you know the API expects a C string, map it explicitly to
#' `cstring` in your custom mapper.
#'
#' @examples
#' \dontrun{
#' tcc_map_c_type_to_ffi("int")
#' tcc_map_c_type_to_ffi("const char *")
#' }
tcc_map_c_type_to_ffi <- function(c_type) {
  x <- trimws(as.character(c_type))
  x <- gsub("\\s+", " ", x)
  x <- gsub("\\b(const|volatile|restrict)\\b", "", x)
  x <- gsub("\\s+", " ", trimws(x))

  strip_trailing_identifier <- function(type_str) {
    parts <- strsplit(type_str, "\\s+")[[1]]
    if (length(parts) < 2) {
      return(type_str)
    }
    last <- parts[length(parts)]
    if (last %in% c("*", "**") || grepl("^\\*+", last)) {
      return(type_str)
    }
    type_tokens <- c(
      "void",
      "bool",
      "_Bool",
      "SEXP",
      "sexp",
      "char",
      "short",
      "int",
      "long",
      "signed",
      "unsigned",
      "float",
      "double",
      "size_t",
      "ssize_t",
      "ptrdiff_t",
      "intptr_t",
      "uintptr_t",
      "off_t",
      "int8_t",
      "int16_t",
      "int32_t",
      "int64_t",
      "uint8_t",
      "uint16_t",
      "uint32_t",
      "uint64_t",
      "__int32",
      "__int64"
    )
    if (last %in% type_tokens) {
      return(type_str)
    }
    paste(parts[-length(parts)], collapse = " ")
  }

  x <- strip_trailing_identifier(x)
  x <- trimws(x)
  ffi_c_type_map_rule(
    x,
    grepl("char\\s*\\*", x),
    grepl("\\*", x)
  )
}

#' Generate bindings from a header
#'
#' @param header Character scalar containing C declarations.
#' @param mapper Function to map C types to FFI types.
#' @param ffi Optional `tcc_ffi` object. When provided, returns an updated
#'   FFI object with generated bindings.
#' @param functions Logical; generate `tcc_bind()` specs for functions.
#' @param structs Logical; generate `tcc_struct()` helpers.
#' @param unions Logical; generate `tcc_union()` helpers.
#' @param enums Logical; generate `tcc_enum()` helpers.
#' @param globals Logical; generate `tcc_global()` getters/setters.
#' @param bitfield_type FFI type to use for bitfields.
#' @param include_bitfields Whether to include bitfields.
#' @return Named list suitable for `tcc_bind()` when `ffi` is NULL, otherwise
#'   an updated `tcc_ffi` object.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "double sqrt(double x);"
#' symbols <- tcc_treesitter_bindings(header)
#' }
tcc_treesitter_bindings <- function(
  header,
  mapper = tcc_map_c_type_to_ffi,
  ffi = NULL,
  functions = TRUE,
  structs = FALSE,
  unions = FALSE,
  enums = FALSE,
  globals = FALSE,
  bitfield_type = "u8",
  include_bitfields = TRUE
) {
  if (!is.null(ffi)) {
    return(tcc_generate_bindings(
      ffi,
      header,
      mapper = mapper,
      functions = functions,
      structs = structs,
      unions = unions,
      enums = enums,
      globals = globals,
      bitfield_type = bitfield_type,
      include_bitfields = include_bitfields
    ))
  }

  if (structs || unions || enums || globals) {
    stop(
      "Provide `ffi` to generate struct/union/enum/global bindings",
      call. = FALSE
    )
  }

  if (!functions) {
    return(list())
  }

  funcs <- tcc_treesitter_functions(header)
  if (nrow(funcs) == 0) {
    return(list())
  }

  normalize_params <- function(params) {
    if (is.list(params)) {
      params <- params[[1]]
    }
    params <- as.character(params)
    if (length(params) == 1 && grepl(",", params)) {
      params <- unlist(strsplit(params, ","))
    }
    params <- trimws(params)
    if (
      length(params) == 1 && (is.na(params) || params == "" || params == "void")
    ) {
      return(character(0))
    }
    params
  }

  signature_to_bind <- function(row) {
    params <- normalize_params(row$params)
    args <- if (length(params) > 0) {
      vapply(params, mapper, character(1), USE.NAMES = FALSE)
    } else {
      character(0)
    }

    ret <- row$return_type
    if (is.list(ret)) {
      ret <- ret[[1]]
    }
    ret <- as.character(ret)
    if (length(ret) > 1) {
      ret <- ret[[1]]
    }

    list(args = as.list(args), returns = mapper(ret))
  }

  stats::setNames(
    lapply(seq_len(nrow(funcs)), function(i) signature_to_bind(funcs[i, ])),
    funcs$text
  )
}

#' Generate tcc_struct() accessors from header structs
#'
#' @param header Character scalar containing C declarations.
#' @param mapper Function to map C types to FFI types.
#' @param bitfield_type FFI type to use for bitfields.
#' @param include_bitfields Whether to include bitfields.
#' @return Named list of accessors by struct name. Bitfields are returned as
#'   lists with `type`, `bitfield = TRUE`, and `width`. Named nested struct
#'   fields are returned as `"struct:<name>"`; ambiguous or anonymous nested
#'   structs fall back to `"ptr"`.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "struct point { double x; double y; };"
#' tcc_treesitter_struct_accessors(header)
#' }
tcc_treesitter_struct_accessors <- function(
  header,
  mapper = tcc_map_c_type_to_ffi,
  bitfield_type = "u8",
  include_bitfields = TRUE
) {
  members <- tcc_treesitter_struct_members(header)
  if (nrow(members) == 0) {
    return(list())
  }

  split_members <- split(members, members$struct_name)
  out <- lapply(split_members, function(df) {
    acc <- list()
    for (i in seq_len(nrow(df))) {
      row <- df[i, ]
      name <- row$member_name
      if (is.na(name) || !nzchar(name)) {
        next
      }
      is_bitfield <- !is.na(row$bitfield) && nzchar(row$bitfield)
      if (is_bitfield && !include_bitfields) {
        next
      }

      mtype <- row$member_type
      if (is_bitfield) {
        ffi_type <- list(
          type = bitfield_type,
          bitfield = TRUE,
          width = as.integer(row$bitfield)
        )
      } else if (is.na(mtype) || !nzchar(mtype)) {
        stop(
          "Missing struct member type for '",
          name,
          "' in struct '",
          row$struct_name,
          "'",
          call. = FALSE
        )
      } else if (grepl("^struct\\s+[A-Za-z_][A-Za-z0-9_]*$", mtype)) {
        ffi_type <- paste0("struct:", sub("^struct\\s+", "", mtype))
      } else if (grepl("^struct\\b", mtype)) {
        ffi_type <- "ptr"
      } else {
        ffi_type <- mapper(mtype)
      }

      acc[[name]] <- ffi_type
    }
    acc
  })

  out[!vapply(out, function(x) length(x) == 0, logical(1))]
}

#' Generate tcc_union() accessors from header unions
#'
#' @param header Character scalar containing C declarations.
#' @param mapper Function to map C types to FFI types.
#' @param bitfield_type FFI type to use for bitfields.
#' @param include_bitfields Whether to include bitfields.
#' @return Named list of accessors by union name. Bitfields are returned as
#'   lists with `type`, `bitfield = TRUE`, and `width`. Nested struct members
#'   are returned as `list(type = "struct", struct_name = <name>)` when the
#'   struct name is available, otherwise `list(type = "struct")`.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "union data { int i; double d; };"
#' tcc_treesitter_union_accessors(header)
#' }
tcc_treesitter_union_accessors <- function(
  header,
  mapper = tcc_map_c_type_to_ffi,
  bitfield_type = "u8",
  include_bitfields = TRUE
) {
  members <- tcc_treesitter_union_members(header)
  if (nrow(members) == 0) {
    return(list())
  }

  split_members <- split(members, members$union_name)
  out <- lapply(split_members, function(df) {
    acc <- list()
    for (i in seq_len(nrow(df))) {
      row <- df[i, ]
      name <- row$member_name
      if (is.na(name) || !nzchar(name)) {
        next
      }
      is_bitfield <- !is.na(row$bitfield) && nzchar(row$bitfield)
      if (is_bitfield && !include_bitfields) {
        next
      }

      mtype <- row$member_type
      if (is_bitfield) {
        ffi_type <- list(
          type = bitfield_type,
          bitfield = TRUE,
          width = as.integer(row$bitfield)
        )
      } else if (is.na(mtype) || !nzchar(mtype)) {
        stop(
          "Missing union member type for '",
          name,
          "' in union '",
          row$union_name,
          "'",
          call. = FALSE
        )
      } else if (
        grepl(
          "^struct\\s+[A-Za-z_][A-Za-z0-9_]*\\s+[A-Za-z_][A-Za-z0-9_]*$",
          mtype
        )
      ) {
        ffi_type <- list(
          type = "struct",
          struct_name = sub(
            "^struct\\s+([A-Za-z_][A-Za-z0-9_]*)\\s+[A-Za-z_][A-Za-z0-9_]*$",
            "\\1",
            mtype
          )
        )
      } else if (
        mtype %in%
          c("struct", "struct (anonymous)") ||
          grepl("^struct\\s+[^*]+$", mtype)
      ) {
        ffi_type <- list(type = "struct")
      } else {
        ffi_type <- mapper(mtype)
      }

      acc[[name]] <- ffi_type
    }
    acc
  })

  out[!vapply(out, function(x) length(x) == 0, logical(1))]
}

#' Apply tcc_struct() bindings from a header
#'
#' @param ffi A tcc_ffi object.
#' @param header Character scalar containing C declarations.
#' @param ... Passed to `tcc_treesitter_struct_accessors()`.
#' @return Updated tcc_ffi object.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "struct point { double x; double y; };"
#' ffi <- tcc_ffi()
#' ffi <- tcc_treesitter_struct_bindings(ffi, header)
#' }
tcc_treesitter_struct_bindings <- function(ffi, header, ...) {
  accessors <- tcc_treesitter_struct_accessors(header, ...)
  if (length(accessors) == 0) {
    return(ffi)
  }
  for (nm in names(accessors)) {
    ffi <- tcc_struct(ffi, nm, accessors[[nm]])
  }
  ffi
}

#' Apply tcc_union() bindings from a header
#'
#' @param ffi A tcc_ffi object.
#' @param header Character scalar containing C declarations.
#' @param ... Passed to `tcc_treesitter_union_accessors()`.
#' @return Updated tcc_ffi object.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "union data { int i; double d; };"
#' ffi <- tcc_ffi()
#' ffi <- tcc_treesitter_union_bindings(ffi, header)
#' }
tcc_treesitter_union_bindings <- function(ffi, header, ...) {
  accessors <- tcc_treesitter_union_accessors(header, ...)
  if (length(accessors) == 0) {
    return(ffi)
  }
  for (nm in names(accessors)) {
    ffi <- tcc_union(ffi, nm, accessors[[nm]])
  }
  ffi
}

#' Apply tcc_enum() bindings from a header
#'
#' @param ffi A tcc_ffi object.
#' @param header Character scalar containing C declarations.
#' @param constants Named list of enum constants.
#' @return Updated tcc_ffi object.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "enum status { OK = 0, ERR = 1 };"
#' ffi <- tcc_ffi()
#' ffi <- tcc_treesitter_enum_bindings(ffi, header, constants = list(status = c("OK", "ERR")))
#' }
tcc_treesitter_enum_bindings <- function(ffi, header, constants = NULL) {
  enums <- tcc_treesitter_enums(header)
  if (nrow(enums) == 0 || is.null(constants)) {
    return(ffi)
  }
  for (i in seq_len(nrow(enums))) {
    nm <- enums$text[i]
    if (!is.null(constants[[nm]])) {
      ffi <- tcc_enum(ffi, nm, constants = constants[[nm]])
    }
  }
  ffi
}

#' Generate bindings from header declarations
#'
#' @param ffi A tcc_ffi object. If NULL, a new one is created.
#' @param header Character scalar containing C declarations.
#' @param mapper Function to map C types to FFI types.
#' @param functions Logical; generate `tcc_bind()` specs for functions.
#' @param structs Logical; generate `tcc_struct()` helpers.
#' @param unions Logical; generate `tcc_union()` helpers.
#' @param enums Logical; generate `tcc_enum()` helpers.
#' @param globals Logical; generate `tcc_global()` getters/setters.
#' @param bitfield_type FFI type to use for bitfields.
#' @param include_bitfields Whether to include bitfields.
#' @return Updated tcc_ffi object.
#' @export
#'
#' @examples
#' \dontrun{
#' header <- "double sqrt(double x); struct point { double x; double y; };"
#' ffi <- tcc_generate_bindings(tcc_ffi(), header)
#' }
tcc_generate_bindings <- function(
  ffi = NULL,
  header,
  mapper = tcc_map_c_type_to_ffi,
  functions = TRUE,
  structs = TRUE,
  unions = TRUE,
  enums = TRUE,
  globals = TRUE,
  bitfield_type = "u8",
  include_bitfields = TRUE
) {
  if (is.null(ffi)) {
    ffi <- tcc_ffi()
  }
  if (!inherits(ffi, "tcc_ffi")) {
    stop("Expected tcc_ffi object", call. = FALSE)
  }

  if (functions) {
    bindings <- tcc_treesitter_bindings(header, mapper = mapper)
    if (length(bindings) > 0) {
      ffi <- do.call(tcc_bind, c(list(ffi), bindings))
    }
  }

  if (structs) {
    ffi <- tcc_treesitter_struct_bindings(
      ffi,
      header,
      mapper = mapper,
      bitfield_type = bitfield_type,
      include_bitfields = include_bitfields
    )
  }

  if (unions) {
    ffi <- tcc_treesitter_union_bindings(
      ffi,
      header,
      mapper = mapper,
      bitfield_type = bitfield_type,
      include_bitfields = include_bitfields
    )
  }

  if (enums) {
    enum_members <- tcc_treesitter_enum_members(header)
    if (nrow(enum_members) > 0) {
      split_members <- split(enum_members, enum_members$enum_name)
      for (nm in names(split_members)) {
        consts <- split_members[[nm]]$member_name
        consts <- consts[!is.na(consts) & nzchar(consts)]
        if (length(consts) > 0) {
          ffi <- tcc_enum(ffi, nm, constants = consts)
        }
      }
    }
  }

  if (globals) {
    g <- tcc_treesitter_global_types(header)
    if (nrow(g) > 0) {
      for (i in seq_len(nrow(g))) {
        nm <- g$text[i]
        c_type <- g$c_type[i]
        if (is.na(nm) || !nzchar(nm) || is.na(c_type) || !nzchar(c_type)) {
          next
        }
        ffi <- tcc_global(ffi, nm, mapper(c_type))
      }
    }
  }

  ffi
}

# Tests for treesitter helpers

library(tinytest)
library(Rtinycc)

composite_semantics <- Rtinycc:::rtinycc_composite_semantics()
bitfield_type <- composite_semantics$bitfield_native$treesitter_bitfield_type


if (!requireNamespace("treesitter.c", quietly = TRUE)) {
  message("Skipping treesitter helper tests: treesitter.c not installed")
  exit_file(msg = "treesitter.c not available")
} else {
  # check if version >= 0.0.4
  version <- packageVersion("treesitter.c")
  if (version < "0.0.4") {
    message("Skipping treesitter helper tests: treesitter.c version too old")
    exit_file(msg = "treesitter.c version too old")
  }
}

# Test 1: function parsing and bindings
expect_true(
  {
    header <- "double sqrt(double x);\nint add(int a, int b);"
    funcs <- tcc_treesitter_functions(header)
    symbols <- tcc_treesitter_bindings(header)

    has_funcs <- all(c("sqrt", "add") %in% funcs$text)
    has_symbols <- all(c("sqrt", "add") %in% names(symbols))
    add_args <- symbols$add$args
    add_ret <- symbols$add$returns

    has_funcs &&
      has_symbols &&
      length(add_args) == 2 &&
      all(unlist(add_args) == c("i32", "i32")) &&
      add_ret == "i32"
  },
  info = "Parse functions and map to FFI bindings"
)

# Test 1b: conservative mapping keeps char pointers as ptr in generated bindings
expect_true(
  {
    symbols <- tcc_treesitter_bindings("int puts(const char *s);")

    identical(symbols$puts$args[[1]], "ptr") &&
      identical(symbols$puts$returns, "i32")
  },
  info = paste(
    "treesitter-generated bindings keep const char * arguments on the",
    "default conservative ptr mapping"
  )
)

# Test 2: struct members and accessors (including bitfield)
expect_true(
  {
    header <- "struct flags { unsigned int flag : 1; unsigned int code : 6; double x; };"
    members <- tcc_treesitter_struct_members(header)
    accessors <- tcc_treesitter_struct_accessors(
      header,
      bitfield_type = bitfield_type
    )

    has_struct <- "flags" %in% names(accessors)
    acc <- accessors[["flags"]]

    has_struct &&
      all(c("flag", "code", "x") %in% names(acc)) &&
      is.list(acc[["flag"]]) &&
      identical(acc[["flag"]]$type, bitfield_type) &&
      isTRUE(acc[["flag"]]$bitfield) &&
      identical(acc[["flag"]]$width, 1L) &&
      is.list(acc[["code"]]) &&
      identical(acc[["code"]]$type, bitfield_type) &&
      isTRUE(acc[["code"]]$bitfield) &&
      identical(acc[["code"]]$width, 6L) &&
      acc[["x"]] == "f64" &&
      nrow(members) >= 3
  },
  info = "Extract struct members and build accessors"
)

# Test 3: struct bindings integrate with tcc_struct()
expect_true(
  {
    header <- "struct point { double x; double y; };"
    ffi <- tcc_ffi() |>
      tcc_source("struct point { double x; double y; };")

    ffi <- tcc_treesitter_struct_bindings(ffi, header)
    compiled <- tcc_compile(ffi)

    p <- compiled$struct_point_new()
    p <- compiled$struct_point_set_x(p, 1.25)
    p <- compiled$struct_point_set_y(p, 2.5)
    x <- compiled$struct_point_get_x(p)
    y <- compiled$struct_point_get_y(p)
    compiled$struct_point_free(p)

    abs(x - 1.25) < 1e-12 && abs(y - 2.5) < 1e-12
  },
  info = "Generate and use struct bindings from header"
)

# Test 4: enum bindings (constants list provided)
expect_true(
  {
    header <- "enum status { OK = 0, ERR = 1 };"
    ffi <- tcc_ffi() |>
      tcc_source("enum status { OK = 0, ERR = 1 };")

    ffi <- tcc_treesitter_enum_bindings(
      ffi,
      header,
      constants = list(status = c("OK", "ERR"))
    )
    compiled <- tcc_compile(ffi)

    ok <- compiled$enum_status_OK()
    err <- compiled$enum_status_ERR()

    ok == 0 && err == 1
  },
  info = "Generate enum bindings from header"
)

# Test 5: unions, globals, and defines
expect_true(
  {
    header <- "union data { int i; double d; struct { int x; } inner; };\nint global_counter = 0;"
    named_header <- "struct inner { int x; }; union named_data { struct inner payload; int y; };"
    unions <- tcc_treesitter_unions(header)
    globals <- tcc_treesitter_globals(header)
    union_members <- tcc_treesitter_union_members(header)
    union_accessors <- tcc_treesitter_union_accessors(header)
    named_union_accessors <- tcc_treesitter_union_accessors(named_header)
    global_types <- tcc_treesitter_global_types(header)

    tmp <- tempfile(fileext = ".h")
    writeLines(c("#define FOO 1", "#define BAR 2"), tmp)
    defs <- tcc_treesitter_defines(tmp, use_cpp = FALSE)

    has_union <- "data" %in% unions$text
    has_global <- "global_counter" %in% globals$text
    has_union_members <- all(
      c("i", "d", "inner") %in% union_members$member_name
    )
    has_nested_union_type <- is.list(union_accessors$data$inner) &&
      identical(union_accessors$data$inner$type, "struct")
    has_named_nested_union_type <- is.list(
      named_union_accessors$named_data$payload
    ) &&
      identical(named_union_accessors$named_data$payload$type, "struct") &&
      identical(named_union_accessors$named_data$payload$struct_name, "inner")
    has_global_types <- any(
      global_types$text == "global_counter" & grepl("int", global_types$c_type)
    )
    has_defs <- all(c("FOO", "BAR") %in% defs)

    has_union &&
      has_global &&
      has_union_members &&
      has_nested_union_type &&
      has_named_nested_union_type &&
      has_global_types &&
      has_defs
  },
  info = "Parse unions, globals, nested union structs, and macro defines"
)

# Test 6: enum members and generate bindings
expect_true(
  {
    header <- paste(
      "double sqrt(double x);",
      "struct point { double x; double y; };",
      "union data { int i; float f; };",
      "enum status { OK = 0, ERR = 1 };",
      "int global_counter;",
      sep = "\n"
    )

    enum_members <- tcc_treesitter_enum_members(header)
    has_enum_members <- all(c("OK", "ERR") %in% enum_members$member_name)

    ffi <- tcc_ffi() |>
      tcc_source(paste(
        "double sqrt(double x) { return x; }",
        "struct point { double x; double y; };",
        "union data { int i; float f; };",
        "enum status { OK = 0, ERR = 1 };",
        "int global_counter = 7;",
        sep = "\n"
      ))

    ffi <- tcc_generate_bindings(
      ffi,
      header,
      functions = TRUE,
      structs = TRUE,
      unions = TRUE,
      enums = TRUE,
      globals = TRUE
    )

    compiled <- tcc_compile(ffi)

    p <- compiled$struct_point_new()
    p <- compiled$struct_point_set_x(p, 1.0)
    x <- compiled$struct_point_get_x(p)
    compiled$struct_point_free(p)

    u <- compiled$union_data_new()
    u <- compiled$union_data_set_i(u, 42L)
    ui <- compiled$union_data_get_i(u)
    compiled$union_data_free(u)

    ok <- compiled$enum_status_OK()
    gc_before <- compiled$global_global_counter_get()
    compiled$global_global_counter_set(9L)
    gc_after <- compiled$global_global_counter_get()

    has_enum_members &&
      abs(x - 1.0) < 1e-12 &&
      ui == 42L &&
      ok == 0 &&
      gc_before == 7L &&
      gc_after == 9L
  },
  info = "Generate bindings for structs, unions, enums, globals"
)

# Test 7: treesitter-generated bitfield bindings preserve helper metadata end-to-end
expect_true(
  {
    header <- "struct flags { unsigned int flag : 1; unsigned int code : 6; };"
    ffi <- tcc_ffi() |>
      tcc_source(header) |>
      tcc_treesitter_struct_bindings(header, bitfield_type = bitfield_type) |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    s <- ffi$struct_flags_new()
    s <- ffi$struct_flags_set_flag(s, 1L)
    s <- ffi$struct_flags_set_code(s, 17L)
    ok_runtime <- identical(ffi$struct_flags_get_flag(s), 1L) &&
      identical(ffi$struct_flags_get_code(s), 17L)
    ffi$struct_flags_free(s)

    identical(
      Rtinycc:::helper_symbol_operation(helper_specs$struct_flags_get_flag),
      "bitfield_getter"
    ) &&
      identical(
        Rtinycc:::helper_symbol_operation(helper_specs$struct_flags_set_flag),
        "bitfield_setter"
      ) &&
      ok_runtime
  },
  info = "treesitter-generated bitfield bindings preserve bitfield helper semantics end-to-end"
)

# Test 8: treesitter-generated bitfields still reject address-style helpers
expect_error(
  {
    header <- "struct flags { unsigned int flag : 1; };"
    ffi <- tcc_ffi() |>
      tcc_source(header) |>
      tcc_treesitter_struct_bindings(header, bitfield_type = bitfield_type) |>
      tcc_field_addr("flags", "flag")
    ffi
  },
  info = "treesitter-generated bitfields reject field_addr"
)
expect_error(
  {
    header <- "struct flags { unsigned int flag : 1; };"
    ffi <- tcc_ffi() |>
      tcc_source(header) |>
      tcc_treesitter_struct_bindings(header, bitfield_type = bitfield_type) |>
      tcc_container_of("flags", "flag")
    ffi
  },
  info = "treesitter-generated bitfields reject container_of"
)

# Test 9: treesitter-generated nested union struct bindings preserve nested-view metadata end-to-end
expect_true(
  {
    header <- "struct inner { int x; }; union wrapper { struct inner payload; int raw; };"
    ffi <- tcc_ffi() |>
      tcc_source(header) |>
      tcc_treesitter_struct_bindings(header) |>
      tcc_treesitter_union_bindings(header) |>
      tcc_bind() |>
      tcc_compile()

    helper_specs <- get(".helper_specs", envir = ffi, inherits = FALSE)
    u <- ffi$union_wrapper_new()
    inner <- ffi$union_wrapper_get_payload(u)
    inner <- ffi$struct_inner_set_x(inner, 33L)
    ok_runtime <- identical(ffi$struct_inner_get_x(inner), 33L)
    ffi$union_wrapper_free(u)

    identical(
      Rtinycc:::helper_symbol_operation(helper_specs$union_wrapper_get_payload),
      "nested_view"
    ) &&
      ok_runtime
  },
  info = "treesitter-generated nested union struct bindings preserve nested-view semantics end-to-end"
)

# Test 10: treesitter-generated nested struct members still fall back to ptr-like accessors when the parser does not recover the nested type name
expect_true(
  {
    header <- "struct child { int x; }; struct outer { struct child child; int y; };"
    accessors <- tcc_treesitter_struct_accessors(header)
    is.character(accessors$outer$child) &&
      identical(accessors$outer$child, "ptr") &&
      identical(accessors$outer$y, "i32")
  },
  info = "treesitter-generated nested struct members currently fall back to ptr-like accessors when the nested type name is unavailable"
)

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

## usethis namespace: start
#' @useDynLib Rtinycc, .registration = TRUE
#' @importFrom lambda.r %as% UseFunction
## usethis namespace: end
NULL

#' we are using .Call directly, this is to make R CMD check happy
#'
#' @return The result returned by the invoked native routine. This alias is
#'   package-internal and exists to satisfy `R CMD check` native routine
#'   registration requirements.
#' @keywords internal
.RtinyccCall <- base::.Call

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

# Callback Registration and Invocation
# Allow R functions to be passed as callbacks to compiled C code

#' Register an R function as a callback
#'
#' Wraps an R function so it can be passed as a C function pointer
#' to compiled code. The callback will be invoked via a trampoline
#' that marshals arguments between C and R.
#'
#' @details
#' Thread safety: callbacks are executed on the R main thread only. Invoking
#' a callback from a worker thread is unsupported and may crash R. The
#' \code{threadsafe} flag is currently informational only.
#'
#' If a callback raises an error, a warning is emitted and a type-appropriate
#' default value is returned.
#'
#' When binding callbacks with `tcc_bind()`, use a `callback:<signature>`
#' argument type so a synchronous trampoline is generated. The trampoline
#' expects a `void*` user-data pointer as its first argument; pass
#' `tcc_callback_ptr(cb)` as the user-data argument to the C API. For
#' thread-safe usage from worker threads, use `callback_async:<signature>`
#' which schedules the call on the main thread and returns a default value.
#'
#' Pointer arguments (e.g., \code{double*}, \code{int*}) are passed as
#' external pointers. Lengths must be supplied separately if needed.
#'
#' The return type may be any scalar type supported by the FFI mappings
#' (e.g., \code{i32}, \code{f64}, \code{bool}, \code{cstring}), or
#' \code{SEXP} to return an R object directly.
#'
#' Callback lifetime: callbacks are eventually released by finalizers and
#' package unload. Call `tcc_callback_close()` when you want deterministic
#' invalidation and earlier release of the preserved R function.
#'
#' @param fun An R function to be called from C
#' @param signature C function signature string (e.g., "double (*)(int, double)")
#' @param threadsafe Whether to enable thread-safe invocation (experimental)
#' @return A tcc_callback object (externalptr wrapper)
#' @export
tcc_callback <- function(fun, signature, threadsafe = FALSE) {
  if (!is.function(fun)) {
    stop("fun must be a function", call. = FALSE)
  }

  if (!is.character(signature) || length(signature) != 1) {
    stop("signature must be a single string", call. = FALSE)
  }

  # Parse signature to get return type and argument types
  sig <- parse_callback_signature(signature)

  # Register with C runtime
  callback_ext <- .Call(
    RC_register_callback,
    fun,
    sig$return_type,
    sig$arg_types,
    as.integer(threadsafe)
  )

  # Add metadata as attributes
  attr(callback_ext, "signature") <- sig
  attr(callback_ext, "threadsafe") <- threadsafe
  attr(callback_ext, "fun") <- fun
  class(callback_ext) <- c("tcc_callback", "externalptr")

  callback_ext
}

#' Close/unregister a callback
#'
#' Invalidates a callback immediately, releases the preserved R function
#' reference, and cleans up callback resources as early as possible.
#' This is recommended for deterministic lifetime management, but callbacks
#' are also eventually released by finalizers if you simply drop all references.
#'
#' @param callback A tcc_callback object returned by tcc_callback()
#' @return NULL (invisible)
#' @export
tcc_callback_close <- function(callback) {
  if (!inherits(callback, "tcc_callback")) {
    stop("Expected tcc_callback object", call. = FALSE)
  }

  .Call(RC_unregister_callback, callback)
  invisible(NULL)
}

#' Get the C-compatible function pointer
#'
#' Returns an external pointer that can be passed to compiled C code
#' as user data for trampolines. Keep this handle (and the original
#' `tcc_callback`) alive for as long as C may call back.
#' The pointer handle keeps the underlying token storage alive until it is
#' garbage collected. Closing the original callback still invalidates the
#' callback registry entry, so C must not continue invoking it after
#' `tcc_callback_close()`.
#'
#' Pointer arguments and return values are treated as external pointers.
#' Use \code{tcc_read_bytes()}, \code{tcc_read_u8()}, or \code{tcc_read_f64()}
#' to inspect pointed data when needed.
#'
#' @param callback A tcc_callback object
#' @return An external pointer (address of the callback token)
#' @export
tcc_callback_ptr <- function(callback) {
  if (!inherits(callback, "tcc_callback")) {
    stop("Expected tcc_callback object", call. = FALSE)
  }

  # Get the token address as external pointer
  .Call(RC_get_callback_ptr, callback)
}

#' Print tcc_callback object
#'
#' @param x A tcc_callback object
#' @param ... Ignored
#' @return The input `tcc_callback` object, invisibly. Called for its
#'   side effect of printing the callback signature, thread-safety flag,
#'   and validity status.
#' @export
print.tcc_callback <- function(x, ...) {
  sig <- attr(x, "signature")
  threadsafe <- attr(x, "threadsafe")

  cat("<tcc_callback>\n")
  cat("  Signature: ", format_signature(sig), "\n", sep = "")
  cat("  Thread-safe: ", if (threadsafe) "yes" else "no", "\n", sep = "")
  cat(
    "  Status: ",
    if (is_callback_valid(x)) "valid" else "invalid",
    "\n",
    sep = ""
  )
  invisible(x)
}

#' Check if callback is still valid
#'
#' @param callback A tcc_callback object
#' @return Logical indicating if callback can be invoked
#' @export
tcc_callback_valid <- function(callback) {
  if (!inherits(callback, "tcc_callback")) {
    return(FALSE)
  }
  is_callback_valid(callback)
}

# Internal function to check validity
is_callback_valid <- function(callback) {
  tryCatch(
    {
      .Call(RC_callback_is_valid, callback)
    },
    error = function(e) FALSE
  )
}

new_rtinycc_callback_signature <- function(
  return_type,
  arg_types,
  raw,
  mode = "sync"
) {
  structure(
    list(
      return_type = return_type,
      arg_types = arg_types,
      raw = raw,
      mode = mode
    ),
    class = c(
      paste0("rtinycc_callback_signature_", mode),
      "rtinycc_callback_signature"
    )
  )
}

is_rtinycc_callback_signature <- function(x) {
  inherits(x, "rtinycc_callback_signature")
}

# Parse a C function signature string
# Returns classed signature object with return_type and arg_types.
parse_callback_signature <- function(signature, mode = "sync") {
  # Remove whitespace
  sig <- gsub("\\s+", " ", trimws(signature))

  # Pattern: return_type (*)(arg1, arg2, ...)
  # or:      return_type (*name)(arg1, arg2, ...)
  # or:      return_type(arg1, arg2, ...)
  patterns <- c(
    "^(.*?)\\s*\\(\\*[^)]*\\)\\s*\\((.*)\\)$",
    "^(.*?)\\((.*)\\)$"
  )

  for (pat in patterns) {
    m <- regexec(pat, sig, perl = TRUE)
    if (m[[1]][1] != -1) {
      matches <- regmatches(sig, m)[[1]]
      return_type <- trimws(matches[2])
      args_str <- trimws(matches[3])

      # Parse argument types
      if (args_str == "" || args_str == "void") {
        arg_types <- character(0)
      } else {
        # Split by comma, handling function pointers and nested parens
        arg_types <- parse_arg_list(args_str)
      }

      return(new_rtinycc_callback_signature(
        return_type = return_type,
        arg_types = arg_types,
        raw = signature,
        mode = mode
      ))
    }
  }

  # If no pattern matched, try simple fallback
  stop("Unable to parse callback signature: ", signature, call. = FALSE)
}

# Parse a comma-separated argument list, handling nested structures
parse_arg_list <- function(args_str) {
  if (args_str == "" || args_str == "void") {
    return(character(0))
  }

  args <- character(0)
  depth_paren <- 0L
  depth_bracket <- 0L
  start <- 1L
  n <- nchar(args_str, type = "chars")

  for (i in seq_len(n)) {
    ch <- substr(args_str, i, i)
    if (ch == "," && depth_paren == 0L && depth_bracket == 0L) {
      args <- c(args, trimws(substr(args_str, start, i - 1L)))
      start <- i + 1L
      next
    }

    if (ch == "(") {
      depth_paren <- depth_paren + 1L
    } else if (ch == ")") {
      depth_paren <- max(0L, depth_paren - 1L)
    } else if (ch == "[") {
      depth_bracket <- depth_bracket + 1L
    } else if (ch == "]") {
      depth_bracket <- max(0L, depth_bracket - 1L)
    }
  }

  args <- c(args, trimws(substr(args_str, start, n)))
  args <- args[nzchar(args)]

  # Remove parameter names (keep only type)
  # e.g., "int x" -> "int", "double* ptr" -> "double*"
  result <- character(length(args))
  for (i in seq_along(args)) {
    arg <- args[i]

    # Handle function pointer types: void (*)(int)
    # Keep them as-is, they'll be treated as "callback" type
    if (grepl("\\(", arg)) {
      result[i] <- "callback"
    } else {
      result[i] <- strip_callback_arg_name(arg)
    }
  }

  result
}

strip_callback_arg_name <- function(arg) {
  arg <- trimws(gsub("\\s+", " ", arg))
  if (!nzchar(arg)) {
    return(arg)
  }

  # Pointer declarators often attach the parameter name directly to the last
  # asterisk, for example `const char *name` or `int **out`.
  ptr_match <- regexec(
    "^(.*?)(\\*+)\\s*([A-Za-z_][A-Za-z0-9_]*)$",
    arg,
    perl = TRUE
  )[[1]]
  if (ptr_match[[1]] != -1) {
    pieces <- regmatches(arg, list(ptr_match))[[1]]
    return(trimws(paste0(pieces[[2]], pieces[[3]])))
  }

  parts <- strsplit(arg, "\\s+")[[1]]
  if (
    length(parts) > 1 &&
      grepl("^[A-Za-z_][A-Za-z0-9_]*$", parts[[length(parts)]])
  ) {
    parts <- parts[-length(parts)]
  }

  trimws(paste(parts, collapse = " "))
}

# Format signature for display
format_signature <- function(sig) {
  if (is.null(sig)) {
    return("unknown")
  }

  args <- if (length(sig$arg_types) == 0) {
    "void"
  } else {
    paste(sig$arg_types, collapse = ", ")
  }
  paste0(sig$return_type, " (*)(", args, ")")
}

# ============================================================================
# FFI Integration - Allow callbacks in tcc_bind()
# ============================================================================

#' Check if a type represents a callback
#'
#' @param type Type string to check
#' @return Logical
is_callback_type <- function(type) {
  if (is_rtinycc_ffi_type(type)) {
    return(ffi_type_family(type) %in% c("callback", "callback_async"))
  }
  if (!is.character(type)) {
    return(FALSE)
  }
  type == "callback" ||
    startsWith(type, "callback:") ||
    startsWith(type, "callback_async:")
}

is_callback_async_type <- function(type) {
  if (is_rtinycc_ffi_type(type)) {
    return(identical(ffi_type_family(type), "callback_async"))
  }
  if (!is.character(type)) {
    return(FALSE)
  }
  startsWith(type, "callback_async:")
}

#' Parse callback type specification
#'
#' @param type Type string like "callback:double(int,int)"
#' @return Parsed signature list or NULL
parse_callback_type <- function(type) {
  if (is_rtinycc_ffi_type(type)) {
    type <- type$name
  }

  if (type == "callback") {
    # Generic callback without signature - will need runtime determination
    return(NULL)
  }

  if (startsWith(type, "callback:")) {
    # Extract signature after "callback:"
    sig_str <- sub("^callback:", "", type)
    return(parse_callback_signature(sig_str, mode = "sync"))
  }

  if (startsWith(type, "callback_async:")) {
    sig_str <- sub("^callback_async:", "", type)
    return(parse_callback_signature(sig_str, mode = "async"))
  }

  NULL
}

#' Schedule a callback to run on the main thread
#'
#' Enqueue a callback for main-thread execution. Arguments must be basic
#' scalars or external pointers.
#'
#' @param callback A tcc_callback object
#' @param args List of arguments to pass to the callback
#' @return NULL (invisible)
#' @export
tcc_callback_async_schedule <- function(callback, args = list()) {
  if (!inherits(callback, "tcc_callback")) {
    stop("Expected tcc_callback object", call. = FALSE)
  }
  if (!is.list(args)) {
    stop("args must be a list", call. = FALSE)
  }
  .Call(RC_callback_async_schedule, callback, args)
  invisible(NULL)
}

#' Drain the async callback queue
#'
#' Execute any pending async callbacks immediately on the main R thread.
#' Normally callbacks fire automatically via R's event loop (input
#' handler on POSIX, message pump on Windows), so explicit draining is
#' only needed in test harnesses or tight batch loops that never yield
#' to R's event loop.
#'
#' TCC-compiled C code running on the main thread can call
#' \code{RC_callback_async_drain_c()} directly instead of returning to R.
#'
#' @return NULL (invisible)
#' @export
tcc_callback_async_drain <- function() {
  .Call(RC_callback_async_drain)
  invisible(NULL)
}

#' Generate trampoline code for a callback argument
#'
#' @param trampoline_name Name of the trampoline function
#' @param sig Parsed signature
#' @return C code string for the trampoline
generate_trampoline <- function(trampoline_name, sig) {
  # Generate a trampoline function that:
  # 1. Takes a user data pointer + C arguments
  # 2. Calls RC_invoke_callback_id() to invoke the R function
  # 3. Returns the result

  n_args <- length(sig$arg_types)

  # Build argument list for C function signature
  c_args <- c("void* cb")
  if (n_args > 0) {
    for (i in seq_len(n_args)) {
      c_args <- c(
        c_args,
        sprintf("%s arg%d", callback_c_decl_type(sig$arg_types[i]), i)
      )
    }
  }

  # Build the trampoline
  lines <- c(
    sprintf("// Trampoline %s", trampoline_name),
    sprintf(
      "%s %s(%s) {",
      callback_c_decl_type(sig$return_type),
      trampoline_name,
      if (length(c_args) > 0) paste(c_args, collapse = ", ") else "void"
    ),
    "  callback_token_t* tok = (callback_token_t*)cb;",
    "  if (!tok || tok->id < 0) {",
    sprintf(
      "    Rf_warning(\"Invalid callback token in %s\");",
      trampoline_name
    ),
    get_c_default_return(sig$return_type, indent = 4L),
    "  }"
  )

  # Build argument conversion
  if (n_args > 0) {
    lines <- c(
      lines,
      sprintf("  SEXP args = PROTECT(allocVector(VECSXP, %d));", n_args)
    )
    for (i in seq_len(n_args)) {
      lines <- c(
        lines,
        sprintf(
          "  SET_VECTOR_ELT(args, %d, %s);",
          i - 1,
          get_sexp_constructor_call(sig$arg_types[i], paste0("arg", i))
        )
      )
    }
  } else {
    lines <- c(lines, "  SEXP args = R_NilValue;")
  }

  # Call the runtime
  lines <- c(lines, "  SEXP result = RC_invoke_callback_id(tok->id, args);")

  if (n_args > 0) {
    lines <- c(lines, "  UNPROTECT(1);")
  }

  # Convert result back to C type
  lines <- c(
    lines,
    get_r_to_c_return_lines(sig$return_type, "result", indent = 2L)
  )

  lines <- c(lines, "}")

  paste(lines, collapse = "\n")
}

# Generate async trampoline that schedules callbacks on the main thread.
#
# void return    -> fire-and-forget via RC_callback_async_schedule_c
#                   (PostMessage on Windows, write-to-pipe on Linux).
#                   Auto-drains through R's input handler / message pump.
# non-void return -> synchronous via RC_callback_async_schedule_sync_c
#                   (SendMessage on Windows, pthread_cond on Linux).
#                   Blocks the worker thread until the main thread
#                   services the callback.  C code on the main thread
#                   can call RC_callback_async_drain_c() in a loop
#                   rather than returning to R.
generate_async_trampoline <- function(trampoline_name, sig) {
  unsupported <- vapply(sig$arg_types, async_type_unsupported, logical(1))
  if (any(unsupported)) {
    bad <- unique(trimws(sig$arg_types[unsupported]))
    stop(
      "callback_async unsupported argument type(s): ",
      paste(bad, collapse = ", "),
      call. = FALSE
    )
  }

  is_void_return <- identical(sig$return_type, "void")

  if (!is_void_return) {
    ret_kind <- async_result_kind(sig$return_type)
    if (is.null(ret_kind)) {
      stop(
        "callback_async unsupported return type: ",
        sig$return_type,
        " (supported: void, int variants, double/float, bool/logical, ptr)",
        call. = FALSE
      )
    }
  }

  n_args <- length(sig$arg_types)

  c_args <- c("void* cb")
  if (n_args > 0) {
    for (i in seq_len(n_args)) {
      c_args <- c(
        c_args,
        sprintf("%s arg%d", callback_c_decl_type(sig$arg_types[i]), i)
      )
    }
  }

  lines <- c(
    sprintf("// Async trampoline %s", trampoline_name),
    sprintf(
      "%s %s(%s) {",
      callback_c_decl_type(sig$return_type),
      trampoline_name,
      if (length(c_args) > 0) paste(c_args, collapse = ", ") else "void"
    ),
    "  callback_token_t* tok = (callback_token_t*)cb;",
    "  if (!tok || tok->id < 0) {",
    sprintf(
      "    Rf_warning(\"Invalid callback token in %s\");",
      trampoline_name
    ),
    get_c_default_return(sig$return_type, indent = 4L),
    "  }"
  )

  if (n_args > 0) {
    lines <- c(lines, sprintf("  cb_arg_t args[%d];", n_args))
    for (i in seq_len(n_args)) {
      lines <- c(lines, generate_cb_arg_assignment(i, sig$arg_types[i]))
    }
  } else {
    lines <- c(lines, "  cb_arg_t* args = NULL;")
  }

  if (is_void_return) {
    lines <- c(
      lines,
      sprintf(
        "  int rc = RC_callback_async_schedule_c(tok->id, %d, %s);",
        n_args,
        if (n_args > 0) "args" else "NULL"
      ),
      "  if (rc != 0) {",
      sprintf(
        "    Rf_warning(\"Async schedule failed (rc=%%d) in %s\", rc);",
        trampoline_name
      ),
      "  }",
      "  return;"
    )
  } else {
    lines <- c(
      lines,
      "  cb_result_t result;",
      sprintf(
        "  int rc = RC_callback_async_schedule_sync_c(tok->id, %d, %s, &result);",
        n_args,
        if (n_args > 0) "args" else "NULL"
      ),
      "  if (rc != 0) {",
      sprintf(
        "    Rf_warning(\"Async schedule failed (rc=%%d) in %s\", rc);",
        trampoline_name
      ),
      get_c_default_return(sig$return_type, indent = 4L),
      "  }",
      get_async_result_return(sig$return_type)
    )
  }

  lines <- c(lines, "}")

  paste(lines, collapse = "\n")
}

# Determine the CB_RESULT_* kind that a C return type maps to.
# Returns NULL if the type cannot be returned from an async sync callback.
async_result_kind <- function(return_type) {
  rt <- trimws(return_type)
  if (
    rt %in%
      c(
        "int",
        "i32",
        "int32_t",
        "i16",
        "int16_t",
        "i8",
        "int8_t",
        "u8",
        "uint8_t",
        "u16",
        "uint16_t"
      )
  ) {
    return("int")
  }
  if (
    rt %in%
      c(
        "double",
        "f64",
        "float",
        "f32",
        "i64",
        "int64_t",
        "u32",
        "uint32_t",
        "u64",
        "uint64_t"
      )
  ) {
    return("real")
  }
  if (rt %in% c("bool", "logical")) {
    return("logical")
  }
  if (is_ptr_type(rt)) {
    return("ptr")
  }
  NULL
}

# Generate the C return statement that extracts a value from cb_result_t.
get_async_result_return <- function(return_type) {
  kind <- async_result_kind(return_type)
  rt <- callback_c_decl_type(return_type)
  switch(
    kind,
    "int" = sprintf("  return (%s)result.v.i;", rt),
    "real" = sprintf("  return (%s)result.v.d;", rt),
    "logical" = sprintf("  return (%s)result.v.i;", rt),
    "ptr" = sprintf("  return (%s)result.v.p;", rt),
    stop(
      "BUG: unhandled async result kind for return type: ",
      return_type,
      call. = FALSE
    )
  )
}

async_type_unsupported <- function(c_type) {
  key <- normalize_cb_kind_key(c_type)
  if (key %in% c("ptr", "cstring", "real", "logical")) {
    return(FALSE)
  }
  if (!identical(key, "int")) {
    return(TRUE)
  }
  type_name <- trimws(c_type)
  !type_name %in%
    c(
      "int",
      "int32_t",
      "i32",
      "int16_t",
      "i16",
      "int8_t",
      "i8",
      "uint8_t",
      "u8",
      "uint16_t",
      "u16"
    )
}

generate_cb_arg_assignment <- function(index, c_type) {
  kind <- map_c_to_cb_arg_kind(c_type)
  value <- map_c_to_cb_arg_value(c_type, paste0("arg", index))
  c(
    sprintf("  args[%d].kind = %s;", index - 1, kind),
    sprintf("  args[%d].v.%s = %s;", index - 1, value$field, value$expr)
  )
}

is_ptr_type <- function(c_type) {
  c_type <- trimws(c_type)
  grepl("\\*", c_type) && !is_cstring_type(c_type)
}

is_cstring_type <- function(c_type) {
  c_type <- trimws(c_type)
  grepl("^const\\s+char\\s*\\*$", c_type) || grepl("^char\\s*\\*$", c_type)
}

normalize_cb_kind_key <- function(c_type) {
  cb_kind_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

normalize_sexp_type_key <- function(c_type) {
  sexp_type_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

normalize_r_type_key <- function(c_type) {
  c_r_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

normalize_sexp_ctor_key <- function(c_type) {
  sexp_ctor_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

normalize_r_to_c_key <- function(c_type) {
  r_to_c_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

normalize_default_key <- function(c_type) {
  default_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

normalize_return_key <- function(c_type) {
  return_key_rule(trimws(c_type), is_ptr_type(trimws(c_type)))
}

# Normalize callback signature spellings into concrete C declaration types.
# This keeps trampoline ABI faithful to the declared callback type while still
# accepting internal alias forms like i8/f32/sexp in callback signatures.
callback_c_decl_type <- function(c_type) {
  type_name <- trimws(gsub("\\s+", " ", c_type))
  if (!nzchar(type_name)) {
    return(type_name)
  }

  alias_map <- c(
    callback = "void*",
    ptr = "void*",
    cstring = "const char*",
    string = "const char*",
    sexp = "SEXP",
    logical = "bool",
    i8 = "int8_t",
    i16 = "int16_t",
    i32 = "int32_t",
    i64 = "int64_t",
    u8 = "uint8_t",
    u16 = "uint16_t",
    u32 = "uint32_t",
    u64 = "uint64_t",
    f32 = "float",
    f64 = "double"
  )

  if (type_name %in% names(alias_map)) {
    mapped <- unname(alias_map[[type_name]])
    return(mapped)
  }

  type_name
}

map_c_to_cb_arg_kind <- function(c_type) {
  cb_arg_kind_rule(normalize_cb_kind_key(c_type))
}

map_c_to_cb_arg_value <- function(c_type, arg_expr) {
  cb_arg_value_rule(normalize_cb_kind_key(c_type), arg_expr)
}

# Get default C return statement for a type
get_c_default_return <- function(c_type, indent = 2L) {
  c_default_return_rule(normalize_default_key(c_type), indent)
}

# Map C types to R SEXP types for trampoline arguments
map_c_to_sexp_type <- function(c_type) {
  c_sexp_type_rule(normalize_sexp_type_key(c_type))
}

# Map C types to FFI types
map_c_to_r_type <- function(c_type) {
  c_r_type_rule(normalize_r_type_key(c_type))
}

# Get SEXP constructor for C type
get_sexp_constructor <- function(c_type) {
  sexp_constructor_rule(normalize_sexp_ctor_key(c_type))
}

# Get SEXP constructor call for a C expression
get_sexp_constructor_call <- function(c_type, arg_expr) {
  sexp_constructor_call_rule(normalize_sexp_ctor_key(c_type), arg_expr)
}

# Get R to C converter function
get_r_to_c_converter <- function(c_type) {
  r_to_c_converter_rule(normalize_r_to_c_key(c_type))
}

get_r_to_c_return_lines <- function(c_type, result_var, indent = 2L) {
  r_to_c_return_lines_rule(
    normalize_return_key(c_type),
    c_type,
    result_var,
    indent
  )
}

# Test suite for R Callbacks (tcc_callback, tcc_callback_close, tcc_callback_ptr)
# These tests verify the callback registration system

library(tinytest)
library(Rtinycc)

callback_abi_specs <- Rtinycc:::rtinycc_callback_abi_specs()

# ============================================================================
# Test 1: Basic callback creation
# ============================================================================

# Create a simple callback
counter <- 0
cb <- tcc_callback(
  function() {
    counter <<- counter + 1
    counter
  },
  signature = "int (*)(void)"
)

expect_true(inherits(cb, "tcc_callback"), info = "Callback has correct class")
expect_true(inherits(cb, "externalptr"), info = "Callback is an externalptr")
expect_true(tcc_callback_valid(cb), info = "Callback is valid after creation")

# ============================================================================
# Test 2: Callback with return value
# ============================================================================
cb2 <- tcc_callback(
  function(x) {
    x * 2
  },
  signature = "double (*)(double)"
)

expect_true(tcc_callback_valid(cb2), info = "Double callback is valid")

# ============================================================================
# Test 3: Callback with multiple arguments
# ============================================================================
cb3 <- tcc_callback(
  function(a, b, c) {
    a + b + c
  },
  signature = "int (*)(int, int, int)"
)

expect_true(tcc_callback_valid(cb3), info = "Multi-arg callback is valid")

# ============================================================================
# Test 4: Error handling - non-function argument
# ============================================================================
expect_error(
  tcc_callback("not a function", signature = "void (*)(void)"),
  info = "tcc_callback rejects non-function"
)

# ============================================================================
# Test 5: Error handling - invalid signature
# ============================================================================
expect_error(
  tcc_callback(function() {}, signature = 123),
  info = "tcc_callback rejects non-character signature"
)

# ============================================================================
# Test 6: Error handling - multiple signatures
# ============================================================================
expect_error(
  tcc_callback(function() {}, signature = c("void (*)(void)", "int (*)(void)")),
  info = "tcc_callback rejects multiple signatures"
)

expect_true(
  Rtinycc:::is_callback_type(Rtinycc:::check_ffi_type(
    "callback:void*(void*)",
    "test"
  )),
  info = "classed callback ffi type is recognized as callback"
)
expect_true(
  Rtinycc:::is_callback_async_type(Rtinycc:::check_ffi_type(
    "callback_async:void*(void*)",
    "test"
  )),
  info = "classed async callback ffi type is recognized as async callback"
)

# ============================================================================
# Test 7: Close callback
# ============================================================================
cb_temp <- tcc_callback(function() 42, signature = "int (*)(void)")
expect_true(
  tcc_callback_valid(cb_temp),
  info = "Temp callback is valid before close"
)
tcc_callback_close(cb_temp)
expect_false(
  tcc_callback_valid(cb_temp),
  info = "Callback is invalid after close"
)

# ============================================================================
# Test 8: Double close error
# ============================================================================
cb_temp2 <- tcc_callback(function() 42, signature = "int (*)(void)")
tcc_callback_close(cb_temp2)
expect_error(
  tcc_callback_close(cb_temp2),
  info = "Double close should error"
)

# ============================================================================
# Test 9: Callback pointer extraction
# ============================================================================
cb4 <- tcc_callback(function() 1, signature = "int (*)(void)")
ptr <- tcc_callback_ptr(cb4)
expect_equal(
  typeof(ptr),
  "externalptr",
  info = "Callback ptr is externalptr (typeof)"
)
expect_true(
  inherits(ptr, "tcc_callback_ptr"),
  info = "Callback ptr has correct class"
)
expect_false(
  .Call("RC_ptr_is_owned", ptr, PACKAGE = "Rtinycc"),
  info = "Callback ptr wrapper is not tagged as owned"
)
expect_error(
  tcc_free(ptr),
  info = "Callback ptr wrapper is not explicitly freeable"
)

# ============================================================================
# Test 10: Callback pointer for closed callback
# ============================================================================
cb_temp3 <- tcc_callback(function() 1, signature = "int (*)(void)")
tcc_callback_close(cb_temp3)
expect_error(
  tcc_callback_ptr(cb_temp3),
  info = "Getting ptr from closed callback should error"
)

# ============================================================================
# Test 11: Print method
# ============================================================================
cb5 <- tcc_callback(function(x) x, signature = "int (*)(int)")
# Just verify print doesn't error
expect_silent(print(cb5), info = "Print method works for callback")

# ============================================================================
# Test 12: Thread-safe option
# ============================================================================
cb_thread <- tcc_callback(
  function() 1,
  signature = "int (*)(void)",
  threadsafe = TRUE
)
expect_true(
  tcc_callback_valid(cb_thread),
  info = "Thread-safe callback is valid"
)
expect_true(
  attr(cb_thread, "threadsafe"),
  info = "Threadsafe attribute is TRUE"
)
tcc_callback_close(cb_thread)

# ============================================================================
# Test 13: Non-threadsafe (default) option
# ============================================================================
cb_nothread <- tcc_callback(
  function() 1,
  signature = "int (*)(void)",
  threadsafe = FALSE
)
expect_false(
  attr(cb_nothread, "threadsafe"),
  info = "Threadsafe attribute is FALSE by default"
)
tcc_callback_close(cb_nothread)

# ============================================================================
# Test 14: Various C types in signatures
# ============================================================================
cb_int <- tcc_callback(function(x) as.integer(x), signature = "int (*)(int)")
cb_double <- tcc_callback(function(x) x, signature = "double (*)(double)")
cb_bool <- tcc_callback(function(x) as.logical(x), signature = "bool (*)(bool)")
cb_void <- tcc_callback(
  function(x) invisible(NULL),
  signature = "void (*)(int)"
)
cb_ptr <- tcc_callback(function(x) x, signature = "void* (*)(void*)")

expect_true(tcc_callback_valid(cb_int), info = "Int callback valid")
expect_true(tcc_callback_valid(cb_double), info = "Double callback valid")
expect_true(tcc_callback_valid(cb_bool), info = "Bool callback valid")
expect_true(tcc_callback_valid(cb_void), info = "Void callback valid")
expect_true(tcc_callback_valid(cb_ptr), info = "Ptr callback valid")
expect_false(
  .Call("RC_ptr_is_owned", cb_ptr, PACKAGE = "Rtinycc"),
  info = "Callback object itself is not tcc_free-owned"
)
expect_error(
  tcc_free(cb_ptr),
  info = "Callback object cannot be freed with tcc_free"
)

# Close all
tcc_callback_close(cb_int)
tcc_callback_close(cb_double)
tcc_callback_close(cb_bool)
tcc_callback_close(cb_void)
tcc_callback_close(cb_ptr)

# ============================================================================
# Test 15: Cleanup remaining callbacks
# ============================================================================
tcc_callback_close(cb)
tcc_callback_close(cb2)
tcc_callback_close(cb3)
tcc_callback_close(cb4)
tcc_callback_close(cb5)

# rm(ptr)
# gc()

# Final validation
expect_false(tcc_callback_valid(cb), info = "All callbacks cleaned up")

# ==========================================================================
# Test 16: Trampoline codegen safeguards
# ==========================================================================
sig_cstr <- list(return_type = "const char*", arg_types = c("const char*"))
tramp_cstr <- Rtinycc:::generate_trampoline("tramp_cstr", sig_cstr)
expect_true(
  grepl("? mkString", tramp_cstr, fixed = TRUE),
  info = "Trampoline guards NULL cstring args"
)
expect_true(
  grepl("Rf_translateCharUTF8", tramp_cstr),
  info = "Trampoline translates cstring returns"
)
expect_true(
  grepl("NA_STRING", tramp_cstr),
  info = "Trampoline handles NA cstring returns"
)

sig_bool <- list(return_type = "bool", arg_types = character())
tramp_bool <- Rtinycc:::generate_trampoline("tramp_bool", sig_bool)
expect_true(
  grepl("asLogical", tramp_bool),
  info = "Trampoline converts logical returns"
)
expect_true(
  grepl("NA_LOGICAL", tramp_bool),
  info = "Trampoline guards NA logical returns"
)

for (case in callback_abi_specs$trampoline) {
  sig <- Rtinycc:::parse_callback_signature(case$signature)
  tramp <- Rtinycc:::generate_trampoline(case$name, sig)
  expect_true(grepl(case$pattern, tramp), info = case$info)
}

for (case in callback_abi_specs$wrapper) {
  symbols <- setNames(
    list(list(args = case$args, returns = case$returns)),
    case$name
  )
  code <- Rtinycc:::generate_ffi_code(
    symbols = symbols,
    c_code = case$c_code
  )

  for (pattern in case$patterns) {
    expect_true(grepl(pattern$pattern, code), info = pattern$info)
  }

  forbidden <- case$forbidden
  if (is.null(forbidden)) {
    forbidden <- list()
  }

  for (pattern in forbidden) {
    expect_false(grepl(pattern$pattern, code), info = pattern$info)
  }
}

# ==========================================================================
# Test 17: Async trampoline validation (supported args, pointer handling)
# ==========================================================================
mk_tramps <- function(arg_type) {
  Rtinycc:::generate_callback_trampolines(
    list(
      run = list(
        args = list(arg_type),
        returns = "i32"
      )
    )
  )
}

expect_silent(
  mk_tramps("callback_async:double(int)"),
  info = "Async callback allows non-void return signatures"
)

expect_silent(
  mk_tramps("callback_async:void(int64_t)"),
  info = "Async callback accepts i64 arguments through the real slot"
)
expect_silent(
  mk_tramps("callback_async:void(uint32_t)"),
  info = "Async callback accepts u32 arguments through the real slot"
)
expect_silent(
  mk_tramps("callback_async:void(uint64_t)"),
  info = "Async callback accepts u64 arguments through the real slot"
)
wide_arg_tramp <- mk_tramps("callback_async:void(int64_t)")$code
expect_true(
  grepl("9007199254740992LL", wide_arg_tramp, fixed = TRUE),
  info = "Async i64 callback arguments guard the exact R numeric range"
)
expect_true(
  grepl("RC_callback_async_note_failure_c(-7)", wide_arg_tramp, fixed = TRUE),
  info = "Inexact async wide arguments record a dispatch failure"
)

expect_silent(
  mk_tramps("callback_async:void(int, double, const char*, void*)"),
  info = "Async callback accepts supported argument types"
)

sig_ptrptr <- Rtinycc:::parse_callback_type("callback_async:void(char **)")
tramp_ptrptr <- Rtinycc:::generate_async_trampoline("tramp_ptrptr", sig_ptrptr)
expect_true(
  grepl("CB_ARG_PTR", tramp_ptrptr, fixed = TRUE),
  info = "Async callback treats char** as pointer"
)
expect_true(
  grepl(".v.p = arg1;", tramp_ptrptr, fixed = TRUE),
  info = "Async callback stores char** in pointer slot"
)

for (case in callback_abi_specs$async_trampoline) {
  sig <- Rtinycc:::parse_callback_type(case$signature)
  tramp <- Rtinycc:::generate_async_trampoline(case$name, sig)

  for (pattern in case$patterns) {
    expect_true(grepl(pattern$pattern, tramp), info = pattern$info)
  }

  forbidden <- case$forbidden
  if (is.null(forbidden)) {
    forbidden <- list()
  }

  for (pattern in forbidden) {
    expect_false(grepl(pattern$pattern, tramp), info = pattern$info)
  }
}

expect_error(
  mk_tramps("callback_async:SEXP(SEXP)"),
  info = "Async callback rejects SEXP arguments and returns"
)
expect_error(
  mk_tramps("callback_async:SEXP(int)"),
  info = "Async callback rejects SEXP return type"
)
expect_error(
  mk_tramps("callback_async:void(SEXP)"),
  info = "Async callback rejects SEXP argument type"
)

library(tinytest)
library(Rtinycc)

sig_ptr <- Rtinycc:::parse_callback_signature(
  "int (*)(const char *name, int **out)"
)

expect_true(Rtinycc:::is_rtinycc_callback_signature(sig_ptr))
expect_identical(sig_ptr$mode, "sync")
expect_identical(sig_ptr$return_type, "int")
expect_identical(sig_ptr$arg_types, c("const char *", "int **"))

sig_plain <- Rtinycc:::parse_callback_signature(
  "double (*)(double x, int count)"
)

expect_identical(sig_plain$arg_types, c("double", "int"))

sig_cb <- Rtinycc:::parse_callback_signature(
  "void (*)(void (*fun)(int), void *ctx)"
)

expect_identical(sig_cb$arg_types, c("callback", "void *"))

sig_async <- Rtinycc:::parse_callback_type("callback_async:void*(void*)")
expect_true(Rtinycc:::is_rtinycc_callback_signature(sig_async))
expect_identical(sig_async$mode, "async")
expect_identical(sig_async$return_type, "void*")
expect_identical(sig_async$arg_types, "void*")

library(tinytest)
library(Rtinycc)

close_if_valid <- function(cb) {
  if (inherits(cb, "tcc_callback") && isTRUE(tcc_callback_valid(cb))) {
    tcc_callback_close(cb)
  }
}

# Test: register an R callback, compile C code that invokes it, and call it
cb <- tcc_callback(function(x) 42, signature = "double (*)(double)")
cb_ptr <- tcc_callback_ptr(cb)

code <- "\n#define _Complex\n\ndouble call_cb(double (*cb)(void* ctx, double), void* ctx, double x) {\n  return cb(ctx, x);\n}\n"

ffi <- tcc_ffi() |>
  tcc_source(code) |>
  tcc_bind(
    call_cb = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

expect_true(inherits(ffi, "tcc_compiled"), info = "Compiled FFI object")

res <- ffi$call_cb(cb, cb_ptr, 21.0)
expect_true(isTRUE(all.equal(res, 42.0, tolerance = 1e-12)))
close_if_valid(cb)

# Test: callback error yields warning and default
cb_err <- tcc_callback(
  function(x) stop("boom"),
  signature = "double (*)(double)"
)
cb_ptr_err <- tcc_callback_ptr(cb_err)

code_err <- "\n#define _Complex\n\ndouble call_cb_err(double (*cb)(void* ctx, double), void* ctx, double x) {\n  return cb(ctx, x);\n}\n"

ffi_err <- tcc_ffi() |>
  tcc_source(code_err) |>
  tcc_bind(
    call_cb_err = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

warned <- FALSE
res <- NULL
tmp <- tempfile()
con <- file(tmp, open = "wt")
sink(con)
sink(con, type = "message")
res <- tryCatch(
  withCallingHandlers(
    ffi_err$call_cb_err(cb_err, cb_ptr_err, 1.0),
    warning = function(w) {
      warned <<- TRUE
      invokeRestart("muffleWarning")
    }
  ),
  finally = {
    sink(type = "message")
    sink()
    close(con)
    unlink(tmp)
  }
)

expect_true(
  isTRUE(warned && is.na(res)),
  info = "Callback errors yield warning and NA"
)
close_if_valid(cb_err)

# Test: pointer return and pointer args use externalptr
ptr_arg_seen_externalptr <- FALSE
ptr_arg_seen_unowned <- FALSE
cb_ptr_rt <- tcc_callback(
  function(x) {
    ptr_arg_seen_externalptr <<- inherits(x, "externalptr")
    ptr_arg_seen_unowned <<- !.Call("RC_ptr_is_owned", x, PACKAGE = "Rtinycc")
    x
  },
  signature = "void* (*)(void*)"
)
cb_ptr_handle <- tcc_callback_ptr(cb_ptr_rt)

code_ptr <- "\n#define _Complex\n\nvoid* echo_ptr(void* (*cb)(void* ctx, void* x), void* ctx, void* x) {\n  return cb(ctx, x);\n}\n"

ffi_ptr <- tcc_ffi() |>
  tcc_source(code_ptr) |>
  tcc_bind(
    echo_ptr = list(
      args = list("callback:void*(void*)", "ptr", "ptr"),
      returns = "ptr"
    )
  ) |>
  tcc_compile()

buf <- tcc_malloc(8)
out <- ffi_ptr$echo_ptr(cb_ptr_rt, cb_ptr_handle, buf)
expect_true(inherits(out, "externalptr"))
expect_true(
  ptr_arg_seen_externalptr,
  info = "Pointer callback receives externalptr wrapper"
)
expect_true(
  ptr_arg_seen_unowned,
  info = "Pointer callback receives non-owned pointer wrapper"
)
expect_false(
  .Call("RC_ptr_is_owned", out, PACKAGE = "Rtinycc"),
  info = "Pointer callback return wrapper is not owned"
)
expect_error(
  tcc_free(out),
  info = "Pointer callback return wrapper is not explicitly freeable"
)
tcc_free(buf)
close_if_valid(cb_ptr_rt)

# Test: SEXP callbacks pass SEXP objects through directly
cb_sexp <- tcc_callback(function(x) x, signature = "SEXP (*)(SEXP)")
cb_ptr_sexp <- tcc_callback_ptr(cb_sexp)

code_sexp <- "\n#define _Complex\n\nSEXP echo_sexp(SEXP (*cb)(void* ctx, SEXP x), void* ctx, SEXP x) {\n  return cb(ctx, x);\n}\n"

ffi_sexp <- tcc_ffi() |>
  tcc_source(code_sexp) |>
  tcc_bind(
    echo_sexp = list(
      args = list("callback:SEXP(SEXP)", "ptr", "sexp"),
      returns = "sexp"
    )
  ) |>
  tcc_compile()

payload <- list(alpha = 1:3, beta = "ok")
out_sexp <- ffi_sexp$echo_sexp(cb_sexp, cb_ptr_sexp, payload)
expect_identical(
  out_sexp,
  payload,
  info = "SEXP callback path passes objects through directly"
)
close_if_valid(cb_sexp)

# Test: ptr handle stays alive but callback is invalid after close
cb_closed <- tcc_callback(function(x) x, signature = "double (*)(double)")
cb_ptr_closed <- tcc_callback_ptr(cb_closed)

code_closed <- "\n#define _Complex\n\ndouble call_cb_closed(double (*cb)(void* ctx, double), void* ctx, double x) {\n  return cb(ctx, x);\n}\n"

ffi_closed <- tcc_ffi() |>
  tcc_source(code_closed) |>
  tcc_bind(
    call_cb_closed = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

tcc_callback_close(cb_closed)
warned <- FALSE
res <- withCallingHandlers(
  ffi_closed$call_cb_closed(cb_closed, cb_ptr_closed, 1.0),
  warning = function(w) {
    warned <<- TRUE
    invokeRestart("muffleWarning")
  }
)
expect_true(
  isTRUE(warned && is.na(res)),
  info = "Closed callback yields warning and NA"
)
close_if_valid(cb_closed)

# Test: async scheduling from worker thread (cross-platform)
hits <- 0L
cb_async <- tcc_callback(
  function(x) {
    hits <<- hits + x
    NULL
  },
  signature = "void (*)(int)"
)
cb_ptr_async <- tcc_callback_ptr(cb_async)

code_async <- "\n#define _Complex\n\nstruct task { void (*cb)(void* ctx, int); void* ctx; int value; };\n\n#ifdef _WIN32\n#include <windows.h>\n\nstatic DWORD WINAPI worker(LPVOID data) {\n  struct task* t = (struct task*) data;\n  t->cb(t->ctx, t->value);\n  return 0;\n}\n\nint spawn_async(void (*cb)(void* ctx, int), void* ctx, int value) {\n  if (!cb || !ctx) return -1;\n  struct task t;\n  t.cb = cb;\n  t.ctx = ctx;\n  t.value = value;\n  HANDLE th = CreateThread(NULL, 0, worker, &t, 0, NULL);\n  if (!th) return -2;\n  WaitForSingleObject(th, INFINITE);\n  CloseHandle(th);\n  return 0;\n}\n#else\n#include <pthread.h>\n\nstatic void* worker(void* data) {\n  struct task* t = (struct task*) data;\n  t->cb(t->ctx, t->value);\n  return NULL;\n}\n\nint spawn_async(void (*cb)(void* ctx, int), void* ctx, int value) {\n  if (!cb || !ctx) return -1;\n  struct task t;\n  t.cb = cb;\n  t.ctx = ctx;\n  t.value = value;\n  pthread_t th;\n  if (pthread_create(&th, NULL, worker, &t) != 0) return -2;\n  pthread_join(th, NULL);\n  return 0;\n}\n#endif\n"

ffi_async <- tcc_ffi() |>
  tcc_source(code_async)
if (.Platform$OS.type != "windows") {
  ffi_async <- tcc_library(ffi_async, "pthread")
}
ffi_async <- ffi_async |>
  tcc_bind(
    spawn_async = list(
      args = list("callback_async:void(int)", "ptr", "i32"),
      returns = "i32"
    )
  ) |>
  tcc_compile()

rc <- ffi_async$spawn_async(cb_async, cb_ptr_async, 2L)
tcc_callback_async_drain()

expect_true(
  isTRUE(rc == 0L && hits == 2L),
  info = "Async callback scheduled from worker thread"
)
close_if_valid(cb_async)

# Test: direct C scheduling from the main R thread executes immediately.
# This exercises the main-thread fast path used to avoid queue-and-wait
# deadlocks when a TCC-side main-thread shim schedules an async callback.
hits_direct <- 0L
cb_direct <- tcc_callback(
  function(x) {
    hits_direct <<- hits_direct + x
    NULL
  },
  signature = "void (*)(int)"
)
cb_ptr_direct <- tcc_callback_ptr(cb_direct)

code_direct_schedule <- "
#define _Complex

typedef enum { CB_ARG_INT = 0, CB_ARG_REAL = 1, CB_ARG_LOGICAL = 2, CB_ARG_PTR = 3, CB_ARG_CSTRING = 4 } cb_arg_kind_t;
typedef struct { cb_arg_kind_t kind; union { int i; double d; void* p; char* s; } v; } cb_arg_t;
typedef struct {
  int id;
  int refs;
  int origin_id;
  unsigned int generation;
} callback_token_t;
int RC_callback_async_schedule_c(int id, unsigned int generation,
                                 int n_args, const cb_arg_t *args);

int schedule_direct(void* ctx, int value) {
  callback_token_t* tok = (callback_token_t*)ctx;
  cb_arg_t args[1];
  if (!tok || tok->id < 0) return -99;
  args[0].kind = CB_ARG_INT;
  args[0].v.i = value;
  return RC_callback_async_schedule_c(
    tok->id, tok->generation, 1, args
  );
}
"

ffi_direct_schedule <- tcc_ffi() |>
  tcc_source(code_direct_schedule) |>
  tcc_bind(
    schedule_direct = list(args = list("ptr", "i32"), returns = "i32")
  ) |>
  tcc_compile()

expect_equal(
  ffi_direct_schedule$schedule_direct(cb_ptr_direct, 4L),
  0L,
  info = "Direct main-thread async schedule succeeds"
)
expect_equal(
  hits_direct,
  4L,
  info = "Direct main-thread async schedule executes without waiting for event-loop auto-drain"
)
close_if_valid(cb_direct)

# A warning promoted to an error must be caught before it can unwind through
# the TCC caller on the main-thread fast path.
cb_direct_error <- tcc_callback(
  function(x) stop("direct callback failure"),
  signature = "void (*)(int)"
)
old_options <- options(warn = 2)
direct_error_rc <- NULL
invisible(capture.output(
  direct_error_rc <- tryCatch(
    ffi_direct_schedule$schedule_direct(
      tcc_callback_ptr(cb_direct_error),
      1L
    ),
    finally = options(old_options)
  ),
  type = "message"
))
expect_equal(
  direct_error_rc,
  -5L,
  info = "Main-thread async callback non-local jumps become dispatch failures"
)
close_if_valid(cb_direct_error)

# Test: queued R-level async scheduling auto-fires during R event-loop activity.
# This exercises the input-handler/message pump path without a manual drain call.
hits_event_loop <- 0L
cb_event_loop <- tcc_callback(
  function(x) {
    hits_event_loop <<- hits_event_loop + x
    NULL
  },
  signature = "void (*)(int)"
)
tcc_callback_async_schedule(cb_event_loop, list(5L))
for (i in seq_len(100)) {
  if (hits_event_loop == 5L) {
    break
  }
  Sys.sleep(0.02)
}
expect_equal(
  hits_event_loop,
  5L,
  info = "Queued async callback auto-fires during R event-loop activity without manual drain"
)
close_if_valid(cb_event_loop)

# A queued task must not target a new callback that reuses the same registry
# slot after the original callback is closed.
old_hits <- 0L
new_hits <- 0L
queue_then_replace <- function() {
  old <- tcc_callback(
    function(x) {
      old_hits <<- old_hits + x
      NULL
    },
    signature = "void (*)(int)"
  )
  tcc_callback_async_schedule(old, list(7L))
  tcc_callback_close(old)

  replacement <- tcc_callback(
    function(x) {
      new_hits <<- new_hits + x
      NULL
    },
    signature = "void (*)(int)"
  )
  on.exit(close_if_valid(replacement), add = TRUE)
  tcc_callback_async_drain()
}
queue_then_replace()
expect_equal(
  old_hits,
  0L,
  info = "Closing a callback cancels its queued fire-and-forget invocation"
)
expect_equal(
  new_hits,
  0L,
  info = "A queued callback cannot jump to a replacement registry entry"
)

# Test: async trampoline error paths return the declared C default without
# attempting to call R APIs from the worker-capable trampoline body.
# A closed callback keeps the callback-pointer token alive with id = -1, which
# exercises the early invalid-token branch. RC_cleanup_callbacks() invalidates
# the registry while leaving a non-negative token id, which exercises the
# schedule-failure branch.
code_invalid_async <- "
#define _Complex

int call_async_i32(int (*cb)(void* ctx, int), void* ctx, int x) {
  return cb(ctx, x);
}

void call_async_void(void (*cb)(void* ctx, int), void* ctx, int x) {
  cb(ctx, x);
}
"

ffi_invalid_async <- tcc_ffi() |>
  tcc_source(code_invalid_async) |>
  tcc_bind(
    call_async_i32 = list(
      args = list("callback_async:int(int)", "ptr", "i32"),
      returns = "i32"
    ),
    call_async_void = list(
      args = list("callback_async:void(int)", "ptr", "i32"),
      returns = "void"
    )
  ) |>
  tcc_compile()

.Call("RC_callback_async_failure_reset", PACKAGE = "Rtinycc")

invalid_hits <- 0L
cb_invalid_i32 <- tcc_callback(
  function(x) {
    invalid_hits <<- invalid_hits + 1L
    x + 1L
  },
  signature = "int (*)(int)"
)
cb_ptr_invalid_i32 <- tcc_callback_ptr(cb_invalid_i32)
tcc_callback_close(cb_invalid_i32)
invalid_res <- NULL
expect_silent(
  invalid_res <- ffi_invalid_async$call_async_i32(cb_invalid_i32, cb_ptr_invalid_i32, 7L),
  info = "Closed async callback token follows invalid-token error path without warning"
)
expect_true(
  is.na(invalid_res),
  info = "Invalid-token async callback returns default integer sentinel"
)
expect_equal(
  invalid_hits,
  0L,
  info = "Invalid-token async callback does not invoke the released R function"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(1L, -2L),
  info = "Invalid-token async callback records an observable failure code"
)
expect_silent(
  ffi_invalid_async$call_async_void(cb_invalid_i32, cb_ptr_invalid_i32, 7L),
  info = "Void async invalid-token path returns without warning"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(2L, -2L),
  info = "Void async invalid-token path records an observable failure code"
)

cb_stale_i32 <- tcc_callback(function(x) x + 10L, signature = "int (*)(int)")
cb_ptr_stale_i32 <- tcc_callback_ptr(cb_stale_i32)
.Call("RC_cleanup_callbacks", PACKAGE = "Rtinycc")
stale_res <- NULL
expect_silent(
  stale_res <- ffi_invalid_async$call_async_i32(cb_stale_i32, cb_ptr_stale_i32, 8L),
  info = "Stale async callback token follows schedule-failure error path without warning"
)
expect_true(
  is.na(stale_res),
  info = "Schedule-failure async callback returns default integer sentinel"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(3L, -2L),
  info = "Schedule-failure async callback records an observable failure code"
)

# Test: non-void async callback returns the real computed value (synchronous path).
#
# The worker thread blocks on the cond/SendMessage until R drains.
# We must NOT join the worker while calling drain — that deadlocks.
# Pattern: start worker (non-blocking) -> drain loop -> join -> check result.
cb_async_int <- tcc_callback(
  function(x) x * 3L,
  signature = "int (*)(int)"
)
# RC_cleanup_callbacks() above made the stale slot reusable. Finalizing its old
# token must not invalidate the new callback occupying that slot.
tcc_callback_close(cb_stale_i32)
expect_true(
  tcc_callback_valid(cb_async_int),
  info = "A stale callback token cannot invalidate a reused registry slot"
)
cb_ptr_async_int <- tcc_callback_ptr(cb_async_int)

code_async_sync <- "
#define _Complex

#ifdef _WIN32
#include <windows.h>

struct itask { int (*cb)(void*,int); void* ctx; int in; volatile int out; volatile int done; };
static struct itask g_it;
static HANDLE g_ith = NULL;

static DWORD WINAPI iworker(LPVOID p) {
  struct itask* t = (struct itask*) p;
  t->out = t->cb(t->ctx, t->in);
  t->done = 1;
  return 0;
}
int  start_int_worker(int (*cb)(void*,int), void* ctx, int x) {
  g_it.cb = cb; g_it.ctx = ctx; g_it.in = x; g_it.out = -999; g_it.done = 0;
  g_ith = CreateThread(NULL, 0, iworker, &g_it, 0, NULL);
  return g_ith ? 0 : -1;
}
int  is_int_worker_done(void) { return g_it.done; }
int  join_int_worker(void) {
  if (g_ith) { WaitForSingleObject(g_ith, INFINITE); CloseHandle(g_ith); g_ith = NULL; }
  return g_it.out;
}

#else
#include <pthread.h>

struct itask { int (*cb)(void*,int); void* ctx; int in; volatile int out; volatile int done; };
static struct itask g_it;
static pthread_t g_ith;

static void* iworker(void* p) {
  struct itask* t = (struct itask*) p;
  t->out = t->cb(t->ctx, t->in);
  t->done = 1;
  return NULL;
}
int  start_int_worker(int (*cb)(void*,int), void* ctx, int x) {
  g_it.cb = cb; g_it.ctx = ctx; g_it.in = x; g_it.out = -999; g_it.done = 0;
  return pthread_create(&g_ith, NULL, iworker, &g_it) == 0 ? 0 : -1;
}
int  is_int_worker_done(void) { return g_it.done; }
int  join_int_worker(void) { pthread_join(g_ith, NULL); return g_it.out; }
#endif
"

ffi_async_sync <- tcc_ffi() |>
  tcc_source(code_async_sync)
if (.Platform$OS.type != "windows") {
  ffi_async_sync <- tcc_library(ffi_async_sync, "pthread")
}
ffi_async_sync <- ffi_async_sync |>
  tcc_bind(
    start_int_worker = list(
      args = list("callback_async:int(int)", "ptr", "i32"),
      returns = "i32"
    ),
    is_int_worker_done = list(args = list(), returns = "i32"),
    join_int_worker = list(args = list(), returns = "i32")
  ) |>
  tcc_compile()

ffi_async_sync$start_int_worker(cb_async_int, cb_ptr_async_int, 7L)
for (i in seq_len(50)) {
  tcc_callback_async_drain()
  if (ffi_async_sync$is_int_worker_done() != 0L) {
    break
  }
  Sys.sleep(0.01)
}
result_val <- ffi_async_sync$join_int_worker()
expect_true(
  isTRUE(result_val == 21L),
  info = "Non-void async callback returns real computed value (7 * 3 = 21)"
)
close_if_valid(cb_async_int)

# Closing a callback from inside its own R function must not invalidate the
# return-type metadata still needed by the active trampoline.
cb_self_close <- NULL
cb_self_close <- tcc_callback(
  function(x) {
    tcc_callback_close(cb_self_close)
    x + 1L
  },
  signature = "int (*)(int)"
)
self_close_result <- ffi_async_sync$start_int_worker(
  cb_self_close,
  tcc_callback_ptr(cb_self_close),
  41L
)
for (i in seq_len(50)) {
  tcc_callback_async_drain()
  if (ffi_async_sync$is_int_worker_done() != 0L) {
    break
  }
  Sys.sleep(0.01)
}
self_close_result <- ffi_async_sync$join_int_worker()
expect_equal(
  self_close_result,
  42L,
  info = "An async callback may close itself without invalidating its result"
)
expect_false(
  tcc_callback_valid(cb_self_close),
  info = "A self-closed callback remains invalid after returning"
)

# A non-local jump while the main thread executes a queued task must still
# signal the blocked worker and return a typed failure instead of deadlocking.
.Call("RC_callback_async_failure_reset", PACKAGE = "Rtinycc")
cb_async_jump <- tcc_callback(
  function(x) stop("queued callback failure"),
  signature = "int (*)(int)"
)
old_options <- options(warn = 2)
queued_jump_result <- NULL
invisible(capture.output(
  tryCatch(
    {
      ffi_async_sync$start_int_worker(
        cb_async_jump,
        tcc_callback_ptr(cb_async_jump),
        1L
      )
      for (i in seq_len(50)) {
        tcc_callback_async_drain()
        if (ffi_async_sync$is_int_worker_done() != 0L) {
          break
        }
        Sys.sleep(0.01)
      }
      queued_jump_result <- ffi_async_sync$join_int_worker()
    },
    finally = options(old_options)
  ),
  type = "message"
))
expect_true(
  is.na(queued_jump_result),
  info = "Queued callback non-local jump returns the declared default"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(1L, -5L),
  info = "Queued callback non-local jump is recorded as a dispatch failure"
)
close_if_valid(cb_async_jump)

# Test: non-void async callback (double return) from worker thread
cb_async_dbl <- tcc_callback(
  function(x) x + 0.5,
  signature = "double (*)(double)"
)
cb_ptr_async_dbl <- tcc_callback_ptr(cb_async_dbl)

code_async_dbl <- "
#define _Complex

#ifdef _WIN32
#include <windows.h>

struct dtask { double (*cb)(void*,double); void* ctx; double in; volatile double out; volatile int done; };
static struct dtask g_dt;
static HANDLE g_dth = NULL;

static DWORD WINAPI dworker(LPVOID p) {
  struct dtask* t = (struct dtask*) p;
  t->out = t->cb(t->ctx, t->in);
  t->done = 1;
  return 0;
}
int    start_dbl_worker(double (*cb)(void*,double), void* ctx, double x) {
  g_dt.cb = cb; g_dt.ctx = ctx; g_dt.in = x; g_dt.out = -999.0; g_dt.done = 0;
  g_dth = CreateThread(NULL, 0, dworker, &g_dt, 0, NULL);
  return g_dth ? 0 : -1;
}
int    is_dbl_worker_done(void) { return g_dt.done; }
double join_dbl_worker(void) {
  if (g_dth) { WaitForSingleObject(g_dth, INFINITE); CloseHandle(g_dth); g_dth = NULL; }
  return g_dt.out;
}

#else
#include <pthread.h>

struct dtask { double (*cb)(void*,double); void* ctx; double in; volatile double out; volatile int done; };
static struct dtask g_dt;
static pthread_t g_dth;

static void* dworker(void* p) {
  struct dtask* t = (struct dtask*) p;
  t->out = t->cb(t->ctx, t->in);
  t->done = 1;
  return NULL;
}
int    start_dbl_worker(double (*cb)(void*,double), void* ctx, double x) {
  g_dt.cb = cb; g_dt.ctx = ctx; g_dt.in = x; g_dt.out = -999.0; g_dt.done = 0;
  return pthread_create(&g_dth, NULL, dworker, &g_dt) == 0 ? 0 : -1;
}
int    is_dbl_worker_done(void) { return g_dt.done; }
double join_dbl_worker(void) { pthread_join(g_dth, NULL); return g_dt.out; }
#endif
"

ffi_async_dbl <- tcc_ffi() |>
  tcc_source(code_async_dbl)
if (.Platform$OS.type != "windows") {
  ffi_async_dbl <- tcc_library(ffi_async_dbl, "pthread")
}
ffi_async_dbl <- ffi_async_dbl |>
  tcc_bind(
    start_dbl_worker = list(
      args = list("callback_async:double(double)", "ptr", "f64"),
      returns = "i32"
    ),
    is_dbl_worker_done = list(args = list(), returns = "i32"),
    join_dbl_worker = list(args = list(), returns = "f64")
  ) |>
  tcc_compile()

ffi_async_dbl$start_dbl_worker(cb_async_dbl, cb_ptr_async_dbl, 2.5)
for (i in seq_len(50)) {
  tcc_callback_async_drain()
  if (ffi_async_dbl$is_dbl_worker_done() != 0L) {
    break
  }
  Sys.sleep(0.01)
}
result_dbl <- ffi_async_dbl$join_dbl_worker()
expect_true(
  isTRUE(all.equal(result_dbl, 3.0, tolerance = 1e-12)),
  info = "Non-void async callback returns real double value (2.5 + 0.5 = 3.0)"
)
close_if_valid(cb_async_dbl)

# Async result transport must check the active union kind instead of reading
# whichever member the binding expected.
.Call("RC_callback_async_failure_reset", PACKAGE = "Rtinycc")
ffi_async_result_checks <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdint.h>
    int call_async_i32(int (*cb)(void*, int), void *ctx, int x) {
      return cb(ctx, x);
    }
    uint8_t call_async_u8(uint8_t (*cb)(void*, uint8_t),
                          void *ctx, uint8_t x) {
      return cb(ctx, x);
    }
    int64_t call_async_i64(int64_t (*cb)(void*, int64_t),
                           void *ctx, int64_t x) {
      return cb(ctx, x);
    }
    int64_t call_async_i64_max(int64_t (*cb)(void*, int64_t), void *ctx) {
      return cb(ctx, INT64_MAX);
    }
    "
  ) |>
  tcc_bind(
    call_async_i32 = list(
      args = list("callback_async:int(int)", "ptr", "i32"),
      returns = "i32"
    ),
    call_async_u8 = list(
      args = list("callback_async:uint8_t(uint8_t)", "ptr", "u8"),
      returns = "u8"
    ),
    call_async_i64 = list(
      args = list("callback_async:int64_t(int64_t)", "ptr", "i64"),
      returns = "i64"
    ),
    call_async_i64_max = list(
      args = list("callback_async:int64_t(int64_t)", "ptr"),
      returns = "i64"
    )
  ) |>
  tcc_compile()

cb_wrong_kind <- tcc_callback(
  function(x) 42.5,
  signature = "double (*)(double)"
)
wrong_kind_result <- ffi_async_result_checks$call_async_i32(
  cb_wrong_kind,
  tcc_callback_ptr(cb_wrong_kind),
  7L
)
expect_true(
  is.na(wrong_kind_result),
  info = "Async callback result-kind mismatch returns the declared default"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(1L, -5L),
  info = "Async callback result-kind mismatch is recorded"
)
close_if_valid(cb_wrong_kind)

cb_u8_bad <- tcc_callback(
  function(x) -1L,
  signature = "uint8_t (*)(uint8_t)"
)
u8_bad_result <- ffi_async_result_checks$call_async_u8(
  cb_u8_bad,
  tcc_callback_ptr(cb_u8_bad),
  0L
)
expect_equal(
  u8_bad_result,
  0L,
  info = "Out-of-range async u8 callback result uses the declared default"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(2L, -6L),
  info = "Out-of-range async narrow callback result is recorded"
)
close_if_valid(cb_u8_bad)

cb_i64_bad <- tcc_callback(
  function(x) 1.5,
  signature = "int64_t (*)(int64_t)"
)
i64_bad_result <- ffi_async_result_checks$call_async_i64(
  cb_i64_bad,
  tcc_callback_ptr(cb_i64_bad),
  0
)
expect_equal(
  i64_bad_result,
  -2147483648,
  info = "Non-integral async i64 callback result uses the declared default"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(3L, -6L),
  info = "Invalid async wide callback result is recorded"
)
close_if_valid(cb_i64_bad)

wide_arg_hits <- 0L
cb_i64_arg <- tcc_callback(
  function(x) {
    wide_arg_hits <<- wide_arg_hits + 1L
    x
  },
  signature = "int64_t (*)(int64_t)"
)
i64_arg_result <- ffi_async_result_checks$call_async_i64_max(
  cb_i64_arg,
  tcc_callback_ptr(cb_i64_arg)
)
expect_equal(
  i64_arg_result,
  -2147483648,
  info = "Inexact async i64 callback argument uses the declared default"
)
expect_equal(
  wide_arg_hits,
  0L,
  info = "Inexact async i64 callback argument is not delivered to R"
)
expect_equal(
  unname(.Call("RC_callback_async_failure_status", PACKAGE = "Rtinycc")),
  c(4L, -7L),
  info = "Inexact async wide callback argument is recorded"
)
close_if_valid(cb_i64_arg)

# Test: non-void async pointer callback preserves non-ownership semantics.
ptr_async_seen_externalptr <- FALSE
ptr_async_seen_unowned <- FALSE
cb_async_ptr <- tcc_callback(
  function(x) {
    ptr_async_seen_externalptr <<- inherits(x, "externalptr")
    ptr_async_seen_unowned <<- !.Call("RC_ptr_is_owned", x, PACKAGE = "Rtinycc")
    x
  },
  signature = "void* (*)(void*)"
)
cb_ptr_async_ptr <- tcc_callback_ptr(cb_async_ptr)

code_async_ptr <- "
#define _Complex

#ifdef _WIN32
#include <windows.h>

struct ptask { void* (*cb)(void*,void*); void* ctx; void* in; volatile void* out; volatile int done; };
static struct ptask g_pt;
static HANDLE g_pth = NULL;

static DWORD WINAPI pworker(LPVOID p) {
  struct ptask* t = (struct ptask*) p;
  t->out = t->cb(t->ctx, t->in);
  t->done = 1;
  return 0;
}
int start_ptr_worker(void* (*cb)(void*,void*), void* ctx, void* x) {
  g_pt.cb = cb; g_pt.ctx = ctx; g_pt.in = x; g_pt.out = NULL; g_pt.done = 0;
  g_pth = CreateThread(NULL, 0, pworker, &g_pt, 0, NULL);
  return g_pth ? 0 : -1;
}
int is_ptr_worker_done(void) { return g_pt.done; }
int join_ptr_worker(void) {
  if (g_pth) { WaitForSingleObject(g_pth, INFINITE); CloseHandle(g_pth); g_pth = NULL; }
  return g_pt.out == g_pt.in;
}
#else
#include <pthread.h>

struct ptask { void* (*cb)(void*,void*); void* ctx; void* in; volatile void* out; volatile int done; };
static struct ptask g_pt;
static pthread_t g_pth;

static void* pworker(void* p) {
  struct ptask* t = (struct ptask*) p;
  t->out = t->cb(t->ctx, t->in);
  t->done = 1;
  return NULL;
}
int start_ptr_worker(void* (*cb)(void*,void*), void* ctx, void* x) {
  g_pt.cb = cb; g_pt.ctx = ctx; g_pt.in = x; g_pt.out = NULL; g_pt.done = 0;
  return pthread_create(&g_pth, NULL, pworker, &g_pt) == 0 ? 0 : -1;
}
int is_ptr_worker_done(void) { return g_pt.done; }
int join_ptr_worker(void) { pthread_join(g_pth, NULL); return g_pt.out == g_pt.in; }
#endif
"

ffi_async_ptr <- tcc_ffi() |>
  tcc_source(code_async_ptr)
if (.Platform$OS.type != "windows") {
  ffi_async_ptr <- tcc_library(ffi_async_ptr, "pthread")
}
ffi_async_ptr <- ffi_async_ptr |>
  tcc_bind(
    start_ptr_worker = list(
      args = list("callback_async:void*(void*)", "ptr", "ptr"),
      returns = "i32"
    ),
    is_ptr_worker_done = list(args = list(), returns = "i32"),
    join_ptr_worker = list(args = list(), returns = "i32")
  ) |>
  tcc_compile()

buf_async_ptr <- tcc_malloc(8)
ffi_async_ptr$start_ptr_worker(cb_async_ptr, cb_ptr_async_ptr, buf_async_ptr)
for (i in seq_len(50)) {
  tcc_callback_async_drain()
  if (ffi_async_ptr$is_ptr_worker_done() != 0L) {
    break
  }
  Sys.sleep(0.01)
}
ptr_same <- ffi_async_ptr$join_ptr_worker()
expect_true(
  ptr_async_seen_externalptr,
  info = "Async pointer callback receives externalptr wrapper"
)
expect_true(
  ptr_async_seen_unowned,
  info = "Async pointer callback receives non-owned pointer wrapper"
)
expect_equal(
  ptr_same,
  1L,
  info = "Async pointer callback returns the same native address"
)
tcc_free(buf_async_ptr)
close_if_valid(cb_async_ptr)

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

# FFI Type System - R/C focused type mapping
# Designed for C FFI (not Zig), optimized for R data types
#
# Key insight: R has native vector types that map directly to C arrays:
# - raw() → uint8_t* (byte buffers)
# - integer() → int32_t* (via INTEGER())
# - numeric() → double* (via REAL())
# - logical() → int* (via LOGICAL())
#
# We support both scalar types (with coercion) and array types (zero-copy)

new_rtinycc_ffi_type <- function(name, ..., family = name) {
  fields <- list(name = name, family = family, ...)
  structure(
    fields,
    class = c(
      paste0("rtinycc_ffi_type_", fields$kind),
      "rtinycc_ffi_type"
    )
  )
}

is_rtinycc_ffi_type <- function(x) {
  inherits(x, "rtinycc_ffi_type")
}

ffi_type_family <- function(x) {
  if (!is_rtinycc_ffi_type(x)) {
    stop("Expected rtinycc_ffi_type object", call. = FALSE)
  }
  x$family
}

FFI_TYPES <- list(
  # Scalar integer types (with range checking)
  i8 = list(c_type = "int8_t", r_type = "integer", size = 1L, kind = "scalar"),
  i16 = list(
    c_type = "int16_t",
    r_type = "integer",
    size = 2L,
    kind = "scalar"
  ),
  i32 = list(
    c_type = "int32_t",
    r_type = "integer",
    size = 4L,
    kind = "scalar"
  ),
  i64 = list(
    c_type = "int64_t",
    r_type = "numeric",
    size = 8L,
    kind = "scalar"
  ),

  # Scalar unsigned integers
  u8 = list(c_type = "uint8_t", r_type = "integer", size = 1L, kind = "scalar"),
  u16 = list(
    c_type = "uint16_t",
    r_type = "integer",
    size = 2L,
    kind = "scalar"
  ),
  u32 = list(
    c_type = "uint32_t",
    r_type = "numeric",
    size = 4L,
    kind = "scalar"
  ),
  u64 = list(
    c_type = "uint64_t",
    r_type = "numeric",
    size = 8L,
    kind = "scalar"
  ),

  # Scalar float types
  f32 = list(c_type = "float", r_type = "numeric", size = 4L, kind = "scalar"),
  f64 = list(c_type = "double", r_type = "numeric", size = 8L, kind = "scalar"),

  # Boolean scalar
  bool = list(c_type = "bool", r_type = "logical", size = 1L, kind = "scalar"),

  # String types
  cstring = list(
    c_type = "char*",
    r_type = "character",
    size = NA_integer_,
    kind = "scalar"
  ),

  # Pointer (opaque externalptr)
  ptr = list(
    c_type = "void*",
    r_type = "externalptr",
    size = NA_integer_,
    kind = "scalar"
  ),

  # Array types - R native vector types (zero-copy)
  # R raw vector → uint8_t* (byte buffer)
  raw = list(
    c_type = "uint8_t*",
    r_type = "raw",
    size = NA_integer_,
    kind = "array",
    r_accessor = "RAW",
    c_element = "uint8_t"
  ),

  # R integer vector → int32_t* (zero-copy via INTEGER())
  integer_array = list(
    c_type = "int32_t*",
    r_type = "integer",
    size = NA_integer_,
    kind = "array",
    r_accessor = "INTEGER",
    c_element = "int32_t"
  ),

  # R numeric vector → double* (zero-copy via REAL())
  numeric_array = list(
    c_type = "double*",
    r_type = "numeric",
    size = NA_integer_,
    kind = "array",
    r_accessor = "REAL",
    c_element = "double"
  ),

  # R logical vector → int* (zero-copy via LOGICAL())
  logical_array = list(
    c_type = "int*",
    r_type = "logical",
    size = NA_integer_,
    kind = "array",
    r_accessor = "LOGICAL",
    c_element = "int"
  ),

  # R character vector -> read-only character-vector elements via STRING_PTR_RO
  character_array = list(
    c_type = "const SEXP*",
    r_type = "character",
    size = NA_integer_,
    kind = "array",
    r_accessor = "STRING_PTR_RO",
    c_element = "SEXP (CHARSXP cell)"
  ),

  # R character vector → const char** (allocated with R_alloc in wrapper)
  cstring_array = list(
    c_type = "const char**",
    r_type = "character",
    size = NA_integer_,
    kind = "array",
    r_accessor = NA_character_,
    c_element = "const char*"
  ),

  # R-specific: pass R object directly (SEXP)
  sexp = list(
    c_type = "SEXP",
    r_type = "ANY",
    size = NA_integer_,
    kind = "scalar"
  ),

  # Void (return only)
  void = list(c_type = "void", r_type = "NULL", size = 0L, kind = "scalar"),

  # Callback type - function pointer passed from R
  callback = list(
    c_type = "void*",
    r_type = "externalptr",
    size = NA_integer_,
    kind = "scalar"
  )
)

# Valid FFI type names
VALID_FFI_TYPES <- names(FFI_TYPES)

# Get array types
ARRAY_TYPES <- names(Filter(function(x) x$kind == "array", FFI_TYPES))
SCALAR_TYPES <- names(Filter(function(x) x$kind == "scalar", FFI_TYPES))

# Validate FFI type
check_ffi_type <- function(type, context = "argument") {
  # Check for enum:type pattern
  if (startsWith(type, "enum:")) {
    # Enum types are always i32 (int)
    return(new_rtinycc_ffi_type(
      name = type,
      family = "enum",
      c_type = "int",
      r_type = "integer",
      size = 4L,
      kind = "scalar"
    ))
  }

  # Check for callback:type pattern (e.g., callback:double(int,int))
  if (startsWith(type, "callback_async:")) {
    return(new_rtinycc_ffi_type(
      name = type,
      family = "callback_async",
      c_type = "void*",
      r_type = "externalptr",
      size = NA_integer_,
      kind = "scalar"
    ))
  }
  if (startsWith(type, "callback")) {
    return(new_rtinycc_ffi_type(
      name = type,
      family = "callback",
      c_type = "void*",
      r_type = "externalptr",
      size = NA_integer_,
      kind = "scalar"
    ))
  }

  if (!type %in% VALID_FFI_TYPES) {
    stop(
      "Invalid FFI type '",
      type,
      "' for ",
      context,
      "\nValid types: ",
      paste(VALID_FFI_TYPES, collapse = ", "),
      call. = FALSE
    )
  }
  do.call(new_rtinycc_ffi_type, c(list(name = type), FFI_TYPES[[type]]))
}

# Check if type is an array type
is_array_type <- function(type) {
  type %in% ARRAY_TYPES
}

# Check if type is a scalar type
is_scalar_type <- function(type) {
  type %in% SCALAR_TYPES
}

# Test FFI Type System
# Real tests for the Bun-style FFI type mappings

library(Rtinycc)

# Test 1: FFI type validation works and returns a classed type object
type_info <- Rtinycc:::check_ffi_type("i32", "test")
expect_true(Rtinycc:::is_rtinycc_ffi_type(type_info))
expect_equal(type_info$name, "i32")
expect_equal(Rtinycc:::ffi_type_family(type_info), "i32")
expect_equal(type_info$c_type, "int32_t")
expect_equal(type_info$r_type, "integer")
expect_equal(type_info$kind, "scalar")

# Test 1b: scalar ffi types record the intended R-side carrier types
expect_equal(Rtinycc:::check_ffi_type("i8", "test")$r_type, "integer")
expect_equal(Rtinycc:::check_ffi_type("i16", "test")$r_type, "integer")
expect_equal(Rtinycc:::check_ffi_type("i32", "test")$r_type, "integer")
expect_equal(Rtinycc:::check_ffi_type("u8", "test")$r_type, "integer")
expect_equal(Rtinycc:::check_ffi_type("u16", "test")$r_type, "integer")
expect_equal(Rtinycc:::check_ffi_type("u32", "test")$r_type, "numeric")
expect_equal(Rtinycc:::check_ffi_type("i64", "test")$r_type, "numeric")
expect_equal(Rtinycc:::check_ffi_type("u64", "test")$r_type, "numeric")
expect_equal(Rtinycc:::check_ffi_type("f32", "test")$r_type, "numeric")
expect_equal(Rtinycc:::check_ffi_type("f64", "test")$r_type, "numeric")
expect_equal(Rtinycc:::check_ffi_type("bool", "test")$r_type, "logical")
expect_equal(Rtinycc:::check_ffi_type("cstring", "test")$r_type, "character")

# Test 2: Array types have proper accessors
type_info <- Rtinycc:::check_ffi_type("raw", "test")
expect_equal(type_info$r_accessor, "RAW")
expect_equal(type_info$kind, "array")

type_info <- Rtinycc:::check_ffi_type("integer_array", "test")
expect_equal(type_info$r_accessor, "INTEGER")
expect_equal(type_info$c_type, "int32_t*")

type_info <- Rtinycc:::check_ffi_type("character_array", "test")
expect_equal(type_info$r_accessor, "STRING_PTR_RO")
expect_equal(type_info$c_type, "const SEXP*")
expect_equal(type_info$c_element, "SEXP (CHARSXP cell)")

# Test 3: Invalid types throw errors
expect_error(Rtinycc:::check_ffi_type("invalid_type", "test"))

# Test 4: All expected types exist
expected_types <- c(
  "i8",
  "i16",
  "i32",
  "i64",
  "u8",
  "u16",
  "u32",
  "u64",
  "f32",
  "f64",
  "bool",
  "cstring",
  "ptr",
  "raw",
  "integer_array",
  "numeric_array",
  "logical_array",
  "sexp",
  "void"
)
for (t in expected_types) {
  info <- Rtinycc:::check_ffi_type(t, "test")
  expect_true(!is.null(info))
}

# Test 5: is_array_type and is_scalar_type work
expect_true(Rtinycc:::is_array_type("raw"))
expect_true(Rtinycc:::is_array_type("integer_array"))
expect_false(Rtinycc:::is_array_type("i32"))
expect_true(Rtinycc:::is_scalar_type("i32"))
expect_false(Rtinycc:::is_scalar_type("raw"))

# Test 6: special type families are classed, not bare stringly metadata
expect_equal(
  Rtinycc:::ffi_type_family(Rtinycc:::check_ffi_type("enum:color", "test")),
  "enum"
)
expect_equal(
  Rtinycc:::ffi_type_family(Rtinycc:::check_ffi_type(
    "callback:double(int)",
    "test"
  )),
  "callback"
)
expect_equal(
  Rtinycc:::ffi_type_family(Rtinycc:::check_ffi_type(
    "callback_async:void*(void*)",
    "test"
  )),
  "callback_async"
)

# Test 7: all declared C type aliases map to the expected canonical FFI types

mapping_cases <- list(
  void = c("void"),
  bool = c("bool", "_Bool"),
  sexp = c("SEXP", "sexp"),
  i32 = c("int", "signed", "int32_t", "__int32", "signed int"),
  i16 = c("short", "short int", "signed short", "signed short int", "int16_t"),
  i8 = c("char", "signed char", "int8_t"),
  i64 = c(
    "long long",
    "long long int",
    "signed long long",
    "signed long long int",
    "int64_t",
    "__int64",
    "off_t"
  ),
  u32 = c("unsigned int", "unsigned", "uint32_t", "unsigned __int32"),
  u16 = c("unsigned short", "unsigned short int", "uint16_t"),
  u8 = c("unsigned char", "uint8_t"),
  u64 = c(
    "unsigned long long",
    "unsigned long long int",
    "uint64_t",
    "unsigned __int64"
  ),
  f64 = c("double"),
  f32 = c("float")
)

for (ffi_type in names(mapping_cases)) {
  for (c_type in mapping_cases[[ffi_type]]) {
    expect_equal(
      tcc_map_c_type_to_ffi(c_type),
      ffi_type,
      info = paste("C type", shQuote(c_type), "maps to", ffi_type)
    )
  }
}

long_type <- if (.Machine$sizeof.long == 8L) "i64" else "i32"
unsigned_long_type <- if (.Machine$sizeof.long == 8L) "u64" else "u32"
pointer_signed_type <- if (.Machine$sizeof.pointer == 8L) "i64" else "i32"
pointer_unsigned_type <- if (.Machine$sizeof.pointer == 8L) "u64" else "u32"

for (c_type in c("long", "long int", "signed long", "signed long int")) {
  expect_equal(
    tcc_map_c_type_to_ffi(c_type),
    long_type,
    info = paste("C type", shQuote(c_type), "follows sizeof(long)")
  )
}
for (c_type in c("unsigned long", "unsigned long int")) {
  expect_equal(
    tcc_map_c_type_to_ffi(c_type),
    unsigned_long_type,
    info = paste("C type", shQuote(c_type), "follows sizeof(unsigned long)")
  )
}
for (c_type in c("intptr_t", "ptrdiff_t", "ssize_t")) {
  expect_equal(
    tcc_map_c_type_to_ffi(c_type),
    pointer_signed_type,
    info = paste("C type", shQuote(c_type), "follows pointer width")
  )
}
for (c_type in c("size_t", "uintptr_t")) {
  expect_equal(
    tcc_map_c_type_to_ffi(c_type),
    pointer_unsigned_type,
    info = paste("C type", shQuote(c_type), "follows pointer width")
  )
}

# Test 8: pointer spellings stay pointer-like under conservative mapping
#
# Regression coverage for a bug where trailing-identifier stripping could drop
# the `*` from declarations such as `const char *`, causing them to collapse to
# the scalar `i8` mapping instead of the intended conservative pointer mapping.
pointer_cases <- c(
  "const char *",
  "char *",
  "const char *name",
  "char *buf",
  "void **out",
  "const void *ctx",
  "volatile char *cursor",
  "char **argv"
)

for (c_type in pointer_cases) {
  expect_equal(
    tcc_map_c_type_to_ffi(c_type),
    "ptr",
    info = paste("pointer type", shQuote(c_type), "stays on conservative ptr mapping")
  )
}

# Test 9: whitespace and qualifiers are normalized before mapping
expect_equal(tcc_map_c_type_to_ffi(" const   unsigned   long  long   int "), "u64")
expect_equal(tcc_map_c_type_to_ffi("const volatile restrict double"), "f64")
expect_equal(tcc_map_c_type_to_ffi("const volatile char * name"), "ptr")
expect_error(
  tcc_map_c_type_to_ffi("long double"),
  "no safe Rtinycc scalar ABI mapping",
  info = "long double is rejected instead of using the double calling convention"
)
expect_error(
  tcc_map_c_type_to_ffi("application_word"),
  "unsupported C scalar type",
  info = "unknown by-value scalar types require an explicit mapper"
)

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

#' Pointer and Buffer Utilities for FFI
#'
#' Generic helper functions for common FFI operations inspired by Bun's FFI.
#' These utilities handle pointer creation, buffer management, and memory operations
#' that are commonly needed when working with external libraries.
#'
#' @return No return value. This is a documentation topic that groups the
#'   pointer and buffer utility functions described below.
#' @name tcc_ptr_utils
NULL

#' Create a NULL pointer
#'
#' Creates a NULL pointer equivalent for use with FFI functions that expect
#' optional pointer arguments.
#'
#' @return An external pointer with NULL address
#' @export
tcc_null_ptr <- function() {
  .Call(RC_null_pointer)
}

#' Create a C-style string pointer
#'
#' Convert R character strings to C-style null-terminated string pointers.
#' This handles UTF-8 encoding and null termination automatically.
#'
#' @param str Character string
#' @return An external pointer tagged `"rtinycc_owned"` pointing to a
#'   malloc'd copy of the string. Freed on garbage collection or via
#'   [tcc_free()].
#' @export
tcc_cstring <- function(str) {
  .Call(RC_create_cstring, as.character(str))
}

#' Read C-style string from pointer
#'
#' Convert a C-style null-terminated string pointer back to R character string.
#' Handles UTF-8 decoding automatically.
#'
#' @param ptr External pointer to C string
#' @param max_bytes Optional maximum number of bytes to read (fixed-length read).
#' @param null_action Behavior when ptr is NULL: one of "na", "empty", "error".
#'   Only effective when `max_bytes` is provided; without it a NULL pointer
#'   returns `""`.
#' @return Character string, or `NA`/`""` for NULL pointers depending on
#'   `null_action`.
#' @export
tcc_read_cstring <- function(
  ptr,
  max_bytes = NULL,
  null_action = c("na", "empty", "error")
) {
  null_action <- match.arg(null_action)

  if (is.null(max_bytes)) {
    out <- .Call(RC_read_cstring, ptr)
  } else {
    out <- .Call(RC_read_cstring_n, ptr, as.integer(max_bytes))
  }

  if (is.na(out) && null_action == "empty") {
    return("")
  }
  if (is.na(out) && null_action == "error") {
    stop("C string pointer is NULL", call. = FALSE)
  }
  out
}

#' Read raw bytes from a pointer
#'
#' Read a fixed number of bytes from an external pointer into a raw vector.
#'
#' @param ptr External pointer
#' @param nbytes Number of bytes to read
#' @return Raw vector
#' @export
tcc_read_bytes <- function(ptr, nbytes) {
  .Call(RC_read_bytes, ptr, as.integer(nbytes))
}

#' Write raw bytes to a pointer
#'
#' Write a raw vector into memory pointed to by an external pointer.
#'
#' @param ptr External pointer
#' @param raw Raw vector to write
#' @return `NULL`.
#' @export
tcc_write_bytes <- function(ptr, raw) {
  .Call(RC_write_bytes, ptr, raw)
}

#' Dereference a pointer-to-pointer
#'
#' Treats `ptr_ref` as a pointer to a pointer and returns the pointed address
#' as an external pointer. This is useful for fields like `void**` or `T**`.
#'
#' @param ptr_ref External pointer to a pointer value (e.g., address of a field).
#' @return An external pointer tagged `"rtinycc_borrowed"`. Not owned by
#'   Rtinycc and never freed on garbage collection. Do not pass to
#'   [tcc_free()].
#' @export
tcc_data_ptr <- function(ptr_ref) {
  .Call(RC_data_ptr, ptr_ref)
}

#' Set a pointer-to-pointer value
#'
#' Assigns the address in `ptr_value` to the location pointed to by `ptr_ref`.
#'
#' @param ptr_ref External pointer to a pointer value (e.g., address of a field).
#' @param ptr_value External pointer to store.
#' @return The updated pointer reference (invisibly).
#' @export
tcc_ptr_set <- function(ptr_ref, ptr_value) {
  .Call(RC_ptr_set, ptr_ref, ptr_value)
}

#' Free the pointed memory and set to NULL
#'
#' Frees the memory pointed to by `ptr_ref` and sets the pointer to NULL.
#' Use this only when the pointed memory is not already owned by another
#' external pointer with its own finalizer.
#'
#' @param ptr_ref External pointer to a pointer value.
#' @return The updated pointer reference (invisibly).
#' @export
tcc_ptr_free_set_null <- function(ptr_ref) {
  .Call(RC_ptr_free_set_null, ptr_ref)
}

#' Read unsigned 8-bit values from a pointer
#'
#' @param ptr External pointer
#' @param n Number of values to read (legacy vectorised interface).
#'   If provided, reads `n` consecutive u8 values starting at byte 0.
#' @param offset Byte offset from `ptr` (scalar interface). Ignored when
#'   `n` is supplied.
#' @return Integer scalar (offset form) or integer vector (n form).
#' @export
tcc_read_u8 <- function(ptr, n = NULL, offset = 0L) {
  if (!is.null(n)) {
    raw <- tcc_read_bytes(ptr, n)
    return(readBin(raw, integer(), n = length(raw), size = 1, signed = FALSE))
  }
  .Call(RC_read_u8_typed, ptr, offset)
}

#' Read signed 8-bit integer
#'
#' @inheritParams tcc_read_u8
#' @return Integer scalar
#' @export
tcc_read_i8 <- function(ptr, offset = 0L) {
  .Call(RC_read_i8, ptr, offset)
}

#' Read signed 16-bit integer
#'
#' @inheritParams tcc_read_u8
#' @return Integer scalar
#' @export
tcc_read_i16 <- function(ptr, offset = 0L) {
  .Call(RC_read_i16, ptr, offset)
}

#' Read unsigned 16-bit integer
#'
#' @inheritParams tcc_read_u8
#' @return Integer scalar
#' @export
tcc_read_u16 <- function(ptr, offset = 0L) {
  .Call(RC_read_u16, ptr, offset)
}

#' Read signed 32-bit integers from a pointer
#'
#' @param ptr External pointer
#' @param n Number of values to read (legacy vectorised interface).
#'   If provided, reads `n` consecutive i32 values starting at byte 0.
#' @param offset Byte offset from `ptr` (scalar interface). Ignored when
#'   `n` is supplied.
#' @return Integer scalar (offset form) or integer vector (n form).
#' @export
tcc_read_i32 <- function(ptr, n = NULL, offset = 0L) {
  if (!is.null(n)) {
    raw <- tcc_read_bytes(ptr, n * 4L)
    return(readBin(raw, integer(), n = n, size = 4, signed = TRUE))
  }
  .Call(RC_read_i32_typed, ptr, offset)
}

#' Read unsigned 32-bit integer
#'
#' @inheritParams tcc_read_u8
#' @return Numeric scalar (double, exact up to 2^32-1).
#' @export
tcc_read_u32 <- function(ptr, offset = 0L) {
  .Call(RC_read_u32, ptr, offset)
}

#' Read signed 64-bit integer
#'
#' @inheritParams tcc_read_u8
#' @return Numeric scalar (double, exact up to 2^53).
#' @export
tcc_read_i64 <- function(ptr, offset = 0L) {
  .Call(RC_read_i64, ptr, offset)
}

#' Read unsigned 64-bit integer
#'
#' @inheritParams tcc_read_u8
#' @return Numeric scalar (double, exact up to 2^53).
#' @export
tcc_read_u64 <- function(ptr, offset = 0L) {
  .Call(RC_read_u64, ptr, offset)
}

#' Read 32-bit float
#'
#' @inheritParams tcc_read_u8
#' @return Numeric scalar (promoted to double).
#' @export
tcc_read_f32 <- function(ptr, offset = 0L) {
  .Call(RC_read_f32, ptr, offset)
}

#' Read 64-bit doubles from a pointer
#'
#' @param ptr External pointer
#' @param n Number of values to read (legacy vectorised interface).
#'   If provided, reads `n` consecutive f64 values starting at byte 0.
#' @param offset Byte offset from `ptr` (scalar interface). Ignored when
#'   `n` is supplied.
#' @return Numeric scalar (offset form) or numeric vector (n form).
#' @export
tcc_read_f64 <- function(ptr, n = NULL, offset = 0L) {
  if (!is.null(n)) {
    raw <- tcc_read_bytes(ptr, n * 8L)
    return(readBin(raw, numeric(), n = n, size = 8, endian = .Platform$endian))
  }
  .Call(RC_read_f64_typed, ptr, offset)
}

#' Read a pointer at byte offset
#'
#' Dereferences a `void*` at the given byte offset from `ptr`.
#' Equivalent to `*(void**)(ptr + offset)`. The returned pointer
#' is tagged `"rtinycc_borrowed"` and will not be freed by the
#' garbage collector.
#'
#' @inheritParams tcc_read_u8
#' @return External pointer
#' @export
tcc_read_ptr <- function(ptr, offset = 0L) {
  .Call(RC_read_ptr, ptr, offset)
}

#' Write a signed 8-bit integer
#' @param ptr External pointer
#' @param offset Byte offset
#' @param value Integer value to write
#' @return `NULL` (invisibly).
#' @export
tcc_write_i8 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_i8, ptr, offset, value))
}

#' Write an unsigned 8-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   unsigned 8-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_u8 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_u8, ptr, offset, value))
}

#' Write a signed 16-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   signed 16-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_i16 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_i16, ptr, offset, value))
}

#' Write an unsigned 16-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   unsigned 16-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_u16 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_u16, ptr, offset, value))
}

#' Write a signed 32-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   signed 32-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_i32 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_i32, ptr, offset, value))
}

#' Write an unsigned 32-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   unsigned 32-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_u32 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_u32, ptr, offset, value))
}

#' Write a signed 64-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   signed 64-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_i64 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_i64, ptr, offset, value))
}

#' Write an unsigned 64-bit integer
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   unsigned 64-bit value into native memory at `ptr + offset`.
#' @export
tcc_write_u64 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_u64, ptr, offset, value))
}

#' Write a 32-bit float
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   32-bit floating-point value into native memory at `ptr + offset`.
#' @export
tcc_write_f32 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_f32, ptr, offset, value))
}

#' Write a 64-bit double
#' @inheritParams tcc_write_i8
#' @return `NULL` (invisibly). Called for its side effect of writing one
#'   64-bit floating-point value into native memory at `ptr + offset`.
#' @export
tcc_write_f64 <- function(ptr, offset, value) {
  invisible(.Call(RC_write_f64, ptr, offset, value))
}

#' Write a pointer at byte offset
#' @param ptr External pointer (destination buffer)
#' @param offset Byte offset
#' @param value External pointer to write
#' @return `NULL` (invisibly).
#' @export
tcc_write_ptr <- function(ptr, offset, value) {
  invisible(.Call(RC_write_ptr, ptr, offset, value))
}

#' Allocate memory buffer
#'
#' Allocate a memory buffer of specified size, equivalent to C malloc.
#' Returns an external pointer that can be passed to FFI functions.
#'
#' @param size Number of bytes to allocate
#' @return An external pointer tagged `"rtinycc_owned"` with an R finalizer.
#'   Freed on garbage collection or explicitly via [tcc_free()].
#' @export
tcc_malloc <- function(size) {
  .Call(RC_malloc, as.integer(size))
}

#' Free owned memory
#'
#' Free memory whose external pointer is tagged `"rtinycc_owned"` (e.g.
#' from [tcc_malloc()] or [tcc_cstring()]). Errors on struct pointers
#' (use the generated `struct_<name>_free()`) or borrowed pointers from
#' [tcc_data_ptr()].
#'
#' @param ptr External pointer to free
#' @return `NULL`.
#' @export
tcc_free <- function(ptr) {
  .Call(RC_free, ptr)
}

#' Get pointer address as integer
#'
#' Get the numeric address of an external pointer, useful for debugging
#' and when APIs require pointer addresses as integers. Optional hex mode available.
#'
#' @param ptr External pointer
#' @param hex Whether to display in hexadecimal (default: FALSE)
#' @return Character representation of pointer address (hex if requested, decimal otherwise)
#' @export
tcc_ptr_addr <- function(ptr, hex = FALSE) {
  if (hex) {
    .Call(RC_get_external_ptr_hex, ptr)
  } else {
    addr <- get_external_ptr_addr(ptr)
    as.character(addr)
  }
}

#' Check whether an external pointer is NULL
#'
#' Returns TRUE if the external pointer address is NULL, FALSE otherwise.
#'
#' @param ptr External pointer
#' @return Logical scalar
#' @export
tcc_ptr_is_null <- function(ptr) {
  stopifnot(inherits(ptr, "externalptr"))
  !.Call(RC_libtcc_ptr_valid, ptr)
}

#' Check for the `"rtinycc_owned"` tag
#'
#' Returns `TRUE` only for pointers created by [tcc_malloc()] or
#' [tcc_cstring()]. Struct pointers (tagged `"struct_<name>"`) and
#' borrowed pointers return `FALSE`.
#'
#' @param ptr External pointer
#' @return Logical scalar
#' @export
tcc_ptr_is_owned <- function(ptr) {
  stopifnot(inherits(ptr, "externalptr"))
  .Call(RC_ptr_is_owned, ptr)
}

# Test pointer utilities
library(tinytest)
library(Rtinycc)

# Test null pointer creation
null_ptr <- tcc_null_ptr()
addr <- tcc_ptr_addr(null_ptr)
expect_equal(addr, "0")
expect_true(is.numeric(addr) || is.character(addr))

# Test C string creation and reading
test_str <- "Hello, World!"
str_ptr <- tcc_cstring(test_str)
expect_true(inherits(str_ptr, "externalptr"))
expect_true(tcc_ptr_addr(str_ptr) > 0)

# Read back the string
read_str <- tcc_read_cstring(str_ptr)
expect_equal(read_str, test_str)

# Fixed-length read (truncate)
read_str_n <- tcc_read_cstring(str_ptr, max_bytes = 5)
expect_equal(read_str_n, "Hello")

# Test memory allocation
mem_ptr <- tcc_malloc(1024)
expect_true(inherits(mem_ptr, "externalptr"))
expect_true(tcc_ptr_addr(mem_ptr) > 0)

# Write and read raw bytes
bytes <- as.raw(0:15)
tcc_write_bytes(mem_ptr, bytes)
read_bytes <- tcc_read_bytes(mem_ptr, length(bytes))
expect_equal(read_bytes, bytes)

# Typed reads
expect_equal(tcc_read_u8(mem_ptr, 4), c(0L, 1L, 2L, 3L))

# Test memory free
result <- tcc_free(mem_ptr)
expect_true(is.null(result))

# Test pointer-to-pointer helpers
ptr_size <- if (!is.null(.Machine$sizeof.pointer)) {
  .Machine$sizeof.pointer
} else {
  8L
}
ptr_ref <- tcc_malloc(ptr_size)
target <- tcc_malloc(16)

tcc_ptr_set(ptr_ref, target)
data <- tcc_data_ptr(ptr_ref)
expect_equal(tcc_ptr_addr(data), tcc_ptr_addr(target))
expect_true(tcc_ptr_is_owned(target))

tcc_ptr_set(ptr_ref, tcc_null_ptr())
expect_true(tcc_ptr_is_null(tcc_data_ptr(ptr_ref)))

expect_true(tcc_ptr_is_null(tcc_data_ptr(ptr_ref)))

# Only use ptr_free_set_null when the pointed memory is unmanaged
tcc_ptr_free_set_null(ptr_ref)
expect_true(tcc_ptr_is_null(tcc_data_ptr(ptr_ref)))

tcc_free(ptr_ref)

# Test multiple string creation
strings <- c("first", "second", "third")
for (i in seq_along(strings)) {
  ptr <- tcc_cstring(strings[i])
  read_back <- tcc_read_cstring(ptr)
  expect_equal(read_back, strings[i])
}

# owned flag is reserved but should not error
owned_obj <- Rtinycc:::tcc_cstring_object(
  tcc_cstring("owned"),
  clone = TRUE,
  owned = TRUE
)
expect_true(inherits(owned_obj, "tcc_cstring"))
expect_equal(as.character(owned_obj), "owned")

# Test typed read/write helpers (Bun-style)
library(tinytest)
library(Rtinycc)

buf <- tcc_malloc(128)

# --- i8 / u8 ---------------------------------------------------------------
tcc_write_i8(buf, 0L, -42L)
expect_equal(tcc_read_i8(buf, 0L), -42L, info = "i8 round-trip")

tcc_write_u8(buf, 1L, 200L)
expect_equal(tcc_read_u8(buf, offset = 1L), 200L, info = "u8 round-trip")

# --- i16 / u16 -------------------------------------------------------------
tcc_write_i16(buf, 2L, -1234L)
expect_equal(tcc_read_i16(buf, 2L), -1234L, info = "i16 round-trip")

tcc_write_u16(buf, 4L, 50000L)
expect_equal(tcc_read_u16(buf, 4L), 50000L, info = "u16 round-trip")

# --- i32 / u32 -------------------------------------------------------------
tcc_write_i32(buf, 8L, -100000L)
expect_equal(tcc_read_i32(buf, offset = 8L), -100000L, info = "i32 round-trip")

tcc_write_u32(buf, 12L, 3000000000)
expect_equal(tcc_read_u32(buf, 12L), 3000000000, info = "u32 round-trip")

# --- i64 / u64 (stored as double, exact up to 2^53) ------------------------
tcc_write_i64(buf, 16L, -123456789)
expect_equal(tcc_read_i64(buf, 16L), -123456789, info = "i64 round-trip")

tcc_write_u64(buf, 24L, 2^52)
expect_equal(tcc_read_u64(buf, 24L), 2^52, info = "u64 round-trip")

expect_error(
  tcc_write_i64(buf, 16L, 2^53 + 2),
  info = "i64 writer rejects lossy numeric input"
)
expect_error(
  tcc_write_u64(buf, 24L, -1),
  info = "u64 writer rejects negative input"
)
expect_error(
  tcc_write_u32(buf, 12L, NaN),
  info = "u32 writer rejects NA/NaN-like numeric input"
)

# --- f32 / f64 -------------------------------------------------------------
tcc_write_f32(buf, 32L, 3.14)
expect_true(
  abs(tcc_read_f32(buf, 32L) - 3.14) < 1e-5,
  info = "f32 round-trip (single precision)"
)

tcc_write_f64(buf, 40L, pi)
expect_equal(tcc_read_f64(buf, offset = 40L), pi, info = "f64 round-trip")

# --- ptr --------------------------------------------------------------------
inner <- tcc_malloc(8)
tcc_write_ptr(buf, 48L, inner)
read_back <- tcc_read_ptr(buf, 48L)
expect_equal(
  tcc_ptr_addr(read_back),
  tcc_ptr_addr(inner),
  info = "ptr round-trip"
)
expect_false(
  tcc_ptr_is_owned(read_back),
  info = "read_ptr returns unowned wrapper"
)
expect_error(
  tcc_free(read_back),
  info = "read_ptr wrapper is not explicitly freeable"
)
tcc_free(inner)

# --- legacy vectorised interface (backward compat) --------------------------
tcc_write_bytes(buf, as.raw(c(10, 20, 30, 40)))
expect_equal(
  tcc_read_u8(buf, n = 4),
  c(10L, 20L, 30L, 40L),
  info = "u8 legacy n interface"
)

tcc_write_i32(buf, 0L, 42L)
tcc_write_i32(buf, 4L, 99L)
expect_equal(
  tcc_read_i32(buf, n = 2),
  c(42L, 99L),
  info = "i32 legacy n interface"
)

tcc_write_f64(buf, 0L, 1.5)
tcc_write_f64(buf, 8L, 2.5)
expect_equal(
  tcc_read_f64(buf, n = 2),
  c(1.5, 2.5),
  info = "f64 legacy n interface"
)

# --- default offset = 0 ----------------------------------------------------
tcc_write_u8(buf, 0L, 77L)
expect_equal(tcc_read_u8(buf), 77L, info = "u8 default offset=0")

expect_error(
  tcc_read_u8(buf, offset = -1),
  info = "typed reads reject negative offsets"
)
expect_error(
  tcc_write_i32(buf, offset = 0.5, value = 1L),
  info = "typed writes reject fractional offsets"
)

tcc_free(buf)

# Rtinycc - TinyCC for R
# Copyright (C) 2025-2026 Sounkou Mahamane Toure
# SPDX-License-Identifier: GPL-3.0-or-later

#' Get the address of an external pointer
#'
#' Extract the memory address from an external pointer as a numeric value.
#' This is primarily useful for debugging and inspection purposes.
#'
#' @param ptr An external pointer object (e.g., from `tcc_get_symbol()`).
#' @return The memory address as a numeric value.
#' @export
get_external_ptr_addr <- function(ptr) {
  .Call(RC_get_external_ptr_addr, ptr)
}

#' TinyCC paths
#'
#' Helpers to locate the bundled tinycc installation after the package is installed.
#' @return A character scalar path.
#' @export
tcc_prefix <- function() {
  system.file("tinycc", package = "Rtinycc")
}

#' @rdname tcc_prefix
#' @export
tcc_lib_path <- function() file.path(tcc_prefix(), "lib")

#' @rdname tcc_prefix
#' @export
tcc_lib_paths <- function() {
  prefix <- tcc_prefix()
  # lib/tcc must come first: the C layer passes the first entry to

  # tcc_set_lib_path(), and TCC resolves its own headers as
  # {lib_path}/include (e.g. stdbool.h, stddef.h).
  paths <- c(file.path(prefix, "lib", "tcc"), file.path(prefix, "lib"))
  normalizePath(paths[file.exists(paths)], winslash = "/", mustWork = FALSE)
}

#' @rdname tcc_prefix
#' @export
tcc_include_path <- function() file.path(tcc_prefix(), "include")

#' @rdname tcc_prefix
#' @export
tcc_bin_path <- function() file.path(tcc_prefix(), "bin")

#' @rdname tcc_prefix
#' @export
tcc_cli <- function() {
  exe <- if (.Platform$OS.type == "windows") "tcc.exe" else "tcc"
  candidates <- c(file.path(tcc_bin_path(), exe), file.path(tcc_prefix(), exe))
  existing <- candidates[file.exists(candidates)]
  if (length(existing)) {
    return(existing[[1L]])
  }
  candidates[[1L]]
}

#' Locate the TinyCC executable
#'
#' Returns the platform-specific `tcc` binary path (or `tcc.exe` on Windows), preferring the bundled installation.
#' @return A character scalar path.
#' @export
tcc_path <- function() {
  normalizePath(tcc_cli(), winslash = "/", mustWork = FALSE)
}

#' TinyCC include search paths
#'
#' Returns the include directories used by the bundled TinyCC (top-level include and lib/tcc/include).
#' @return A character vector of include directories.
#' @export
tcc_include_paths <- function() {
  prefix <- tcc_prefix()
  paths <- c(
    file.path(prefix, "include"),
    file.path(prefix, "lib", "tcc", "include")
  )
  if (.Platform$OS.type == "windows") {
    paths <- c(paths, file.path(prefix, "include", "winapi"))
  }
  normalizePath(paths[file.exists(paths)], winslash = "/", mustWork = FALSE)
}

#' @rdname tcc_include_paths
#' @export
tcc_sysinclude_paths <- tcc_include_paths

tcc_parse_runtime_lib_path <- function(x) {
  if (is.null(x) || !length(x)) {
    return(NA_character_)
  }
  s <- as.character(x[[1]])
  if (!nzchar(s) || is.na(s)) {
    return(NA_character_)
  }
  s <- trimws(sub(";.*$", "", s))
  if (!nzchar(s) || identical(s, "NA")) {
    return(NA_character_)
  }
  normalizePath(s, winslash = "/", mustWork = FALSE)
}

#' Report active BLAS/LAPACK runtime information from R
#'
#' Returns the BLAS/LAPACK runtime details as reported by R itself, plus
#' convenience flags indicating whether `Rblas` and `Rlapack` appear available
#' in loaded DLLs/shared objects.
#'
#' @return A named list with fields:
#'   `blas_path`, `lapack_path`, `has_rblas`, `has_rlapack`, `loaded_dlls`.
#' @export
blas_lapack_info <- function() {
  si <- utils::sessionInfo()
  loaded <- getLoadedDLLs()
  loaded_names <- names(loaded)
  loaded_paths <- vapply(
    loaded,
    function(x) {
      p <- x[["path"]]
      if (is.null(p) || !length(p)) "" else as.character(p[[1]])
    },
    character(1)
  )

  blas_path <- tcc_parse_runtime_lib_path(si$BLAS)
  lapack_path <- tryCatch(La_library(), error = function(e) NA_character_)
  lapack_path <- tcc_parse_runtime_lib_path(lapack_path)
  if (is.na(lapack_path) || !nzchar(lapack_path)) {
    lapack_path <- tcc_parse_runtime_lib_path(si$LAPACK)
  }

  path_has_name <- function(path, name) {
    if (
      !is.character(path) || length(path) != 1 || is.na(path) || !nzchar(path)
    ) {
      return(FALSE)
    }
    grepl(name, basename(path), ignore.case = TRUE)
  }
  loaded_has_name <- function(name) {
    any(grepl(name, loaded_names, ignore.case = TRUE)) ||
      any(grepl(name, basename(loaded_paths), ignore.case = TRUE))
  }

  has_rblas <- loaded_has_name("Rblas") || path_has_name(blas_path, "Rblas")
  has_rlapack <- loaded_has_name("Rlapack") ||
    path_has_name(lapack_path, "Rlapack")

  list(
    blas_path = blas_path,
    lapack_path = lapack_path,
    has_rblas = isTRUE(has_rblas),
    has_rlapack = isTRUE(has_rlapack),
    loaded_dlls = loaded_names
  )
}

tcc_output_type <- function(output) {
  output <- match.arg(output, c("memory", "obj", "dll", "exe", "preprocess"))
  tcc_output_type_rule(output)
}

check_cli_exists <- function() {
  path <- tcc_path()
  if (!nzchar(path) || !file.exists(path)) {
    stop("tinycc CLI not found at ", path, call. = FALSE)
  }
  path
}

#' Create a libtcc state
#'
#' Initialize a libtcc compilation state, optionally pointing at the bundled include/lib paths.
#' @param output Output type: one of "memory", "obj", "dll", "exe", "preprocess".
#' @param include_path Path(s) to headers; defaults to the bundled include dirs.
#' @param lib_path Path(s) to libraries; defaults to the bundled lib dirs (lib and lib/tcc).
#' @return An external pointer of class `tcc_state`.
#' @export
tcc_state <- function(
  output = c("memory", "obj", "dll", "exe", "preprocess"),
  include_path = tcc_include_paths(),
  lib_path = tcc_lib_paths()
) {
  .Call(
    RC_libtcc_state_new,
    normalizePath(lib_path, winslash = "/", mustWork = FALSE),
    normalizePath(include_path, winslash = "/", mustWork = FALSE),
    tcc_output_type(output)
  )
}

#' Add a source file to a libtcc state
#' @param state A `tcc_state`.
#' @param path Path to a C source file.
#' @return Integer status code (0 = success).
#' @export
tcc_add_file <- function(state, path) {
  .Call(
    RC_libtcc_add_file,
    state,
    normalizePath(path, winslash = "/", mustWork = TRUE)
  )
}

#' Add an include path to a libtcc state
#' @param state A `tcc_state`.
#' @param path Path to include directory.
#' @return Integer status code (0 = success).
#' @export
tcc_add_include_path <- function(state, path) {
  .Call(
    RC_libtcc_add_include_path,
    state,
    normalizePath(path, winslash = "/", mustWork = FALSE)
  )
}

#' Add a system include path to a libtcc state
#' @param state A `tcc_state`.
#' @param path Path to system include directory.
#' @return Integer status code (0 = success).
#' @export
tcc_add_sysinclude_path <- function(state, path) {
  .Call(
    RC_libtcc_add_sysinclude_path,
    state,
    normalizePath(path, winslash = "/", mustWork = FALSE)
  )
}

#' Add a library path to a libtcc state
#' @param state A `tcc_state`.
#' @param path Path to library directory.
#' @return Integer status code (0 = success).
#' @export
tcc_add_library_path <- function(state, path) {
  .Call(
    RC_libtcc_add_library_path,
    state,
    normalizePath(path, winslash = "/", mustWork = FALSE)
  )
}

#' Add a library to a libtcc state
#' @param state A `tcc_state`.
#' @param library Library name (e.g., "m" for libm, "R" for libR).
#' @return Integer status code (0 = success).
#' @export
tcc_add_library <- function(state, library) {
  .Call(RC_libtcc_add_library, state, library)
}

#' Apply raw TinyCC options to a libtcc state
#'
#' Passes options directly to `tcc_set_options()` for the given state.
#'
#' @param state A `tcc_state`.
#' @param options Character scalar of options (for example `"-O2 -Wall"`).
#' @return Integer status code (`0` on success; negative on parse error).
#' @export
tcc_set_options <- function(state, options) {
  if (
    !is.character(options) || length(options) != 1 || !nzchar(trimws(options))
  ) {
    stop("`options` must be a non-empty character scalar", call. = FALSE)
  }
  .Call(RC_libtcc_set_options, state, options)
}

#' Compile C code from a character string
#' @param state A `tcc_state`.
#' @param code C source code string.
#' @return Integer status code (0 = success).
#' @export
tcc_compile_string <- function(state, code) {
  .Call(RC_libtcc_compile_string, state, code)
}

#' Relocate compiled code
#' @param state A `tcc_state`.
#' @return Integer status code (0 = success).
#' @export
tcc_relocate <- function(state) {
  .Call(RC_libtcc_relocate, state)
}

#' Add a symbol to a libtcc state
#' @param state A `tcc_state`.
#' @param name Symbol name.
#' @param addr External pointer address or symbol value.
#' @return Integer status code (0 = success).
#' @export
tcc_add_symbol <- function(state, name, addr) {
  .Call(RC_libtcc_add_symbol, state, name, addr)
}

#' Get a symbol pointer from a libtcc state
#' @param state A `tcc_state`.
#' @param name Symbol name to look up.
#' @return External pointer of class `tcc_symbol`.
#' @export
tcc_get_symbol <- function(state, name) {
  .Call(RC_libtcc_get_symbol, state, name)
}


#' Call a zero-argument symbol with a specified return type
#' @param state A `tcc_state`.
#' @param name Symbol name to call.
#' @param return One of "int", "double", "void".
#' @return The return value cast to the requested type (NULL for void).
#' @export
tcc_call_symbol <- function(state, name, return = c("int", "double", "void")) {
  return <- match.arg(return)
  .Call(RC_libtcc_call_symbol, state, name, return)
}

#' Check if a tcc_symbol external pointer is valid
#' @param ptr External pointer from `tcc_get_symbol()`.
#' @return TRUE if the pointer address is non-null, FALSE otherwise.
#' @export
tcc_symbol_is_valid <- function(ptr) {
  .Call(RC_libtcc_ptr_valid, ptr)
}

#' Run the tinycc CLI
#' @param args Character vector of CLI arguments (e.g., `c("-c", file, "-o", out)`).
#' @param tcc_path Optional path to the `tcc` binary; defaults to the bundled CLI.
#' @return Integer status from `system2()`.
#' @export
tcc_run_cli <- function(args = character(), tcc_path = check_cli_exists()) {
  tcc_path <- normalizePath(tcc_path, winslash = "/", mustWork = TRUE)
  args <- as.character(args)
  lib_paths <- normalizePath(tcc_lib_paths(), winslash = "/", mustWork = FALSE)
  env <- character()
  if (length(lib_paths)) {
    sysname <- Sys.info()[["sysname"]]
    env_key <- tcc_loader_env_key(sysname)
    env_sep <- tcc_loader_env_sep(sysname)
    env <- sprintf(
      "%s=%s",
      env_key,
      paste(c(lib_paths, Sys.getenv(env_key)), collapse = env_sep)
    )
  }
  res <- system2(tcc_path, args, env = env)
  as.integer(res)
}

# Test FFI Code Generation - With Real Compilation
# Tests that generated code actually compiles and runs

library(Rtinycc)

# Test 1: Generate and compile simple FFI code
symbols <- list(
  forty_two = Rtinycc:::as_rtinycc_bound_symbol(
    "forty_two",
    list(args = list(), returns = "i32")
  )
)

code <- Rtinycc:::generate_ffi_code(
  symbols = symbols,
  headers = NULL,
  c_code = "int forty_two() { return 42; }",
  is_external = FALSE
)

expect_true(grepl("int forty_two", code))
expect_true(grepl("R_wrap_forty_two", code))

r_lib_paths <- file.path(R.home("lib"))
if (.Platform$OS.type == "windows") {
  r_lib_paths <- c(
    r_lib_paths,
    normalizePath(
      file.path(R.home(), "bin", .Platform$r_arch),
      winslash = "/",
      mustWork = FALSE
    )
  )
}
state <- tcc_state(
  output = "memory",
  include_path = c(tcc_include_paths(), file.path(R.home("include"))),
  lib_path = c(tcc_lib_paths(), r_lib_paths)
)
if (.Platform$OS.type == "windows") {
  tcc_add_library(state, "R")
}
result <- tcc_compile_string(state, code)
expect_equal(result, 0L)

relocate_result <- tcc_relocate(state)
expect_equal(relocate_result, 0L)

sym_ptr <- tcc_get_symbol(state, "R_wrap_forty_two")
expect_true(inherits(sym_ptr, "tcc_symbol"))

# Test 2: Verify array type code generation compiles
symbols <- list(
  sum_ints = Rtinycc:::as_rtinycc_bound_symbol(
    "sum_ints",
    list(args = list("integer_array", "i32"), returns = "i64")
  )
)

code <- Rtinycc:::generate_ffi_code(
  symbols = symbols,
  c_code = "
    int64_t sum_ints(int32_t* arr, int32_t n) {
      int64_t sum = 0;
      for(int i = 0; i < n; i++) sum += arr[i];
      return sum;
    }
  ",
  is_external = FALSE
)

expect_true(grepl("sum_ints", code))
expect_true(grepl("R_wrap_sum_ints", code))

r_lib_paths <- file.path(R.home("lib"))
if (.Platform$OS.type == "windows") {
  r_lib_paths <- c(
    r_lib_paths,
    normalizePath(
      file.path(R.home(), "bin", .Platform$r_arch),
      winslash = "/",
      mustWork = FALSE
    )
  )
}
state <- tcc_state(
  output = "memory",
  include_path = c(tcc_include_paths(), file.path(R.home("include"))),
  lib_path = c(tcc_lib_paths(), r_lib_paths)
)
if (.Platform$OS.type == "windows") {
  tcc_add_library(state, "R")
}
result <- tcc_compile_string(state, code)
expect_equal(result, 0L)
expect_equal(tcc_relocate(state), 0L)

library(tinytest)
library(Rtinycc)

Sys.setenv(RTINYCC_DEBUG = "1")

tcc_dir <- tcc_prefix()
expect_true(
  nzchar(tcc_dir) && file.exists(tcc_dir),
  info = "Bundled tinycc not found; run configure to build it."
)
if (!(nzchar(tcc_dir) && file.exists(tcc_dir))) {
  quit(save = "no", status = 1)
}

# libtcc in-memory compile
state <- tcc_state(output = "memory")
# Keep the in-memory test side-effect free.
# printf can be problematic on some platforms/toolchains (macOS, Windows).
code <- "
typedef struct { double r; double i; } Rcomplex;

int forty_two(void) { return 42; }

void mutate_dotc(int *i, double *d, int *l, unsigned char *r, Rcomplex *z, char **s) {
  i[0] += 10;
  i[1] += 20;
  d[0] += 0.5;
  d[1] += 1.5;
  l[0] = 2;
  l[1] = 0;
  r[0] += 1;
  r[1] += 2;
  z[0].r += 1.0;
  z[0].i += 2.0;
  s[0][0] = 'Z';
}

void mutate_single(float *x) {
  x[0] += 1.25f;
}

void accept_int(int *x) {
  (void)x;
}

void accept_double(double *x) {
  (void)x;
}

void accept_complex(Rcomplex *x) {
  (void)x;
}

void mark_na_string(char **s) {
  if (s[0][0] == 'N' && s[0][1] == 'A' && s[0][2] == 0) {
    s[0][0] = 'O';
  }
}

void inspect_list(void **xs, int *out) {
  out[0] = (xs && xs[0] && xs[1]) ? 1 : 0;
}

void inspect_sexp(void *x, int *out) {
  out[0] = x ? 1 : 0;
}

void overrun_int(int *x) {
  x[2] = 99;
}

void overrun_char(char **s) {
  for (int i = 0; i < 132; i++) {
    s[0][i] = 'x';
  }
}

void replace_char_pointer(char **s) {
  s[0] = \"replacement\";
}
"
message("Adding source code...")
expect_equal(tcc_compile_string(state, code), 0L)
message("Relocating code...")
expect_equal(tcc_relocate(state), 0L)
sym_ptr <- tcc_get_symbol(state, "forty_two")
expect_true(inherits(sym_ptr, "tcc_symbol"))
expect_true(tcc_symbol_is_valid(sym_ptr))
addr <- get_external_ptr_addr(sym_ptr)
cat(sprintf("symbol 'forty_two' address: %f\n", addr))
cat(sprintf("address %% 8: %f\n", addr %% 8))
symbols <- tcc_list_symbols(state)
expect_true(is.data.frame(symbols))
expect_true(identical(names(symbols), c("name", "address")))
expect_true(is.character(symbols$name))
expect_true(is.character(symbols$address))
expect_true(all(grepl("^0x[0-9a-fA-F]+$", symbols$address)))
if (!("forty_two" %in% symbols$name)) {
  message(
    "tcc_list_symbols() did not report compiled symbol 'forty_two' on this platform"
  )
}
expect_equal(tcc_call_symbol(state, "forty_two", return = "int"), 42L)
expect_equal(
  tcc_call_symbol(state = state, name = "forty_two", return = "int"),
  42L,
  info = "tcc_call_symbol keeps named state/name compatibility"
)
expect_equal(
  tcc_call_symbol(state, "forty_two", "int"),
  42L,
  info = "tcc_call_symbol keeps positional return-type compatibility"
)

dotc_i <- as.integer(c(1, 2))
dotc_d <- c(1, 2)
dotc_l <- c(TRUE, FALSE)
dotc_r <- as.raw(c(1, 2))
dotc_z <- 1 + 3i
dotc_s <- "abc"
dotc_res <- tcc_call_symbol(
  state,
  "mutate_dotc",
  i = dotc_i,
  d = dotc_d,
  l = dotc_l,
  r = dotc_r,
  z = dotc_z,
  s = dotc_s
)
expect_equal(names(dotc_res), c("i", "d", "l", "r", "z", "s"))
expect_equal(dotc_res$i, as.integer(c(11, 22)))
expect_equal(dotc_res$d, c(1.5, 3.5))
expect_equal(dotc_res$l, c(TRUE, FALSE))
expect_equal(dotc_res$r, as.raw(c(2, 4)))
expect_equal(dotc_res$z, 2 + 5i)
expect_equal(dotc_res$s, "Zbc")
expect_equal(dotc_i, as.integer(c(1, 2)), info = "integer input is not mutated in place")
expect_equal(dotc_d, c(1, 2), info = "double input is not mutated in place")
expect_equal(dotc_l, c(TRUE, FALSE), info = "logical input is not mutated in place")
expect_equal(dotc_r, as.raw(c(1, 2)), info = "raw input is not mutated in place")
expect_equal(dotc_z, 1 + 3i, info = "complex input is not mutated in place")
expect_equal(dotc_s, "abc", info = "character input is not mutated in place")

altrep_i <- 1:2
altrep_res <- tcc_call_symbol(state, "accept_int", altrep_i)
expect_equal(altrep_res[[1]], as.integer(1:2), info = "ALTREP integer input is copied into result")
expect_equal(altrep_i, 1:2, info = "ALTREP integer input remains unchanged")

single_arg <- structure(1, Csingle = TRUE)
single_res <- tcc_call_symbol(state, "mutate_single", single_arg)
expect_equal(single_res[[1]], structure(2.25, Csingle = TRUE))
expect_error(
  tcc_call_symbol(state, "accept_int", as.integer(NA)),
  info = "tcc_call_symbol .C-style calls reject NA by default"
)
expect_equal(
  tcc_call_symbol(state, "accept_int", as.integer(NA), NAOK = TRUE)[[1]],
  as.integer(NA),
  info = "tcc_call_symbol .C-style calls pass integer NA when NAOK = TRUE"
)
expect_error(
  tcc_call_symbol(state, "accept_double", NaN),
  info = "tcc_call_symbol .C-style calls reject NaN by default"
)
expect_equal(
  tcc_call_symbol(state, "accept_double", NaN, NAOK = TRUE)[[1]],
  NaN,
  info = "tcc_call_symbol .C-style calls pass NaN when NAOK = TRUE"
)
expect_error(
  tcc_call_symbol(state, "accept_complex", complex(real = 1, imaginary = NaN)),
  info = "tcc_call_symbol .C-style calls reject non-finite complex values by default"
)
expect_equal(
  tcc_call_symbol(state, "accept_complex", complex(real = 1, imaginary = NaN), NAOK = TRUE)[[1]],
  complex(real = 1, imaginary = NaN),
  info = "tcc_call_symbol .C-style calls pass non-finite complex values when NAOK = TRUE"
)
expect_equal(
  tcc_call_symbol(state, "mark_na_string", NA_character_)[[1]],
  "OA",
  info = "tcc_call_symbol .C-style character path passes NA strings as literal 'NA'"
)
expect_equal(
  tcc_call_symbol(state, "inspect_list", list(1L, "x"), as.integer(0))[[2]],
  1L,
  info = "tcc_call_symbol passes lists through read-only SEXP* path"
)
expect_equal(
  tcc_call_symbol(state, "inspect_sexp", identity, as.integer(0))[[2]],
  1L,
  info = "tcc_call_symbol passes functions/environments/other objects as read-only SEXP"
)
expect_error(
  do.call(tcc_call_symbol, c(list(state, "accept_int"), rep(list(1L), 66L))),
  info = "tcc_call_symbol .C-style calls reject more than 65 arguments"
)
expect_error(
  tcc_call_symbol(state, "overrun_int", as.integer(c(1, 2))),
  info = "tcc_call_symbol checks guard bytes around copied integer buffers"
)
expect_error(
  tcc_call_symbol(state, "overrun_char", "abc"),
  info = "tcc_call_symbol checks guard bytes around copied character buffers"
)
expect_error(
  tcc_call_symbol(state, "replace_char_pointer", "abc"),
  info = "tcc_call_symbol rejects character pointer replacement"
)
# CLI compile to object
# forty_two.c uses stdio.h / printf which are UCRT-inline on Windows,
# so skip this CLI test there.
if (.Platform$OS.type != "windows") {
  src <- system.file("c_examples", "forty_two.c", package = "Rtinycc")
  expect_true(
    file.exists(src),
    info = "example source missing: inst/c_examples/forty_two.c"
  )
  if (!file.exists(src)) {
    quit(save = "no", status = 1)
  }
  out <- tempfile(fileext = ".o")
  on.exit(unlink(out), add = TRUE)
  inc_args <- as.character(paste0("-I", tcc_include_paths()))
  status <- tcc_run_cli(c(inc_args, "-c", src, "-o", out))
  expect_equal(status, 0L)
  expect_true(file.exists(out))
}

library(Rtinycc)

info <- blas_lapack_info()

expect_true(is.list(info))
expect_true(all(
  c(
    "blas_path",
    "lapack_path",
    "has_rblas",
    "has_rlapack",
    "loaded_dlls"
  ) %in%
    names(info)
))

expect_true(is.character(info$blas_path) && length(info$blas_path) == 1L)
expect_true(is.character(info$lapack_path) && length(info$lapack_path) == 1L)
expect_true(is.logical(info$has_rblas) && length(info$has_rblas) == 1L)
expect_true(is.logical(info$has_rlapack) && length(info$has_rlapack) == 1L)
expect_true(is.character(info$loaded_dlls))

# Positive detections should be justified by either a loaded DLL name or the
# runtime library path that R reported.
if (isTRUE(info$has_rblas)) {
  expect_true(
    any(grepl("Rblas", info$loaded_dlls, ignore.case = TRUE)) ||
      grepl("Rblas", basename(info$blas_path), ignore.case = TRUE)
  )
}
if (isTRUE(info$has_rlapack)) {
  expect_true(
    any(grepl("Rlapack", info$loaded_dlls, ignore.case = TRUE)) ||
      grepl("Rlapack", basename(info$lapack_path), ignore.case = TRUE)
  )
}

code_path <- system.file("c_examples", "simd_bytecode.c", package = "Rtinycc")
if (!nzchar(code_path)) {
  candidates <- c(
    file.path("inst", "c_examples", "simd_bytecode.c"),
    file.path("..", "c_examples", "simd_bytecode.c")
  )
  code_path <- candidates[file.exists(candidates)][1L]
}
expect_true(file.exists(code_path))

code <- paste(readLines(code_path, warn = FALSE), collapse = "\n")
state <- tcc_state()
expect_equal(tcc_compile_string(state, code), 0L)
expect_equal(tcc_relocate(state), 0L)

expect_true(tcc_call_symbol(state, "rtcc_cpu_has_sse2", return = "int") %in% 0:1)
expect_true(tcc_call_symbol(state, "rtcc_cpu_has_avx2", return = "int") %in% 0:1)
expect_true(tcc_call_symbol(state, "rtcc_cpu_has_avx512f_bw_vl", return = "int") %in% 0:1)

sse2 <- tcc_call_symbol(
  state,
  "rtcc_sse2_add4_i32",
  a = as.integer(1:4),
  b = as.integer(c(10, 20, 30, 40)),
  out = integer(4)
)
expect_equal(sse2$out, as.integer(c(11, 22, 33, 44)))

if (identical(tcc_call_symbol(state, "rtcc_cpu_has_avx2", return = "int"), 1L)) {
  avx2 <- tcc_call_symbol(
    state,
    "rtcc_avx2_add8_i32",
    a = as.integer(1:8),
    b = as.integer(seq(10, 80, by = 10)),
    out = integer(8)
  )
  expect_equal(avx2$out, as.integer(c(11, 22, 33, 44, 55, 66, 77, 88)))
} else {
  expect_true(TRUE)
}

#!/usr/bin/env Rscript
# vendor tinycc: download or unpack tinycc tarball from specific commit

args <- commandArgs(trailingOnly = TRUE)
mode <- if (length(args) > 0) args[[1]] else "download"

# Use specific commit hash
commit_hash <- "34eed88a70fc8238cd5b623cc96b8bf0e40392ed"
url <- paste0(
  "https://github.com/TinyCC/tinycc/archive/",
  commit_hash,
  ".tar.gz"
)

dest_dir <- "../src"
tarball_path <- paste0("../src/tinycc-", substr(commit_hash, 1, 7), ".tar.gz")

getScriptPath <- function() {
  cmd.args <- commandArgs()
  m <- regexpr("(?<=^--file=).+", cmd.args, perl = TRUE)
  script.dir <- dirname(regmatches(cmd.args, m))
  if (length(script.dir) == 0) {
    stop("can't determine script dir: please call the script with Rscript")
  }
  if (length(script.dir) > 1) {
    stop("can't determine script dir: more than one '--file' argument detected")
  }
  return(script.dir)
}
script_dir <- getScriptPath()
dest_dir <- file.path(script_dir, dest_dir) |> normalizePath()
tarball_path <- file.path(script_dir, tarball_path)

if (mode == "download") {
  message(
    "Downloading tinycc tarball (commit ",
    substr(commit_hash, 1, 7),
    ") to ",
    tarball_path
  )
  download.file(url, tarball_path, quiet = FALSE, method = "auto")
  message("tinycc tarball downloaded.")
  quit(save = "no")
}

if (mode == "unpack") {
  if (!file.exists(tarball_path)) {
    stop(
      "tinycc tarball not found: ",
      tarball_path,
      "\nRun with 'download' mode first."
    )
  }

  # Remove any existing tinycc* directories in dest_dir
  old_dirs <- list.dirs(dest_dir, full.names = TRUE, recursive = FALSE)
  old_dirs <- old_dirs[grepl("^tinycc", basename(old_dirs))]
  if (length(old_dirs) > 0) {
    message("Removing old tinycc directories...")
    unlink(old_dirs, recursive = TRUE)
  }

  message("Unpacking tinycc tarball...")
  utils::untar(tarball_path, exdir = dest_dir)

  # Find the extracted directory (GitHub archives as tinycc-<hash>)
  new_dirs <- list.dirs(dest_dir, full.names = TRUE, recursive = FALSE)
  tinycc_dir <- new_dirs[grepl("^tinycc", basename(new_dirs))]

  if (length(tinycc_dir) == 0) {
    stop("Could not find extracted tinycc directory in ", dest_dir)
  }

  # Rename to simply 'tinycc'
  target_dir <- file.path(dest_dir, "tinycc")
  if (!basename(tinycc_dir[1]) == "tinycc") {
    message("Renaming ", basename(tinycc_dir[1]), " to tinycc")
    file.rename(tinycc_dir[1], target_dir)
    tinycc_dir <- target_dir
  }

  message("tinycc unpacked to ", tinycc_dir)
  quit(save = "no")
}

stop("Unknown mode: ", mode, ". Use 'download' or 'unpack'.")

#!/bin/sh
# Run after configure: build and test minimal TCC JIT example
# Usage: sh run_tcc_minimal_test.sh
set -e
cd "$(dirname "$0")"

# Path to TCC binary and include/lib paths
TCC_BIN="../inst/tinycc/bin/tcc"
TCC_INC="../inst/tinycc/include"
TCC_LIB="../inst/tinycc/lib"

# Build the minimal test
"$TCC_BIN" -I"$TCC_INC" -L"$TCC_LIB" -ltcc -o tcc_minimal_test tcc_minimal_test.c

# Set library path for both Linux and macOS
export LD_LIBRARY_PATH="$TCC_LIB${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"
export DYLD_LIBRARY_PATH="$TCC_LIB${DYLD_LIBRARY_PATH:+:$DYLD_LIBRARY_PATH}"
./tcc_minimal_test

---
title: "Advanced FFI Types"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Advanced FFI Types}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_callback <- Rtinycc::tcc_callback
tcc_callback_async_drain <- Rtinycc::tcc_callback_async_drain
tcc_callback_close <- Rtinycc::tcc_callback_close
tcc_callback_ptr <- Rtinycc::tcc_callback_ptr
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_global <- Rtinycc::tcc_global
tcc_source <- Rtinycc::tcc_source
tcc_struct <- Rtinycc::tcc_struct
```

This article covers the more structured parts of the FFI surface: array-return
specifications, callbacks, globals, and generated struct helpers.

## Array Returns

Array returns use a structured return specification rather than a bare type
string. The implemented form is:

- `type`: one of the supported array types
- `length_arg`: which argument provides the returned length
- `free`: whether the wrapper should `free()` the returned C buffer after copying

```{r}
ffi <- tcc_ffi() |>
  tcc_source(
    "
    #include <stdlib.h>

    int* dup_array(int* x, int n) {
      if (n <= 0) return NULL;
      int* out = (int*)malloc(sizeof(int) * n);
      for (int i = 0; i < n; ++i) out[i] = x[i] * 2;
      return out;
    }
    "
  ) |>
  tcc_bind(
    dup_array = list(
      args = list("integer_array", "i32"),
      returns = list(type = "integer_array", length_arg = 2, free = TRUE)
    )
  ) |>
  tcc_compile()

ffi$dup_array(as.integer(c(1, 2, 3)), 3L)
```

The wrapper copies the returned buffer into a fresh R vector. It does not hand
that C buffer to R by reference.

## Callbacks

Callback arguments use `callback:<signature>` or `callback_async:<signature>`.

```{r}
cb <- tcc_callback(function(x) x * 3, "double (*)(double)")
cb_ptr <- tcc_callback_ptr(cb)

ffi_cb <- tcc_ffi() |>
  tcc_source(
    "
    double apply_cb(double (*cb)(void* ctx, double), void* ctx, double x) {
      return cb(ctx, x);
    }
    "
  ) |>
  tcc_bind(
    apply_cb = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi_cb$apply_cb(cb, cb_ptr, 5)
tcc_callback_close(cb)
```

The callback object owns the registered R function. The callback pointer is the
user-data token passed into the generated trampoline.

For `callback_async:<signature>` specifically:

- the registry currently supports up to 256 live callbacks at once
- `i64`, `u32`, and `u64` async arguments and returns travel through R numeric
  (`double`), so only exact integer values up to `2^53` round-trip exactly
- async callbacks run when the main thread services R's event loop, or when you
  call `tcc_callback_async_drain()` explicitly in tight loops and tests

## Globals

`tcc_global()` generates getter and setter wrappers for C globals.

```{r}
ffi_global <- tcc_ffi() |>
  tcc_source(
    "
    int global_counter = 7;
    "
  ) |>
  tcc_global("global_counter", "i32") |>
  tcc_compile()

ffi_global$global_global_counter_get()
ffi_global$global_global_counter_set(9L)
ffi_global$global_global_counter_get()
```

Globals are limited to scalar FFI types. Array globals are rejected by the API.

## Struct Helpers

For C structs, `Rtinycc` can generate allocation, getter, setter, and free
helpers.

```{r}
ffi_struct <- tcc_ffi() |>
  tcc_source(
    "
    struct point {
      double x;
      double y;
    };
    "
  ) |>
  tcc_struct("point", accessors = c(x = "f64", y = "f64")) |>
  tcc_compile()

pt <- ffi_struct$struct_point_new()
pt <- ffi_struct$struct_point_set_x(pt, 1.5)
pt <- ffi_struct$struct_point_set_y(pt, 2.5)
ffi_struct$struct_point_get_x(pt)
ffi_struct$struct_point_free(pt)
```

These helpers are separate from `tcc_bind()` because they expose storage and
layout rather than just call signatures.

---
title: "FFI Boundary Semantics"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{FFI Boundary Semantics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This article is about what actually crosses the R/C boundary in `Rtinycc`:

- when values are copied
- when they are borrowed
- when they stay as raw addresses
- when wrappers allocate temporary storage

The statements below are based on the implemented wrapper generator and runtime
helpers.

## Scalar Inputs Are Converted

Scalar inputs are converted at the boundary. For example:

- `i8`, `i16`, `i32`, `u8`, `u16` use integer coercion plus range checks
- `i64`, `u32`, `u64` use numeric coercion plus integer-value checks
- `bool` rejects `NA`
- `f32` and `f64` are read from R numerics

So scalar arguments are not zero-copy views into R objects. They become C
scalars inside the wrapper.

## Vector Inputs Are Usually Borrowed

The array input types:

- `raw`
- `integer_array`
- `numeric_array`
- `logical_array`

are passed as direct pointers into the underlying R vector storage.

That means:

- no extra buffer is allocated by the wrapper
- C sees the existing vector data
- mutation from C writes into the same memory region

This is the main zero-copy part of the FFI boundary.

## `cstring_array` Is Rebuilt Per Call

`cstring_array` is different. The wrapper allocates a temporary `const char **`
with `R_alloc()` and fills it by translating each R string element.

So:

- the pointer array itself is allocated for the call
- each element points at translated string data
- this is not the same as passing a pre-existing C array through unchanged

## Returned Arrays Are Copied into Fresh R Vectors

Array returns are always copied into a newly allocated R vector. The wrapper
uses the declared `length_arg` to size the R result, then `memcpy()` copies the
returned C buffer into that vector.

If `free = TRUE`, the wrapper also frees the original returned buffer after the
copy.

So array returns are not borrowed views into C memory.

## Returned `cstring` Values Are Copied

For `cstring` returns, the wrapper creates an R string with `mkString()` when
the returned pointer is non-NULL.

That means the resulting R value is a copy in R-managed memory, not a retained
external pointer to the original C string.

## Returned `ptr` Values Stay as Pointers

For `ptr` returns, the wrapper constructs an external pointer around the raw
address.

That means:

- no pointee copy is made
- ownership is not implied
- the pointer may dangle if the underlying C storage goes away

The same distinction matters for globals and struct fields.

## `sexp` Passes Through Directly

`sexp` is the most direct boundary mode:

- input `sexp` arguments are passed through as `SEXP`
- returned `sexp` values are returned directly

This is useful when you want the R C API contract rather than the stricter FFI
conversion layer.

## Owned vs Borrowed Helper Pointers

At the helper level:

- `tcc_malloc()` and `tcc_cstring()` create owned external pointers
- `tcc_data_ptr()` and `tcc_read_ptr()` return borrowed external pointers
- struct field address helpers and many raw pointer returns are borrowed views
- named nested struct getters such as `struct_outer_get_child()` return borrowed
  nested views into the owning struct storage

Use `tcc_ptr_is_owned()` when you need to distinguish these cases in R code.

## Bitfields Are Scalar Helpers, Not Addressable Views

Bitfield helpers behave like scalar getter/setter helpers at the R boundary, but
that does **not** make them ordinary addressable fields.

In particular:

- bitfield getters return copied scalar values
- bitfield setters write scalar values back through the compiler-managed bitfield
  storage
- `tcc_field_addr()` and `tcc_container_of()` reject bitfield members

So bitfields are intentionally excluded from the borrowed-address helper model.

## Serialization Boundary

Compiled `tcc_compiled` objects store enough recipe information to recompile
after `serialize()` / `unserialize()` or `readRDS()`.

Raw pointers and raw `tcc_state` objects do not gain that behavior. After
serialization they are just dead addresses or invalid states, not
auto-reconstructed resources. The same applies to callback tokens,
struct/union external pointers, and helper allocations from `tcc_malloc()` or
`tcc_cstring()`: they do not serialize as live native resources.

---
title: "Compilation and Call Overhead"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Compilation and Call Overhead}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
knitr::knit_engines$set(rtinycc = Rtinycc:::rtinycc_engine)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_source <- Rtinycc::tcc_source
has_callme <- requireNamespace("callme", quietly = TRUE)
has_bench <- requireNamespace("bench", quietly = TRUE)
has_profmem <- isTRUE(capabilities("profmem"))
```

```{css, echo=FALSE}
.rtinycc         { background-color: #E3F2FD; }
pre.rtinycc span { background-color: #E3F2FD; }
```

This article measures two different costs:

- compilation latency for a tiny module
- call overhead once the code is already compiled

The comparison target is the [`callme`](https://cran.r-project.org/package=callme)
package, which builds ordinary `.Call()` entry points with `R CMD SHLIB`. That
means it goes through the platform compiler toolchain (`gcc`/`clang` on the
usual Unix-like targets), so we should expect stronger optimization than TinyCC
for steady-state machine code. That does not make the comparison useless, but it
does mean the runtime results combine two effects:

- direct `.Call()` entry points and direct R C API allocation in `callme`
- better backend optimization from the system compiler

The point is not that the two packages expose identical APIs. They do not.
Instead, the comparison asks a narrower question:

- how much compile-time latency does in-memory TinyCC avoid?
- what is the extra runtime cost of Rtinycc's generated wrapper layer?
- how much does an extra copy matter when Rtinycc has to convert a returned C
  buffer into an R vector?

## Three Minimal Cases

We use three small workloads:

- `noop()`: takes nothing, returns nothing
- `fill_rand(out, n)`: fills a caller-provided numeric buffer in place
- `rand_unif(n)`: generates `n` random doubles

The `fill_rand()` case is the fairer array-oriented comparison:

- `Rtinycc` receives a `numeric_array`, so the wrapper borrows the backing
  `REAL()` storage of an already-materialized R vector directly
- `callme` takes an R numeric vector and writes into `REAL(vec)` directly

The `rand_unif()` case intentionally stresses the extra copy path:

- `callme` allocates the final R vector directly with the R C API
- `Rtinycc` returns a heap-allocated `double*`, and the generated wrapper copies
  that buffer into a fresh R numeric vector before freeing the original C
  allocation

```{rtinycc, object = "rtinycc_code"}
#include <R.h>
#include <Rinternals.h>
#include <Rmath.h>
#include <stdlib.h>

void noop(void) {}

void fill_rand(double* out, int n) {
  if (n < 0) {
    Rf_error("n must be non-negative");
  }

  GetRNGstate();
  for (int i = 0; i < n; ++i) {
    out[i] = unif_rand();
  }
  PutRNGstate();
}

double* rand_unif(int n) {
  if (n < 0) {
    Rf_error("n must be non-negative");
  }
  if (n == 0) {
    return (double*) malloc(sizeof(double));
  }

  double *out = (double*) malloc(sizeof(double) * (size_t) n);
  if (!out) {
    Rf_error("malloc failed");
  }

  GetRNGstate();
  for (int i = 0; i < n; ++i) {
    out[i] = unif_rand();
  }
  PutRNGstate();
  return out;
}
```

```{rtinycc, object = "callme_code"}
#include <R.h>
#include <Rinternals.h>
#include <Rmath.h>

SEXP noop(void) {
  return R_NilValue;
}

SEXP fill_rand(SEXP out_, SEXP n_) {
  int n = asInteger(n_);
  if (n < 0) {
    Rf_error("n must be non-negative");
  }

  if (TYPEOF(out_) != REALSXP) {
    Rf_error("out must be a numeric vector");
  }

  if (XLENGTH(out_) < n) {
    Rf_error("out is shorter than n");
  }

  double *out = REAL(out_);
  GetRNGstate();
  for (int i = 0; i < n; ++i) {
    out[i] = unif_rand();
  }
  PutRNGstate();

  return out_;
}

SEXP rand_unif(SEXP n_) {
  int n = asInteger(n_);
  if (n < 0) {
    Rf_error("n must be non-negative");
  }

  SEXP out = PROTECT(allocVector(REALSXP, n));
  double *ptr = REAL(out);

  GetRNGstate();
  for (int i = 0; i < n; ++i) {
    ptr[i] = unif_rand();
  }
  PutRNGstate();

  UNPROTECT(1);
  return out;
}
```

```{r}
build_rtinycc_module <- function() {
  tcc_ffi() |>
    tcc_source(rtinycc_code) |>
    tcc_bind(
      noop = list(args = list(), returns = "void"),
      fill_rand = list(args = list("numeric_array", "i32"), returns = "void"),
      rand_unif = list(
        args = list("i32"),
        returns = list(type = "numeric_array", length_arg = 1, free = TRUE)
      )
    ) |>
    tcc_compile()
}

build_callme_module <- function() {
  before <- names(getLoadedDLLs())
  mod <- callme::compile(callme_code, env = NULL, verbosity = 0)
  dlls <- getLoadedDLLs()
  new_names <- setdiff(names(dlls), before)
  new_names <- new_names[startsWith(new_names, "callme_")]
  attr(mod, "dll_paths") <- unname(vapply(
    dlls[new_names],
    function(x) x[["path"]],
    character(1)
  ))
  mod
}

unload_callme_dlls <- function(dll_paths) {
  dll_paths <- rev(unique(dll_paths))
  if (is.null(dll_paths) || !length(dll_paths)) {
    return(invisible(NULL))
  }
  for (dll_path in dll_paths) {
    if (is.character(dll_path) && nzchar(dll_path) && file.exists(dll_path)) {
      try(dyn.unload(dll_path), silent = TRUE)
    }
  }
  invisible(NULL)
}

build_and_dispose_callme_module <- function() {
  mod <- build_callme_module()
  dll_paths <- attr(mod, "dll_paths", exact = TRUE)
  rm(mod)
  gc()
  unload_callme_dlls(dll_paths)
  invisible(NULL)
}

callme_runtime_reason <- NULL
can_run_callme <- FALSE

if (!has_callme) {
  callme_runtime_reason <- "`callme` is not installed."
} else if (.Platform$OS.type == "windows") {
  callme_runtime_reason <- paste(
    "`callme` comparisons are skipped on Windows during vignette builds",
    "because the helper DLL compilation step is not reliable in CI."
  )
} else {
  callme_probe <- tryCatch(
    {
      build_and_dispose_callme_module()
      NULL
    },
    error = identity
  )

  if (inherits(callme_probe, "error")) {
    callme_runtime_reason <- paste(
      "`callme` comparisons were skipped because runtime compilation failed:",
      conditionMessage(callme_probe)
    )
  } else {
    can_run_callme <- TRUE
  }
}

can_run_benchmarks <- can_run_callme && has_bench && has_profmem

if (is.null(callme_runtime_reason) && !has_bench) {
  callme_runtime_reason <- "`bench` is not installed."
} else if (is.null(callme_runtime_reason) && !has_profmem) {
  callme_runtime_reason <- paste(
    "`bench` runtime comparisons are skipped because memory profiling",
    "is not available in this R build."
  )
} else if (is.null(callme_runtime_reason)) {
  callme_runtime_reason <- "Executable comparisons are enabled."
}

with_benchmark_modules <- function(fun) {
  rt_mod <- build_rtinycc_module()
  cm_mod <- build_callme_module()
  dll_paths <- attr(cm_mod, "dll_paths", exact = TRUE)

  on.exit({
    rm(rt_mod, cm_mod)
    gc()
    unload_callme_dlls(dll_paths)
  }, add = TRUE)

  fun(rt_mod, cm_mod)
}

median_elapsed <- function(expr, times = 3L) {
  expr <- substitute(expr)
  env <- parent.frame()
  stats::median(replicate(
    times,
    {
      gc()
      t0 <- proc.time()[["elapsed"]]
      eval(expr, envir = env)
      proc.time()[["elapsed"]] - t0
    }
  ))
}

run_noop <- function(fun, n) {
  for (i in seq_len(n)) {
    fun()
  }
  invisible(NULL)
}

run_rand <- function(fun, n, reps) {
  for (i in seq_len(reps)) {
    invisible(fun(n))
  }
  invisible(NULL)
}

run_fill <- function(fun, n, reps) {
  for (i in seq_len(reps)) {
    out <- numeric(n)
    invisible(fun(out, n))
  }
  invisible(NULL)
}

rtinycc_recipe <- tcc_ffi() |>
  tcc_source(rtinycc_code) |>
  tcc_bind(
    noop = list(args = list(), returns = "void"),
    fill_rand = list(args = list("numeric_array", "i32"), returns = "void"),
    rand_unif = list(
      args = list("i32"),
      returns = list(type = "numeric_array", length_arg = 1, free = TRUE)
    )
  )

generated_code <- Rtinycc:::generate_ffi_code(
  symbols = rtinycc_recipe$symbols,
  headers = rtinycc_recipe$headers,
  c_code = rtinycc_recipe$c_code,
  is_external = FALSE,
  structs = rtinycc_recipe$structs,
  unions = rtinycc_recipe$unions,
  enums = rtinycc_recipe$enums,
  globals = rtinycc_recipe$globals,
  container_of = rtinycc_recipe$container_of,
  field_addr = rtinycc_recipe$field_addr,
  struct_raw_access = rtinycc_recipe$struct_raw_access,
  introspect = rtinycc_recipe$introspect
)
```

## Availability

```{r}
has_callme
```

If `callme`, `bench`, or R memory profiling is unavailable, or if the current
build environment cannot compile the temporary `callme` helper DLL, the
executable comparisons below are skipped.

```{r}
has_bench
```

```{r}
has_profmem
```

```{r}
can_run_callme
```

```{r}
can_run_benchmarks
```

Current comparison status:

```{r}
callme_runtime_reason
```

## Compilation Latency

This measures module build time, not call time.

```{r, eval = can_run_callme}
compile_times <- data.frame(
  implementation = c("Rtinycc", "callme"),
  seconds = c(
    median_elapsed(build_rtinycc_module(), times = 3L),
    median_elapsed(build_and_dispose_callme_module(), times = 3L)
  )
)

compile_times$milliseconds <- round(compile_times$seconds * 1000, 1)
compile_times
```

The expected pattern is:

- `Rtinycc` wins clearly on tiny compile latency because it stays in-process and
  does not shell out to `R CMD SHLIB`
- `callme` pays the ordinary shared-library toolchain cost

## Generated Wrapper Code

The generated code makes the extra return-path work explicit. In particular, the
`rand_unif()` wrapper allocates an R vector, `memcpy()`s the native `double*`
buffer into it, then `free()`s the original buffer. In contrast, `fill_rand()`
uses the borrowed `numeric_array` input path.

```{r, results='asis'}
Rtinycc:::rtinycc_c_block(generated_code)
```

## `noop()` Call Overhead

This is the smallest useful call path. It approximates the lower bound on call
overhead above a plain `.Call()` entry point.

```{r, eval = can_run_benchmarks}
noop_bench <- with_benchmark_modules(function(rt_mod, cm_mod) {
  n_noop <- 1000L

  bench::mark(
    Rtinycc = run_noop(rt_mod$noop, n_noop),
    callme = run_noop(cm_mod$noop, n_noop),
    iterations = 20,
    check = TRUE,
    memory = has_profmem,
    filter_gc = FALSE
  )
})

noop_bench
```

Interpretation:

- the `callme` path is close to the cost of a conventional `.Call()` wrapper
- the `Rtinycc` path adds the generated wrapper layer and external-pointer call
  target
- the difference here is mostly boundary overhead, not useful computation
- `check = TRUE` is appropriate here because both expressions always return
  `NULL`
- `bench` also exposes allocation and GC differences directly, which is useful
  for understanding the cost of boxing and copying

## `fill_rand(out, n)` And Zero-Copy Arrays

This is the fairer vector comparison because both implementations fill an
existing R numeric vector instead of returning a newly allocated result.

```{r, eval = can_run_benchmarks}
fill_bench_n4096 <- with_benchmark_modules(function(rt_mod, cm_mod) {
  bench::mark(
    Rtinycc = run_fill(rt_mod$fill_rand, 4096L, 100L),
    callme = run_fill(cm_mod$fill_rand, 4096L, 100L),
    iterations = 20,
    check = FALSE,
    memory = has_profmem,
    filter_gc = FALSE
  )
})

fill_bench_n4096
```

Interpretation:

- both sides now write into ordinary R numeric storage
- this removes the return-copy penalty from the comparison
- the remaining gap is mostly call boundary overhead plus backend code quality

## `rand_unif(n)` And Copy Cost

Here the implementation work is still small, but the return path differs:

- `callme` fills the final R vector directly
- `Rtinycc` fills a native buffer, then the wrapper copies into a fresh R vector

We time both a tiny and a larger return size.

```{r, eval = can_run_benchmarks}
rand_results <- with_benchmark_modules(function(rt_mod, cm_mod) {
  rand_bench_n1 <- bench::mark(
    Rtinycc = run_rand(rt_mod$rand_unif, 1L, 1000L),
    callme = run_rand(cm_mod$rand_unif, 1L, 1000L),
    iterations = 20,
    check = FALSE,
    memory = has_profmem,
    filter_gc = FALSE
  )

  rand_bench_n4096 <- bench::mark(
    Rtinycc = run_rand(rt_mod$rand_unif, 4096L, 100L),
    callme = run_rand(cm_mod$rand_unif, 4096L, 100L),
    iterations = 20,
    check = FALSE,
    memory = has_profmem,
    filter_gc = FALSE
  )

  list(rand_bench_n1 = rand_bench_n1, rand_bench_n4096 = rand_bench_n4096)
})

rand_results$rand_bench_n1
rand_results$rand_bench_n4096
```

The usual pattern is:

- for `fill_rand()`, the comparison is much closer to `Rtinycc`'s intended
  array-oriented usage
- for `n = 1`, wrapper overhead and return-path mechanics dominate
- for larger `n`, the copy still matters, but more of the time is spent in the
  actual loop and RNG generation

## What These Numbers Mean

The benchmark gives a reasonable mental model:

- `Rtinycc` is optimized for low compilation latency and direct interactive use
- for very small scalar calls, a traditional `.Call()` entry point has lower
  overhead
- when `Rtinycc` must copy returned buffers into R vectors, that copy is real
  and measurable
- part of the runtime gap is also expected backend quality: `callme` is using
  the system compiler, while `Rtinycc` is using TinyCC
- the main way to amortize the boundary cost is to do more work per call

So the package is usually strongest when:

- compile latency matters
- you want to bind plain C signatures quickly
- you batch work into array-oriented or coarse-grained calls

It is less ideal when:

- every microsecond of scalar-call overhead matters
- you can already afford and manage a regular shared-library toolchain
- you need a direct `.Call()` entry point that writes its final result straight
  into R-managed objects

---
title: "Header Parsing with treesitter.c"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Header Parsing with treesitter.c}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_generate_bindings <- Rtinycc::tcc_generate_bindings
tcc_map_c_type_to_ffi <- Rtinycc::tcc_map_c_type_to_ffi
tcc_source <- Rtinycc::tcc_source
tcc_struct <- Rtinycc::tcc_struct
tcc_treesitter_bindings <- Rtinycc::tcc_treesitter_bindings
tcc_treesitter_functions <- Rtinycc::tcc_treesitter_functions
tcc_treesitter_struct_accessors <- Rtinycc::tcc_treesitter_struct_accessors
has_treesitter <- requireNamespace("treesitter.c", quietly = TRUE) &&
  packageVersion("treesitter.c") >= "0.0.4"
```

`Rtinycc` can build FFI bindings from C declarations instead of requiring every
signature to be written manually. The parsing layer is provided by the optional
`treesitter.c` package, while `Rtinycc` is responsible for mapping parsed C
types into its own FFI type system.

This workflow is useful when:

- you already have a C header snippet
- you want a first-pass binding specification quickly
- you want to inspect declarations before deciding what to expose

## Availability

These helpers require the optional `treesitter.c` package.

```{r}
has_treesitter
```

If `FALSE`, the parsing functions are unavailable and the executable examples in
this vignette are skipped.

## Start from a Header Snippet

```{r}
header <- paste(
  "double sqrt(double x);",
  "int add(int a, int b);",
  "struct point { double x; double y; };",
  "enum status { OK = 0, ERR = 1 };",
  sep = "\n"
)
```

## Inspect Parsed Functions

The lowest-level helpers return parsed declarations from the header:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
tcc_treesitter_functions(header)
```

For quick inspection of functions only, `tcc_treesitter_bindings()` converts the
parsed signatures into `tcc_bind()`-ready specs:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
tcc_treesitter_bindings(header)
```

## Generate a Working FFI Object

For a fuller workflow, use `tcc_generate_bindings()` on a `tcc_ffi` object that
already has matching C source attached:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
ffi <- tcc_ffi() |>
  tcc_source(
    "
    double sqrt(double x) { return x < 0 ? 0 : x; }
    int add(int a, int b) { return a + b; }
    struct point { double x; double y; };
    enum status { OK = 0, ERR = 1 };
    "
  )

ffi <- tcc_generate_bindings(
  ffi,
  header,
  functions = TRUE,
  structs = TRUE,
  enums = TRUE,
  unions = FALSE,
  globals = FALSE
)

compiled <- tcc_compile(ffi)

compiled$add(2L, 3L)
compiled$enum_status_OK()
```

This keeps the parsing step separate from the actual compilation step, which is
important when you want to inspect or edit the generated binding plan first.

## Struct Helpers

You can also extract just the generated struct accessors:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
tcc_treesitter_struct_accessors("struct point { double x; double y; };")
```

That output is what `Rtinycc` feeds into `tcc_struct()` when you ask it to
generate struct bindings from a header.

Bitfields now stay explicit in the accessor metadata rather than collapsing to a
bare scalar type:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
tcc_treesitter_struct_accessors(
  "struct flags { unsigned int flag : 1; unsigned int code : 6; };"
)
```

Nested struct fields in structs currently still fall back to ptr-like
accessors:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
tcc_treesitter_struct_accessors(
  "struct child { int x; }; struct outer { struct child child; int y; };"
)
```

For unions, nested struct members preserve `list(type = "struct", ...)` so the
generated helper remains a borrowed nested view rather than an opaque raw
pointer.

## Conservative Type Mapping

The default mapper is intentionally conservative. In particular, pointer types
are not automatically treated as C strings unless that is semantically safe.

```{r}
tcc_map_c_type_to_ffi("int")
tcc_map_c_type_to_ffi("double")
tcc_map_c_type_to_ffi("const char *")
```

If you know a specific API uses `const char *` as a real NUL-terminated string,
you can override the mapping:

```{r, eval = requireNamespace("treesitter.c", quietly = TRUE) && packageVersion("treesitter.c") >= "0.0.4"}
string_mapper <- function(type) {
  if (trimws(type) == "const char *") {
    return("cstring")
  }
  tcc_map_c_type_to_ffi(type)
}

tcc_treesitter_bindings(
  "int puts(const char *s);",
  mapper = string_mapper
)
```

This is the intended extension point: keep the default mapper strict, then
relax specific cases where you know the source API contract.

---
title: "FFI Helper APIs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{FFI Helper APIs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_cstring <- Rtinycc::tcc_cstring
tcc_data_ptr <- Rtinycc::tcc_data_ptr
tcc_free <- Rtinycc::tcc_free
tcc_malloc <- Rtinycc::tcc_malloc
tcc_null_ptr <- Rtinycc::tcc_null_ptr
tcc_ptr_addr <- Rtinycc::tcc_ptr_addr
tcc_ptr_is_null <- Rtinycc::tcc_ptr_is_null
tcc_ptr_is_owned <- Rtinycc::tcc_ptr_is_owned
tcc_ptr_set <- Rtinycc::tcc_ptr_set
tcc_read_bytes <- Rtinycc::tcc_read_bytes
tcc_read_cstring <- Rtinycc::tcc_read_cstring
tcc_read_f64 <- Rtinycc::tcc_read_f64
tcc_read_i32 <- Rtinycc::tcc_read_i32
tcc_write_bytes <- Rtinycc::tcc_write_bytes
tcc_write_f64 <- Rtinycc::tcc_write_f64
tcc_write_i32 <- Rtinycc::tcc_write_i32
```

The helper APIs are the low-level tools you use when a binding needs explicit
memory management, pointer inspection, or typed reads and writes.

## Allocation and Ownership

The owned-memory helpers are `tcc_malloc()` and `tcc_cstring()`.

```{r}
buf <- tcc_malloc(16)
str_ptr <- tcc_cstring("hello")

tcc_ptr_is_owned(buf)
tcc_ptr_is_owned(str_ptr)
```

Both return owned external pointers. They can be freed explicitly with
`tcc_free()` and also carry finalizers.

```{r}
tcc_free(buf)
tcc_free(str_ptr)
```

## Reading and Writing Raw Bytes

```{r}
buf <- tcc_malloc(8)
tcc_write_bytes(buf, as.raw(1:8))
tcc_read_bytes(buf, 8)
tcc_free(buf)
```

## Typed Reads and Writes

The helper layer also exposes typed accessors for primitive values at byte
offsets.

```{r}
buf <- tcc_malloc(16)
tcc_write_i32(buf, 0L, 42L)
tcc_write_f64(buf, 8L, 3.5)

tcc_read_i32(buf, offset = 0L)
tcc_read_f64(buf, offset = 8L)

tcc_free(buf)
```

These helpers are useful for manual struct-like layouts, output buffers, or
pointer-heavy APIs that do not map cleanly onto ordinary R vectors.

## Pointer-to-Pointer Helpers

Some C APIs fill outputs through `T **` or `void **`. `Rtinycc` exposes a small
set of helpers for that pattern.

```{r}
ptr_size <- if (!is.null(.Machine$sizeof.pointer)) .Machine$sizeof.pointer else 8L
ptr_ref <- tcc_malloc(ptr_size)
target <- tcc_malloc(4)

tcc_ptr_set(ptr_ref, target)
tcc_ptr_addr(tcc_data_ptr(ptr_ref))
tcc_ptr_addr(target)

tcc_ptr_set(ptr_ref, tcc_null_ptr())
tcc_ptr_is_null(tcc_data_ptr(ptr_ref))

tcc_free(target)
tcc_free(ptr_ref)
```

`tcc_data_ptr()` returns a borrowed pointer view. It is not an owned allocation.

## C String Helpers

```{r}
str_ptr <- tcc_cstring("Hello, world")
tcc_read_cstring(str_ptr)
tcc_read_cstring(str_ptr, max_bytes = 5)
tcc_free(str_ptr)
```

This helper is the safe way to allocate a C-owned NUL-terminated string when
the callee expects a mutable or longer-lived `char *`.

---
title: "FFI Objects, Structs, and Callbacks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{FFI Objects, Structs, and Callbacks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_callback <- Rtinycc::tcc_callback
tcc_callback_async_drain <- Rtinycc::tcc_callback_async_drain
tcc_callback_close <- Rtinycc::tcc_callback_close
tcc_callback_ptr <- Rtinycc::tcc_callback_ptr
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_source <- Rtinycc::tcc_source
tcc_struct <- Rtinycc::tcc_struct
```

This vignette covers two patterns that come up quickly in real bindings:

- exposing C structs through generated helper methods
- passing R functions into compiled code as callbacks

## Working with Struct Helpers

Struct helpers are generated from a declarative description. In the example
below, `Rtinycc` creates allocation, field getter, field setter, and free
helpers for a simple `struct point`.

```{r}
ffi_struct <- tcc_ffi() |>
  tcc_source(
    "
    struct point {
      double x;
      double y;
    };

    double point_norm2(struct point* p) {
      return p->x * p->x + p->y * p->y;
    }
    "
  ) |>
  tcc_struct("point", accessors = c(x = "f64", y = "f64")) |>
  tcc_bind(
    point_norm2 = list(args = list("ptr"), returns = "f64")
  ) |>
  tcc_compile()

pt <- ffi_struct$struct_point_new()
pt <- ffi_struct$struct_point_set_x(pt, 3)
pt <- ffi_struct$struct_point_set_y(pt, 4)

ffi_struct$point_norm2(pt)
ffi_struct$struct_point_free(pt)
```

This keeps the C layout explicit while still giving you a usable R-facing
surface.

Named nested struct fields can also be modeled directly with `struct:<name>`.
Those getters return borrowed nested views and setters copy bytes from a source
struct object of the matching nested type.

## Registering Callbacks

Callbacks let compiled C code invoke an R function through a generated
trampoline. The callback object and the callback pointer play different roles:

- the `tcc_callback` object owns the registered R function
- `tcc_callback_ptr()` returns the user-data handle that C trampolines expect

```{r}
cb <- tcc_callback(
  function(x) x * 2,
  signature = "double (*)(double)"
)
cb_ptr <- tcc_callback_ptr(cb)

ffi_cb <- tcc_ffi() |>
  tcc_source(
    "
    double apply_cb(double (*cb)(void* ctx, double), void* ctx, double x) {
      return cb(ctx, x);
    }
    "
  ) |>
  tcc_bind(
    apply_cb = list(
      args = list("callback:double(double)", "ptr", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi_cb$apply_cb(cb, cb_ptr, 5)
tcc_callback_close(cb)
```

`tcc_callback_close()` is recommended when you want deterministic invalidation
and prompt release of the preserved R function. If you simply drop all
references, finalizers will still clean up the callback eventually.

## Async Callback Caveats

`callback_async:<signature>` is the safe path for worker-thread callbacks, but
its contract is narrower than the synchronous trampoline path:

- the callback registry currently holds at most 256 live callbacks at once
- `i64`, `u32`, and `u64` async arguments and returns are marshalled through
  R numeric (`double`), so only exact integer values up to `2^53` are exact
- queued async callbacks run when the main thread services R's event loop; in
  tests or tight compute loops, call `tcc_callback_async_drain()` explicitly

## Soundness Notes

The callback contract is deliberately explicit:

- the callback signature must match what the C code expects
- pointer arguments are passed through as external pointers
- callback errors are converted into warnings plus a type-appropriate default return value

That explicitness is part of what keeps `Rtinycc` predictable as a systems
interface rather than a partial compiler front-end.

---
title: "FFI Types"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{FFI Types}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_free <- Rtinycc::tcc_free
tcc_malloc <- Rtinycc::tcc_malloc
tcc_source <- Rtinycc::tcc_source
```

This article describes the FFI type names implemented in `Rtinycc`.
The mapping below is taken from the package type table and wrapper-generation
code, not from a separate design document.

## Scalar Types

The core scalar types are:

- signed integers: `i8`, `i16`, `i32`, `i64`
- unsigned integers: `u8`, `u16`, `u32`, `u64`
- floating point: `f32`, `f64`
- logical: `bool`
- string: `cstring`
- opaque pointer: `ptr`
- direct R object: `sexp`
- return-only sentinel: `void`

`Rtinycc` converts R values to these C types inside generated wrappers, but the
R-side scalar carriers depend on what R can represent directly:

- `i8`, `i16`, `i32`, `u8`, and `u16` are mediated through R integer scalars
- `u32`, `i64`, `u64`, `f32`, and `f64` are mediated through R numeric
  (`double`) coercion and boxing
- `bool` uses R logical
- `cstring` uses an R character scalar

So the FFI names are C-side type names, not promises that R has a matching
native scalar type for every width. Integer-like paths are validated, not just
cast silently, and `i64` / `u64` are only exact up to `2^53` on the R side.

## A Minimal Example

```{r}
ffi <- tcc_ffi() |>
  tcc_source(
    "
    int add_i32(int a, int b) { return a + b; }
    double mul_f64(double x, double y) { return x * y; }
    int negate_bool(_Bool x) { return !x; }
    "
  ) |>
  tcc_bind(
    add_i32 = list(args = list("i32", "i32"), returns = "i32"),
    mul_f64 = list(args = list("f64", "f64"), returns = "f64"),
    negate_bool = list(args = list("bool"), returns = "bool")
  ) |>
  tcc_compile()

ffi$add_i32(2L, 3L)
ffi$mul_f64(2, 4)
ffi$negate_bool(TRUE)
```

## String and Pointer Types

`cstring` and `ptr` are intentionally different:

- `cstring` converts an R character scalar into a C `char *` view for the call
- `ptr` passes an external pointer address through unchanged

```{r}
ffi_str <- tcc_ffi() |>
  tcc_source(
    "
    const char* echo_cstring(const char* s) { return s; }
    void* echo_ptr(void* p) { return p; }
    "
  ) |>
  tcc_bind(
    echo_cstring = list(args = list("cstring"), returns = "cstring"),
    echo_ptr = list(args = list("ptr"), returns = "ptr")
  ) |>
  tcc_compile()

ptr <- tcc_malloc(8)

ffi_str$echo_cstring("hello")
inherits(ffi_str$echo_ptr(ptr), "externalptr")

tcc_free(ptr)
```

The important semantic difference is discussed in the boundary-semantics
article: returned `cstring` values are copied into R strings, while returned
`ptr` values stay as raw addresses.

## Array Types

The implemented array input types are:

- `raw`
- `integer_array`
- `numeric_array`
- `logical_array`
- `character_array`
- `cstring_array`

The first four map directly onto R vectors. `character_array` passes the
underlying `CHARSXP` cells as a read-only `SEXP *`, not a `char **`.
Use `cstring_array` when the C side expects a temporary `const char **`.

```{r}
ffi_arr <- tcc_ffi() |>
  tcc_source(
    "
    int first_int(int* x) { return x[0]; }
    double second_num(double* x) { return x[1]; }
    "
  ) |>
  tcc_bind(
    first_int = list(args = list("integer_array"), returns = "i32"),
    second_num = list(args = list("numeric_array"), returns = "f64")
  ) |>
  tcc_compile()

ffi_arr$first_int(as.integer(c(10, 20, 30)))
ffi_arr$second_num(c(1.5, 2.5))
```

## Direct R Objects with `sexp`

`sexp` passes the R object through the wrapper without conversion.

```{r}
ffi_sexp <- tcc_ffi() |>
  tcc_source(
    "
    #include <Rinternals.h>

    SEXP id_sexp(SEXP x) { return x; }
    "
  ) |>
  tcc_bind(
    id_sexp = list(args = list("sexp"), returns = "sexp")
  ) |>
  tcc_compile()

ffi_sexp$id_sexp(list(a = 1, b = 2))
```

This is the lowest-friction way to cross the boundary when you want to work in
terms of the R C API directly rather than the stricter scalar/vector FFI types.

---
title: "Getting Started with Rtinycc"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with Rtinycc}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_compile <- Rtinycc::tcc_compile
tcc_ffi <- Rtinycc::tcc_ffi
tcc_global <- Rtinycc::tcc_global
tcc_include_paths <- Rtinycc::tcc_include_paths
tcc_lib_paths <- Rtinycc::tcc_lib_paths
tcc_link <- Rtinycc::tcc_link
tcc_path <- Rtinycc::tcc_path
tcc_source <- Rtinycc::tcc_source
```

`Rtinycc` bundles TinyCC and exposes two layers:

- low-level helpers for working with the bundled compiler and `libtcc`
- a higher-level FFI interface for compiling small C fragments into callable R functions

The package is designed around explicit data conversion. You declare the C
signature you want, `Rtinycc` generates the wrapper code, and the compiled
object exposes ordinary R-callable functions.

## Toolchain Paths

The bundled TinyCC installation is discoverable from R:

```{r}
tcc_path()
tcc_include_paths()
tcc_lib_paths()
```

These helpers are useful when you need to inspect the bundled toolchain or
bridge `Rtinycc` with other build logic.

## A Minimal FFI Example

The usual entry point is `tcc_ffi()`. You add C source, declare the function
signatures you want to expose, and compile once:

```{r}
ffi <- tcc_ffi() |>
  tcc_source(
    "
    static double affine(double x, double slope, double intercept) {
      return slope * x + intercept;
    }
    "
  ) |>
  tcc_bind(
    affine = list(
      args = list("f64", "f64", "f64"),
      returns = "f64"
    )
  ) |>
  tcc_compile()

ffi$affine(2, 3, 4)
```

The wrapper takes care of converting R numerics to `double` and converting the
C return value back to an R scalar.

## Managing Globals

You can also expose C globals through generated getter and setter helpers:

```{r}
ffi_globals <- tcc_ffi() |>
  tcc_source(
    "
    int shared_counter = 1;

    void bump_counter(void) {
      shared_counter += 1;
    }
    "
  ) |>
  tcc_bind(
    bump_counter = list(args = list(), returns = "void")
  ) |>
  tcc_global("shared_counter", "i32") |>
  tcc_compile()

ffi_globals$global_shared_counter_get()
ffi_globals$bump_counter()
ffi_globals$global_shared_counter_get()
```

This pattern is useful for wrapping small stateful C helpers without dropping
to the lower-level `libtcc` API.

## When to Use `tcc_link()`

`tcc_compile()` is for source you provide directly. `tcc_link()` is for binding
to functions that already exist in a shared library. The high-level binding
format is the same, so you can move between generated code and linked code
without changing the R-side calling convention. See the
`Linking External Libraries` vignette for the external-linking workflow and the
platform-specific runtime notes.

---
title: "Linking External Libraries"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Linking External Libraries}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_compile <- Rtinycc::tcc_compile
tcc_link <- Rtinycc::tcc_link
is_windows <- .Platform$OS.type == "windows"
find_system_math_library <- function() {
  if (is_windows) {
    return(NULL)
  }

  sysname <- as.character(unname(Sys.info()[["sysname"]]))
  paths <- unique(Rtinycc:::tcc_library_search_paths(sysname))
  # On Linux prefer the unversioned `libm.so`, which on glibc systems is a
  # GNU ld script that pulls in libm.so.6 plus libmvec.so.1; TinyCC handles
  # those scripts. Fall back to the SONAME-versioned file only if the dev
  # symlink is absent (some minimal containers).
  patterns <- switch(
    sysname,
    Linux = c("^libm\\.so$", "^libm\\.so(\\.[0-9]+)+$"),
    Darwin = c("^libSystem\\.B\\.dylib$", "^libm\\.dylib$"),
    character(0)
  )

  for (pattern in patterns) {
    for (path in paths) {
      if (!dir.exists(path)) {
        next
      }
      candidates <- list.files(
        path,
        pattern = pattern,
        full.names = TRUE
      )
      if (length(candidates) > 0L) {
        return(candidates[[1]])
      }
    }
  }

  NULL
}
math_lib_path <- find_system_math_library()
math_link_ok <- !is.null(math_lib_path) && tryCatch(
  {
    Rtinycc::tcc_link(
      math_lib_path,
      symbols = list(sqrt = list(args = list("f64"), returns = "f64"))
    )
    TRUE
  },
  error = function(e) FALSE
)
```

`tcc_link()` is the external-library counterpart to `tcc_compile()`.
Instead of compiling only the C source you provide, it generates wrappers for
symbols that already exist in a shared library and links those wrappers with
TinyCC.

The important contract is:

- you describe the C signatures you want to call from R
- `Rtinycc` generates the conversion layer
- TinyCC links that conversion layer against the target shared library

## What `tcc_link()` Is For

Use `tcc_link()` when the implementation already exists elsewhere and you want
an R-facing wrapper with the same calling style as `tcc_compile()`.

Typical uses are:

- binding a few functions from a system library
- wrapping part of a third-party C API
- compiling a small helper function next to external symbols

## A Minimal Linking Example

On Unix-like systems, the math library is a simple example because common
functions such as `sqrt()` and `cos()` are usually resolved from the system
math runtime. For robustness, this vignette resolves an actual runtime library
file instead of relying on a short name. On Linux that is typically a versioned
`libm.so.*` file, while on macOS the resolver prefers `libSystem.B.dylib`.

```{r, eval = !is_windows && math_link_ok}
math <- tcc_link(
  math_lib_path,
  symbols = list(
    sqrt = list(args = list("f64"), returns = "f64"),
    cos = list(args = list("f64"), returns = "f64")
  )
)

math$sqrt(25)
math$cos(0)
```

On Windows, the same family of CRT functions is typically provided through the
UCRT. In `Rtinycc`, that runtime detail is handled by the bundled TinyCC setup
and `configure.win`; user-facing code should generally think in terms of “CRT
functions are available” rather than depending on a specific historical import
library name.

## Adding Helper Code Beside External Symbols

You can include a small amount of helper code in the same TinyCC unit while
still linking against an external library:

```{r, eval = !is_windows && math_link_ok}
math_helpers <- tcc_link(
  math_lib_path,
  symbols = list(
    sqrt = list(args = list("f64"), returns = "f64"),
    hypot2_checked = list(args = list("f64", "f64"), returns = "f64")
  ),
  user_code = "
    #include <math.h>

    double hypot2_checked(double x, double y) {
      if (x < 0 || y < 0) {
        return NAN;
      }
      return sqrt(x * x + y * y);
    }
  "
)

math_helpers$hypot2_checked(3, 4)
is.nan(math_helpers$hypot2_checked(-1, 4))
```

This is often enough for:

- argument checks
- small adapter functions
- convenience wrappers around an existing library API

## Linking by Name or by Path

`tcc_link()` accepts either:

- a short library name such as `"m"` or `"sqlite3"`
- a full path to a shared library file

When given a full path, `Rtinycc` adds the containing directory and asks TinyCC
to link that exact file name. This matters for versioned runtime libraries such
as `libm.so.6`, which are not interchangeable with a generic development
symlink such as `libm.so` on every Linux distribution. When given a short name,
TinyCC resolves it through its configured library search paths.

This means the R-side binding declaration stays the same even when deployment
changes from “use the system copy” to “use this exact shared object”.

## Relationship to `tcc_compile()`

Use `tcc_compile()` when your primary artifact is new C code shipped from R.
Use `tcc_link()` when the primary implementation already lives in a shared
library and you want `Rtinycc` to generate the wrapper layer around it.

---
title: "Internals"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Internals}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tcc_bind <- Rtinycc::tcc_bind
tcc_callback_close <- Rtinycc::tcc_callback_close
tcc_compile <- Rtinycc::tcc_compile
tcc_cstring <- Rtinycc::tcc_cstring
tcc_data_ptr <- Rtinycc::tcc_data_ptr
tcc_ffi <- Rtinycc::tcc_ffi
tcc_get_symbol <- Rtinycc::tcc_get_symbol
tcc_link <- Rtinycc::tcc_link
tcc_malloc <- Rtinycc::tcc_malloc
tcc_relocate <- Rtinycc::tcc_relocate
tcc_set_options <- Rtinycc::tcc_set_options
tcc_source <- Rtinycc::tcc_source
```

This article is about how `Rtinycc` works internally. It is not the stable
user-facing API contract. The pieces described here are current implementation
choices and may change as the package evolves.

At a high level, the package is built as a pipeline:

1. an R-side recipe is accumulated in a `tcc_ffi` object
2. that recipe is turned into a generated C translation unit
3. TinyCC compiles and relocates the generated code in memory
4. wrapper pointers are recovered and exposed back to R as closures

## The `tcc_ffi` Object Is a Recipe

`tcc_ffi()` does not compile anything by itself. It creates a plain R object
that accumulates:

- bound symbols
- user headers and user C code
- library and include paths
- extra compiler options
- helper declarations such as structs, unions, enums, globals, and callback use

That state lives in the `tcc_ffi` list object built by `tcc_ffi_object()`. The
important point is that `tcc_compile()` works from this declarative recipe, not
from an already-live TCC process.

```{r}
ffi <- tcc_ffi() |>
  tcc_source("int add(int a, int b) { return a + b; }") |>
  tcc_bind(add = list(args = list("i32", "i32"), returns = "i32"))

names(ffi)
```

## Code Generation Is Central

`tcc_compile()` calls the internal `generate_ffi_code()` helper to assemble one
large C source string. That generated source is the real boundary layer between
R and the target C functions.

Internally, the generated translation unit is assembled in this order:

- a TinyCC workaround for `_Complex`
- `R.h` and `Rinternals.h`
- callback trampoline declarations when needed
- user headers
- external declarations for `tcc_link()`
- user C code
- generated helpers for structs, unions, enums, globals, and raw access
- generated `SEXP` wrappers for each bound symbol

For a small binding:

```{r}
code <- Rtinycc:::generate_ffi_code(
  symbols = ffi$symbols,
  headers = ffi$headers,
  c_code = ffi$c_code,
  is_external = FALSE,
  structs = ffi$structs,
  unions = ffi$unions,
  enums = ffi$enums,
  globals = ffi$globals,
  container_of = ffi$container_of,
  field_addr = ffi$field_addr,
  struct_raw_access = ffi$struct_raw_access,
  introspect = ffi$introspect
)

grepl("SEXP R_wrap_add", code, fixed = TRUE)
```

The wrapper is where input coercion, range checks, callback trampoline setup,
actual C invocation, and return boxing happen.

## How Values Move Between R, The Wrapper, And C

The important internal boundary is not "R calls user C directly". The flow is:

1. an R closure created by `make_callable()` calls `.Call` with the compiled
   wrapper's native symbol external pointer
2. that wrapper receives `SEXP` arguments
3. wrapper code uses the R C API to decode or borrow data from those `SEXP`s
4. the wrapper calls the target C symbol using ordinary C arguments
5. the wrapper converts the C result back into a `SEXP`
6. `.Call` returns that `SEXP` to the R interpreter

So the generated wrapper is the translator between:

- R evaluation and `SEXP` objects on one side
- the target function's plain C signature on the other side

This is why `Rtinycc` includes `R.h` and `Rinternals.h` in every generated
translation unit and why the wrapper code uses constructors and accessors such
as:

- `asInteger()` / `asReal()`
- `RAW()`, `INTEGER()`, `REAL()`, `LOGICAL()`
- `STRING_ELT()` and `Rf_translateCharUTF8()`
- `ScalarInteger()`, `ScalarReal()`, `ScalarLogical()`
- `mkString()` and `R_MakeExternalPtr()`

At the R level, `make_callable()` builds a small closure around the compiled
wrapper pointer. That closure does argument-count validation, checks that the
pointer is still valid, and then hands control to `.Call`.

The wrapper itself is where the actual C API interaction happens.

## Copying Versus Borrowing Happens In The Wrapper

The copy model is mostly determined by the generated conversion code.

Scalar inputs are copied or coerced into local C values:

- integers and booleans go through `asInteger()`
- doubles go through `asReal()`
- range checks happen before the target function is called

These are not zero-copy paths.

Vector inputs are split into two groups:

- `raw`, `integer_array`, `numeric_array`, and `logical_array` borrow writable
  R vector storage directly; for ALTREP inputs, R may materialize the vector
  when the wrapper asks for a C data pointer
- `cstring_array` allocates a temporary pointer array with `R_alloc()` and
  fills it from translated R strings

String and pointer inputs need more care:

- `cstring` uses `STRING_ELT()` plus `Rf_translateCharUTF8()` for the duration
  of the call
- `ptr` reads the raw address from an external pointer with
  `R_ExternalPtrAddr()`
- `sexp` passes the original `SEXP` through unchanged

Returns have their own copy model:

- scalar returns are boxed into fresh R objects
- `cstring` returns are copied into R-managed string memory with `mkString()`
- `ptr` returns stay as external pointers to raw addresses
- array returns allocate a fresh R vector and `memcpy()` the C buffer into it

So the internal design is intentionally mixed:

- borrow when R already has contiguous vector storage that matches the C view
- copy when returning data into R-managed memory
- keep raw pointers raw when the package cannot safely invent ownership

That is the main semantic reason the generated wrapper layer exists.

## Why `lambda.r` Is Used

The large rule file `R/aaa_ffi_codegen_rules.R` uses `lambda.r` as a small
dispatch DSL. The package imports `%as%` and `UseFunction`, and defines rules
like:

- `ffi_input_rule(...)`
- `ffi_return_rule(...)`
- `array_return_alloc_line_rule(...)`
- `c_default_return_rule(...)`
- `ffi_c_type_map_rule(...)`

Those rules are not user-facing metaprogramming. They are an internal way to
register many small code-generation cases without turning
`R/ffi_codegen.R` into one enormous nest of `if` and `switch` statements.

In practice, `generate_c_input()` and `generate_c_return()` delegate into that
rule table:

```{r}
Rtinycc:::generate_c_input("x", "arg1_", "i32")
Rtinycc:::generate_c_return("res", "f64")
```

The main tradeoff is simple:

- `lambda.r` keeps the dispatch table explicit and composable
- the rule file becomes long because many integer, floating-point, and helper
  cases are still written out individually

So `lambda.r` here is being used for internal rule dispatch and code-template
selection, not because the public API depends on functional programming style.

## Wrapper Builders Work at the `SEXP` Boundary

`Rtinycc` is not using a libffi ABI layer. The generated wrappers are normal C
functions with `SEXP` signatures so that R can call them through `.Call`.

The key internal steps are:

- `generate_wrappers()` decides which wrapper variants are needed
- `generate_c_wrapper()` builds the normal synchronous wrapper body
- `generate_async_exec_wrapper()` builds the async execution path for
  `callback_async:` arguments
- `generate_callback_trampolines()` emits trampoline functions for callback
  arguments

For non-variadic bindings, the generated wrapper is named `R_wrap_<symbol>`.
Variadic bindings generate several wrapper variants and dispatch is chosen later
from R based on tail arity or inferred tail types.

This design keeps platform-specific calling conventions inside compiled C rather
than trying to reproduce them from R.

## Protection And Lifetime Rules Matter

Because wrappers use the R C API directly, protection and object lifetime are
part of the internal design.

When wrapper code allocates a fresh R object, it protects that object until the
result is fully built and returned. Typical cases include:

- array returns that allocate `out`
- `cstring` returns that construct an R string
- callback trampolines that build an argument list before calling back into R

Borrowed pointers have a different constraint: they are only sound as long as
the underlying owner stays alive and the wrapper does not invalidate the
assumption by introducing unexpected allocation patterns.

This is especially important for:

- zero-copy vector inputs
- borrowed field-address helpers for structs and unions
- callback token pointers that must remain tied to a live callback registry

The package also uses external pointer metadata and protected slots to encode
lifetime relationships. For example, borrowed field pointers can keep their
owner object alive by storing that owner in the external pointer's protected
field.

## Ownership And Lifetime Semantics In The Main Cases

The main internal cases are easier to reason about if you separate them by who
owns the underlying storage and how long the view is valid.

### Call-scoped borrows from R objects

These values are borrowed from existing R objects and are only intended to be
used during the wrapper call:

- `raw`, `integer_array`, `numeric_array`, and `logical_array` inputs borrow
  writable backing R vector storage; ALTREP inputs may be materialized by R on
  pointer access
- `cstring` input borrows the translated string pointer for the duration of the
  call
- `sexp` input borrows the original R object directly

The wrapper does not transfer ownership of these objects to C. If target C code
stores the pointer and uses it after the call returns, that is outside the safe
contract.

### Owned native allocations

These are heap allocations owned through explicit external-pointer semantics:

- `tcc_malloc()` returns `rtinycc_owned` memory with a finalizer
- `tcc_cstring()` returns a malloc-backed UTF-8 C string with the same owned tag
- generated struct and union constructors allocate native storage and attach
  type-specific finalizers

These objects have a stable native lifetime until:

- they are explicitly freed
- their owner-specific free helper is called
- or their finalizer runs during normal R lifetime

### Borrowed native views

These are external pointers that point into someone else's storage:

- `tcc_data_ptr()` returns a borrowed pointer
- field-address helpers for structs and unions return borrowed pointers
- many plain `ptr` returns are just raw addresses wrapped as external pointers

Borrowed pointers do not imply ownership and must not be freed as if they were
`rtinycc_owned`. Their validity depends entirely on the lifetime of the
underlying storage.

### Returned R objects

When the wrapper returns a scalar, string, or copied array to R, the result is
an ordinary R-managed object:

- scalar returns are fresh boxed R values
- `cstring` returns become fresh R strings
- array returns become fresh R vectors after copying

Once returned, these objects follow the normal R GC lifetime and are no longer
tied to the lifetime of the original C storage.

### Callback registry lifetime

Callbacks have a separate ownership model:

- the callback registry preserves the underlying R function
- callback tokens are external pointers referencing registry entries
- `tcc_callback_close()` releases the preserved function deterministically
- if not closed manually, finalizers and package unload eventually release it

This means the callback object is not just a function pointer. It is a managed
pairing of:

- preserved R function state
- callback metadata
- one or more external-pointer handles to the token

### Compiled object lifetime

A `tcc_compiled` object owns a live TCC state and the wrapper pointers recovered
from that state.

When that state dies, the wrapper pointers are dead as machine-code references
even though the R closures still exist. That is why the package stores a recipe
and recompiles instead of pretending those pointers survive serialization.

## Host Symbol Injection Happens Before Relocation

After the generated code is compiled, `tcc_ffi_compile_state()` calls the C
entry point `RC_libtcc_add_host_symbols()` before `tcc_relocate()`.

That host-injection step registers package-side C helpers with the live TCC
state. This matters most on macOS, where the package cannot rely on the dynamic
linker to expose every host symbol the same way TinyCC expects.

The injected symbols include:

- `RC_free_finalizer`
- callback invocation helpers
- async callback scheduling helpers
- async drain helpers
- the `RC_callback_async_exec_c()` helper used by generated async wrappers

The important semantic point is that some generated C code depends on package
runtime helpers, not just on user code and the R API.

## Callback Round-Trips Cross The Boundary Twice

Callbacks are the clearest example of value exchange between plain C and the R
interpreter.

For synchronous callbacks:

1. generated C trampoline code receives plain C arguments
2. the trampoline boxes them into a `VECSXP` argument list
3. it calls `RC_invoke_callback_id()`
4. the runtime builds and evaluates the R call with `R_tryEvalSilent()`
5. the result is converted back into the declared C return type
6. the trampoline returns that C value to the original compiled code

So a callback call is:

- C values -> boxed into R objects
- evaluated in R
- converted back from R objects -> C values

Async callbacks add one more layer: arguments are first marshaled into a
cross-thread task representation, then rebuilt as fresh R objects on the main
thread before the callback is evaluated.

## State Creation Is Separate from Compilation

The TCC state is created first, then populated and compiled.

Internally:

- `tcc_ffi_create_state()` creates the state with bundled TinyCC include/lib
  paths, user include/lib paths, and R headers/runtime library paths
- user compiler options are applied with `tcc_set_options()`
- `tcc_ffi_compile_state()` adds requested libraries, always links `R`,
  compiles the generated C string, injects host symbols, then relocates

This split is useful because both `tcc_compile()` and `tcc_link()` follow the
same broad pattern even though one starts from user C source and the other
starts from external-library declarations.

## The Compiled Object Is an Environment of Closures

After relocation, `tcc_compiled_object()` recovers wrapper symbols with
`tcc_get_symbol()` and turns them into R callables with `make_callable()`.

That compiled object is an environment, not an S4 class or external pointer
wrapper. The environment stores:

- callable closures for user symbols
- callable closures for generated helpers
- the live TCC state
- metadata such as symbol specs and helper specs

For non-variadic functions, `make_callable()` creates a closure that:

- checks arity
- checks that the wrapper pointer is still valid
- calls the wrapper through `.Call`

For variadic bindings, the closure selects the matching precompiled wrapper
first, then calls that wrapper pointer.

## Serialization Works by Recompiling the Recipe

Compiled wrapper pointers do not survive serialization as usable machine code.
`Rtinycc` handles this by storing the original recipe:

- `tcc_compile()` stores `.ffi` on the compiled object
- `tcc_link()` stores `.link_args`
- `$.tcc_compiled` checks whether the state pointer is still valid
- if not, `recompile_into()` rebuilds a fresh compiled object and copies the
  bindings back into the target environment

So serialization support is not pointer persistence. It is recipe persistence
plus transparent recompilation.

## Where To Read Next

If you want to inspect the implementation directly, the main files are:

- `R/ffi.R`: high-level FFI object, compilation flow, compiled-object assembly
- `R/ffi_codegen.R`: generated wrapper and translation-unit builders
- `R/aaa_ffi_codegen_rules.R`: rule tables for conversions and mapping
- `R/callbacks.R`: callback parsing and trampoline generation helpers
- `src/RC_libtcc.c`: TCC/R bridge, host symbol injection, callback runtime

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_output}
\alias{tcc_output}
\title{Set output type for FFI compilation}
\usage{
tcc_output(ffi, output = c("memory", "dll", "exe"))
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{output}{One of "memory", "dll", "exe"}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Set output type for FFI compilation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_union}
\alias{tcc_union}
\title{Declare union for FFI helper generation}
\usage{
tcc_union(ffi, name, members, active = NULL)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{name}{Union name (as defined in C header)}

\item{members}{Named list of union members with FFI types}

\item{active}{Default active member for accessors}
}
\value{
Updated tcc_ffi object
}
\description{
Generate R-callable helpers for union allocation and member access.
The union must be defined in a header.
}
\examples{
\dontrun{
ffi <- tcc_ffi() |>
  tcc_union("data_variant",
    members = list(as_int = "i32", as_float = "f32"),
    active = "as_int"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_cstring}
\alias{tcc_read_cstring}
\title{Read C-style string from pointer}
\usage{
tcc_read_cstring(
  ptr,
  max_bytes = NULL,
  null_action = c("na", "empty", "error")
)
}
\arguments{
\item{ptr}{External pointer to C string}

\item{max_bytes}{Optional maximum number of bytes to read (fixed-length read).}

\item{null_action}{Behavior when ptr is NULL: one of "na", "empty", "error".
Only effective when \code{max_bytes} is provided; without it a NULL pointer
returns \code{""}.}
}
\value{
Character string, or \code{NA}/\code{""} for NULL pointers depending on
\code{null_action}.
}
\description{
Convert a C-style null-terminated string pointer back to R character string.
Handles UTF-8 decoding automatically.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_null_ptr}
\alias{tcc_null_ptr}
\title{Create a NULL pointer}
\usage{
tcc_null_ptr()
}
\value{
An external pointer with NULL address
}
\description{
Creates a NULL pointer equivalent for use with FFI functions that expect
optional pointer arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_field_addr}
\alias{tcc_field_addr}
\title{Generate field address getter helpers}
\usage{
tcc_field_addr(ffi, struct_name, fields)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{struct_name}{Struct name}

\item{fields}{Character vector of field names}
}
\value{
Updated tcc_ffi object
}
\description{
Creates functions that return pointers to specific struct fields.
Useful for passing field pointers to C functions or for container_of.
}
\examples{
\dontrun{
ffi <- tcc_ffi() |>
  tcc_struct("point", list(x = "f64", y = "f64")) |>
  tcc_field_addr("point", c("x", "y")) # point_x_addr(), point_y_addr()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_library_path}
\alias{tcc_library_path}
\title{Add library path to FFI context}
\usage{
tcc_library_path(ffi, path)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{path}{Library directory path}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Add library path to FFI context
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_symbol_is_valid}
\alias{tcc_symbol_is_valid}
\title{Check if a tcc_symbol external pointer is valid}
\usage{
tcc_symbol_is_valid(ptr)
}
\arguments{
\item{ptr}{External pointer from \code{tcc_get_symbol()}.}
}
\value{
TRUE if the pointer address is non-null, FALSE otherwise.
}
\description{
Check if a tcc_symbol external pointer is valid
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_generate_bindings}
\alias{tcc_generate_bindings}
\title{Generate bindings from header declarations}
\usage{
tcc_generate_bindings(
  ffi = NULL,
  header,
  mapper = tcc_map_c_type_to_ffi,
  functions = TRUE,
  structs = TRUE,
  unions = TRUE,
  enums = TRUE,
  globals = TRUE,
  bitfield_type = "u8",
  include_bitfields = TRUE
)
}
\arguments{
\item{ffi}{A tcc_ffi object. If NULL, a new one is created.}

\item{header}{Character scalar containing C declarations.}

\item{mapper}{Function to map C types to FFI types.}

\item{functions}{Logical; generate \code{tcc_bind()} specs for functions.}

\item{structs}{Logical; generate \code{tcc_struct()} helpers.}

\item{unions}{Logical; generate \code{tcc_union()} helpers.}

\item{enums}{Logical; generate \code{tcc_enum()} helpers.}

\item{globals}{Logical; generate \code{tcc_global()} getters/setters.}

\item{bitfield_type}{FFI type to use for bitfields.}

\item{include_bitfields}{Whether to include bitfields.}
}
\value{
Updated tcc_ffi object.
}
\description{
Generate bindings from header declarations
}
\examples{
\dontrun{
header <- "double sqrt(double x); struct point { double x; double y; };"
ffi <- tcc_generate_bindings(tcc_ffi(), header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_run_cli}
\alias{tcc_run_cli}
\title{Run the tinycc CLI}
\usage{
tcc_run_cli(args = character(), tcc_path = check_cli_exists())
}
\arguments{
\item{args}{Character vector of CLI arguments (e.g., \code{c("-c", file, "-o", out)}).}

\item{tcc_path}{Optional path to the \code{tcc} binary; defaults to the bundled CLI.}
}
\value{
Integer status from \code{system2()}.
}
\description{
Run the tinycc CLI
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_helpers}
\alias{tcc_treesitter_helpers}
\title{Treesitter helpers for header-driven bindings}
\arguments{
\item{header}{Character scalar containing C declarations.}

\item{...}{Additional arguments passed to treesitter.c helpers.}
}
\value{
A data frame of nodes or a named list of binding specs.
}
\description{
Convenience wrappers around \code{treesitter.c} that map C types to the
Rtinycc FFI type system and return \code{tcc_bind()}-ready specs.
}
\section{treesitter.c}{

These helpers require the optional treesitter.c package.
}

\examples{
\dontrun{
header <- "double sqrt(double x);\nint add(int a, int b);"
tcc_treesitter_functions(header)
tcc_treesitter_bindings(header)
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_write_bytes}
\alias{tcc_write_bytes}
\title{Write raw bytes to a pointer}
\usage{
tcc_write_bytes(ptr, raw)
}
\arguments{
\item{ptr}{External pointer}

\item{raw}{Raw vector to write}
}
\value{
\code{NULL}.
}
\description{
Write a raw vector into memory pointed to by an external pointer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_get_symbol}
\alias{tcc_get_symbol}
\title{Get a symbol pointer from a libtcc state}
\usage{
tcc_get_symbol(state, name)
}
\arguments{
\item{state}{A \code{tcc_state}.}

\item{name}{Symbol name to look up.}
}
\value{
External pointer of class \code{tcc_symbol}.
}
\description{
Get a symbol pointer from a libtcc state
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_library}
\alias{tcc_library}
\title{Add library to link against}
\usage{
tcc_library(ffi, library)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{library}{Library name (e.g., "m", "sqlite3") or a path to a
shared library (e.g., "libm.so.6"). When a path or platform library file
name is provided, the library directory is added automatically and TinyCC
is asked to link that exact file name. This keeps versioned runtime
libraries such as \code{libm.so.6} distinct from generic linker names such as
\code{m}/\code{libm.so}.}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Add library to link against
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_set_options}
\alias{tcc_set_options}
\title{Apply raw TinyCC options to a libtcc state}
\usage{
tcc_set_options(state, options)
}
\arguments{
\item{state}{A \code{tcc_state}.}

\item{options}{Character scalar of options (for example \code{"-O2 -Wall"}).}
}
\value{
Integer status code (\code{0} on success; negative on parse error).
}
\description{
Passes options directly to \code{tcc_set_options()} for the given state.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{is_callback_type}
\alias{is_callback_type}
\title{Check if a type represents a callback}
\usage{
is_callback_type(type)
}
\arguments{
\item{type}{Type string to check}
}
\value{
Logical
}
\description{
Check if a type represents a callback
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_global}
\alias{tcc_global}
\title{Declare a global variable getter}
\usage{
tcc_global(ffi, name, type)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{name}{Global symbol name}

\item{type}{FFI type for the global (scalar types only)}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Register a global C symbol so the compiled object exposes getter/setter
functions \verb{global_<name>_get()} and \verb{global_<name>_set()}.
}
\details{
Globals are limited to scalar FFI types. Array types are rejected.
Scalar conversions follow the same rules as wrapper arguments:
\itemize{
\item Integer inputs (\code{i8}, \code{i16}, \code{i32}, \code{u8}, \code{u16}) must be finite and
within range; \code{NA} values error.
\item Large integer types (\code{i64}, \code{u32}, \code{u64}) are mediated through R numeric
(double). Values must be integer-valued and within range; for \code{i64}/\code{u64}
only exact integers up to $2^53$ are accepted.
\item Getter wrappers for \code{i64}/\code{u64} warn when the stored value exceeds R's
exact integer range for numeric vectors.
\item \code{bool} rejects \code{NA} logicals.
}

Ownership notes:
\itemize{
\item \code{ptr} globals store the raw address from an external pointer. If the
external pointer owns memory, keep it alive; otherwise the pointer may
be freed while the global still points to it.
\item \code{cstring} globals store a borrowed pointer to R's string data
(UTF-8 translation). Do not free it; for C-owned strings prefer a \code{ptr}
global and manage lifetime explicitly (e.g., with \code{tcc_cstring()}).
}
}
\note{
Global helpers are generated inside the compiled TCC unit. Recompiling
creates a new instance of the global variable; existing compiled objects
continue to refer to their own copy.
}
\examples{
ffi <- tcc_ffi() |>
  tcc_source("int global_counter = 7;") |>
  tcc_global("global_counter", "i32") |>
  tcc_compile()
ffi$global_global_counter_get()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_include_paths}
\alias{tcc_include_paths}
\alias{tcc_sysinclude_paths}
\title{TinyCC include search paths}
\usage{
tcc_include_paths()

tcc_sysinclude_paths()
}
\value{
A character vector of include directories.
}
\description{
Returns the include directories used by the bundled TinyCC (top-level include and lib/tcc/include).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{.RtinyccCall}
\alias{.RtinyccCall}
\title{we are using .Call directly, this is to make R CMD check happy}
\usage{
.RtinyccCall()
}
\value{
The result returned by the invoked native routine. This alias is
package-internal and exists to satisfy \verb{R CMD check} native routine
registration requirements.
}
\description{
we are using .Call directly, this is to make R CMD check happy
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_bind}
\alias{tcc_bind}
\title{Bind symbols with type specifications}
\usage{
tcc_bind(ffi, ...)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{...}{Named list of symbol definitions. Each definition is a list with:
\itemize{
\item args: List of fixed FFI argument types (e.g., list("i32", "f64"))
\item returns: FFI type for return value (e.g., "f64", "cstring")
\item variadic: Set TRUE for C varargs functions
\item varargs: Legacy typed variadic tail (exact/prefix mode)
\item varargs_types: Allowed scalar FFI types for true variadic tails
\item varargs_min: Minimum number of trailing varargs
\item varargs_max: Maximum number of trailing varargs (required for
true variadic mode, defaults to \code{varargs_min})
\item code: Optional C code for the symbol (for embedded functions)
}
Callback arguments should use the form \code{callback:<signature>} (e.g.,
\code{callback:double(double)}). The generated trampoline expects a
\code{tcc_callback_ptr(cb)} to the corresponding user-data parameter in
the C API. For thread-safe scheduling, use
\code{callback_async:<signature>} which enqueues the call on the main
thread and returns a default value immediately.}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Define symbols with Bun-style type specifications for API mode.
This is the core of the declarative FFI API.
}
\examples{
ffi <- tcc_ffi() |>
  tcc_bind(
    add = list(args = list("i32", "i32"), returns = "i32"),
    greet = list(args = list("cstring"), returns = "cstring")
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_ptr}
\alias{tcc_read_ptr}
\title{Read a pointer at byte offset}
\usage{
tcc_read_ptr(ptr, offset = 0L)
}
\arguments{
\item{ptr}{External pointer}

\item{offset}{Byte offset from \code{ptr} (scalar interface). Ignored when
\code{n} is supplied.}
}
\value{
External pointer
}
\description{
Dereferences a \verb{void*} at the given byte offset from \code{ptr}.
Equivalent to \verb{*(void**)(ptr + offset)}. The returned pointer
is tagged \code{"rtinycc_borrowed"} and will not be freed by the
garbage collector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_map_c_type_to_ffi}
\alias{tcc_map_c_type_to_ffi}
\title{Map a C type string to an Rtinycc FFI type}
\usage{
tcc_map_c_type_to_ffi(c_type)
}
\arguments{
\item{c_type}{C type string (e.g., "int", "double", "char *").}
}
\value{
A single FFI type string.
}
\description{
Map a C type string to an Rtinycc FFI type
}
\details{
This mapper is intentionally conservative for pointer types. \verb{char*} is
treated as a raw pointer (\code{ptr}) because C does not guarantee NUL-terminated
strings. If you know the API expects a C string, map it explicitly to
\code{cstring} in your custom mapper.
}
\examples{
\dontrun{
tcc_map_c_type_to_ffi("int")
tcc_map_c_type_to_ffi("const char *")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_compile_string}
\alias{tcc_compile_string}
\title{Compile C code from a character string}
\usage{
tcc_compile_string(state, code)
}
\arguments{
\item{state}{A \code{tcc_state}.}

\item{code}{C source code string.}
}
\value{
Integer status code (0 = success).
}
\description{
Compile C code from a character string
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_ptr_set}
\alias{tcc_ptr_set}
\title{Set a pointer-to-pointer value}
\usage{
tcc_ptr_set(ptr_ref, ptr_value)
}
\arguments{
\item{ptr_ref}{External pointer to a pointer value (e.g., address of a field).}

\item{ptr_value}{External pointer to store.}
}
\value{
The updated pointer reference (invisibly).
}
\description{
Assigns the address in \code{ptr_value} to the location pointed to by \code{ptr_ref}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_header}
\alias{tcc_header}
\title{Add C headers}
\usage{
tcc_header(ffi, header)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{header}{Header string or include directive}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Add C headers
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_enum}
\alias{tcc_enum}
\title{Declare enum for FFI helper generation}
\usage{
tcc_enum(ffi, name, constants = NULL, export_constants = FALSE)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{name}{Enum name (as defined in C header)}

\item{constants}{Character vector of constant names to export}

\item{export_constants}{Whether to export enum constants as R functions}
}
\value{
Updated tcc_ffi object
}
\description{
Generate R-callable helpers for enum constants and type conversions.
The enum must be defined in a header.
}
\examples{
\dontrun{
ffi <- tcc_ffi() |>
  tcc_header("#include <errors.h>") |>
  tcc_enum("error_code", constants = c("OK", "ERROR"), export_constants = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_call_symbol}
\alias{tcc_call_symbol}
\title{Call a zero-argument symbol with a specified return type}
\usage{
tcc_call_symbol(state, name, return = c("int", "double", "void"))
}
\arguments{
\item{state}{A \code{tcc_state}.}

\item{name}{Symbol name to call.}

\item{return}{One of "int", "double", "void".}
}
\value{
The return value cast to the requested type (NULL for void).
}
\description{
Call a zero-argument symbol with a specified return type
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_container_of}
\alias{tcc_container_of}
\title{Generate container_of helper for struct member}
\usage{
tcc_container_of(ffi, struct_name, member_name)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{struct_name}{Struct name}

\item{member_name}{Member field name to compute offset from}
}
\value{
Updated tcc_ffi object
}
\description{
Creates a function that recovers the parent struct pointer from
a pointer to one of its members. This is the classic Linux kernel
container_of macro made accessible from R.
}
\examples{
\dontrun{
ffi <- tcc_ffi() |>
  tcc_struct("student", list(id = "i32", marks = "i32")) |>
  tcc_container_of("student", "marks") # Creates struct_student_from_marks()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_write_ptr}
\alias{tcc_write_ptr}
\title{Write a pointer at byte offset}
\usage{
tcc_write_ptr(ptr, offset, value)
}
\arguments{
\item{ptr}{External pointer (destination buffer)}

\item{offset}{Byte offset}

\item{value}{External pointer to write}
}
\value{
\code{NULL} (invisibly).
}
\description{
Write a pointer at byte offset
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_union_accessors}
\alias{tcc_treesitter_union_accessors}
\title{Generate tcc_union() accessors from header unions}
\usage{
tcc_treesitter_union_accessors(
  header,
  mapper = tcc_map_c_type_to_ffi,
  bitfield_type = "u8",
  include_bitfields = TRUE
)
}
\arguments{
\item{header}{Character scalar containing C declarations.}

\item{mapper}{Function to map C types to FFI types.}

\item{bitfield_type}{FFI type to use for bitfields.}

\item{include_bitfields}{Whether to include bitfields.}
}
\value{
Named list of accessors by union name. Bitfields are returned as
lists with \code{type}, \code{bitfield = TRUE}, and \code{width}. Nested struct members
are returned as \verb{list(type = "struct", struct_name = <name>)} when the
struct name is available, otherwise \code{list(type = "struct")}.
}
\description{
Generate tcc_union() accessors from header unions
}
\examples{
\dontrun{
header <- "union data { int i; double d; };"
tcc_treesitter_union_accessors(header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_struct}
\alias{tcc_struct}
\title{Declare struct for FFI helper generation}
\usage{
tcc_struct(ffi, name, accessors)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{name}{Struct name (as defined in C header)}

\item{accessors}{Named list of field accessors where
names are field names and values are FFI types (e.g., list(x="f64", y="f64")).
Named nested struct fields can use \code{"struct:<name>"} to generate borrowed
nested-view getters and copy-in setters (for example \code{child = "struct:child"}).}
}
\value{
Updated tcc_ffi object
}
\description{
Generate R-callable helpers for struct allocation, field access,
and pointer management. The struct must be defined in a header.
}
\examples{
\dontrun{
ffi <- tcc_ffi() |>
  tcc_header("#include <point.h>") |>
  tcc_struct("point", list(x = "f64", y = "f64", id = "i32"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_link}
\alias{tcc_link}
\title{Link an external shared library with Bun-style FFI bindings}
\usage{
tcc_link(
  path,
  symbols,
  headers = NULL,
  libs = character(0),
  lib_paths = character(0),
  include_paths = character(0),
  user_code = NULL,
  verbose = FALSE
)
}
\arguments{
\item{path}{Library short name (e.g., \code{"m"}, \code{"sqlite3"}) or full path to
the shared library. Short names stay on the normal linker-name path
(\verb{-l<name>}). File names such as \code{libm.so.6} or full paths are resolved
through the configured library search paths when needed and linked as exact
files rather than collapsed to generic names like \code{m}.}

\item{symbols}{Named list of symbol definitions with:
\itemize{
\item args: List of FFI types for arguments
\item returns: FFI type for return value
}}

\item{headers}{Optional C headers to include}

\item{libs}{Library names to link (e.g., "sqlite3")}

\item{lib_paths}{Additional library search paths}

\item{include_paths}{Additional include search paths}

\item{user_code}{Optional custom C code to include in the compilation}

\item{verbose}{Print debug information}
}
\value{
A tcc_compiled object with callable functions
}
\description{
Link a system library (like libsqlite3) and generate type-safe
wrappers automatically using TinyCC JIT compilation (API mode).
Unlike dlopen(), this uses TinyCC to compile bindings that handle
type conversion between R and C automatically.
}
\examples{
\dontrun{
# Link SQLite with type-safe bindings
sqlite <- tcc_link(
  "libsqlite3.so",
  symbols = list(
    sqlite3_libversion = list(args = list(), returns = "cstring"),
    sqlite3_open = list(args = list("cstring", "ptr"), returns = "i32")
  ),
  libs = "sqlite3"
)

# Call directly - type conversion happens automatically
sqlite$sqlite3_libversion()

# Example with custom user code for helper functions
math_with_helpers <- tcc_link(
  "m",
  symbols = list(
    sqrt = list(args = list("f64"), returns = "f64"),
    safe_sqrt = list(args = list("f64"), returns = "f64")
  ),
  user_code = "
    #include <math.h>

    // Helper function that validates input before calling sqrt
    double safe_sqrt(double x) {
      if (x < 0) {
        return NAN;
      }
      return sqrt(x);
    }
  ",
  libs = "m"
)
math_with_helpers$safe_sqrt(16.0)
math_with_helpers$safe_sqrt(-4.0) # Returns NaN for negative input
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_enum_members}
\alias{tcc_treesitter_enum_members}
\title{Parse enum members with treesitter.c}
\usage{
tcc_treesitter_enum_members(header, ...)
}
\arguments{
\item{header}{Character scalar containing C declarations.}

\item{...}{Additional arguments passed to \code{treesitter.c::get_enum_members_from_root()}.}
}
\value{
A data frame of enum members.
}
\description{
Parse enum members with treesitter.c
}
\examples{
\dontrun{
header <- "enum status { OK = 0, ERR = 1 };"
tcc_treesitter_enum_members(header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_struct_accessors}
\alias{tcc_treesitter_struct_accessors}
\title{Generate tcc_struct() accessors from header structs}
\usage{
tcc_treesitter_struct_accessors(
  header,
  mapper = tcc_map_c_type_to_ffi,
  bitfield_type = "u8",
  include_bitfields = TRUE
)
}
\arguments{
\item{header}{Character scalar containing C declarations.}

\item{mapper}{Function to map C types to FFI types.}

\item{bitfield_type}{FFI type to use for bitfields.}

\item{include_bitfields}{Whether to include bitfields.}
}
\value{
Named list of accessors by struct name. Bitfields are returned as
lists with \code{type}, \code{bitfield = TRUE}, and \code{width}. Named nested struct
fields are returned as \code{"struct:<name>"}; ambiguous or anonymous nested
structs fall back to \code{"ptr"}.
}
\description{
Generate tcc_struct() accessors from header structs
}
\examples{
\dontrun{
header <- "struct point { double x; double y; };"
tcc_treesitter_struct_accessors(header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_relocate}
\alias{tcc_relocate}
\title{Relocate compiled code}
\usage{
tcc_relocate(state)
}
\arguments{
\item{state}{A \code{tcc_state}.}
}
\value{
Integer status code (0 = success).
}
\description{
Relocate compiled code
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_struct_raw_access}
\alias{tcc_struct_raw_access}
\title{Enable raw byte access for struct}
\usage{
tcc_struct_raw_access(ffi, struct_name)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{struct_name}{Struct name}
}
\value{
Updated tcc_ffi object
}
\description{
Generates helper functions to read/write raw bytes from struct memory.
Useful for bitwise operations, debugging, or manual serialization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_prefix}
\alias{tcc_prefix}
\alias{tcc_lib_path}
\alias{tcc_lib_paths}
\alias{tcc_include_path}
\alias{tcc_bin_path}
\alias{tcc_cli}
\title{TinyCC paths}
\usage{
tcc_prefix()

tcc_lib_path()

tcc_lib_paths()

tcc_include_path()

tcc_bin_path()

tcc_cli()
}
\value{
A character scalar path.
}
\description{
Helpers to locate the bundled tinycc installation after the package is installed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_state}
\alias{tcc_state}
\title{Create a libtcc state}
\usage{
tcc_state(
  output = c("memory", "obj", "dll", "exe", "preprocess"),
  include_path = tcc_include_paths(),
  lib_path = tcc_lib_paths()
)
}
\arguments{
\item{output}{Output type: one of "memory", "obj", "dll", "exe", "preprocess".}

\item{include_path}{Path(s) to headers; defaults to the bundled include dirs.}

\item{lib_path}{Path(s) to libraries; defaults to the bundled lib dirs (lib and lib/tcc).}
}
\value{
An external pointer of class \code{tcc_state}.
}
\description{
Initialize a libtcc compilation state, optionally pointing at the bundled include/lib paths.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{as.character.tcc_cstring}
\alias{as.character.tcc_cstring}
\title{Convert a \code{tcc_cstring} object to an R string}
\usage{
\method{as.character}{tcc_cstring}(x, ...)
}
\arguments{
\item{x}{A \code{tcc_cstring} object.}

\item{...}{Ignored.}
}
\value{
A character scalar containing the string value. Returns the
cached R copy when available; otherwise reads the current NUL-terminated
C string from \code{x$ptr}.
}
\description{
Convert a \code{tcc_cstring} object to an R string
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_struct_bindings}
\alias{tcc_treesitter_struct_bindings}
\title{Apply tcc_struct() bindings from a header}
\usage{
tcc_treesitter_struct_bindings(ffi, header, ...)
}
\arguments{
\item{ffi}{A tcc_ffi object.}

\item{header}{Character scalar containing C declarations.}

\item{...}{Passed to \code{tcc_treesitter_struct_accessors()}.}
}
\value{
Updated tcc_ffi object.
}
\description{
Apply tcc_struct() bindings from a header
}
\examples{
\dontrun{
header <- "struct point { double x; double y; };"
ffi <- tcc_ffi()
ffi <- tcc_treesitter_struct_bindings(ffi, header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_struct_members}
\alias{tcc_treesitter_struct_members}
\title{Parse struct members (including bitfields) with treesitter.c}
\usage{
tcc_treesitter_struct_members(header, ...)
}
\arguments{
\item{header}{Character scalar containing C declarations.}

\item{...}{Additional arguments passed to \code{treesitter.c::get_struct_members()}.}
}
\value{
A data frame of struct members.
}
\description{
Parse struct members (including bitfields) with treesitter.c
}
\examples{
\dontrun{
header <- "struct point { double x; double y; };"
tcc_treesitter_struct_members(header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{print.tcc_compiled}
\alias{print.tcc_compiled}
\title{Print tcc_compiled object}
\usage{
\method{print}{tcc_compiled}(x, ...)
}
\arguments{
\item{x}{A tcc_compiled object}

\item{...}{Ignored}
}
\value{
The input \code{tcc_compiled} object, invisibly. Called for its side
effect of printing the compilation output mode and the status of
compiled callable symbols.
}
\description{
Print tcc_compiled object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_source}
\alias{tcc_source}
\title{Add C source code}
\usage{
tcc_source(ffi, code)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{code}{C source code string}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Add C source code
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_ptr_addr}
\alias{tcc_ptr_addr}
\title{Get pointer address as integer}
\usage{
tcc_ptr_addr(ptr, hex = FALSE)
}
\arguments{
\item{ptr}{External pointer}

\item{hex}{Whether to display in hexadecimal (default: FALSE)}
}
\value{
Character representation of pointer address (hex if requested, decimal otherwise)
}
\description{
Get the numeric address of an external pointer, useful for debugging
and when APIs require pointer addresses as integers. Optional hex mode available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_ptr_is_owned}
\alias{tcc_ptr_is_owned}
\title{Check for the \code{"rtinycc_owned"} tag}
\usage{
tcc_ptr_is_owned(ptr)
}
\arguments{
\item{ptr}{External pointer}
}
\value{
Logical scalar
}
\description{
Returns \code{TRUE} only for pointers created by \code{\link[=tcc_malloc]{tcc_malloc()}} or
\code{\link[=tcc_cstring]{tcc_cstring()}}. Struct pointers (tagged \code{"struct_<name>"}) and
borrowed pointers return \code{FALSE}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_i32}
\alias{tcc_read_i32}
\title{Read signed 32-bit integers from a pointer}
\usage{
tcc_read_i32(ptr, n = NULL, offset = 0L)
}
\arguments{
\item{ptr}{External pointer}

\item{n}{Number of values to read (legacy vectorised interface).
If provided, reads \code{n} consecutive i32 values starting at byte 0.}

\item{offset}{Byte offset from \code{ptr} (scalar interface). Ignored when
\code{n} is supplied.}
}
\value{
Integer scalar (offset form) or integer vector (n form).
}
\description{
Read signed 32-bit integers from a pointer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_u8}
\alias{tcc_read_u8}
\title{Read unsigned 8-bit values from a pointer}
\usage{
tcc_read_u8(ptr, n = NULL, offset = 0L)
}
\arguments{
\item{ptr}{External pointer}

\item{n}{Number of values to read (legacy vectorised interface).
If provided, reads \code{n} consecutive u8 values starting at byte 0.}

\item{offset}{Byte offset from \code{ptr} (scalar interface). Ignored when
\code{n} is supplied.}
}
\value{
Integer scalar (offset form) or integer vector (n form).
}
\description{
Read unsigned 8-bit values from a pointer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{tcc_callback_close}
\alias{tcc_callback_close}
\title{Close/unregister a callback}
\usage{
tcc_callback_close(callback)
}
\arguments{
\item{callback}{A tcc_callback object returned by tcc_callback()}
}
\value{
NULL (invisible)
}
\description{
Invalidates a callback immediately, releases the preserved R function
reference, and cleans up callback resources as early as possible.
This is recommended for deterministic lifetime management, but callbacks
are also eventually released by finalizers if you simply drop all references.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_cstring_object}
\alias{tcc_cstring_object}
\title{CString S3 Class}
\usage{
tcc_cstring_object(ptr, clone = TRUE, owned = FALSE)
}
\arguments{
\item{ptr}{External pointer to C string}

\item{clone}{Whether to clone the string immediately (safe for R use)}

\item{owned}{Currently unused. Reserved for future finalizer support.}
}
\value{
A tcc_cstring object
}
\description{
Wrapper around a C string pointer with an optional cached R copy.
Ownership follows the underlying external pointer; this wrapper does not add
finalizer or freeing behavior on top of that pointer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{get_external_ptr_addr}
\alias{get_external_ptr_addr}
\title{Get the address of an external pointer}
\usage{
get_external_ptr_addr(ptr)
}
\arguments{
\item{ptr}{An external pointer object (e.g., from \code{tcc_get_symbol()}).}
}
\value{
The memory address as a numeric value.
}
\description{
Extract the memory address from an external pointer as a numeric value.
This is primarily useful for debugging and inspection purposes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_malloc}
\alias{tcc_malloc}
\title{Allocate memory buffer}
\usage{
tcc_malloc(size)
}
\arguments{
\item{size}{Number of bytes to allocate}
}
\value{
An external pointer tagged \code{"rtinycc_owned"} with an R finalizer.
Freed on garbage collection or explicitly via \code{\link[=tcc_free]{tcc_free()}}.
}
\description{
Allocate a memory buffer of specified size, equivalent to C malloc.
Returns an external pointer that can be passed to FFI functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_ffi}
\alias{tcc_ffi}
\title{Create a new FFI compilation context}
\usage{
tcc_ffi()
}
\value{
A tcc_ffi object with chaining support
}
\description{
Initialize a Bun-style FFI context for API-mode compilation.
This is the entry point for the modern FFI API.
}
\examples{
ffi <- tcc_ffi()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{tcc_callback_async_drain}
\alias{tcc_callback_async_drain}
\title{Drain the async callback queue}
\usage{
tcc_callback_async_drain()
}
\value{
NULL (invisible)
}
\description{
Execute any pending async callbacks immediately on the main R thread.
Normally callbacks fire automatically via R's event loop (input
handler on POSIX, message pump on Windows), so explicit draining is
only needed in test harnesses or tight batch loops that never yield
to R's event loop.
}
\details{
TCC-compiled C code running on the main thread can call
\code{RC_callback_async_drain_c()} directly instead of returning to R.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_i16}
\alias{tcc_read_i16}
\title{Read signed 16-bit integer}
\usage{
tcc_read_i16(ptr, offset = 0L)
}
\arguments{
\item{ptr}{External pointer}

\item{offset}{Byte offset from \code{ptr} (scalar interface). Ignored when
\code{n} is supplied.}
}
\value{
Integer scalar
}
\description{
Read signed 16-bit integer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_data_ptr}
\alias{tcc_data_ptr}
\title{Dereference a pointer-to-pointer}
\usage{
tcc_data_ptr(ptr_ref)
}
\arguments{
\item{ptr_ref}{External pointer to a pointer value (e.g., address of a field).}
}
\value{
An external pointer tagged \code{"rtinycc_borrowed"}. Not owned by
Rtinycc and never freed on garbage collection. Do not pass to
\code{\link[=tcc_free]{tcc_free()}}.
}
\description{
Treats \code{ptr_ref} as a pointer to a pointer and returns the pointed address
as an external pointer. This is useful for fields like \verb{void**} or \verb{T**}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_ptr_utils}
\alias{tcc_ptr_utils}
\title{Pointer and Buffer Utilities for FFI}
\value{
No return value. This is a documentation topic that groups the
pointer and buffer utility functions described below.
}
\description{
Generic helper functions for common FFI operations inspired by Bun's FFI.
These utilities handle pointer creation, buffer management, and memory operations
that are commonly needed when working with external libraries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{print.tcc_ffi}
\alias{print.tcc_ffi}
\title{Print tcc_ffi object}
\usage{
\method{print}{tcc_ffi}(x, ...)
}
\arguments{
\item{x}{A tcc_ffi object}

\item{...}{Ignored}
}
\value{
The input \code{tcc_ffi} object, invisibly. Called for its side
effect of printing the configured output mode, registered symbols,
and selected libraries/include paths.
}
\description{
Print tcc_ffi object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_cstring}
\alias{tcc_cstring}
\title{Create a C-style string pointer}
\usage{
tcc_cstring(str)
}
\arguments{
\item{str}{Character string}
}
\value{
An external pointer tagged \code{"rtinycc_owned"} pointing to a
malloc'd copy of the string. Freed on garbage collection or via
\code{\link[=tcc_free]{tcc_free()}}.
}
\description{
Convert R character strings to C-style null-terminated string pointers.
This handles UTF-8 encoding and null termination automatically.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{tcc_callback_ptr}
\alias{tcc_callback_ptr}
\title{Get the C-compatible function pointer}
\usage{
tcc_callback_ptr(callback)
}
\arguments{
\item{callback}{A tcc_callback object}
}
\value{
An external pointer (address of the callback token)
}
\description{
Returns an external pointer that can be passed to compiled C code
as user data for trampolines. Keep this handle (and the original
\code{tcc_callback}) alive for as long as C may call back.
The pointer handle keeps the underlying token storage alive until it is
garbage collected. Closing the original callback still invalidates the
callback registry entry, so C must not continue invoking it after
\code{tcc_callback_close()}.
}
\details{
Pointer arguments and return values are treated as external pointers.
Use \code{tcc_read_bytes()}, \code{tcc_read_u8()}, or \code{tcc_read_f64()}
to inspect pointed data when needed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_enum_bindings}
\alias{tcc_treesitter_enum_bindings}
\title{Apply tcc_enum() bindings from a header}
\usage{
tcc_treesitter_enum_bindings(ffi, header, constants = NULL)
}
\arguments{
\item{ffi}{A tcc_ffi object.}

\item{header}{Character scalar containing C declarations.}

\item{constants}{Named list of enum constants.}
}
\value{
Updated tcc_ffi object.
}
\description{
Apply tcc_enum() bindings from a header
}
\examples{
\dontrun{
header <- "enum status { OK = 0, ERR = 1 };"
ffi <- tcc_ffi()
ffi <- tcc_treesitter_enum_bindings(ffi, header, constants = list(status = c("OK", "ERR")))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_options}
\alias{tcc_options}
\title{Add TinyCC compiler options to FFI context}
\usage{
tcc_options(ffi, options)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{options}{Character vector of option fragments}
}
\value{
Updated tcc_ffi object (for chaining)
}
\description{
Append raw options that are passed to \code{tcc_set_options()} before compiling
generated wrappers (for example \code{"-O2"} or \code{"-Wall"}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{generate_trampoline}
\alias{generate_trampoline}
\title{Generate trampoline code for a callback argument}
\usage{
generate_trampoline(trampoline_name, sig)
}
\arguments{
\item{trampoline_name}{Name of the trampoline function}

\item{sig}{Parsed signature}
}
\value{
C code string for the trampoline
}
\description{
Generate trampoline code for a callback argument
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{blas_lapack_info}
\alias{blas_lapack_info}
\title{Report active BLAS/LAPACK runtime information from R}
\usage{
blas_lapack_info()
}
\value{
A named list with fields:
\code{blas_path}, \code{lapack_path}, \code{has_rblas}, \code{has_rlapack}, \code{loaded_dlls}.
}
\description{
Returns the BLAS/LAPACK runtime details as reported by R itself, plus
convenience flags indicating whether \code{Rblas} and \code{Rlapack} appear available
in loaded DLLs/shared objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_introspect}
\alias{tcc_introspect}
\title{Enable introspection helpers}
\usage{
tcc_introspect(ffi)
}
\arguments{
\item{ffi}{A tcc_ffi object}
}
\value{
Updated tcc_ffi object
}
\description{
Generates sizeof, alignof, and offsetof helper functions for
structs, unions, and enums. Useful for debugging or when you need
to know C layout information from R.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_compile}
\alias{tcc_compile}
\title{Compile FFI bindings}
\usage{
tcc_compile(ffi, verbose = FALSE)
}
\arguments{
\item{ffi}{A tcc_ffi object}

\item{verbose}{Print compilation info}
}
\value{
A tcc_compiled object with callable functions
}
\description{
Compile the defined symbols into callable functions.
This generates C wrapper code and compiles it with TinyCC.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{tcc_recompile}
\alias{tcc_recompile}
\title{Recompile a tcc_compiled object}
\usage{
tcc_recompile(compiled)
}
\arguments{
\item{compiled}{A tcc_compiled object}
}
\value{
The recompiled tcc_compiled object (invisibly, same environment)
}
\description{
Explicitly recompile from the stored FFI recipe.
Useful after deserialization (\code{readRDS}, \code{unserialize})
or to force a fresh compilation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_unions}
\alias{tcc_treesitter_unions}
\title{Parse union declarations with treesitter.c}
\usage{
tcc_treesitter_unions(header, ...)
}
\arguments{
\item{header}{Character scalar containing C declarations.}

\item{...}{Additional arguments passed to \code{treesitter.c::get_union_nodes()}.}
}
\value{
A data frame of union nodes.
}
\description{
Parse union declarations with treesitter.c
}
\examples{
\dontrun{
header <- "union data { int i; double d; };"
tcc_treesitter_unions(header)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{tcc_callback_async_schedule}
\alias{tcc_callback_async_schedule}
\title{Schedule a callback to run on the main thread}
\usage{
tcc_callback_async_schedule(callback, args = list())
}
\arguments{
\item{callback}{A tcc_callback object}

\item{args}{List of arguments to pass to the callback}
}
\value{
NULL (invisible)
}
\description{
Enqueue a callback for main-thread execution. Arguments must be basic
scalars or external pointers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_path}
\alias{tcc_path}
\title{Locate the TinyCC executable}
\usage{
tcc_path()
}
\value{
A character scalar path.
}
\description{
Returns the platform-specific \code{tcc} binary path (or \code{tcc.exe} on Windows), preferring the bundled installation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{tcc_callback_valid}
\alias{tcc_callback_valid}
\title{Check if callback is still valid}
\usage{
tcc_callback_valid(callback)
}
\arguments{
\item{callback}{A tcc_callback object}
}
\value{
Logical indicating if callback can be invoked
}
\description{
Check if callback is still valid
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{parse_callback_type}
\alias{parse_callback_type}
\title{Parse callback type specification}
\usage{
parse_callback_type(type)
}
\arguments{
\item{type}{Type string like "callback:double(int,int)"}
}
\value{
Parsed signature list or NULL
}
\description{
Parse callback type specification
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_f64}
\alias{tcc_read_f64}
\title{Read 64-bit doubles from a pointer}
\usage{
tcc_read_f64(ptr, n = NULL, offset = 0L)
}
\arguments{
\item{ptr}{External pointer}

\item{n}{Number of values to read (legacy vectorised interface).
If provided, reads \code{n} consecutive f64 values starting at byte 0.}

\item{offset}{Byte offset from \code{ptr} (scalar interface). Ignored when
\code{n} is supplied.}
}
\value{
Numeric scalar (offset form) or numeric vector (n form).
}
\description{
Read 64-bit doubles from a pointer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_ptr_is_null}
\alias{tcc_ptr_is_null}
\title{Check whether an external pointer is NULL}
\usage{
tcc_ptr_is_null(ptr)
}
\arguments{
\item{ptr}{External pointer}
}
\value{
Logical scalar
}
\description{
Returns TRUE if the external pointer address is NULL, FALSE otherwise.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_write_u32}
\alias{tcc_write_u32}
\title{Write an unsigned 32-bit integer}
\usage{
tcc_write_u32(ptr, offset, value)
}
\arguments{
\item{ptr}{External pointer}

\item{offset}{Byte offset}

\item{value}{Integer value to write}
}
\value{
\code{NULL} (invisibly). Called for its side effect of writing one
unsigned 32-bit value into native memory at \code{ptr + offset}.
}
\description{
Write an unsigned 32-bit integer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_free}
\alias{tcc_free}
\title{Free owned memory}
\usage{
tcc_free(ptr)
}
\arguments{
\item{ptr}{External pointer to free}
}
\value{
\code{NULL}.
}
\description{
Free memory whose external pointer is tagged \code{"rtinycc_owned"} (e.g.
from \code{\link[=tcc_malloc]{tcc_malloc()}} or \code{\link[=tcc_cstring]{tcc_cstring()}}). Errors on struct pointers
(use the generated \verb{struct_<name>_free()}) or borrowed pointers from
\code{\link[=tcc_data_ptr]{tcc_data_ptr()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_read_bytes}
\alias{tcc_read_bytes}
\title{Read raw bytes from a pointer}
\usage{
tcc_read_bytes(ptr, nbytes)
}
\arguments{
\item{ptr}{External pointer}

\item{nbytes}{Number of bytes to read}
}
\value{
Raw vector
}
\description{
Read a fixed number of bytes from an external pointer into a raw vector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinycc.R
\name{tcc_add_library_path}
\alias{tcc_add_library_path}
\title{Add a library path to a libtcc state}
\usage{
tcc_add_library_path(state, path)
}
\arguments{
\item{state}{A \code{tcc_state}.}

\item{path}{Path to library directory.}
}
\value{
Integer status code (0 = success).
}
\description{
Add a library path to a libtcc state
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcc_ptr_utils.R
\name{tcc_ptr_free_set_null}
\alias{tcc_ptr_free_set_null}
\title{Free the pointed memory and set to NULL}
\usage{
tcc_ptr_free_set_null(ptr_ref)
}
\arguments{
\item{ptr_ref}{External pointer to a pointer value.}
}
\value{
The updated pointer reference (invisibly).
}
\description{
Frees the memory pointed to by \code{ptr_ref} and sets the pointer to NULL.
Use this only when the pointed memory is not already owned by another
external pointer with its own finalizer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{$.tcc_compiled}
\alias{$.tcc_compiled}
\title{Access a compiled FFI symbol}
\usage{
\method{$}{tcc_compiled}(x, name)
}
\arguments{
\item{x}{A tcc_compiled object}

\item{name}{Symbol name to access}
}
\value{
The callable function or metadata field
}
\description{
Overrides \code{$} to detect dead pointers after deserialization
and recompile transparently from the stored recipe.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{print.tcc_callback}
\alias{print.tcc_callback}
\title{Print tcc_callback object}
\usage{
\method{print}{tcc_callback}(x, ...)
}
\arguments{
\item{x}{A tcc_callback object}

\item{...}{Ignored}
}
\value{
The input \code{tcc_callback} object, invisibly. Called for its
side effect of printing the callback signature, thread-safety flag,
and validity status.
}
\description{
Print tcc_callback object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treesitter_helpers.R
\name{tcc_treesitter_defines}
\alias{tcc_treesitter_defines}
\title{Extract macro defines from a header file}
\usage{
tcc_treesitter_defines(
  file,
  use_cpp = TRUE,
  cc = treesitter.c::r_cc(),
  ccflags = NULL
)
}
\arguments{
\item{file}{Path to a header file.}

\item{use_cpp}{Logical; use the C preprocessor if available.}

\item{cc}{Compiler string; passed to \code{system2()} if \code{use_cpp = TRUE}.}

\item{ccflags}{Additional flags for the compiler.}
}
\value{
Character vector of macro names defined in \code{file}.
}
\description{
Extract macro defines from a header file
}
\examples{
\dontrun{
tcc_treesitter_defines("/usr/include/math.h")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ffi.R
\name{print.tcc_cstring}
\alias{print.tcc_cstring}
\title{Print a \code{tcc_cstring} object}
\usage{
\method{print}{tcc_cstring}(x, ...)
}
\arguments{
\item{x}{A \code{tcc_cstring} object.}

\item{...}{Ignored.}
}
\value{
The input \code{tcc_cstring} object, invisibly. Called for its side
effect of printing the current string value.
}
\description{
Print a \code{tcc_cstring} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callbacks.R
\name{tcc_callback}
\alias{tcc_callback}
\title{Register an R function as a callback}
\usage{
tcc_callback(fun, signature, threadsafe = FALSE)
}
\arguments{
\item{fun}{An R function to be called from C}

\item{signature}{C function signature string (e.g., "double (*)(int, double)")}

\item{threadsafe}{Whether to enable thread-safe invocation (experimental)}
}
\value{
A tcc_callback object (externalptr wrapper)
}
\description{
Wraps an R function so it can be passed as a C function pointer
to compiled code. The callback will be invoked via a trampoline
that marshals arguments between C and R.
}
\details{
Thread safety: callbacks are executed on the R main thread only. Invoking
a callback from a worker thread is unsupported and may crash R. The
\code{threadsafe} flag is currently informational only.

If a callback raises an error, a warning is emitted and a type-appropriate
default value is returned.

When binding callbacks with \code{tcc_bind()}, use a \verb{callback:<signature>}
argument type so a synchronous trampoline is generated. The trampoline
expects a \verb{void*} user-data pointer as its first argument; pass
\code{tcc_callback_ptr(cb)} as the user-data argument to the C API. For
thread-safe usage from worker threads, use \verb{callback_async:<signature>}
which schedules the call on the main thread and returns a default value.

Pointer arguments (e.g., \code{double*}, \code{int*}) are passed as
external pointers. Lengths must be supplied separately if needed.

The return type may be any scalar type supported by the FFI mappings
(e.g., \code{i32}, \code{f64}, \code{bool}, \code{cstring}), or
\code{SEXP} to return an R object directly.

Callback lifetime: callbacks are eventually released by finalizers and
package unload. Call \code{tcc_callback_close()} when you want deterministic
invalidation and earlier release of the preserved R function.
}
